/* **********************************************************
 * Copyright (c) 2011-2018 Google, Inc.  All rights reserved.
 * **********************************************************/

/* Dr. Memory: the memory debugger
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation;
 * version 2.1 of the License, and no later version.

 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Library General Public License for more details.

 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

/* Many syscall wrappers in user32, gdi32, and kernel32 are not exported.
 * Rather than requiring the user to retrieve symbols, or complicating
 * startup by having the front-end check for symbols, we use a table.
 * Xref i#388.
 */

/* Numbers are for these platforms ("w11" means Windows 10-1511,
 * "w12" means Windows 10-1607, "w13" means Windows 10-1703,
 * "w14" means Windows 10-1709):
 * USER32(name, w2k, xpx86, w2003, xpwow, xp64, vistax86, vistawow, vista64,
 *        w7x86, w7wow, w7x64, w8x86, w8wow, w8x64, w81x86, w81wow, w81x64,
 *        w10x86, w10wow, w10x64, w11x86, w11wow, w11x64, w12x86, w12wow, w12x64,
 *        w13x86, w13wow, w13x64, w14x86, w14wow, w14x64, w15x86, w15wow, w15x64)
 *
 * Data source: win32lore/ syscall lists generated by winsysnums
 * DR client operating on dlls+pdbs.
 *
 * To turn into this format, run tools/sysnums.pl on the proper order
 * of "syscalls" files.  That script will auto-collapse the duplicate
 * names, taking the union of the numbers among them.
 * The script also appends to the existing numbers.
 * An example:
 * % tools/sysnums.pl USER32 <path-to-drmemory>/drsyscall/drsyscall_numx.h user32.dll/10.0.10586.20-x86/syscalls user32.dll/10.0.10586.20-wow64/syscalls user32.dll/10.0.10586.20-x64/syscalls
 * % tools/sysnums.pl IMM32 <path-to-drmemory>/drsyscall/drsyscall_numx.h imm32.dll/10.0.10586.0-x86/syscalls imm32.dll/10.0.10586.0-wow64/syscalls imm32.dll/10.0.10586.0-x64/syscalls
 * % tools/sysnums.pl GDI32 <path-to-drmemory>/drsyscall/drsyscall_numx.h gdi32.dll/10.0.10586.0-x86/syscalls gdi32.dll/10.0.10586.0-wow64/syscalls gdi32.dll/10.0.10586.0-x64/syscalls
 * % tools/sysnums.pl KERNEL32 <path-to-drmemory>/drsyscall/drsyscall_numx.h kernel32.dll/10.0.10586.0-x86/syscalls kernel32.dll/10.0.10586.0-wow64/syscalls kernel32.dll/10.0.10586.0-x64/syscalls
 * % tools/sysnums.pl NTDLL <path-to-drmemory>/drsyscall/drsyscall_numx.h ntdll.dll/10.0.10586.20-x86/syscalls ntdll.dll/10.0.10586.20-wow64/syscalls ntdll.dll/10.0.10586.20-x64/syscalls
 *
 * I did make some manual changes: win2K has NtUserEndTouchOperation in gdi32 but I
 * put its number in the user32 entry (b/c no duplicates allowed): currently
 * we don't care which dll it comes from.
 * Ditto for 7 wrappers in imm32 which on all other platforms duplicate the
 * user32 wrapper of the same name: but on win2k there is no user32 wrapper.
 * These are:
 *   NtUserCallOneParam
 *   NtUserGetImeHotKey
 *   NtUserGetThreadState
 *   NtUserNotifyIMEStatus
 *   NtUserQueryWindow
 *   NtUserUpdateInputContext
 *   NtUserValidateHandleSecure
 * Pre-Win8, NtUserSelectPalette is in gdi32, but I put it in user32 under the
 * assumption that future versions will be only in user32 like Win8 is.
 * Update: running sysnums.pl on gdi32 with USER32 prefix generates the right line,
 * so Win8 does seem to be in gdi32 as well.
 *
 * XXX: imm32 on x64 seems to have a complete set of user32 wrappers!
 * for now not looking at x64 imm32 but may need to eventually do automated
 * cross-dll de-duplication.
 * Also, recent platforms seem to have NtGdi wrappers in user32.dll, and NtUser wrappers
 * in gdi32.dll.
 *
 * Some other manual changes needed beyond sysnums.pl:
 * + GetThreadDesktop vs NtUserGetThreadDesktop: the former has 2 entries of its own.
 *   Xref i#487.
 *   sysnums.pl now automatically fixes this up.
 * + NtUserCreateWindowStation syscall is missing from win8-x86 lists (i#1518).
 *   sysnums.pl now appends, so new syscalls can be added w/o a manual step.
 *
 * Note that some of these are exported without the NtUser or NtGdi prefix.
 */

/* Spot-check syscalls: duplicates of entries below.
 * The number of these must match NUM_SPOT_CHECKS in drsyscall_windows.c.
 */
/* NTDLL(name,                                          w2k, xpx86,w2003,xpwow, xp64,vis86,viwow,vis64,w7x86,w7wow,w7x64,w8x86,   w8wow,w8x64,w81x86, w81wow,w81x64,w10x86,w10wow,w10x64,w11x86,w11wow,w11x64,w12x86,w12wow,w12x64,w13x86,w13wow,w13x64,w14x86,w14wow,w14x64,w15x86,w15wow,w15x64)*/
NTDLL(NtAllocateVirtualMemory                           ,0x10,0x011,0x012,0x015,0x015,0x012,0x015,0x015,0x013,0x015,0x015,0x196,0x000016,0x016,0x19b,0x000017,0x017,0x1a3,0x000018,0x018,0x1a6,0x000018,0x018,0x1a8,0x000018,0x018,0x1ad,0x000018,0x018,0x1b1,0x000018,0x018,0x1b1,0x000018,0x018)
NTDLL(NtGetContextThread                                ,0x49,0x055,0x059,0x09d,0x09d,0x097,0x0c7,0x0c7,0x087,0x0ca,0x0ca,0x113,0x0000dd,0x0dd,0x116,0x0000e0,0x0e0,0x11b,0x0000e3,0x0e3,0x11d,0x0000e4,0x0e4,0x11f,0x0000e6,0x0e6,0x122,0x0000e9,0x0e9,0x125,0x0000ea,0x0ea,0x125,0x0000eb,0x0eb)
NTDLL(NtContinue                                        ,0x1c,0x020,0x022,0x040,0x040,0x037,0x040,0x040,0x03c,0x040,0x040,0x16a,0x000041,0x041,0x16f,0x000042,0x042,0x175,0x000043,0x043,0x178,0x000043,0x043,0x17a,0x000043,0x043,0x17e,0x000043,0x043,0x181,0x000043,0x043,0x181,0x000043,0x043)
NTDLL(NtCallbackReturn                                  ,0x13,0x014,0x016,0x002,0x002,0x02b,0x002,0x002,0x02c,0x002,0x002,0x17b,0x000003,0x003,0x180,0x000004,0x004,0x187,0x000005,0x005,0x18a,0x000005,0x005,0x18c,0x000005,0x005,0x191,0x000005,0x005,0x195,0x000005,0x005,0x195,0x000005,0x005)

/* USER32(name,                                              w2k, xpx86, w2003, xpwow,  xp64,vist86,vistwow,vist64,w7x86, w7wow, w7x64, w8x86,   w8wow, w8x64,w81x86,  w81wow,w81x64,w10x86,  w10wow,w10x64,w11x86,  w11wow,w11x64,w12x86,  w12wow,w12x64,w13x86,  w13wow,w13x64,w14x86,  w14wow,w14x64,w15x86,  w15wow,w15x64)*/
USER32(NtBindCompositionSurface                          ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,0x111f,  NONE,    NONE,0x1120,0x13cd,0x001121,0x1121,0x13cf,0x001121,0x1121,0x13de,0x001120,0x1120,0x13b0,0x001115,0x1115,0x13cb,0x001115,0x1115,0x13dd,0x001115,0x1115)
USER32(NtCloseCompositionInputSink                       ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x13f3,0x001116,0x1116)
USER32(NtCompositionInputThread                          ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,0x1121,0x13e5,0x001122,0x1122,0x13e7,0x001122,0x1122,0x13f8,0x001121,0x1121,0x13cb,0x001116,0x1116,0x13e6,0x001116,0x1116,0x13f8,0x001117,0x1117)
USER32(NtCompositionSetDropTarget                        ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x13ea,0x001123,0x1123,0x13fb,0x001122,0x1122,0x13ce,0x001117,0x1117,0x13e9,0x001117,0x1117,0x13fb,0x001118,0x1118)
USER32(NtCreateCompositionInputSink                      ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,0x1122,0x13de,0x001123,0x1123,0x13e0,0x001124,0x1124,0x13f1,0x001123,0x1123,0x13c4,0x001118,0x1118,0x13df,0x001118,0x1118,0x13f0,0x001119,0x1119)
USER32(NtCreateCompositionSurfaceHandle                  ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,0x1120,  NONE,    NONE,0x1123,0x13cb,0x001124,0x1124,0x13cd,0x001125,0x1125,0x13dc,0x001124,0x1124,0x13ae,0x001119,0x1119,0x13c9,0x001119,0x1119,0x13db,0x00111a,0x111a)
USER32(NtCreateImplicitCompositionInputSink              ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x13df,0x001125,0x1125,0x13e1,0x001126,0x1126,0x13f2,0x001125,0x1125,0x13c5,0x00111a,0x111a,0x13e0,0x00111a,0x111a,0x13f1,0x00111b,0x111b)
USER32(NtDCompositionAddCrossDeviceVisualChild           ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,0x1121,  NONE,    NONE,0x1124,0x140e,0x001126,0x1126,0x1411,0x001127,0x1127,0x141e,0x001126,0x1126,0x13f1,0x00111b,0x111b,0x140c,0x00111b,0x111b,0x1420,0x00111c,0x111c)
USER32(NtDCompositionAddVisualChild                      ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,0x1122,  NONE,    NONE,0x1125,0x140f,0x001127,0x1127,0x1412,0x001128,0x1128,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
USER32(NtDCompositionAttachMouseWheelToHwnd              ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1430,0x001129,0x1129,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
USER32(NtDCompositionBeginFrame                          ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,0x1123,  NONE,    NONE,0x1126,0x13f1,0x001128,0x1128,0x13f4,0x00112a,0x112a,0x1403,0x001127,0x1127,0x13d6,0x00111c,0x111c,0x13f1,0x00111c,0x111c,0x1405,0x00111d,0x111d)
USER32(NtDCompositionCapturePointer                      ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x142a,0x001129,0x1129,0x142c,0x00112b,0x112b,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
USER32(NtDCompositionCommitChannel                       ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,0x1124,  NONE,    NONE,0x1127,0x1400,0x00112a,0x112a,0x1403,0x00112c,0x112c,0x1412,0x001128,0x1128,0x13e5,0x00111d,0x111d,0x1400,0x00111d,0x111d,0x1414,0x00111e,0x111e)
USER32(NtDCompositionCommitSynchronizationObject         ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1401,0x00111e,0x111e,0x141c,0x00111e,0x111e,0x1432,0x00111f,0x111f)
USER32(NtDCompositionConfirmFrame                        ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,0x1125,  NONE,    NONE,0x1128,0x13f2,0x00112b,0x112b,0x13f5,0x00112d,0x112d,0x1404,0x001129,0x1129,0x13d7,0x00111f,0x111f,0x13f2,0x00111f,0x111f,0x1406,0x001120,0x1120)
USER32(NtDCompositionConnectPipe                         ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,0x1126,  NONE,    NONE,0x1129,0x1423,0x00112c,0x112c,0x1425,0x00112e,0x112e,0x1426,0x00112a,0x112a,0x13f9,0x001120,0x1120,0x1414,0x001120,0x1120,0x1428,0x001121,0x1121)
USER32(NtDCompositionCreateAndBindSharedSection          ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,0x112a,0x1420,0x00112d,0x112d,0x1422,0x00112f,0x112f,0x1423,0x00112b,0x112b,0x13f6,0x001121,0x1121,0x1411,0x001121,0x1121,0x1425,0x001122,0x1122)
USER32(NtDCompositionCreateChannel                       ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,0x1127,  NONE,    NONE,0x112b,0x141c,0x00112e,0x112e,0x141f,0x001130,0x1130,0x1420,0x00112c,0x112c,0x13f3,0x001122,0x1122,0x140e,0x001122,0x1122,0x1422,0x001123,0x1123)
USER32(NtDCompositionCreateConnection                    ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,0x112c,0x13fc,0x00112f,0x112f,0x13ff,0x001131,0x1131,0x140d,0x00112d,0x112d,0x13e0,0x001123,0x1123,0x13fb,0x001123,0x1123,0x140f,0x001124,0x1124)
USER32(NtDCompositionCreateConnectionContext             ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,0x1128,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
USER32(NtDCompositionCreateDwmChannel                    ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,0x1129,  NONE,    NONE,0x112d,0x141b,0x001130,0x1130,0x141e,0x001132,0x1132,0x141f,0x00112e,0x112e,0x13f2,0x001124,0x1124,0x140d,0x001124,0x1124,0x1421,0x001125,0x1125)
USER32(NtDCompositionCreateResource                      ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,0x112a,  NONE,    NONE,0x112e,0x1418,0x001131,0x1131,0x141b,0x001133,0x1133,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
USER32(NtDCompositionCreateSharedVisualHandle            ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x13c3,0x001125,0x1125,0x13de,0x001125,0x1125,0x13ef,0x001126,0x1126)
USER32(NtDCompositionCreateSynchronizationObject         ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1431,0x001127,0x1127)
USER32(NtDCompositionCurrentBatchId                      ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,0x112b,  NONE,    NONE,0x112f,0x140a,0x001132,0x1132,0x140d,0x001134,0x1134,0x141c,0x00112f,0x112f,0x13ef,0x001126,0x1126,0x140a,0x001126,0x1126,  NONE,    NONE,  NONE)
USER32(NtDCompositionDestroyChannel                      ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,0x112c,  NONE,    NONE,0x1130,0x13fd,0x001133,0x1133,0x1400,0x001135,0x1135,0x140e,0x001130,0x1130,0x13e1,0x001127,0x1127,0x13fc,0x001127,0x1127,0x1410,0x001128,0x1128)
USER32(NtDCompositionDestroyConnection                   ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,0x1131,0x13f7,0x001134,0x1134,0x13fa,0x001136,0x1136,0x1409,0x001131,0x1131,0x13dc,0x001128,0x1128,0x13f7,0x001128,0x1128,0x140b,0x001129,0x1129)
USER32(NtDCompositionDestroyConnectionContext            ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,0x112d,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
USER32(NtDCompositionDiscardFrame                        ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,0x112e,  NONE,    NONE,0x1132,0x13f4,0x001135,0x1135,0x13f7,0x001137,0x1137,0x1406,0x001132,0x1132,0x13d9,0x001129,0x1129,0x13f4,0x001129,0x1129,0x1408,0x00112a,0x112a)
USER32(NtDCompositionDuplicateHandleToProcess            ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,0x1133,0x1426,0x001136,0x1136,0x1428,0x001138,0x1138,0x1429,0x001133,0x1133,0x13fc,0x00112a,0x112a,0x1417,0x00112a,0x112a,0x142b,0x00112b,0x112b)
USER32(NtDCompositionDuplicateSwapchainHandleToDwm       ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x142d,0x001137,0x1137,0x142f,0x001139,0x1139,0x142c,0x001134,0x1134,0x13ff,0x00112b,0x112b,0x141a,0x00112b,0x112b,0x142f,0x00112c,0x112c)
USER32(NtDCompositionDwmSyncFlush                        ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,0x112f,  NONE,    NONE,0x1134,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
USER32(NtDCompositionEnableDDASupport                    ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x13dd,0x001138,0x1138,0x13df,0x00113a,0x113a,0x13f0,0x001135,0x1135,0x13c2,0x00112c,0x112c,0x13dd,0x00112c,0x112c,0x13ee,0x00112d,0x112d)
USER32(NtDCompositionEnableMMCSS                         ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x13db,0x001139,0x1139,0x13dd,0x00113b,0x113b,0x13ee,0x001136,0x1136,0x13c0,0x00112d,0x112d,0x13db,0x00112d,0x112d,0x13ec,0x00112e,0x112e)
USER32(NtDCompositionGetAnimationTime                    ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1428,0x00113a,0x113a,0x142a,0x00113c,0x113c,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
USER32(NtDCompositionGetBatchId                          ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x141e,0x00112f,0x112f)
USER32(NtDCompositionGetChannels                         ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,0x1130,  NONE,    NONE,0x1135,0x1422,0x00113b,0x113b,0x1424,0x00113d,0x113d,0x1425,0x001137,0x1137,0x13f8,0x00112e,0x112e,0x1413,0x00112e,0x112e,0x1427,0x001130,0x1130)
USER32(NtDCompositionGetConnectionBatch                  ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,0x1136,0x13f8,0x00113c,0x113c,0x13fb,0x00113e,0x113e,0x140a,0x001138,0x1138,0x13dd,0x00112f,0x112f,0x13f8,0x00112f,0x112f,0x140c,0x001131,0x1131)
USER32(NtDCompositionGetConnectionContextBatch           ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,0x1131,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
USER32(NtDCompositionGetDeletedResources                 ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,0x1132,  NONE,    NONE,0x1137,0x13fa,0x00113d,0x113d,0x13fd,0x00113f,0x113f,0x140c,0x001139,0x1139,0x13df,0x001130,0x1130,0x13fa,0x001130,0x1130,0x140e,0x001132,0x1132)
USER32(NtDCompositionGetFrameLegacyTokens                ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,0x1133,  NONE,    NONE,0x1138,0x13f6,0x00113e,0x113e,0x13f9,0x001140,0x1140,0x1408,0x00113a,0x113a,0x13db,0x001131,0x1131,0x13f6,0x001131,0x1131,0x140a,0x001133,0x1133)
USER32(NtDCompositionGetFrameStatistics                  ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,0x1134,  NONE,    NONE,0x1139,0x13f9,0x00113f,0x113f,0x13fc,0x001141,0x1141,0x140b,0x00113b,0x113b,0x13de,0x001132,0x1132,0x13f9,0x001132,0x1132,0x140d,0x001134,0x1134)
USER32(NtDCompositionGetFrameSurfaceUpdates              ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,0x1135,  NONE,    NONE,0x113a,0x13f5,0x001140,0x1140,0x13f8,0x001142,0x1142,0x1407,0x00113c,0x113c,0x13da,0x001133,0x1133,0x13f5,0x001133,0x1133,0x1409,0x001135,0x1135)
USER32(NtDCompositionOpenSharedResource                  ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,0x113b,0x1419,0x001141,0x1141,0x141c,0x001143,0x1143,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
USER32(NtDCompositionOpenSharedResourceHandle            ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,0x113c,0x141a,0x001142,0x1142,0x141d,0x001144,0x1144,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
USER32(NtDCompositionProcessChannelBatchBuffer           ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1411,0x00113d,0x113d,0x13e4,0x001134,0x1134,0x13ff,0x001134,0x1134,0x1413,0x001136,0x1136)
USER32(NtDCompositionReferenceSharedResourceOnDwmChannel ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,0x113d,0x141e,0x001143,0x1143,0x1421,0x001145,0x1145,0x1422,0x00113e,0x113e,0x13f5,0x001135,0x1135,0x1410,0x001135,0x1135,0x1424,0x001137,0x1137)
USER32(NtDCompositionRegisterThumbnailVisual             ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,0x113e,0x1424,0x001144,0x1144,0x1426,0x001146,0x1146,0x1427,0x00113f,0x113f,0x13fa,0x001136,0x1136,0x1415,0x001136,0x1136,0x1429,0x001138,0x1138)
USER32(NtDCompositionRegisterVirtualDesktopVisual        ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1425,0x001145,0x1145,0x1427,0x001147,0x1147,0x1428,0x001140,0x1140,0x13fb,0x001137,0x1137,0x1416,0x001137,0x1137,0x142a,0x001139,0x1139)
USER32(NtDCompositionReleaseAllResources                 ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,0x1136,  NONE,    NONE,0x113f,0x13fe,0x001146,0x1146,0x1401,0x001148,0x1148,0x140f,0x001141,0x1141,0x13e2,0x001138,0x1138,0x13fd,0x001138,0x1138,0x1411,0x00113a,0x113a)
USER32(NtDCompositionReleaseResource                     ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,0x1137,  NONE,    NONE,0x1140,0x140b,0x001147,0x1147,0x140e,0x001149,0x1149,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
USER32(NtDCompositionRemoveCrossDeviceVisualChild        ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,0x1138,  NONE,    NONE,0x1141,0x140c,0x001148,0x1148,0x140f,0x00114a,0x114a,0x141d,0x001142,0x1142,0x13f0,0x001139,0x1139,0x140b,0x001139,0x1139,0x141f,0x00113b,0x113b)
USER32(NtDCompositionRemoveVisualChild                   ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,0x1139,  NONE,    NONE,0x1142,0x140d,0x001149,0x1149,0x1410,0x00114b,0x114b,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
USER32(NtDCompositionReplaceVisualChildren               ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,0x113a,  NONE,    NONE,0x1143,0x1410,0x00114a,0x114a,0x1413,0x00114c,0x114c,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
USER32(NtDCompositionRetireFrame                         ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,0x113b,  NONE,    NONE,0x1144,0x13f3,0x00114b,0x114b,0x13f6,0x00114d,0x114d,0x1405,0x001143,0x1143,0x13d8,0x00113a,0x113a,0x13f3,0x00113a,0x113a,0x1407,0x00113c,0x113c)
USER32(NtDCompositionSetChannelCallbackId                ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x142b,0x00114c,0x114c,0x142d,0x00114e,0x114e,0x142b,0x001144,0x1144,0x13fe,0x00113b,0x113b,0x1419,0x00113b,0x113b,0x142d,0x00113d,0x113d)
USER32(NtDCompositionSetChannelCommitCompletionEvent     ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,0x113c,  NONE,    NONE,0x1145,0x1402,0x00114d,0x114d,0x1405,0x00114f,0x114f,0x1414,0x001145,0x1145,0x13e7,0x00113c,0x113c,0x1402,0x00113c,0x113c,0x1416,0x00113e,0x113e)
USER32(NtDCompositionSetChildRootVisual                  ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1400,0x00113d,0x113d,0x141b,0x00113d,0x113d,0x1430,0x00113f,0x113f)
USER32(NtDCompositionSetDebugCounter                     ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,0x1146,0x1421,0x00114e,0x114e,0x1423,0x001150,0x1150,0x1424,0x001146,0x1146,0x13f7,0x00113e,0x113e,0x1412,0x00113e,0x113e,0x1426,0x001140,0x1140)
USER32(NtDCompositionSetResourceAnimationProperty        ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,0x113d,  NONE,    NONE,0x1147,0x1411,0x00114f,0x114f,0x1414,0x001151,0x1151,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
USER32(NtDCompositionSetResourceBufferProperty           ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,0x113e,  NONE,    NONE,0x1148,0x1414,0x001150,0x1150,0x1417,0x001152,0x1152,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
USER32(NtDCompositionSetResourceCallbackId               ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x142c,0x001151,0x1151,0x142e,0x001153,0x1153,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
USER32(NtDCompositionSetResourceDeletedNotificationTag   ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,0x113f,  NONE,    NONE,0x1149,0x13fb,0x001152,0x1152,0x13fe,0x001154,0x1154,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
USER32(NtDCompositionSetResourceFloatProperty            ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,0x1140,  NONE,    NONE,0x114a,0x1416,0x001153,0x1153,0x1419,0x001155,0x1155,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
USER32(NtDCompositionSetResourceHandleProperty           ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,0x114b,0x1417,0x001154,0x1154,0x141a,0x001156,0x1156,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
USER32(NtDCompositionSetResourceIntegerProperty          ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,0x1141,  NONE,    NONE,0x114c,0x1415,0x001155,0x1155,0x1418,0x001157,0x1157,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
USER32(NtDCompositionSetResourceReferenceArrayProperty   ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,0x1142,  NONE,    NONE,0x114d,0x1412,0x001156,0x1156,0x1415,0x001158,0x1158,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
USER32(NtDCompositionSetResourceReferenceProperty        ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,0x1143,  NONE,    NONE,0x114e,0x1413,0x001157,0x1157,0x1416,0x001159,0x1159,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
USER32(NtDCompositionSetVisualInputSink                  ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1427,0x001158,0x1158,0x1429,0x00115a,0x115a,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
USER32(NtDCompositionSignalGpuFence                      ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,0x1144,  NONE,    NONE,0x114f,0x141f,0x001159,0x1159,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
USER32(NtDCompositionSubmitDWMBatch                      ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,0x1145,  NONE,    NONE,0x1150,0x13ff,0x00115a,0x115a,0x1402,0x00115b,0x115b,0x1410,0x001147,0x1147,0x13e3,0x00113f,0x113f,0x13fe,0x00113f,0x113f,0x1412,0x001141,0x1141)
USER32(NtDCompositionSuspendAnimations                   ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x142e,0x001142,0x1142)
USER32(NtDCompositionSynchronize                         ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,0x1146,  NONE,    NONE,0x1151,0x141d,0x00115b,0x115b,0x1420,0x00115c,0x115c,0x1421,0x001148,0x1148,0x13f4,0x001140,0x1140,0x140f,0x001140,0x1140,0x1423,0x001143,0x1143)
USER32(NtDCompositionTelemetryAnimationScenarioBegin     ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,0x1152,0x1407,0x00115c,0x115c,0x140a,0x00115d,0x115d,0x1419,0x001149,0x1149,0x13ec,0x001141,0x1141,0x1407,0x001141,0x1141,0x141b,0x001144,0x1144)
USER32(NtDCompositionTelemetryAnimationScenarioReference ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,0x1153,0x1408,0x00115d,0x115d,0x140b,0x00115e,0x115e,0x141a,0x00114a,0x114a,0x13ed,0x001142,0x1142,0x1408,0x001142,0x1142,0x141c,0x001145,0x1145)
USER32(NtDCompositionTelemetryAnimationScenarioUnreference,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,0x1154,0x1409,0x00115e,0x115e,0x140c,0x00115f,0x115f,0x141b,0x00114b,0x114b,0x13ee,0x001143,0x1143,0x1409,0x001143,0x1143,0x141d,0x001146,0x1146)
USER32(NtDCompositionTelemetrySetApplicationId           ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,0x1155,0x1406,0x00115f,0x115f,0x1409,0x001160,0x1160,0x1418,0x00114c,0x114c,0x13eb,0x001144,0x1144,0x1406,0x001144,0x1144,0x141a,0x001147,0x1147)
USER32(NtDCompositionTelemetryTouchInteractionBegin      ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,0x1147,  NONE,    NONE,0x1156,0x1403,0x001160,0x1160,0x1406,0x001161,0x1161,0x1415,0x00114d,0x114d,0x13e8,0x001145,0x1145,0x1403,0x001145,0x1145,0x1417,0x001148,0x1148)
USER32(NtDCompositionTelemetryTouchInteractionEnd        ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,0x1148,  NONE,    NONE,0x1157,0x1405,0x001161,0x1161,0x1408,0x001162,0x1162,0x1417,0x00114e,0x114e,0x13ea,0x001146,0x1146,0x1405,0x001146,0x1146,0x1419,0x001149,0x1149)
USER32(NtDCompositionTelemetryTouchInteractionUpdate     ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,0x1149,  NONE,    NONE,0x1158,0x1404,0x001162,0x1162,0x1407,0x001163,0x1163,0x1416,0x00114f,0x114f,0x13e9,0x001147,0x1147,0x1404,0x001147,0x1147,0x1418,0x00114a,0x114a)
USER32(NtDCompositionUpdatePointerCapture                ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1429,0x001163,0x1163,0x142b,0x001164,0x1164,0x142a,0x001150,0x1150,0x13fd,0x001148,0x1148,0x1418,0x001148,0x1148,0x142c,0x00114b,0x114b)
USER32(NtDCompositionValidateAndReferenceSystemVisualForHwndTarget,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,0x114a,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
USER32(NtDCompositionWaitForChannel                      ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,0x114b,  NONE,    NONE,0x1159,0x1401,0x001164,0x1164,0x1404,0x001165,0x1165,0x1413,0x001151,0x1151,0x13e6,0x001149,0x1149,0x1401,0x001149,0x1149,0x1415,0x00114c,0x114c)
USER32(NtDWMBindCursorToOutputConfig                     ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1493,0x00114a,0x114a,0x14c1,0x00114d,0x114d)
USER32(NtDWMCommitInputSystemOutputConfig                ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1492,0x00114b,0x114b,0x14c0,0x00114e,0x114e)
USER32(NtDWMSetCursorOrientation                         ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1494,0x00114c,0x114c,0x14c2,0x00114f,0x114f)
USER32(NtDWMSetInputSystemOutputConfig                   ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1491,0x00114d,0x114d,0x14bf,0x001150,0x1150)
USER32(NtDesktopCaptureBits                              ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x13da,0x001165,0x1165,0x13dc,0x001166,0x1166,0x13ed,0x001152,0x1152,0x13bf,0x00114a,0x114a,0x13da,0x00114e,0x114e,0x13eb,0x001151,0x1151)
USER32(NtDuplicateCompositionInputSink                   ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,0x115a,0x13e0,0x001166,0x1166,0x13e2,0x001167,0x1167,0x13f3,0x001153,0x1153,0x13c6,0x00114b,0x114b,0x13e1,0x00114f,0x114f,0x13f2,0x001152,0x1152)
USER32(NtDxgkGetProcessList                              ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x13be,0x001153,0x1153)
USER32(NtDxgkRegisterVailProcess                         ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x13ba,0x001154,0x1154)
USER32(NtDxgkSubmitPresentBltToHwQueue                   ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x139f,0x001155,0x1155)
USER32(NtDxgkVailConnect                                 ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x13bb,0x001156,0x1156)
USER32(NtDxgkVailDisconnect                              ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x13bc,0x001157,0x1157)
USER32(NtDxgkVailPromoteCompositionSurface               ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x13bd,0x001158,0x1158)
USER32(NtEnableOneCoreTransformMode                      ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x14c3,0x001159,0x1159)
USER32(NtFlipObjectAddContent                            ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1437,0x00115a,0x115a)
USER32(NtFlipObjectAddPoolBuffer                         ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x141f,0x001150,0x1150,0x1435,0x00115b,0x115b)
USER32(NtFlipObjectConsumerAcquirePresent                ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1440,0x00115c,0x115c)
USER32(NtFlipObjectConsumerAdjustUsageReference          ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1443,0x00115d,0x115d)
USER32(NtFlipObjectConsumerBeginProcessPresent           ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x143f,0x00115e,0x115e)
USER32(NtFlipObjectConsumerEndProcessPresent             ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1441,0x00115f,0x115f)
USER32(NtFlipObjectConsumerPostMessage                   ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1444,0x001160,0x1160)
USER32(NtFlipObjectConsumerQueryBufferInfo               ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1442,0x001161,0x1161)
USER32(NtFlipObjectCreate                                ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x141d,0x001151,0x1151,0x1433,0x001162,0x1162)
USER32(NtFlipObjectDisconnectEndpoint                    ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x143e,0x001163,0x1163)
USER32(NtFlipObjectOpen                                  ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x141e,0x001152,0x1152,0x1434,0x001164,0x1164)
USER32(NtFlipObjectQueryBufferAvailableEvent             ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x143a,0x001165,0x1165)
USER32(NtFlipObjectQueryEndpointConnected                ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x143d,0x001166,0x1166)
USER32(NtFlipObjectQueryNextMessageToProducer            ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x143b,0x001167,0x1167)
USER32(NtFlipObjectReadNextMessageToProducer             ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x143c,0x001168,0x1168)
USER32(NtFlipObjectRemoveContent                         ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1438,0x001169,0x1169)
USER32(NtFlipObjectRemovePoolBuffer                      ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1420,0x001153,0x1153,0x1436,0x00116a,0x116a)
USER32(NtFlipObjectSetContent                            ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1439,0x00116b,0x116b)
USER32(NtHWCursorUpdatePointer                           ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1467,0x001320,0x1320,0x146c,0x001324,0x1324,0x1467,0x00131b,0x131b,0x1453,0x0012ec,0x12ec,0x1477,0x001306,0x1306,0x14a2,0x001320,0x1320)
USER32(NtIsOneCoreTransformMode                          ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x14c4,0x001321,0x1321)
USER32(NtMITActivateInputProcessing                      ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x145e,0x0012ed,0x12ed,0x1482,0x001307,0x1307,0x14ad,0x001322,0x1322)
USER32(NtMITBindInputTypeToMonitors                      ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1466,0x0012ee,0x12ee,0x148a,0x001308,0x1308,0x14b6,0x001323,0x1323)
USER32(NtMITCoreMsgKGetConnectionHandle                  ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1462,0x0012ef,0x12ef,0x1486,0x001309,0x1309,0x14b2,0x001324,0x1324)
USER32(NtMITCoreMsgKOpenConnectionTo                     ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1464,0x0012f0,0x12f0,0x1488,0x00130a,0x130a,0x14b4,0x001325,0x1325)
USER32(NtMITCoreMsgKSend                                 ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1463,0x0012f1,0x12f1,0x1487,0x00130b,0x130b,0x14b3,0x001326,0x1326)
USER32(NtMITDeactivateInputProcessing                    ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1460,0x0012f2,0x12f2,0x1484,0x00130c,0x130c,0x14b0,0x001327,0x1327)
USER32(NtMITDisableMouseIntercept                        ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1468,0x0012f3,0x12f3,0x148c,0x00130d,0x130d,0x14b8,0x001328,0x1328)
USER32(NtMITDispatchCompletion                           ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x14af,0x001329,0x1329)
USER32(NtMITEnableMouseIntercept                         ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1467,0x0012f4,0x12f4,0x148b,0x00130e,0x130e,0x14b7,0x00132a,0x132a)
USER32(NtMITGetCursorUpdateHandle                        ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1490,0x00130f,0x130f,0x14bc,0x00132b,0x132b)
USER32(NtMITSetInputCallbacks                            ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1461,0x0012f5,0x12f5,0x1485,0x001310,0x1310,0x14b1,0x00132c,0x132c)
USER32(NtMITSetInputDelegationMode                       ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x14bd,0x00132d,0x132d)
USER32(NtMITSetLastInputRecipient                        ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x14be,0x00132e,0x132e)
USER32(NtMITSynthesizeMouseInput                         ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x146a,0x0012f6,0x12f6,0x148e,0x001311,0x1311,0x14ba,0x00132f,0x132f)
USER32(NtMITSynthesizeMouseWheel                         ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x146b,0x0012f7,0x12f7,0x148f,0x001312,0x1312,0x14bb,0x001330,0x1330)
USER32(NtMITSynthesizeTouchInput                         ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1469,0x0012f8,0x12f8,0x148d,0x001313,0x1313,0x14b9,0x001331,0x1331)
USER32(NtMITUpdateInputGlobals                           ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1465,0x0012f9,0x12f9,0x1489,0x001314,0x1314,0x14b5,0x001332,0x1332)
USER32(NtMITWaitForMultipleObjectsEx                     ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x145f,0x0012fa,0x12fa,0x1483,0x001315,0x1315,0x14ae,0x001333,0x1333)
USER32(NtMapVisualRelativePoints                         ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x14c5,0x001334,0x1334)
USER32(NtNotifyPresentToCompositionSurface               ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,0x12ce,  NONE,    NONE,0x12e7,0x13d0,0x001321,0x1321,0x13d2,0x001325,0x1325,0x13e1,0x00131c,0x131c,0x13b3,0x0012fb,0x12fb,0x13ce,0x001316,0x1316,0x13e0,0x001335,0x1335)
USER32(NtOpenCompositionSurfaceDirtyRegion               ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,0x12cf,  NONE,    NONE,0x12e8,0x13d5,0x001322,0x1322,0x13d7,0x001326,0x1326,0x13e6,0x00131d,0x131d,0x13b8,0x0012fc,0x12fc,0x13d3,0x001317,0x1317,0x13e5,0x001336,0x1336)
USER32(NtOpenCompositionSurfaceSectionInfo               ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,0x12d0,  NONE,    NONE,0x12e9,0x13d2,0x001323,0x1323,0x13d4,0x001327,0x1327,0x13e3,0x00131e,0x131e,0x13b5,0x0012fd,0x12fd,0x13d0,0x001318,0x1318,0x13e2,0x001337,0x1337)
USER32(NtOpenCompositionSurfaceSwapChainHandleInfo       ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,0x12d1,  NONE,    NONE,0x12ea,0x13d3,0x001324,0x1324,0x13d5,0x001328,0x1328,0x13e4,0x00131f,0x131f,0x13b6,0x0012fe,0x12fe,0x13d1,0x001319,0x1319,0x13e3,0x001338,0x1338)
USER32(NtQueryCompositionInputIsImplicit                 ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x13e7,0x001325,0x1325,0x13e9,0x001329,0x1329,0x13fa,0x001320,0x1320,0x13cd,0x0012ff,0x12ff,0x13e8,0x00131a,0x131a,0x13fa,0x001339,0x1339)
USER32(NtQueryCompositionInputQueueAndTransform          ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x13e6,0x001326,0x1326,0x13e8,0x00132a,0x132a,0x13f9,0x001321,0x1321,0x13cc,0x001300,0x1300,0x13e7,0x00131b,0x131b,0x13f9,0x00133a,0x133a)
USER32(NtQueryCompositionInputSink                       ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,0x12eb,0x13e1,0x001327,0x1327,0x13e3,0x00132b,0x132b,0x13f4,0x001322,0x1322,0x13c7,0x001301,0x1301,0x13e2,0x00131c,0x131c,0x13f4,0x00133b,0x133b)
USER32(NtQueryCompositionInputSinkLuid                   ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,0x12ec,0x13e2,0x001328,0x1328,0x13e4,0x00132c,0x132c,0x13f5,0x001323,0x1323,0x13c8,0x001302,0x1302,0x13e3,0x00131d,0x131d,0x13f5,0x00133c,0x133c)
USER32(NtQueryCompositionInputSinkViewId                 ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x13e3,0x001329,0x1329,0x13e5,0x00132d,0x132d,0x13f6,0x001324,0x1324,0x13c9,0x001303,0x1303,0x13e4,0x00131e,0x131e,0x13f6,0x00133d,0x133d)
USER32(NtQueryCompositionSurfaceBinding                  ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,0x12d2,  NONE,    NONE,0x12ed,0x13cf,0x00132a,0x132a,0x13d1,0x00132e,0x132e,0x13e0,0x001325,0x1325,0x13b2,0x001304,0x1304,0x13cd,0x00131f,0x131f,0x13df,0x00133e,0x133e)
USER32(NtQueryCompositionSurfaceHDRMetaData              ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x13e7,0x001326,0x1326,0x13b9,0x001305,0x1305,0x13d4,0x001320,0x1320,0x13e6,0x00133f,0x133f)
USER32(NtQueryCompositionSurfaceRenderingRealization     ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,0x12d3,  NONE,    NONE,0x12ee,0x13d4,0x00132b,0x132b,0x13d6,0x00132f,0x132f,0x13e5,0x001327,0x1327,0x13b7,0x001306,0x1306,0x13d2,0x001321,0x1321,0x13e4,0x001340,0x1340)
USER32(NtQueryCompositionSurfaceStatistics               ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,0x12d4,  NONE,    NONE,0x12ef,0x13d1,0x00132c,0x132c,0x13d3,0x001330,0x1330,0x13e2,0x001328,0x1328,0x13b4,0x001307,0x1307,0x13cf,0x001322,0x1322,0x13e1,0x001341,0x1341)
USER32(NtRIMAddInputObserver                             ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x145f,0x00132d,0x132d,0x1464,0x001331,0x1331,0x145f,0x001329,0x1329,0x144a,0x001308,0x1308,0x146d,0x001323,0x1323,0x1498,0x001342,0x1342)
USER32(NtRIMAreSiblingDevices                            ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1441,0x001309,0x1309,0x1464,0x001324,0x1324,0x148f,0x001343,0x1343)
USER32(NtRIMDeviceIoControl                              ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1445,0x00130a,0x130a,0x1468,0x001325,0x1325,0x1493,0x001344,0x1344)
USER32(NtRIMEnableMonitorMappingForDevice                ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1473,0x001326,0x1326,0x149e,0x001345,0x1345)
USER32(NtRIMFreeInputBuffer                              ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1442,0x00130b,0x130b,0x1465,0x001327,0x1327,0x1490,0x001346,0x1346)
USER32(NtRIMGetDevicePreparsedData                       ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x143f,0x00130c,0x130c,0x1462,0x001328,0x1328,0x148d,0x001347,0x1347)
USER32(NtRIMGetDevicePreparsedDataLockfree               ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1463,0x00132e,0x132e,0x1468,0x001332,0x1332,0x1463,0x00132a,0x132a,0x144e,0x00130d,0x130d,0x1471,0x001329,0x1329,0x149c,0x001348,0x1348)
USER32(NtRIMGetDeviceProperties                          ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1440,0x00130e,0x130e,0x1463,0x00132a,0x132a,0x148e,0x001349,0x1349)
USER32(NtRIMGetDevicePropertiesLockfree                  ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x144f,0x00130f,0x130f,0x1472,0x00132b,0x132b,0x149d,0x00134a,0x134a)
USER32(NtRIMGetPhysicalDeviceRect                        ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1448,0x001310,0x1310,0x146b,0x00132c,0x132c,0x1496,0x00134b,0x134b)
USER32(NtRIMGetSourceProcessId                           ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1449,0x001311,0x1311,0x146c,0x00132d,0x132d,0x1497,0x00134c,0x134c)
USER32(NtRIMObserveNextInput                             ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1462,0x00132f,0x132f,0x1467,0x001333,0x1333,0x1462,0x00132b,0x132b,0x144d,0x001312,0x1312,0x1470,0x00132e,0x132e,0x149b,0x00134d,0x134d)
USER32(NtRIMOnPnpNotification                            ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1443,0x001313,0x1313,0x1466,0x00132f,0x132f,0x1491,0x00134e,0x134e)
USER32(NtRIMOnTimerNotification                          ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1444,0x001314,0x1314,0x1467,0x001330,0x1330,0x1492,0x00134f,0x134f)
USER32(NtRIMReadInput                                    ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x143e,0x001315,0x1315,0x1461,0x001331,0x1331,0x148c,0x001350,0x1350)
USER32(NtRIMRegisterForInput                             ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x143d,0x001316,0x1316,0x1460,0x001332,0x1332,0x148b,0x001351,0x1351)
USER32(NtRIMRemoveInputObserver                          ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1460,0x001330,0x1330,0x1465,0x001334,0x1334,0x1460,0x00132c,0x132c,0x144b,0x001317,0x1317,0x146e,0x001333,0x1333,0x1499,0x001352,0x1352)
USER32(NtRIMSetTestModeStatus                            ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1447,0x001318,0x1318,0x146a,0x001334,0x1334,0x1495,0x001353,0x1353)
USER32(NtRIMUnregisterForInput                           ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1446,0x001319,0x1319,0x1469,0x001335,0x1335,0x1494,0x001354,0x1354)
USER32(NtRIMUpdateInputObserverRegistration              ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1461,0x001331,0x1331,0x1466,0x001335,0x1335,0x1461,0x00132d,0x132d,0x144c,0x00131a,0x131a,0x146f,0x001336,0x1336,0x149a,0x001355,0x1355)
USER32(NtSetCompositionSurfaceAnalogExclusive            ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x13f0,0x001332,0x1332,0x13f3,0x001336,0x1336,0x1402,0x00132e,0x132e,0x13d5,0x00131b,0x131b,0x13f0,0x001337,0x1337,0x1404,0x001356,0x1356)
USER32(NtSetCompositionSurfaceBufferCompositionMode      ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,0x12f0,0x13d8,0x001333,0x1333,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
USER32(NtSetCompositionSurfaceBufferCompositionModeAndOrientation,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x13da,0x001337,0x1337,0x13eb,0x00132f,0x132f,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
USER32(NtSetCompositionSurfaceBufferUsage                ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x13bd,0x00131c,0x131c,0x13d8,0x001338,0x1338,0x13e9,0x001357,0x1357)
USER32(NtSetCompositionSurfaceDirectFlipState            ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x13e9,0x001330,0x1330,0x13bb,0x00131d,0x131d,0x13d6,0x001339,0x1339,0x13e7,0x001358,0x1358)
USER32(NtSetCompositionSurfaceHDRMetaData                ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x13e8,0x001331,0x1331,0x13ba,0x00131e,0x131e,0x13d5,0x00133a,0x133a,  NONE,    NONE,  NONE)
USER32(NtSetCompositionSurfaceIndependentFlipInfo        ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,0x12f1,0x13d9,0x001334,0x1334,0x13db,0x001338,0x1338,0x13ec,0x001332,0x1332,0x13be,0x00131f,0x131f,0x13d9,0x00133b,0x133b,0x13ea,0x001359,0x1359)
USER32(NtSetCompositionSurfaceOutOfFrameDirectFlipNotification,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,0x12d5,  NONE,    NONE,0x12f2,0x13d6,0x001335,0x1335,0x13d8,0x001339,0x1339,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
USER32(NtSetCompositionSurfaceStatistics                 ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,0x12d6,  NONE,    NONE,0x12f3,0x13d7,0x001336,0x1336,0x13d9,0x00133a,0x133a,0x13ea,0x001333,0x1333,0x13bc,0x001320,0x1320,0x13d7,0x00133c,0x133c,0x13e8,0x00135a,0x135a)
USER32(NtTokenManagerConfirmOutstandingAnalogToken       ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x13ef,0x001337,0x1337,0x13f2,0x00133b,0x133b,0x1401,0x001334,0x1334,0x13d4,0x001321,0x1321,0x13ef,0x00133d,0x133d,0x1403,0x00135b,0x135b)
USER32(NtTokenManagerCreateCompositionTokenHandle        ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,0x12f4,0x13ec,0x001338,0x1338,0x13ef,0x00133c,0x133c,0x13fe,0x001335,0x1335,0x13d1,0x001322,0x1322,0x13ec,0x00133e,0x133e,0x13fe,0x00135c,0x135c)
USER32(NtTokenManagerCreateFlipObjectReturnTokenHandle   ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1400,0x00135d,0x135d)
USER32(NtTokenManagerCreateFlipObjectTokenHandle         ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x13ff,0x00135e,0x135e)
USER32(NtTokenManagerDeleteOutstandingDirectFlipTokens   ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,0x12f5,0x13eb,0x001339,0x1339,0x13ee,0x00133d,0x133d,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
USER32(NtTokenManagerGetAnalogExclusiveSurfaceUpdates    ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x13ee,0x00133a,0x133a,0x13f1,0x00133e,0x133e,0x1400,0x001336,0x1336,0x13d3,0x001323,0x1323,0x13ee,0x00133f,0x133f,0x1402,0x00135f,0x135f)
USER32(NtTokenManagerGetAnalogExclusiveTokenEvent        ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x13ed,0x00133b,0x133b,0x13f0,0x00133f,0x133f,0x13ff,0x001337,0x1337,0x13d2,0x001324,0x1324,0x13ed,0x001340,0x1340,0x1401,0x001360,0x1360)
USER32(NtTokenManagerGetOutOfFrameDirectFlipSurfaceUpdates,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,0x12d7,  NONE,    NONE,0x12f6,0x13ea,0x00133c,0x133c,0x13ed,0x001340,0x1340,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
USER32(NtTokenManagerOpenEvent                           ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,0x12d8,  NONE,    NONE,0x12f7,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
USER32(NtTokenManagerOpenSection                         ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,0x12f8,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
USER32(NtTokenManagerOpenSectionAndEvents                ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x13e8,0x00133d,0x133d,0x13eb,0x001341,0x1341,0x13fc,0x001338,0x1338,0x13cf,0x001325,0x1325,0x13ea,0x001341,0x1341,0x13fc,0x001361,0x1361)
USER32(NtTokenManagerThread                              ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,0x12d9,  NONE,    NONE,0x12f9,0x13e9,0x00133e,0x133e,0x13ec,0x001342,0x1342,0x13fd,0x001339,0x1339,0x13d0,0x001326,0x1326,0x13eb,0x001342,0x1342,0x13fd,0x001362,0x1362)
USER32(NtUnBindCompositionSurface                        ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,0x12da,  NONE,    NONE,0x12fa,0x13ce,0x00133f,0x133f,0x13d0,0x001343,0x1343,0x13df,0x00133a,0x133a,0x13b1,0x001327,0x1327,0x13cc,0x001343,0x1343,0x13de,0x001363,0x1363)
USER32(NtUpdateInputSinkTransforms                       ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,0x12fb,0x13e4,0x001340,0x1340,0x13e6,0x001344,0x1344,0x13f7,0x00133b,0x133b,0x13ca,0x001328,0x1328,0x13e5,0x001344,0x1344,0x13f7,0x001364,0x1364)
USER32(NtUserAcquireIAMKey                               ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x13c7,0x0012db,0x12db,0x13ee,0x0012fc,0x12fc,0x1439,0x001341,0x1341,0x143c,0x001345,0x1345,0x1438,0x00133c,0x133c,0x1413,0x001329,0x1329,0x1434,0x001345,0x1345,0x145a,0x001365,0x1365)
USER32(NtUserAcquireInteractiveControlBackgroundAccess   ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1468,0x00133d,0x133d,0x1454,0x00132a,0x132a,0x1478,0x001346,0x1346,0x14a3,0x001366,0x1366)
USER32(NtUserActivateKeyboardLayout                      ,0x1126,0x1130,0x112f,0x1126,0x1126,0x1139,0x1127,0x1127,0x113a,0x111e,0x111e,0x1171,0x07111e,0x111e,0x1173,0x07111f,0x111f,0x1174,0x071120,0x1120,0x1176,0x071120,0x1120,0x1178,0x07111f,0x111f,0x1136,0x071114,0x1114,0x1137,0x071114,0x1114,0x113b,0x071114,0x1114)
USER32(NtUserAddClipboardFormatListener                  ,  NONE,  NONE,  NONE,  NONE,  NONE,0x113a,0x1274,0x1274,0x113b,0x1281,0x1281,0x1170,0x0012dc,0x12dc,0x1172,0x0012fd,0x12fd,0x1173,0x001342,0x1342,0x1175,0x001346,0x1346,0x1177,0x00133e,0x133e,0x1135,0x00132b,0x132b,0x1136,0x001347,0x1347,0x113a,0x001367,0x1367)
USER32(NtUserAlterWindowStyle                            ,  NONE,0x1131,0x1130,0x10d7,0x10d7,0x113b,0x10d8,0x10d8,0x113c,0x10d3,0x10d3,0x116f,0x0c10d3,0x10d3,0x1171,0x0c10d4,0x10d4,0x1172,0x0c10d5,0x10d5,0x1174,0x0c10d5,0x10d5,0x1176,0x0c10d4,0x10d4,0x1134,0x0c10c9,0x10c9,0x1135,0x0c10c9,0x10c9,0x1139,0x0c10c9,0x10c9)
USER32(NtUserAttachThreadInput                           ,0x1129,0x1133,0x1132,0x10f0,0x10f0,0x113d,0x10f1,0x10f1,0x113e,0x10ec,0x10ec,0x116d,0x0a10ec,0x10ec,0x116f,0x0a10ed,0x10ed,0x1170,0x0a10ee,0x10ee,0x1172,0x0a10ee,0x10ee,0x1174,0x0a10ed,0x10ed,0x1132,0x0a10e2,0x10e2,0x1133,0x0a10e2,0x10e2,0x1137,0x0a10e2,0x10e2)
USER32(NtUserAutoPromoteMouseInPointer                   ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x13d1,0x0012de,0x12de,0x13fa,0x0012ff,0x12ff,0x1447,0x001344,0x1344,0x144b,0x001348,0x1348,0x1447,0x001340,0x1340,0x1422,0x00132d,0x132d,0x1443,0x001349,0x1349,0x146e,0x001369,0x1369)
USER32(NtUserAutoRotateScreen                            ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,0x12df,  NONE,    NONE,0x1300,0x1438,0x001345,0x1345,0x143b,0x001349,0x1349,0x1437,0x001341,0x1341,0x1412,0x00132e,0x132e,0x1433,0x00134a,0x134a,0x1459,0x00136a,0x136a)
USER32(NtUserBeginLayoutUpdate                           ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1409,0x00132f,0x132f,0x142a,0x00134b,0x134b,0x144e,0x00136b,0x136b)
USER32(NtUserBeginPaint                                  ,0x112a,0x1134,0x1133,0x1016,0x1016,0x113e,0x1017,0x1017,0x113f,0x1017,0x1017,0x116c,0x001018,0x1018,0x116e,0x001019,0x1019,0x116f,0x00101a,0x101a,0x1171,0x00101a,0x101a,0x1173,0x00101a,0x101a,0x1131,0x00101a,0x101a,0x1132,0x00101a,0x101a,0x1136,0x00101a,0x101a)
USER32(NtUserBitBltSysBmp                                ,  NONE,0x1135,0x1134,0x10cf,0x10cf,0x113f,0x10d0,0x10d0,0x1140,0x10cb,0x10cb,0x116b,0x0010cb,0x10cb,0x116d,0x0010cc,0x10cc,0x116e,0x0010cd,0x10cd,0x1170,0x0010cd,0x10cd,0x1172,0x0010cd,0x10cd,0x1130,0x0010c2,0x10c2,0x1131,0x0010c2,0x10c2,0x1134,0x0010c2,0x10c2)
USER32(NtUserBlockInput                                  ,0x112c,0x1136,0x1135,0x1226,0x1226,0x1140,0x1276,0x1276,0x1141,0x1283,0x1283,0x116a,0x0012e0,0x12e0,0x116c,0x001301,0x1301,0x116d,0x001346,0x1346,0x116f,0x00134a,0x134a,0x1171,0x001342,0x1342,0x112f,0x001330,0x1330,0x1130,0x00134c,0x134c,0x1133,0x00136c,0x136c)
USER32(NtUserBroadcastThemeChangeEvent                   ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x116c,0x001343,0x1343,0x112a,0x001331,0x1331,0x112b,0x00134d,0x134d,0x112e,0x00136d,0x136d)
USER32(NtUserBuildHwndList                               ,  NONE,0x1138,0x1137,0x101b,0x101b,0x1142,0x101c,0x101c,0x1143,0x101c,0x101c,0x1168,0x00101d,0x101d,0x116a,0x00101e,0x101e,0x116b,0x00101f,0x101f,0x116d,0x00101f,0x101f,0x116f,0x00101f,0x101f,0x112d,0x00101f,0x101f,0x112e,0x00101f,0x101f,0x1131,0x00101f,0x101f)
USER32(NtUserBuildNameList                               ,  NONE,0x1139,0x1138,0x10b2,0x10b2,0x1143,0x10b3,0x10b3,0x1144,0x10ae,0x10ae,0x1167,0x0010ae,0x10ae,0x1169,0x0010af,0x10af,0x116a,0x0010b0,0x10b0,0x116c,0x0010b0,0x10b0,0x116e,0x0010b0,0x10b0,0x112c,0x0010a9,0x10a9,0x112d,0x0010a9,0x10a9,0x1130,0x0010a9,0x10a9)
USER32(NtUserBuildPropList                               ,  NONE,0x113a,0x1139,0x1228,0x1228,0x1144,0x1278,0x1278,0x1145,0x1285,0x1285,0x1166,0x0012e2,0x12e2,0x1168,0x001303,0x1303,0x1169,0x001348,0x1348,0x116b,0x00134c,0x134c,0x116d,0x001345,0x1345,0x112b,0x001333,0x1333,0x112c,0x00134f,0x134f,0x112f,0x00136f,0x136f)
USER32(NtUserCalcMenuBar                                 ,  NONE,0x1236,0x1232,0x1098,0x1098,0x124e,0x1099,0x1099,0x125b,0x1097,0x1097,0x1219,0x001097,0x1097,0x121c,0x001098,0x1098,0x121f,0x001099,0x1099,0x1221,0x001099,0x1099,0x1224,0x001099,0x1099,0x11e3,0x001097,0x1097,0x11e7,0x001097,0x1097,0x11ee,0x001097,0x1097)
USER32(NtUserCalculatePopupWindowPosition                ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x125a,0x1286,0x1286,0x121a,0x0012e3,0x12e3,0x121d,0x001304,0x1304,0x1220,0x001349,0x1349,0x1222,0x00134d,0x134d,0x1225,0x001346,0x1346,0x11e4,0x001334,0x1334,0x11e8,0x001350,0x1350,0x11ef,0x001370,0x1370)
USER32(NtUserCallHwnd                                    ,  NONE,0x113b,0x113a,0x1112,0x1112,0x1145,0x1113,0x1113,0x1146,0x110c,0x110c,0x1165,0x00110c,0x110c,0x1167,0x00110d,0x110d,0x1168,0x00110e,0x110e,0x116a,0x00110e,0x110e,0x116b,0x00110d,0x110d,0x1129,0x001102,0x1102,0x112a,0x001102,0x1102,0x112d,0x001102,0x1102)
USER32(NtUserCallHwndLock                                ,  NONE,0x113c,0x113b,0x1020,0x1020,0x1146,0x1021,0x1021,0x1147,0x1021,0x1021,0x1164,0x001022,0x1022,0x1166,0x001023,0x1023,0x1167,0x001024,0x1024,0x1169,0x001024,0x1024,0x116a,0x001024,0x1024,0x1128,0x001024,0x1024,0x1129,0x001024,0x1024,0x112c,0x001024,0x1024)
USER32(NtUserCallHwndOpt                                 ,  NONE,0x113d,0x113c,0x1229,0x1229,0x1147,0x1279,0x1279,0x1148,0x1287,0x1287,0x1163,0x0012e4,0x12e4,0x1165,0x001305,0x1305,0x1166,0x00134a,0x134a,0x1168,0x00134e,0x134e,0x1169,0x001347,0x1347,0x1127,0x001335,0x1335,0x1128,0x001351,0x1351,0x112b,0x001371,0x1371)
USER32(NtUserCallHwndParam                               ,  NONE,0x113e,0x113d,0x109f,0x109f,0x1148,0x10a0,0x10a0,0x1149,0x109e,0x109e,0x1162,0x00109e,0x109e,0x1164,0x00109f,0x109f,0x1165,0x0010a0,0x10a0,0x1167,0x0010a0,0x10a0,0x1168,0x0010a0,0x10a0,0x1126,0x00109e,0x109e,0x1127,0x00109e,0x109e,0x112a,0x00109e,0x109e)
USER32(NtUserCallHwndParamLock                           ,  NONE,0x113f,0x113e,0x1026,0x1026,0x1149,0x1027,0x1027,0x114a,0x1027,0x1027,0x1161,0x001028,0x1028,0x1163,0x001029,0x1029,0x1164,0x00102a,0x102a,0x1166,0x00102a,0x102a,0x1167,0x00102a,0x102a,0x1125,0x00102a,0x102a,0x1126,0x00102a,0x102a,0x1129,0x00102a,0x102a)
USER32(NtUserCallMsgFilter                               ,  NONE,0x1140,0x113f,0x1014,0x1014,0x114a,0x1015,0x1015,0x114b,0x1015,0x1015,0x1160,0x001016,0x1016,0x1162,0x001017,0x1017,0x1163,0x001018,0x1018,0x1165,0x001018,0x1018,0x1166,0x001018,0x1018,0x1124,0x001018,0x1018,0x1125,0x001018,0x1018,0x1128,0x001018,0x1018)
USER32(NtUserCallNextHookEx                              ,  NONE,0x1141,0x1140,0x101d,0x101d,0x114b,0x101e,0x101e,0x114c,0x101e,0x101e,0x115f,0x00101f,0x101f,0x1161,0x001020,0x1020,0x1162,0x001021,0x1021,0x1164,0x001021,0x1021,0x1165,0x001021,0x1021,0x1123,0x001021,0x1021,0x1124,0x001021,0x1021,0x1127,0x001021,0x1021)
USER32(NtUserCallNoParam                                 ,  NONE,0x1142,0x1141,0x1005,0x1005,0x114c,0x1005,0x1005,0x114d,0x1005,0x1005,0x115e,0x001006,0x1006,0x1160,0x001007,0x1007,0x1161,0x001008,0x1008,0x1163,0x001008,0x1008,0x1164,0x001008,0x1008,0x1122,0x001008,0x1008,0x1123,0x001008,0x1008,0x1126,0x001008,0x1008)
USER32(NtUserCallOneParam                                ,0x1139,0x1143,0x1142,0x1002,0x1002,0x114d,0x1002,0x1002,0x114e,0x1002,0x1002,0x115d,0x001003,0x1003,0x115f,0x001004,0x1004,0x1160,0x001005,0x1005,0x1162,0x001005,0x1005,0x1163,0x001005,0x1005,0x1121,0x001005,0x1005,0x1122,0x001005,0x1005,0x1125,0x001005,0x1005)
USER32(NtUserCallTwoParam                                ,  NONE,0x1144,0x1143,0x1029,0x1029,0x114e,0x102a,0x102a,0x114f,0x102a,0x102a,0x115c,0x00102b,0x102b,0x115e,0x00102c,0x102c,0x115f,0x00102d,0x102d,0x1161,0x00102d,0x102d,0x1162,0x00102d,0x102d,0x1120,0x00102d,0x102d,0x1121,0x00102d,0x102d,0x1124,0x00102d,0x102d)
USER32(NtUserCallUserpExitWindowsEx                      ,  NONE,  NONE,  NONE,0x2057,  NONE,  NONE,0x205d,  NONE,  NONE,0x205b,  NONE,  NONE,0x002000,  NONE,  NONE,0x002001,  NONE,  NONE,0x002001,  NONE,  NONE,0x002001,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
USER32(NtUserCallUserpRegisterLogonProcess               ,  NONE,  NONE,  NONE,0x205a,  NONE,  NONE,0x2060,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
USER32(NtUserCanBrokerForceForeground                    ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x1140,0x0012e5,0x12e5,0x1142,0x001306,0x1306,0x1143,0x00134b,0x134b,0x1145,0x00134f,0x134f,0x1146,0x001348,0x1348,0x1103,0x001336,0x1336,0x1104,0x001352,0x1352,0x1104,0x001372,0x1372)
USER32(NtUserChangeClipboardChain                        ,0x113b,0x1145,0x1144,0x111f,0x111f,0x114f,0x1120,0x1120,0x1150,0x1119,0x1119,0x115b,0x081119,0x1119,0x115d,0x08111a,0x111a,0x115e,0x08111b,0x111b,0x1160,0x08111b,0x111b,0x1161,0x08111a,0x111a,0x111f,0x08110f,0x110f,0x1120,0x08110f,0x110f,0x1123,0x08110f,0x110f)
USER32(NtUserChangeDisplaySettings                       ,  NONE,0x1146,0x1145,0x122a,0x122a,0x1150,0x127a,0x127a,0x1151,0x1288,0x1288,0x115a,0x0012e6,0x12e6,0x115c,0x001307,0x1307,0x115d,0x00134c,0x134c,0x115f,0x001350,0x1350,0x1160,0x001349,0x1349,0x111e,0x001337,0x1337,0x111f,0x001353,0x1353,0x111f,0x001373,0x1373)
USER32(NtUserChangeWindowMessageFilterEx                 ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x1291,0x1289,0x1289,0x12d9,0x0012e7,0x12e7,0x12e2,0x001308,0x1308,0x12ee,0x00134d,0x134d,0x12ef,0x001351,0x1351,0x12f5,0x00134a,0x134a,0x12b6,0x001338,0x1338,0x12bf,0x001354,0x1354,0x12c7,0x001374,0x1374)
USER32(NtUserCheckAccessForIntegrityLevel                ,  NONE,  NONE,  NONE,  NONE,  NONE,0x1151,0x127b,0x127b,0x1157,0x128a,0x128a,0x1154,0x0012e8,0x12e8,0x1156,0x001309,0x1309,0x1157,0x00134e,0x134e,0x1159,0x001352,0x1352,0x115a,0x00134b,0x134b,0x1117,0x001339,0x1339,0x1118,0x001355,0x1355,0x1118,0x001375,0x1375)
USER32(NtUserCheckDesktopByThreadId                      ,  NONE,  NONE,  NONE,  NONE,  NONE,0x1152,0x127c,0x127c,0x1158,0x128b,0x128b,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
USER32(NtUserCheckImeHotKey                              ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x1075,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
USER32(NtUserCheckMenuItem                               ,  NONE,0x1148,0x1147,0x1108,0x1108,0x1155,0x1109,0x1109,0x115a,0x1103,0x1103,0x1151,0x0a1103,0x1103,0x1153,0x0a1104,0x1104,0x1154,0x0a1105,0x1105,0x1156,0x0a1105,0x1105,0x1157,0x0a1104,0x1104,0x1114,0x0a10f9,0x10f9,0x1115,0x0a10f9,0x10f9,0x1115,0x0a10f9,0x10f9)
USER32(NtUserCheckProcessForClipboardAccess              ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x13d5,0x0012e9,0x12e9,0x13fe,0x00130a,0x130a,0x144b,0x00134f,0x134f,0x144f,0x001353,0x1353,0x144b,0x00134c,0x134c,0x1425,0x00133a,0x133a,0x1446,0x001356,0x1356,0x1471,0x001376,0x1376)
USER32(NtUserCheckProcessSession                         ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x1280,0x0012ea,0x12ea,0x1283,0x00130b,0x130b,0x128d,0x001350,0x1350,0x128e,0x001354,0x1354,0x1294,0x00134d,0x134d,0x1255,0x00133b,0x133b,0x125d,0x001357,0x1357,0x1265,0x001377,0x1377)
USER32(NtUserCheckWindowThreadDesktop                    ,  NONE,  NONE,  NONE,  NONE,  NONE,0x1153,0x2064,0x127d,0x1159,0x128c,0x128c,0x1152,0x0012eb,0x12eb,0x1154,0x00130c,0x130c,0x1155,0x001351,0x1351,0x1157,0x001355,0x1355,0x1158,0x00134e,0x134e,0x1115,0x00133c,0x133c,0x1116,0x001358,0x1358,0x1116,0x001378,0x1378)
USER32(NtUserChildWindowFromPointEx                      ,0x113f,0x1149,0x1148,0x122b,0x122b,0x1156,0x127e,0x127e,0x115b,0x128d,0x128d,0x1150,0x0012ec,0x12ec,0x1152,0x00130d,0x130d,0x1153,0x001352,0x1352,0x1155,0x001356,0x1356,0x1156,0x00134f,0x134f,0x1113,0x00133d,0x133d,0x1114,0x001359,0x1359,0x1114,0x001379,0x1379)
USER32(NtUserClearForeground                             ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,0x13f3,0x00130e,0x130e,0x143e,0x001353,0x1353,0x1441,0x001357,0x1357,0x143d,0x001350,0x1350,0x1418,0x00133e,0x133e,0x1439,0x00135a,0x135a,0x145f,0x00137a,0x137a)
USER32(NtUserClipCursor                                  ,0x1140,0x114a,0x1149,0x122c,0x122c,0x1157,0x127f,0x127f,0x115c,0x128e,0x128e,0x114f,0x0412ed,0x12ed,0x1151,0x04130f,0x130f,0x1152,0x041354,0x1354,0x1154,0x041358,0x1358,0x1155,0x041351,0x1351,0x1112,0x04133f,0x133f,0x1113,0x04135b,0x135b,0x1113,0x04137b,0x137b)
USER32(NtUserCloseClipboard                              ,0x1141,0x114b,0x114a,0x10d2,0x10d2,0x1158,0x10d3,0x10d3,0x115d,0x10ce,0x10ce,0x114e,0x0110ce,0x10ce,0x1150,0x0110cf,0x10cf,0x1151,0x0110d0,0x10d0,0x1153,0x0110d0,0x10d0,0x1154,0x0110d0,0x10d0,0x1111,0x0110c5,0x10c5,0x1112,0x0110c5,0x10c5,0x1112,0x0110c5,0x10c5)
USER32(NtUserCloseDesktop                                ,0x1142,0x114c,0x114b,0x10aa,0x10aa,0x1159,0x10ab,0x10ab,0x115e,0x10a6,0x10a6,0x114d,0x0410a6,0x10a6,0x114f,0x0410a7,0x10a7,0x1150,0x0410a8,0x10a8,0x1152,0x0410a8,0x10a8,0x1153,0x0410a8,0x10a8,0x1110,0x0410a2,0x10a2,0x1111,0x0410a2,0x10a2,0x1111,0x0410a2,0x10a2)
USER32(NtUserCloseWindowStation                          ,0x1143,0x114d,0x114c,0x10b9,0x10b9,0x115a,0x10ba,0x10ba,0x115f,0x10b5,0x10b5,0x114c,0x0410b5,0x10b5,0x114e,0x0410b6,0x10b6,0x114f,0x0410b7,0x10b7,0x1151,0x0410b7,0x10b7,0x1152,0x0410b7,0x10b7,0x110f,0x0410b0,0x10b0,0x1110,0x0410b0,0x10b0,0x1110,0x0410b0,0x10b0)
USER32(NtUserCompositionInputSinkLuidFromPoint           ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,0x1401,0x001310,0x1310,0x144e,0x001355,0x1355,0x1452,0x001359,0x1359,0x144e,0x001352,0x1352,0x1428,0x001340,0x1340,0x1449,0x00135c,0x135c,0x1474,0x00137c,0x137c)
USER32(NtUserCompositionInputSinkViewInstanceIdFromPoint ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1429,0x001341,0x1341,0x144a,0x00135d,0x135d,0x1475,0x00137d,0x137d)
USER32(NtUserConfirmResizeCommit                         ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x140b,0x001342,0x1342,0x142c,0x00135e,0x135e,0x1450,0x00137e,0x137e)
USER32(NtUserConsoleControl                              ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x10e4,0x1160,0x10df,0x10df,0x114b,0x0010df,0x10df,0x114d,0x0010e0,0x10e0,0x114e,0x0010e1,0x10e1,0x1150,0x0010e1,0x10e1,0x1151,0x0010e0,0x10e0,0x110e,0x0010d5,0x10d5,0x110f,0x0010d5,0x10d5,0x110f,0x0010d5,0x10d5)
USER32(NtUserConvertMemHandle                            ,  NONE,0x114f,0x114e,0x1102,0x1102,0x115c,0x1103,0x1103,0x1161,0x10fd,0x10fd,0x114a,0x0510fd,0x10fd,0x114c,0x0510fe,0x10fe,0x114d,0x0510ff,0x10ff,0x114f,0x0510ff,0x10ff,0x1150,0x0510fe,0x10fe,0x110d,0x0510f3,0x10f3,0x110e,0x0510f3,0x10f3,0x110e,0x0510f3,0x10f3)
USER32(NtUserCopyAcceleratorTable                        ,0x1146,0x1150,0x114f,0x102b,0x102b,0x115d,0x102c,0x102c,0x1162,0x102c,0x102c,0x1149,0x0a102d,0x102d,0x114b,0x0a102e,0x102e,0x114c,0x0a102f,0x102f,0x114e,0x0a102f,0x102f,0x114f,0x0a102f,0x102f,0x110c,0x0a102f,0x102f,0x110d,0x0a102f,0x102f,0x110d,0x0a102f,0x102f)
USER32(NtUserCountClipboardFormats                       ,0x1147,0x1151,0x1150,0x1115,0x1115,0x115e,0x1116,0x1116,0x1163,0x110f,0x110f,0x1148,0x01110f,0x110f,0x114a,0x011110,0x1110,0x114b,0x011111,0x1111,0x114d,0x011111,0x1111,0x114e,0x011110,0x1110,0x110b,0x011105,0x1105,0x110c,0x011105,0x1105,0x110c,0x011105,0x1105)
USER32(NtUserCreateAcceleratorTable                      ,0x1148,0x1152,0x1151,0x10f5,0x10f5,0x115f,0x10f6,0x10f6,0x1164,0x10f1,0x10f1,0x1147,0x0910f1,0x10f1,0x1149,0x0910f2,0x10f2,0x114a,0x0910f3,0x10f3,0x114c,0x0910f3,0x10f3,0x114d,0x0910f2,0x10f2,0x110a,0x0910e7,0x10e7,0x110b,0x0910e7,0x10e7,0x110b,0x0910e7,0x10e7)
USER32(NtUserCreateCaret                                 ,0x1149,0x1153,0x1152,0x1031,0x1031,0x1160,0x1032,0x1032,0x1165,0x1032,0x1032,0x1146,0x141033,0x1033,0x1148,0x141034,0x1034,0x1149,0x141035,0x1035,0x114b,0x141035,0x1035,0x114c,0x141035,0x1035,0x1109,0x141035,0x1035,0x110a,0x141035,0x1035,0x110a,0x141035,0x1035)
USER32(NtUserCreateDCompositionHwndTarget                ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x13bc,0x0012ee,0x12ee,0x13e4,0x001311,0x1311,0x142f,0x001356,0x1356,0x1432,0x00135a,0x135a,0x142e,0x001353,0x1353,0x1403,0x001343,0x1343,0x1422,0x00135f,0x135f,0x1446,0x00137f,0x137f)
USER32(NtUserCreateDesktop                               ,  NONE,0x1154,0x1153,0x122d,0x122d,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
USER32(NtUserCreateDesktopEx                             ,  NONE,  NONE,  NONE,  NONE,  NONE,0x1161,0x1280,0x1280,0x1166,0x128f,0x128f,0x1145,0x0012ef,0x12ef,0x1147,0x001312,0x1312,0x1148,0x001357,0x1357,0x114a,0x00135b,0x135b,0x114b,0x001354,0x1354,0x1108,0x001344,0x1344,0x1109,0x001360,0x1360,0x1109,0x001380,0x1380)
USER32(NtUserCreateEmptyCursorObject                     ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1423,0x001361,0x1361,0x1447,0x001381,0x1381)
USER32(NtUserCreateLocalMemHandle                        ,  NONE,0x1156,0x1155,0x10ef,0x10ef,0x1163,0x10f0,0x10f0,0x1168,0x10eb,0x10eb,0x1143,0x1410eb,0x10eb,0x1145,0x1410ec,0x10ec,0x1146,0x1410ed,0x10ed,0x1148,0x1410ed,0x10ed,0x1149,0x1410ec,0x10ec,0x1106,0x1410e1,0x10e1,0x1107,0x1410e1,0x10e1,0x1107,0x1410e1,0x10e1)
USER32(NtUserCreatePalmRejectionDelayZone                ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x12c8,0x001383,0x1383)
USER32(NtUserCreateWindowEx                              ,  NONE,0x1157,0x1156,0x1077,0x1077,0x1164,0x1078,0x1078,0x1169,0x1076,0x1076,0x1142,0x001076,0x1076,0x1144,0x001077,0x1077,0x1145,0x001078,0x1078,0x1147,0x001078,0x1078,0x1148,0x001078,0x1078,0x1105,0x001077,0x1077,0x1106,0x001077,0x1077,0x1106,0x001077,0x1077)
USER32(NtUserCreateWindowGroup                           ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1466,0x001384,0x1384)
USER32(NtUserCreateWindowStation                         ,  NONE,0x1158,0x1157,0x122f,0x122f,0x1165,0x1282,0x1282,0x116a,0x1291,0x1291,0x1141,0x0012f1,0x12f1,0x1143,0x001314,0x1314,0x1144,0x001359,0x1359,0x1146,0x00135d,0x135d,0x1147,0x001356,0x1356,0x1104,0x001346,0x1346,0x1105,0x001363,0x1363,0x1105,0x001385,0x1385)
USER32(NtUserCtxDisplayIOCtl                             ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x1283,  NONE,  NONE,0x1292,  NONE,    NONE,0x12f2,  NONE,    NONE,0x1315,0x1290,0x00135a,0x135a,0x1291,0x00135e,0x135e,0x1297,0x001357,0x1357,0x1258,0x001347,0x1347,0x1260,0x001364,0x1364,0x1268,0x001386,0x1386)
USER32(NtUserDdeGetQualityOfService                      ,0x114f,0x1159,0x1158,0x1231,0x1231,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
USER32(NtUserDdeInitialize                               ,  NONE,0x115a,0x1159,0x1113,0x1113,0x1166,0x1114,0x1114,0x116b,0x110d,0x110d,0x113f,0x00110d,0x110d,0x1141,0x00110e,0x110e,0x1142,0x00110f,0x110f,0x1144,0x00110f,0x110f,0x1145,0x00110e,0x110e,0x1102,0x001103,0x1103,0x1103,0x001103,0x1103,0x1103,0x001103,0x1103)
USER32(NtUserDdeSetQualityOfService                      ,0x1151,0x115b,0x115a,0x1232,0x1232,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
USER32(NtUserDefSetText                                  ,  NONE,0x115d,0x115c,0x1080,0x1080,0x1168,0x1081,0x1081,0x116d,0x107f,0x107f,0x113d,0x00107f,0x107f,0x113f,0x001080,0x1080,0x1140,0x001081,0x1081,0x1142,0x001081,0x1081,0x1143,0x001081,0x1081,0x1100,0x00107f,0x107f,0x1101,0x00107f,0x107f,0x1101,0x00107f,0x107f)
USER32(NtUserDeferWindowPos                              ,0x1152,0x115c,0x115b,0x1052,0x1052,0x1167,0x1053,0x1053,0x116c,0x1053,0x1053,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
USER32(NtUserDeferWindowPosAndBand                       ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x113e,0x0012f3,0x12f3,0x1140,0x001316,0x1316,0x1141,0x00135b,0x135b,0x1143,0x00135f,0x135f,0x1144,0x001358,0x1358,0x1101,0x001348,0x1348,0x1102,0x001365,0x1365,0x1102,0x001387,0x1387)
USER32(NtUserDelegateCapturePointers                     ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x1172,0x0012f4,0x12f4,0x1174,0x001317,0x1317,0x1175,0x00135c,0x135c,0x1177,0x001360,0x1360,0x1179,0x001359,0x1359,0x1137,0x001349,0x1349,0x1138,0x001366,0x1366,0x113c,0x001388,0x1388)
USER32(NtUserDelegateInput                               ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x1173,0x0012f5,0x12f5,0x1175,0x001318,0x1318,0x1176,0x00135d,0x135d,0x1178,0x001361,0x1361,0x117a,0x00135a,0x135a,0x1138,0x00134a,0x134a,0x1139,0x001367,0x1367,0x113d,0x001389,0x1389)
USER32(NtUserDeleteMenu                                  ,0x1154,0x115e,0x115d,0x10c0,0x10c0,0x1169,0x10c1,0x10c1,0x116e,0x10bc,0x10bc,0x117c,0x0a10bc,0x10bc,0x117e,0x0a10bd,0x10bd,0x117f,0x0a10be,0x10be,0x1181,0x0a10be,0x10be,0x1183,0x0a10be,0x10be,0x1141,0x0a10b7,0x10b7,0x1142,0x0a10b7,0x10b7,0x1146,0x0a10b7,0x10b7)
USER32(NtUserDeleteWindowGroup                           ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1467,0x00138a,0x138a)
USER32(NtUserDesktopHasWatermarkText                     ,0x1170,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
USER32(NtUserDestroyAcceleratorTable                     ,  NONE,0x115f,0x115e,0x1103,0x1103,0x116a,0x1104,0x1104,0x116f,0x10fe,0x10fe,0x117b,0x0410fe,0x10fe,0x117d,0x0410ff,0x10ff,0x117e,0x041100,0x1100,0x1180,0x041100,0x1100,0x1182,0x0410ff,0x10ff,0x1140,0x0410f4,0x10f4,0x1141,0x0410f4,0x10f4,0x1145,0x0410f4,0x10f4)
USER32(NtUserDestroyCursor                               ,  NONE,0x1160,0x115f,0x109d,0x109d,0x116b,0x109e,0x109e,0x1170,0x109c,0x109c,0x117a,0x05109c,0x109c,0x117c,0x05109d,0x109d,0x117d,0x05109e,0x109e,0x117f,0x05109e,0x109e,0x1181,0x05109e,0x109e,0x113f,0x05109c,0x109c,0x1140,0x05109c,0x109c,0x1144,0x05109c,0x109c)
USER32(NtUserDestroyDCompositionHwndTarget               ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x13bb,0x0012f6,0x12f6,0x13e3,0x001319,0x1319,0x142e,0x00135e,0x135e,0x1431,0x001362,0x1362,0x142d,0x00135b,0x135b,0x1402,0x00134b,0x134b,0x1421,0x001368,0x1368,0x1445,0x00138b,0x138b)
USER32(NtUserDestroyMenu                                 ,0x1158,0x1162,0x1161,0x10e1,0x10e1,0x116d,0x10e2,0x10e2,0x1172,0x10dd,0x10dd,0x1178,0x0410dd,0x10dd,0x117a,0x0410de,0x10de,0x117b,0x0410df,0x10df,0x117d,0x0410df,0x10df,0x117f,0x0410de,0x10de,0x113d,0x0410d3,0x10d3,0x113e,0x0410d3,0x10d3,0x1142,0x0410d3,0x10d3)
USER32(NtUserDestroyPalmRejectionDelayZone               ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x12c9,0x00138d,0x138d)
USER32(NtUserDestroyWindow                               ,0x1159,0x1163,0x1162,0x109e,0x109e,0x116e,0x109f,0x109f,0x1173,0x109d,0x109d,0x1177,0x03109d,0x109d,0x1179,0x03109e,0x109e,0x117a,0x03109f,0x109f,0x117c,0x03109f,0x109f,0x117e,0x03109f,0x109f,0x113c,0x03109d,0x109d,0x113d,0x03109d,0x109d,0x1141,0x03109d,0x109d)
USER32(NtUserDeviceEventWorker                           ,  NONE,  NONE,  NONE,0x205d,  NONE,  NONE,0x2065,  NONE,  NONE,0x205e,  NONE,  NONE,0x002003,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
USER32(NtUserDisableImmersiveOwner                       ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x13ca,0x0012f8,0x12f8,0x13f2,0x00131b,0x131b,0x143d,0x001360,0x1360,0x1440,0x001364,0x1364,0x143c,0x00135d,0x135d,0x1417,0x00134d,0x134d,0x1438,0x00136a,0x136a,0x145e,0x00138e,0x138e)
USER32(NtUserDisableProcessWindowFiltering               ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x1175,0x0012f9,0x12f9,0x1177,0x00131c,0x131c,0x1178,0x001361,0x1361,0x117a,0x001365,0x1365,0x117c,0x00135e,0x135e,0x113a,0x00134e,0x134e,0x113b,0x00136b,0x136b,0x113f,0x00138f,0x138f)
USER32(NtUserDisableThreadIme                            ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x1285,  NONE,  NONE,0x1294,  NONE,    NONE,0x12fa,  NONE,    NONE,0x131d,0x1179,0x041362,0x1362,0x117b,0x041366,0x1366,0x117d,0x04135f,0x135f,0x113b,0x04134f,0x134f,0x113c,0x04136c,0x136c,0x1140,0x041390,0x1390)
USER32(NtUserDiscardPointerFrameMessages                 ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x12cf,0x0012fb,0x12fb,0x12d7,0x00131e,0x131e,0x12e2,0x001363,0x1363,0x12e3,0x001367,0x1367,0x12e9,0x001360,0x1360,0x12aa,0x001350,0x1350,0x12b3,0x00136d,0x136d,0x12bb,0x001391,0x1391)
USER32(NtUserDispatchMessage                             ,  NONE,0x1165,0x1164,0x1035,0x1035,0x1170,0x1036,0x1036,0x1175,0x1036,0x1036,0x1174,0x001037,0x1037,0x1176,0x001038,0x1038,0x1177,0x001039,0x1039,0x1179,0x001039,0x1039,0x117b,0x001039,0x1039,0x1139,0x001039,0x1039,0x113a,0x001039,0x1039,0x113e,0x001039,0x1039)
USER32(NtUserDisplayConfigGetDeviceInfo                  ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x1155,0x1295,0x1295,0x1156,0x0012fc,0x12fc,0x1158,0x00131f,0x131f,0x1159,0x001364,0x1364,0x115b,0x001368,0x1368,0x115c,0x001361,0x1361,0x1119,0x001351,0x1351,0x111a,0x00136e,0x136e,0x111a,0x001392,0x1392)
USER32(NtUserDisplayConfigSetDeviceInfo                  ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x1156,0x1296,0x1296,0x1155,0x0012fd,0x12fd,0x1157,0x001320,0x1320,0x1158,0x001365,0x1365,0x115a,0x001369,0x1369,0x115b,0x001362,0x1362,0x1118,0x001352,0x1352,0x1119,0x00136f,0x136f,0x1119,0x001393,0x1393)
USER32(NtUserDoSoundConnect                              ,  NONE,  NONE,  NONE,  NONE,  NONE,0x1171,0x1286,0x1286,0x1176,0x1297,0x1297,0x117e,0x0012fe,0x12fe,0x1180,0x001321,0x1321,0x1181,0x001366,0x1366,0x1183,0x00136a,0x136a,0x1185,0x001363,0x1363,0x1143,0x001353,0x1353,0x1144,0x001370,0x1370,0x1149,0x001394,0x1394)
USER32(NtUserDoSoundDisconnect                           ,  NONE,  NONE,  NONE,  NONE,  NONE,0x1172,0x1287,0x1287,0x1177,0x1298,0x1298,0x117d,0x0012ff,0x12ff,0x117f,0x001322,0x1322,0x1180,0x001367,0x1367,0x1182,0x00136b,0x136b,0x1184,0x001364,0x1364,0x1142,0x001354,0x1354,0x1143,0x001371,0x1371,0x1148,0x001395,0x1395)
USER32(NtUserDragDetect                                  ,0x115c,0x1166,0x1165,0x1235,0x1235,0x1173,0x1288,0x1288,0x1178,0x1299,0x1299,0x11df,0x001300,0x1300,0x11e0,0x001323,0x1323,0x11e2,0x001368,0x1368,0x11e4,0x00136c,0x136c,0x11e6,0x001365,0x1365,0x11a4,0x001355,0x1355,0x11a7,0x001372,0x1372,0x11ad,0x001396,0x1396)
USER32(NtUserDragObject                                  ,0x115d,0x1167,0x1166,0x1236,0x1236,0x1174,0x1289,0x1289,0x1179,0x129a,0x129a,0x11de,0x001301,0x1301,0x11df,0x001324,0x1324,0x11e1,0x001369,0x1369,0x11e3,0x00136d,0x136d,0x11e5,0x001366,0x1366,0x11a3,0x001356,0x1356,0x11a6,0x001373,0x1373,0x11ac,0x001397,0x1397)
USER32(NtUserDrawAnimatedRects                           ,0x115e,0x1168,0x1167,0x1237,0x1237,0x1175,0x128a,0x128a,0x117a,0x129b,0x129b,0x11dd,0x141302,0x1302,0x11de,0x141325,0x1325,0x11e0,0x14136a,0x136a,0x11e2,0x14136e,0x136e,0x11e4,0x141367,0x1367,0x11a2,0x141357,0x1357,0x11a5,0x141374,0x1374,0x11ab,0x141398,0x1398)
USER32(NtUserDrawCaption                                 ,  NONE,0x1169,0x1168,0x1238,0x1238,0x1176,0x128b,0x128b,0x117b,0x129c,0x129c,0x11dc,0x121303,0x1303,0x11dd,0x121326,0x1326,0x11df,0x12136b,0x136b,0x11e1,0x12136f,0x136f,0x11e3,0x121368,0x1368,0x11a1,0x121358,0x1358,0x11a4,0x121375,0x1375,0x11aa,0x121399,0x1399)
USER32(NtUserDrawCaptionTemp                             ,  NONE,0x116a,0x1169,0x1239,0x1239,0x1177,0x128c,0x128c,0x117c,0x129d,0x129d,0x11db,0x001304,0x1304,0x11dc,0x001327,0x1327,0x11de,0x00136c,0x136c,0x11e0,0x001370,0x1370,0x11e2,0x001369,0x1369,0x11a0,0x001359,0x1359,0x11a3,0x001376,0x1376,0x11a9,0x00139a,0x139a)
USER32(NtUserDrawIconEx                                  ,  NONE,0x116b,0x116a,0x105f,0x105f,0x1178,0x1060,0x1060,0x117d,0x1060,0x1060,0x11da,0x001060,0x1060,0x11db,0x001061,0x1061,0x11dd,0x001062,0x1062,0x11df,0x001062,0x1062,0x11e1,0x001062,0x1062,0x119f,0x001062,0x1062,0x11a2,0x001062,0x1062,0x11a8,0x001062,0x1062)
USER32(NtUserDrawMenuBarTemp                             ,  NONE,0x116c,0x116b,0x123a,0x123a,0x1179,0x128d,0x128d,0x117e,0x129e,0x129e,0x11d9,0x001305,0x1305,0x11da,0x001328,0x1328,0x11dc,0x00136d,0x136d,0x11de,0x001371,0x1371,0x11e0,0x00136a,0x136a,0x119e,0x00135a,0x135a,0x11a1,0x001377,0x1377,0x11a7,0x00139b,0x139b)
USER32(NtUserDwmGetDxRgn                                 ,  NONE,  NONE,  NONE,  NONE,  NONE,0x1276,0x128e,0x128e,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
USER32(NtUserDwmGetRemoteSessionOcclusionEvent           ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x127d,0x001306,0x1306,0x1280,0x001329,0x1329,0x1288,0x00136e,0x136e,0x1289,0x001372,0x1372,0x128f,0x00136b,0x136b,0x1251,0x00135b,0x135b,0x1259,0x001378,0x1378,0x1261,0x00139c,0x139c)
USER32(NtUserDwmGetRemoteSessionOcclusionState           ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x127e,0x001307,0x1307,0x1281,0x00132a,0x132a,0x1289,0x00136f,0x136f,0x128a,0x001373,0x1373,0x1290,0x00136c,0x136c,0x1252,0x00135c,0x135c,0x125a,0x001379,0x1379,0x1262,0x00139d,0x139d)
USER32(NtUserDwmHintDxUpdate                             ,  NONE,  NONE,  NONE,  NONE,  NONE,0x1275,0x128f,0x128f,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
USER32(NtUserDwmKernelShutdown                           ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x128a,0x001370,0x1370,0x128b,0x001374,0x1374,0x1291,0x00136d,0x136d,0x1253,0x00135d,0x135d,0x125b,0x00137a,0x137a,0x1263,0x00139e,0x139e)
USER32(NtUserDwmKernelStartup                            ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x128b,0x001371,0x1371,0x128c,0x001375,0x1375,0x1292,0x00136e,0x136e,0x1254,0x00135e,0x135e,0x125c,0x00137b,0x137b,0x1264,0x00139f,0x139f)
USER32(NtUserDwmStartRedirection                         ,  NONE,  NONE,  NONE,  NONE,  NONE,0x1273,0x1290,0x1290,0x127f,0x129f,0x129f,0x127c,0x001308,0x1308,0x127f,0x00132b,0x132b,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
USER32(NtUserDwmStopRedirection                          ,  NONE,  NONE,  NONE,  NONE,  NONE,0x1274,0x1291,0x1291,0x1280,0x12a0,0x12a0,0x127b,0x001309,0x1309,0x127e,0x00132c,0x132c,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
USER32(NtUserDwmValidateWindow                           ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x1153,0x00130a,0x130a,0x1155,0x00132d,0x132d,0x1156,0x001372,0x1372,0x1158,0x001376,0x1376,0x1159,0x00136f,0x136f,0x1116,0x00135f,0x135f,0x1117,0x00137c,0x137c,0x1117,0x0013a0,0x13a0)
USER32(NtUserEmptyClipboard                              ,0x1163,0x116d,0x116c,0x10fc,0x10fc,0x117a,0x10fd,0x10fd,0x117f,0x10f8,0x10f8,0x11d8,0x0110f8,0x10f8,0x11d9,0x0110f9,0x10f9,0x11db,0x0110fa,0x10fa,0x11dd,0x0110fa,0x10fa,0x11df,0x0110f9,0x10f9,0x119d,0x0110ee,0x10ee,0x11a0,0x0110ee,0x10ee,0x11a6,0x0110ee,0x10ee)
USER32(NtUserEnableChildWindowDpiMessage                 ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1245,0x001373,0x1373,0x1247,0x001377,0x1377,0x124b,0x001370,0x1370,0x120a,0x001360,0x1360,0x120f,0x00137d,0x137d,0x1217,0x0013a1,0x13a1)
USER32(NtUserEnableIAMAccess                             ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x13cb,0x00130b,0x130b,0x13f4,0x00132e,0x132e,0x143f,0x001374,0x1374,0x1442,0x001378,0x1378,0x143e,0x001371,0x1371,0x1419,0x001361,0x1361,0x143a,0x00137e,0x137e,0x1460,0x0013a2,0x13a2)
USER32(NtUserEnableMenuItem                              ,  NONE,0x116e,0x116d,0x10d6,0x10d6,0x117b,0x10d7,0x10d7,0x1180,0x10d2,0x10d2,0x11d7,0x0a10d2,0x10d2,0x11d8,0x0a10d3,0x10d3,0x11da,0x0a10d4,0x10d4,0x11dc,0x0a10d4,0x10d4,0x11de,0x0a10d3,0x10d3,0x119c,0x0a10c8,0x10c8,0x119f,0x0a10c8,0x10c8,0x11a5,0x0a10c8,0x10c8)
USER32(NtUserEnableMouseInPointer                        ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x13ce,0x00130c,0x130c,0x13f7,0x00132f,0x132f,0x1444,0x001375,0x1375,0x1448,0x001379,0x1379,0x1444,0x001372,0x1372,0x141f,0x001362,0x1362,0x1440,0x00137f,0x137f,0x146b,0x0013a3,0x13a3)
USER32(NtUserEnableMouseInputForCursorSuppression        ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x13d2,0x00130d,0x130d,0x13fb,0x001330,0x1330,0x1448,0x001376,0x1376,0x144c,0x00137a,0x137a,0x1448,0x001373,0x1373,0x1423,0x001363,0x1363,0x1444,0x001380,0x1380,0x146f,0x0013a4,0x13a4)
USER32(NtUserEnableNonClientDpiScaling                   ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x124d,0x001374,0x1374,0x120c,0x001364,0x1364,0x1211,0x001381,0x1381,0x1219,0x0013a5,0x13a5)
USER32(NtUserEnableResizeLayoutSynchronization           ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1408,0x001365,0x1365,0x1429,0x001382,0x1382,0x144d,0x0013a6,0x13a6)
USER32(NtUserEnableScrollBar                             ,0x1165,0x116f,0x116e,0x10bb,0x10bb,0x117c,0x10bc,0x10bc,0x1181,0x10b7,0x10b7,0x11d6,0x0c10b7,0x10b7,0x11d7,0x0c10b8,0x10b8,0x11d9,0x0c10b9,0x10b9,0x11db,0x0c10b9,0x10b9,0x11dd,0x0c10b9,0x10b9,0x119b,0x0c10b2,0x10b2,0x119e,0x0c10b2,0x10b2,0x11a4,0x0c10b2,0x10b2)
USER32(NtUserEnableSoftwareCursorForScreenCapture        ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1203,0x0013a7,0x13a7)
USER32(NtUserEnableTouchPad                              ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,0x12d3,0x001331,0x1331,0x12de,0x001377,0x1377,0x12df,0x00137b,0x137b,0x12e5,0x001375,0x1375,0x12a6,0x001366,0x1366,0x12af,0x001383,0x1383,0x12b7,0x0013a8,0x13a8)
USER32(NtUserEnableWindowGDIScaledDpiMessage             ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1210,0x001367,0x1367,0x1215,0x001384,0x1384,0x121d,0x0013a9,0x13a9)
USER32(NtUserEnableWindowGroupPolicy                     ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1469,0x0013aa,0x13aa)
USER32(NtUserEnableWindowResizeOptimization              ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1406,0x001368,0x1368,0x1427,0x001385,0x1385,0x144b,0x0013ab,0x13ab)
USER32(NtUserEndDeferWindowPosEx                         ,  NONE,0x1170,0x116f,0x1025,0x1025,0x117d,0x1026,0x1026,0x1182,0x1026,0x1026,0x11d5,0x071027,0x1027,0x11d6,0x071028,0x1028,0x11d8,0x071029,0x1029,0x11da,0x071029,0x1029,0x11dc,0x071029,0x1029,0x119a,0x071029,0x1029,0x119d,0x071029,0x1029,0x11a3,0x071029,0x1029)
USER32(NtUserEndMenu                                     ,0x1167,0x1171,0x1170,0x123b,0x123b,0x117e,0x1292,0x1292,0x1183,0x12a1,0x12a1,0x11d4,0x01130e,0x130e,0x11d5,0x011332,0x1332,0x11d7,0x011378,0x1378,0x11d9,0x01137c,0x137c,0x11db,0x011376,0x1376,0x1199,0x011369,0x1369,0x119c,0x011386,0x1386,0x11a2,0x0113ac,0x13ac)
USER32(NtUserEndPaint                                    ,0x1168,0x1172,0x1171,0x1018,0x1018,0x117f,0x1019,0x1019,0x1184,0x1019,0x1019,0x11d3,0x00101a,0x101a,0x11d4,0x00101b,0x101b,0x11d6,0x00101c,0x101c,0x11d8,0x00101c,0x101c,0x11da,0x00101c,0x101c,0x1198,0x00101c,0x101c,0x119b,0x00101c,0x101c,0x11a1,0x00101c,0x101c)
USER32(NtUserEndTask                                     ,  NONE,  NONE,  NONE,0x2058,  NONE,  NONE,0x205e,  NONE,  NONE,0x205c,  NONE,  NONE,0x002001,  NONE,  NONE,0x002002,  NONE,  NONE,0x002002,  NONE,  NONE,0x002002,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
USER32(NtUserEndTouchOperation                           ,0x1084,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x128f,0x12a2,0x12a2,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
USER32(NtUserEnumDisplayDevices                          ,  NONE,0x1173,0x1172,0x10fb,0x10fb,0x1180,0x10fc,0x10fc,0x1185,0x10f7,0x10f7,0x11d2,0x0010f7,0x10f7,0x11d3,0x0010f8,0x10f8,0x11d5,0x0010f9,0x10f9,0x11d7,0x0010f9,0x10f9,0x11d9,0x0010f8,0x10f8,0x1197,0x0010ed,0x10ed,0x119a,0x0010ed,0x10ed,0x11a0,0x0010ed,0x10ed)
USER32(NtUserEnumDisplayMonitors                         ,0x116a,0x1174,0x1173,0x1049,0x1049,0x1181,0x104a,0x104a,0x1186,0x104a,0x104a,0x11d1,0x00104b,0x104b,0x11d2,0x00104c,0x104c,0x11d4,0x00104d,0x104d,0x11d6,0x00104d,0x104d,0x11d8,0x00104d,0x104d,0x1196,0x00104d,0x104d,0x1199,0x00104d,0x104d,0x119f,0x00104d,0x104d)
USER32(NtUserEnumDisplaySettings                         ,  NONE,0x1175,0x1174,0x111a,0x111a,0x1182,0x111b,0x111b,0x1187,0x1114,0x1114,0x11d0,0x001114,0x1114,0x11d1,0x001115,0x1115,0x11d3,0x001116,0x1116,0x11d5,0x001116,0x1116,0x11d7,0x001115,0x1115,0x1195,0x00110a,0x110a,0x1198,0x00110a,0x110a,0x119e,0x00110a,0x110a)
USER32(NtUserEvent                                       ,  NONE,0x1176,0x1175,0x123c,0x123c,0x1183,0x1293,0x1293,0x1188,0x12a3,0x12a3,0x11cf,0x04130f,0x130f,0x11d0,0x041333,0x1333,0x11d2,0x041379,0x1379,0x11d4,0x04137d,0x137d,0x11d6,0x041377,0x1377,0x1194,0x04136a,0x136a,0x1197,0x041387,0x1387,0x119d,0x0413ad,0x13ad)
USER32(NtUserExcludeUpdateRgn                            ,0x116d,0x1177,0x1176,0x104f,0x104f,0x1184,0x1050,0x1050,0x1189,0x1050,0x1050,0x11ce,0x081051,0x1051,0x11cf,0x081052,0x1052,0x11d1,0x081053,0x1053,0x11d3,0x081053,0x1053,0x11d5,0x081053,0x1053,0x1193,0x081053,0x1053,0x1196,0x081053,0x1053,0x119c,0x081053,0x1053)
USER32(NtUserFillWindow                                  ,  NONE,0x1178,0x1177,0x108a,0x108a,0x1185,0x108b,0x108b,0x118a,0x1089,0x1089,0x11cd,0x171089,0x1089,0x11ce,0x17108a,0x108a,0x11d0,0x17108b,0x108b,0x11d2,0x17108b,0x108b,0x11d4,0x17108b,0x108b,0x1192,0x171089,0x1089,0x1195,0x171089,0x1089,0x119b,0x171089,0x1089)
USER32(NtUserFindExistingCursorIcon                      ,  NONE,0x1179,0x1178,0x103d,0x103d,0x1186,0x103e,0x103e,0x118b,0x103e,0x103e,0x11cc,0x00103f,0x103f,0x11cd,0x001040,0x1040,0x11cf,0x001041,0x1041,0x11d1,0x001041,0x1041,0x11d3,0x001041,0x1041,0x1191,0x001041,0x1041,0x1194,0x001041,0x1041,0x119a,0x001041,0x1041)
USER32(NtUserFindWindowEx                                ,  NONE,0x117a,0x1179,0x106e,0x106e,0x1187,0x106f,0x106f,0x118c,0x106e,0x106e,0x11cb,0x00106e,0x106e,0x11cc,0x00106f,0x106f,0x11ce,0x001070,0x1070,0x11d0,0x001070,0x1070,0x11d2,0x001070,0x1070,0x1190,0x00106f,0x106f,0x1193,0x00106f,0x106f,0x1199,0x00106f,0x106f)
USER32(NtUserFlashWindowEx                               ,0x1171,0x117b,0x117a,0x123d,0x123d,0x1188,0x1294,0x1294,0x118d,0x12a4,0x12a4,0x11ca,0x001310,0x1310,0x11cb,0x001334,0x1334,0x11cd,0x00137a,0x137a,0x11cf,0x00137e,0x137e,0x11d1,0x001378,0x1378,0x118f,0x00136b,0x136b,0x1192,0x001388,0x1388,0x1198,0x0013ae,0x13ae)
USER32(NtUserFrostCrashedWindow                          ,  NONE,  NONE,  NONE,  NONE,  NONE,0x1189,0x1295,0x1295,0x118e,0x12a5,0x12a5,0x11c9,0x001311,0x1311,0x11ca,0x001335,0x1335,0x11cc,0x00137b,0x137b,0x11ce,0x00137f,0x137f,0x11d0,0x001379,0x1379,0x118e,0x00136c,0x136c,0x1191,0x001389,0x1389,0x1197,0x0013af,0x13af)
USER32(NtUserFunctionalizeDisplayConfig                  ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x111c,0x00136d,0x136d,0x111d,0x00138a,0x138a,0x111d,0x0013b0,0x13b0)
USER32(NtUserGetActiveProcessesDpis                      ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x115b,0x00138b,0x138b,0x1160,0x0013b1,0x13b1)
USER32(NtUserGetAltTabInfo                               ,  NONE,0x117c,0x117b,0x10f7,0x10f7,0x118a,0x10f8,0x10f8,0x118f,0x10f3,0x10f3,0x11c8,0x0010f3,0x10f3,0x11c9,0x0010f4,0x10f4,0x11cb,0x0010f5,0x10f5,0x11cd,0x0010f5,0x10f5,0x11cf,0x0010f4,0x10f4,0x118d,0x0010e9,0x10e9,0x1190,0x0010e9,0x10e9,0x1196,0x0010e9,0x10e9)
USER32(NtUserGetAncestor                                 ,0x1173,0x117d,0x117c,0x10b6,0x10b6,0x118b,0x10b7,0x10b7,0x1190,0x10b2,0x10b2,0x11c7,0x0710b2,0x10b2,0x11c8,0x0710b3,0x10b3,0x11ca,0x0710b4,0x10b4,0x11cc,0x0710b4,0x10b4,0x11ce,0x0710b4,0x10b4,0x118c,0x0710ad,0x10ad,0x118f,0x0710ad,0x10ad,0x1195,0x0710ad,0x10ad)
USER32(NtUserGetAsyncKeyState                            ,  NONE,0x117f,0x117e,0x1043,0x1043,0x118d,0x1044,0x1044,0x1192,0x1044,0x1044,0x11c5,0x031045,0x1045,0x11c6,0x031046,0x1046,0x11c8,0x031047,0x1047,0x11ca,0x031047,0x1047,0x11cc,0x031047,0x1047,0x118a,0x031047,0x1047,0x118d,0x031047,0x1047,0x1193,0x031047,0x1047)
USER32(NtUserGetAtomName                                 ,  NONE,0x1180,0x117f,0x10ad,0x10ad,0x118e,0x10ae,0x10ae,0x1193,0x10a9,0x10a9,0x11c4,0x0010a9,0x10a9,0x11c5,0x0010aa,0x10aa,0x11c7,0x0010ab,0x10ab,0x11c9,0x0010ab,0x10ab,0x11cb,0x0010ab,0x10ab,0x1189,0x0010a5,0x10a5,0x118c,0x0010a5,0x10a5,0x1192,0x0010a5,0x10a5)
USER32(NtUserGetAutoRotationState                        ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x13c4,0x001313,0x1313,0x13ec,0x001337,0x1337,0x1437,0x00137d,0x137d,0x143a,0x001381,0x1381,0x1436,0x00137b,0x137b,0x1411,0x00136f,0x136f,0x1432,0x00138d,0x138d,0x1458,0x0013b3,0x13b3)
USER32(NtUserGetCIMSSM                                   ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x11b5,0x001314,0x1314,0x11b6,0x001338,0x1338,0x11b8,0x00137e,0x137e,0x11ba,0x001382,0x1382,0x11bc,0x00137c,0x137c,0x117a,0x001370,0x1370,0x117d,0x00138e,0x138e,0x1183,0x0013b4,0x13b4)
USER32(NtUserGetCPD                                      ,  NONE,0x118e,0x118d,0x1044,0x1044,0x119c,0x1045,0x1045,0x11a1,0x1045,0x1045,0x11b6,0x001046,0x1046,0x11b7,0x001047,0x1047,0x11b9,0x001048,0x1048,0x11bb,0x001048,0x1048,0x11bd,0x001048,0x1048,0x117b,0x001048,0x1048,0x117e,0x001048,0x1048,0x1184,0x001048,0x1048)
USER32(NtUserGetCaretBlinkTime                           ,0x1176,0x1181,0x1180,0x10f8,0x10f8,0x118f,0x10f9,0x10f9,0x1194,0x10f4,0x10f4,0x11c3,0x0110f4,0x10f4,0x11c4,0x0110f5,0x10f5,0x11c6,0x0110f6,0x10f6,0x11c8,0x0110f6,0x10f6,0x11ca,0x0110f5,0x10f5,0x1188,0x0110ea,0x10ea,0x118b,0x0110ea,0x10ea,0x1191,0x0110ea,0x10ea)
USER32(NtUserGetCaretPos                                 ,0x1177,0x1182,0x1181,0x123f,0x123f,0x1190,0x1297,0x1297,0x1195,0x12a7,0x12a7,0x11c2,0x041315,0x1315,0x11c3,0x041339,0x1339,0x11c5,0x04137f,0x137f,0x11c7,0x041383,0x1383,0x11c9,0x04137d,0x137d,0x1187,0x041371,0x1371,0x118a,0x04138f,0x138f,0x1190,0x0413b5,0x13b5)
USER32(NtUserGetClassInfo                                ,  NONE,0x1183,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
USER32(NtUserGetClassInfoEx                              ,  NONE,  NONE,0x1182,0x10bd,0x10bd,0x1191,0x10be,0x10be,0x1196,0x10b9,0x10b9,0x11c1,0x0010b9,0x10b9,0x11c2,0x0010ba,0x10ba,0x11c4,0x0010bb,0x10bb,0x11c6,0x0010bb,0x10bb,0x11c8,0x0010bb,0x10bb,0x1186,0x0010b4,0x10b4,0x1189,0x0010b4,0x10b4,0x118f,0x0010b4,0x10b4)
USER32(NtUserGetClassName                                ,  NONE,0x1184,0x1183,0x107c,0x107c,0x1192,0x107d,0x107d,0x1197,0x107b,0x107b,0x11c0,0x00107b,0x107b,0x11c1,0x00107c,0x107c,0x11c3,0x00107d,0x107d,0x11c5,0x00107d,0x107d,0x11c7,0x00107d,0x107d,0x1185,0x00107c,0x107c,0x1188,0x00107c,0x107c,0x118e,0x00107c,0x107c)
USER32(NtUserGetClipCursor                               ,0x117f,0x118a,0x1189,0x1240,0x1240,0x1198,0x1298,0x1298,0x119d,0x12a8,0x12a8,0x11ba,0x041316,0x1316,0x11bb,0x04133a,0x133a,0x11bd,0x041380,0x1380,0x11bf,0x041384,0x1384,0x11c1,0x04137e,0x137e,0x117f,0x041372,0x1372,0x1182,0x041390,0x1390,0x1188,0x0413b6,0x13b6)
USER32(NtUserGetClipboardAccessToken                     ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x13d6,0x001317,0x1317,0x13ff,0x00133b,0x133b,0x144c,0x001381,0x1381,0x1450,0x001385,0x1385,0x144c,0x00137f,0x137f,0x1426,0x001373,0x1373,0x1447,0x001391,0x1391,0x1472,0x0013b7,0x13b7)
USER32(NtUserGetClipboardData                            ,  NONE,0x1185,0x1184,0x10fd,0x10fd,0x1193,0x10fe,0x10fe,0x1198,0x10f9,0x10f9,0x11bf,0x0010f9,0x10f9,0x11c0,0x0010fa,0x10fa,0x11c2,0x0010fb,0x10fb,0x11c4,0x0010fb,0x10fb,0x11c6,0x0010fa,0x10fa,0x1184,0x0010ef,0x10ef,0x1187,0x0010ef,0x10ef,0x118d,0x0010ef,0x10ef)
USER32(NtUserGetClipboardFormatName                      ,  NONE,0x1186,0x1185,0x10ed,0x10ed,0x1194,0x10ee,0x10ee,0x1199,0x10e9,0x10e9,0x11be,0x0a10e9,0x10e9,0x11bf,0x0a10ea,0x10ea,0x11c1,0x0a10eb,0x10eb,0x11c3,0x0a10eb,0x10eb,0x11c5,0x0a10ea,0x10ea,0x1183,0x0a10df,0x10df,0x1186,0x0a10df,0x10df,0x118c,0x0a10df,0x10df)
USER32(NtUserGetClipboardOwner                           ,0x117c,0x1187,0x1186,0x10cd,0x10cd,0x1195,0x10ce,0x10ce,0x119a,0x10c9,0x10c9,0x11bd,0x0110c9,0x10c9,0x11be,0x0110ca,0x10ca,0x11c0,0x0110cb,0x10cb,0x11c2,0x0110cb,0x10cb,0x11c4,0x0110cb,0x10cb,0x1182,0x0110c0,0x10c0,0x1185,0x0110c0,0x10c0,0x118b,0x0110c0,0x10c0)
USER32(NtUserGetClipboardSequenceNumber                  ,0x117d,0x1188,0x1187,0x1055,0x1055,0x1196,0x1056,0x1056,0x119b,0x1056,0x1056,0x11bc,0x011056,0x1056,0x11bd,0x011057,0x1057,0x11bf,0x011058,0x1058,0x11c1,0x011058,0x1058,0x11c3,0x011058,0x1058,0x1181,0x011058,0x1058,0x1184,0x011058,0x1058,0x118a,0x011058,0x1058)
USER32(NtUserGetClipboardViewer                          ,0x117e,0x1189,0x1188,0x1241,0x1241,0x1197,0x1299,0x1299,0x119c,0x12a9,0x12a9,0x11bb,0x011318,0x1318,0x11bc,0x01133c,0x133c,0x11be,0x011382,0x1382,0x11c0,0x011386,0x1386,0x11c2,0x011380,0x1380,0x1180,0x011374,0x1374,0x1183,0x011392,0x1392,0x1189,0x0113b8,0x13b8)
USER32(NtUserGetComboBoxInfo                             ,0x1180,0x118b,0x118a,0x1242,0x1242,0x1199,0x129a,0x129a,0x119e,0x12aa,0x12aa,0x11b9,0x001319,0x1319,0x11ba,0x00133d,0x133d,0x11bc,0x001383,0x1383,0x11be,0x001387,0x1387,0x11c0,0x001381,0x1381,0x117e,0x001375,0x1375,0x1181,0x001393,0x1393,0x1187,0x0013b9,0x13b9)
USER32(NtUserGetControlBrush                             ,  NONE,0x118c,0x118b,0x107b,0x107b,0x119a,0x107c,0x107c,0x119f,0x107a,0x107a,0x11b8,0x0e107a,0x107a,0x11b9,0x0e107b,0x107b,0x11bb,0x0e107c,0x107c,0x11bd,0x0e107c,0x107c,0x11bf,0x0e107c,0x107c,0x117d,0x0e107b,0x107b,0x1180,0x0e107b,0x107b,0x1186,0x0e107b,0x107b)
USER32(NtUserGetControlColor                             ,  NONE,0x118d,0x118c,0x10e7,0x10e7,0x119b,0x10e8,0x10e8,0x11a0,0x10e3,0x10e3,0x11b7,0x1710e3,0x10e3,0x11b8,0x1710e4,0x10e4,0x11ba,0x1710e5,0x10e5,0x11bc,0x1710e5,0x10e5,0x11be,0x1710e4,0x10e4,0x117c,0x1710d9,0x10d9,0x117f,0x1710d9,0x10d9,0x1185,0x1710d9,0x10d9)
USER32(NtUserGetCurrentInputMessageSource                ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x11b4,0x00131a,0x131a,0x11b5,0x00133e,0x133e,0x11b7,0x001384,0x1384,0x11b9,0x001388,0x1388,0x11bb,0x001382,0x1382,0x1179,0x001376,0x1376,0x117c,0x001394,0x1394,0x1182,0x0013ba,0x13ba)
USER32(NtUserGetCursor                                   ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1454,0x001395,0x1395,0x147f,0x0013bb,0x13bb)
USER32(NtUserGetCursorDims                               ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,0x140a,0x00133f,0x133f,0x1457,0x001385,0x1385,0x145b,0x001389,0x1389,0x1456,0x001383,0x1383,0x1431,0x001377,0x1377,0x1453,0x001396,0x1396,0x147e,0x0013bc,0x13bc)
USER32(NtUserGetCursorFrameInfo                          ,  NONE,0x118f,0x118e,0x10f6,0x10f6,0x119d,0x10f7,0x10f7,0x11a2,0x10f2,0x10f2,0x11b3,0x1110f2,0x10f2,0x11b4,0x1110f3,0x10f3,0x11b6,0x1110f4,0x10f4,0x11b8,0x1110f4,0x10f4,0x11ba,0x1110f3,0x10f3,0x1178,0x1110e8,0x10e8,0x117b,0x1110e8,0x10e8,0x1181,0x1110e8,0x10e8)
USER32(NtUserGetCursorInfo                               ,0x1185,0x1190,0x118f,0x1243,0x1243,0x119e,0x129b,0x129b,0x11a3,0x12ab,0x12ab,0x11b2,0x00131b,0x131b,0x11b3,0x001340,0x1340,0x11b5,0x001386,0x1386,0x11b7,0x00138a,0x138a,0x11b9,0x001384,0x1384,0x1177,0x001378,0x1378,0x117a,0x001397,0x1397,0x1180,0x0013bd,0x13bd)
USER32(NtUserGetDC                                       ,0x1186,0x1191,0x1190,0x100a,0x100a,0x119f,0x100a,0x100a,0x11a4,0x100a,0x100a,0x11b1,0x03100b,0x100b,0x11b2,0x03100c,0x100c,0x11b4,0x03100d,0x100d,0x11b6,0x03100d,0x100d,0x11b8,0x03100d,0x100d,0x1176,0x03100d,0x100d,0x1179,0x03100d,0x100d,0x117f,0x03100d,0x100d)
USER32(NtUserGetDCEx                                     ,0x1187,0x1192,0x1191,0x1093,0x1093,0x11a0,0x1094,0x1094,0x11a5,0x1092,0x1092,0x11b0,0x0c1092,0x1092,0x11b1,0x0c1093,0x1093,0x11b3,0x0c1094,0x1094,0x11b5,0x0c1094,0x1094,0x11b7,0x0c1094,0x1094,0x1175,0x0c1092,0x1092,0x1178,0x0c1092,0x1092,0x117e,0x0c1092,0x1092)
USER32(NtUserGetDManipHookInitFunction                   ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x127e,0x001387,0x1387,0x127f,0x00138b,0x138b,0x1285,0x001385,0x1385,0x1247,0x001379,0x1379,0x124e,0x001398,0x1398,0x1256,0x0013be,0x13be)
USER32(NtUserGetDesktopID                                ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x11af,0x00131c,0x131c,0x11b0,0x001341,0x1341,0x11b2,0x001388,0x1388,0x11b4,0x00138c,0x138c,0x11b6,0x001386,0x1386,0x1174,0x00137a,0x137a,0x1177,0x001399,0x1399,0x117d,0x0013bf,0x13bf)
USER32(NtUserGetDisplayAutoRotationPreferences           ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x13c1,0x00131d,0x131d,0x13e9,0x001342,0x1342,0x1434,0x001389,0x1389,0x1437,0x00138d,0x138d,0x1433,0x001387,0x1387,0x140e,0x00137b,0x137b,0x142f,0x00139a,0x139a,0x1455,0x0013c0,0x13c0)
USER32(NtUserGetDisplayAutoRotationPreferencesByProcessId,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x13c0,0x00131e,0x131e,0x13e8,0x001343,0x1343,0x1433,0x00138a,0x138a,0x1436,0x00138e,0x138e,0x1432,0x001388,0x1388,0x140d,0x00137c,0x137c,0x142e,0x00139b,0x139b,0x1454,0x0013c1,0x13c1)
USER32(NtUserGetDisplayConfigBufferSizes                 ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x1152,0x12ac,0x12ac,0x1159,0x00131f,0x131f,0x115b,0x001344,0x1344,0x115c,0x00138b,0x138b,0x115e,0x00138f,0x138f,0x115f,0x001389,0x1389,0x111d,0x00137d,0x137d,0x111e,0x00139c,0x139c,0x111e,0x0013c2,0x13c2)
USER32(NtUserGetDoubleClickTime                          ,0x1188,0x1193,0x1192,0x10ba,0x10ba,0x11a1,0x10bb,0x10bb,0x11a6,0x10b6,0x10b6,0x11ae,0x0110b6,0x10b6,0x11af,0x0110b7,0x10b7,0x11b1,0x0110b8,0x10b8,0x11b3,0x0110b8,0x10b8,0x11b5,0x0110b8,0x10b8,0x1173,0x0110b1,0x10b1,0x1176,0x0110b1,0x10b1,0x117c,0x0110b1,0x10b1)
USER32(NtUserGetDpiForCurrentProcess                     ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x115c,0x00139d,0x139d,0x1161,0x0013c3,0x13c3)
USER32(NtUserGetDpiForMonitor                            ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,0x1407,0x001345,0x1345,0x1454,0x00138c,0x138c,0x1458,0x001390,0x1390,0x1453,0x00138a,0x138a,0x142e,0x00137e,0x137e,0x1450,0x00139e,0x139e,0x147b,0x0013c4,0x13c4)
USER32(NtUserGetDpiMetrics                               ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x11b0,0x00138d,0x138d,0x11b2,0x001391,0x1391,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
USER32(NtUserGetForegroundWindow                         ,0x1189,0x1194,0x1193,0x103b,0x103b,0x11a2,0x103c,0x103c,0x11a7,0x103c,0x103c,0x11ad,0x01103d,0x103d,0x11ae,0x01103e,0x103e,0x11af,0x01103f,0x103f,0x11b1,0x01103f,0x103f,0x11b4,0x01103f,0x103f,0x1172,0x01103f,0x103f,0x1175,0x01103f,0x103f,0x117b,0x01103f,0x103f)
USER32(NtUserGetGUIThreadInfo                            ,0x118b,0x1196,0x1195,0x1104,0x1104,0x11a4,0x1105,0x1105,0x11a9,0x10ff,0x10ff,0x11aa,0x0010ff,0x10ff,0x11ac,0x001100,0x1100,0x11ad,0x001101,0x1101,0x11af,0x001101,0x1101,0x11b2,0x001100,0x1100,0x1170,0x0010f5,0x10f5,0x1173,0x0010f5,0x10f5,0x1179,0x0010f5,0x10f5)
USER32(NtUserGetGestureConfig                            ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x1297,0x12ad,0x12ad,0x12d4,0x001320,0x1320,0x12dd,0x001346,0x1346,0x12e9,0x00138e,0x138e,0x12ea,0x001392,0x1392,0x12f0,0x00138b,0x138b,0x12b1,0x00137f,0x137f,0x12ba,0x00139f,0x139f,0x12c2,0x0013c5,0x13c5)
USER32(NtUserGetGestureExtArgs                           ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x1294,0x12ae,0x12ae,0x12d6,0x001321,0x1321,0x12df,0x001347,0x1347,0x12eb,0x00138f,0x138f,0x12ec,0x001393,0x1393,0x12f2,0x00138c,0x138c,0x12b3,0x001380,0x1380,0x12bc,0x0013a0,0x13a0,0x12c4,0x0013c6,0x13c6)
USER32(NtUserGetGestureInfo                              ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x1293,0x12af,0x12af,0x12d7,0x001322,0x1322,0x12e0,0x001348,0x1348,0x12ec,0x001390,0x1390,0x12ed,0x001394,0x1394,0x12f3,0x00138d,0x138d,0x12b4,0x001381,0x1381,0x12bd,0x0013a1,0x13a1,0x12c5,0x0013c7,0x13c7)
USER32(NtUserGetGlobalIMEStatus                          ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x11ac,0x001323,0x1323,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
USER32(NtUserGetGuiResources                             ,0x118a,0x1195,0x1194,0x1244,0x1244,0x11a3,0x129c,0x129c,0x11a8,0x12b0,0x12b0,0x11ab,0x071324,0x1324,0x11ad,0x071349,0x1349,0x11ae,0x071391,0x1391,0x11b0,0x071395,0x1395,0x11b3,0x07138e,0x138e,0x1171,0x071382,0x1382,0x1174,0x0713a2,0x13a2,0x117a,0x0713c8,0x13c8)
USER32(NtUserGetHDevName                                 ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x144f,0x0013a3,0x13a3,0x147a,0x0013c9,0x13c9)
USER32(NtUserGetHimetricScaleFactorFromPixelLocation     ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,0x1405,0x00134a,0x134a,0x1452,0x001392,0x1392,0x1456,0x001396,0x1396,0x1452,0x00138f,0x138f,0x142d,0x001383,0x1383,0x144e,0x0013a4,0x13a4,0x1479,0x0013ca,0x13ca)
USER32(NtUserGetIconInfo                                 ,  NONE,0x1197,0x1196,0x104e,0x104e,0x11a5,0x104f,0x104f,0x11aa,0x104f,0x104f,0x11a9,0x001050,0x1050,0x11ab,0x001051,0x1051,0x11ac,0x001052,0x1052,0x11ae,0x001052,0x1052,0x11b1,0x001052,0x1052,0x116f,0x001052,0x1052,0x1172,0x001052,0x1052,0x1178,0x001052,0x1052)
USER32(NtUserGetIconSize                                 ,  NONE,0x1198,0x1197,0x1089,0x1089,0x11a6,0x108a,0x108a,0x11ab,0x1088,0x1088,0x11a8,0x111088,0x1088,0x11aa,0x111089,0x1089,0x11ab,0x11108a,0x108a,0x11ad,0x11108a,0x108a,0x11b0,0x11108a,0x108a,0x116e,0x111088,0x1088,0x1171,0x111088,0x1088,0x1177,0x111088,0x1088)
USER32(NtUserGetImeHotKey                                ,0x118e,0x1199,0x1198,0x1245,0x1245,0x11a7,0x129d,0x129d,0x11ac,0x12b1,0x12b1,0x11a7,0x001325,0x1325,0x11a9,0x00134b,0x134b,0x11aa,0x001393,0x1393,0x11ac,0x001397,0x1397,0x11af,0x001390,0x1390,0x116d,0x001384,0x1384,0x1170,0x0013a5,0x13a5,0x1176,0x0013cb,0x13cb)
USER32(NtUserGetInputLocaleInfo                          ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x11ae,0x12b3,0x12b3,0x11a5,0x001327,0x1327,0x11a7,0x00134d,0x134d,0x11a8,0x001395,0x1395,0x11aa,0x001399,0x1399,0x11ad,0x001392,0x1392,0x116b,0x001386,0x1386,0x116e,0x0013a7,0x13a7,0x1174,0x0013cd,0x13cd)
USER32(NtUserGetInteractiveControlDeviceInfo             ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x146a,0x001393,0x1393,0x1456,0x001387,0x1387,0x147a,0x0013a8,0x13a8,0x14a5,0x0013ce,0x13ce)
USER32(NtUserGetInteractiveControlInfo                   ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1469,0x001394,0x1394,0x1455,0x001388,0x1388,0x1479,0x0013a9,0x13a9,0x14a4,0x0013cf,0x13cf)
USER32(NtUserGetInteractiveCtrlSupportedWaveforms        ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x145b,0x001389,0x1389,0x147f,0x0013aa,0x13aa,0x14aa,0x0013d0,0x13d0)
USER32(NtUserGetInternalWindowPos                        ,0x1190,0x119b,0x119a,0x1247,0x1247,0x11a9,0x129f,0x129f,0x11af,0x12b4,0x12b4,0x11a4,0x0c1328,0x1328,0x11a6,0x0c134e,0x134e,0x11a7,0x0c1396,0x1396,0x11a9,0x0c139a,0x139a,0x11ac,0x0c1395,0x1395,0x116a,0x0c138a,0x138a,0x116d,0x0c13ab,0x13ab,0x1173,0x0c13d1,0x13d1)
USER32(NtUserGetKeyNameText                              ,  NONE,0x119f,0x119e,0x1248,0x1248,0x11ad,0x12a0,0x12a0,0x11b3,0x12b5,0x12b5,0x11a0,0x0a1329,0x1329,0x11a2,0x0a134f,0x134f,0x11a3,0x0a1397,0x1397,0x11a5,0x0a139b,0x139b,0x11a8,0x0a1396,0x1396,0x1166,0x0a138b,0x138b,0x1169,0x0a13ac,0x13ac,0x116f,0x0a13d2,0x13d2)
USER32(NtUserGetKeyState                                 ,  NONE,0x11a0,0x119f,0x1003,0x1003,0x11ae,0x1003,0x1003,0x11b4,0x1003,0x1003,0x119f,0x031004,0x1004,0x11a1,0x031005,0x1005,0x11a2,0x031006,0x1006,0x11a4,0x031006,0x1006,0x11a7,0x031006,0x1006,0x1165,0x031006,0x1006,0x1168,0x031006,0x1006,0x116e,0x031006,0x1006)
USER32(NtUserGetKeyboardLayoutList                       ,0x1191,0x119c,0x119b,0x1058,0x1058,0x11aa,0x1059,0x1059,0x11b0,0x1059,0x1059,0x11a3,0x001059,0x1059,0x11a5,0x00105a,0x105a,0x11a6,0x00105b,0x105b,0x11a8,0x00105b,0x105b,0x11ab,0x00105b,0x105b,0x1169,0x00105b,0x105b,0x116c,0x00105b,0x105b,0x1172,0x00105b,0x105b)
USER32(NtUserGetKeyboardLayoutName                       ,  NONE,0x119d,0x119c,0x1249,0x1249,0x11ab,0x12a1,0x12a1,0x11b1,0x12b6,0x12b6,0x11a2,0x00132a,0x132a,0x11a4,0x001350,0x1350,0x11a5,0x001398,0x1398,0x11a7,0x00139c,0x139c,0x11aa,0x001397,0x1397,0x1168,0x00138c,0x138c,0x116b,0x0013ad,0x13ad,0x1171,0x0013d3,0x13d3)
USER32(NtUserGetKeyboardState                            ,0x1193,0x119e,0x119d,0x1079,0x1079,0x11ac,0x107a,0x107a,0x11b2,0x1078,0x1078,0x11a1,0x041078,0x1078,0x11a3,0x041079,0x1079,0x11a4,0x04107a,0x107a,0x11a6,0x04107a,0x107a,0x11a9,0x04107a,0x107a,0x1167,0x041079,0x1079,0x116a,0x041079,0x1079,0x1170,0x041079,0x1079)
USER32(NtUserGetLayeredWindowAttributes                  ,  NONE,0x1244,0x1240,0x124a,0x124a,0x125c,0x12a2,0x12a2,0x1269,0x12b7,0x12b7,0x1295,0x14132b,0x132b,0x1298,0x141351,0x1351,0x12a2,0x141399,0x1399,0x12a3,0x14139d,0x139d,0x12a9,0x141398,0x1398,0x126a,0x14138d,0x138d,0x1273,0x1413ae,0x13ae,0x127b,0x1413d4,0x13d4)
USER32(NtUserGetListBoxInfo                              ,0x1196,0x11a1,0x11a0,0x124b,0x124b,0x11af,0x12a3,0x12a3,0x11b5,0x12b8,0x12b8,0x119e,0x03132c,0x132c,0x11a0,0x031352,0x1352,0x11a1,0x03139a,0x139a,0x11a3,0x03139e,0x139e,0x11a6,0x031399,0x1399,0x1164,0x03138e,0x138e,0x1167,0x0313af,0x13af,0x116d,0x0313d5,0x13d5)
USER32(NtUserGetMenuBarInfo                              ,0x1197,0x11a2,0x11a1,0x10c5,0x10c5,0x11b0,0x10c6,0x10c6,0x11b6,0x10c1,0x10c1,0x119d,0x0010c1,0x10c1,0x119f,0x0010c2,0x10c2,0x11a0,0x0010c3,0x10c3,0x11a2,0x0010c3,0x10c3,0x11a5,0x0010c3,0x10c3,0x1163,0x0010bc,0x10bc,0x1166,0x0010bc,0x10bc,0x116c,0x0010bc,0x10bc)
USER32(NtUserGetMenuIndex                                ,  NONE,0x11a3,0x11a2,0x124c,0x124c,0x11b1,0x12a4,0x12a4,0x11b7,0x12b9,0x12b9,0x119c,0x05132d,0x132d,0x119e,0x051353,0x1353,0x119f,0x05139b,0x139b,0x11a1,0x05139f,0x139f,0x11a4,0x05139a,0x139a,0x1162,0x05138f,0x138f,0x1165,0x0513b0,0x13b0,0x116b,0x0513d6,0x13d6)
USER32(NtUserGetMenuItemRect                             ,0x1199,0x11a4,0x11a3,0x124d,0x124d,0x11b2,0x12a5,0x12a5,0x11b8,0x12ba,0x12ba,0x119b,0x14132e,0x132e,0x119d,0x141354,0x1354,0x119e,0x14139c,0x139c,0x11a0,0x1413a0,0x13a0,0x11a3,0x14139b,0x139b,0x1161,0x141390,0x1390,0x1164,0x1413b1,0x13b1,0x116a,0x1413d7,0x13d7)
USER32(NtUserGetMessage                                  ,  NONE,0x11a5,0x11a4,0x1006,0x1006,0x11b3,0x1006,0x1006,0x11b9,0x1006,0x1006,0x119a,0x001007,0x1007,0x119c,0x001008,0x1008,0x119d,0x001009,0x1009,0x119f,0x001009,0x1009,0x11a2,0x001009,0x1009,0x1160,0x001009,0x1009,0x1163,0x001009,0x1009,0x1169,0x001009,0x1009)
USER32(NtUserGetMonitorBrightness                        ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1120,0x0013d8,0x13d8)
USER32(NtUserGetMouseMovePointsEx                        ,0x119b,0x11a6,0x11a5,0x124e,0x124e,0x11b4,0x12a6,0x12a6,0x11ba,0x12bb,0x12bb,0x1199,0x00132f,0x132f,0x119b,0x001355,0x1355,0x119c,0x00139d,0x139d,0x119e,0x0013a1,0x13a1,0x11a1,0x00139c,0x139c,0x115f,0x001391,0x1391,0x1162,0x0013b2,0x13b2,0x1168,0x0013d9,0x13d9)
USER32(NtUserGetObjectInformation                        ,0x119c,0x11a7,0x11a6,0x106b,0x106b,0x11b5,0x106c,0x106c,0x11bb,0x106c,0x106c,0x1198,0x00106c,0x106c,0x119a,0x00106d,0x106d,0x119b,0x00106e,0x106e,0x119d,0x00106e,0x106e,0x11a0,0x00106e,0x106e,0x115e,0x00106d,0x106d,0x1161,0x00106d,0x106d,0x1167,0x00106d,0x106d)
USER32(NtUserGetOemBitmapSize                            ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1135,0x0013da,0x13da)
USER32(NtUserGetOpenClipboardWindow                      ,0x119d,0x11a8,0x11a7,0x10dc,0x10dc,0x11b6,0x10dd,0x10dd,0x11bc,0x10d8,0x10d8,0x1197,0x0110d8,0x10d8,0x1199,0x0110d9,0x10d9,0x119a,0x0110da,0x10da,0x119c,0x0110da,0x10da,0x119f,0x0110d9,0x10d9,0x115d,0x0110ce,0x10ce,0x1160,0x0110ce,0x10ce,0x1166,0x0110ce,0x10ce)
USER32(NtUserGetOwnerTransformedMonitorRect              ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,0x140b,0x001356,0x1356,0x1000,0x001000,0x1000,0x1000,0x001000,0x1000,0x1000,0x001000,0x1000,0x1000,0x001000,0x1000,0x1000,0x001000,0x1000,0x1000,0x001000,0x1000)
USER32(NtUserGetPhysicalDeviceRect                       ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,0x12ca,0x001357,0x1357,0x12d5,0x00139e,0x139e,0x12d6,0x0013a2,0x13a2,0x12dc,0x00139d,0x139d,0x129d,0x001392,0x1392,0x12a6,0x0013b3,0x13b3,0x12ae,0x0013db,0x13db)
USER32(NtUserGetPointerCursorId                          ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x12d2,0x001330,0x1330,0x12db,0x001358,0x1358,0x12e7,0x00139f,0x139f,0x12e8,0x0013a3,0x13a3,0x12ee,0x00139e,0x139e,0x12af,0x001393,0x1393,0x12b8,0x0013b4,0x13b4,0x12c0,0x0013dc,0x13dc)
USER32(NtUserGetPointerDevice                            ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x12cc,0x001331,0x1331,0x12d1,0x001359,0x1359,0x12dc,0x0013a0,0x13a0,0x12dd,0x0013a4,0x13a4,0x12e3,0x00139f,0x139f,0x12a4,0x001394,0x1394,0x12ad,0x0013b5,0x13b5,0x12b5,0x0013dd,0x13dd)
USER32(NtUserGetPointerDeviceCursors                     ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x12c8,0x001332,0x1332,0x12cd,0x00135a,0x135a,0x12d8,0x0013a1,0x13a1,0x12d9,0x0013a5,0x13a5,0x12df,0x0013a0,0x13a0,0x12a0,0x001395,0x1395,0x12a9,0x0013b6,0x13b6,0x12b1,0x0013de,0x13de)
USER32(NtUserGetPointerDeviceProperties                  ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x12cb,0x001333,0x1333,0x12d0,0x00135b,0x135b,0x12db,0x0013a2,0x13a2,0x12dc,0x0013a6,0x13a6,0x12e2,0x0013a1,0x13a1,0x12a3,0x001396,0x1396,0x12ac,0x0013b7,0x13b7,0x12b4,0x0013df,0x13df)
USER32(NtUserGetPointerDeviceRects                       ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x12c9,0x001334,0x1334,0x12ce,0x00135c,0x135c,0x12d9,0x0013a3,0x13a3,0x12da,0x0013a7,0x13a7,0x12e0,0x0013a2,0x13a2,0x12a1,0x001397,0x1397,0x12aa,0x0013b8,0x13b8,0x12b2,0x0013e0,0x13e0)
USER32(NtUserGetPointerDevices                           ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x12cd,0x001335,0x1335,0x12d2,0x00135d,0x135d,0x12dd,0x0013a4,0x13a4,0x12de,0x0013a8,0x13a8,0x12e4,0x0013a3,0x13a3,0x12a5,0x001398,0x1398,0x12ae,0x0013b9,0x13b9,0x12b6,0x0013e1,0x13e1)
USER32(NtUserGetPointerFrameArrivalTimes                 ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x12e4,0x0013a5,0x13a5,0x12e5,0x0013a9,0x13a9,0x12eb,0x0013a4,0x13a4,0x12ac,0x001399,0x1399,0x12b5,0x0013ba,0x13ba,0x12bd,0x0013e2,0x13e2)
USER32(NtUserGetPointerInfoList                          ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x12d1,0x001336,0x1336,0x12da,0x00135e,0x135e,0x12e6,0x0013a6,0x13a6,0x12e7,0x0013aa,0x13aa,0x12ed,0x0013a5,0x13a5,0x12ae,0x00139a,0x139a,0x12b7,0x0013bb,0x13bb,0x12bf,0x0013e3,0x13e3)
USER32(NtUserGetPointerInputTransform                    ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,0x12d9,0x00135f,0x135f,0x12e5,0x0013a7,0x13a7,0x12e6,0x0013ab,0x13ab,0x12ec,0x0013a6,0x13a6,0x12ad,0x00139b,0x139b,0x12b6,0x0013bc,0x13bc,0x12be,0x0013e4,0x13e4)
USER32(NtUserGetPointerType                              ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x12d3,0x001337,0x1337,0x12dc,0x001360,0x1360,0x12e8,0x0013a8,0x13a8,0x12e9,0x0013ac,0x13ac,0x12ef,0x0013a7,0x13a7,0x12b0,0x00139c,0x139c,0x12b9,0x0013bd,0x13bd,0x12c1,0x0013e5,0x13e5)
USER32(NtUserGetPrecisionTouchPadConfiguration           ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,0x12d4,0x001361,0x1361,0x12df,0x0013a9,0x13a9,0x12e0,0x0013ad,0x13ad,0x12e6,0x0013a8,0x13a8,0x12a7,0x00139d,0x139d,0x12b0,0x0013be,0x13be,0x12b8,0x0013e6,0x13e6)
USER32(NtUserGetPriorityClipboardFormat                  ,0x119e,0x11a9,0x11a8,0x124f,0x124f,0x11b7,0x12a7,0x12a7,0x11bd,0x12bc,0x12bc,0x1196,0x091338,0x1338,0x1198,0x091362,0x1362,0x1199,0x0913aa,0x13aa,0x119b,0x0913ae,0x13ae,0x119e,0x0913a9,0x13a9,0x115c,0x09139e,0x139e,0x115f,0x0913bf,0x13bf,0x1165,0x0913e7,0x13e7)
USER32(NtUserGetProcessDpiAwareness                      ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,0x1406,0x001363,0x1363,0x1453,0x0013ab,0x13ab,0x1457,0x0013af,0x13af,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
USER32(NtUserGetProcessDpiAwarenessContext               ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x119c,0x0013aa,0x13aa,0x115a,0x00139f,0x139f,0x115d,0x0013c0,0x13c0,0x1162,0x0013e8,0x13e8)
USER32(NtUserGetProcessUIContextInformation              ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x13cc,0x001339,0x1339,0x13f5,0x001364,0x1364,0x1440,0x0013ac,0x13ac,0x1443,0x0013b0,0x13b0,0x143f,0x0013ab,0x13ab,0x141a,0x0013a0,0x13a0,0x143b,0x0013c1,0x13c1,0x1461,0x0013e9,0x13e9)
USER32(NtUserGetProcessWindowStation                     ,0x119f,0x11aa,0x11a9,0x1021,0x1021,0x11b8,0x1022,0x1022,0x11be,0x1022,0x1022,0x1195,0x011023,0x1023,0x1197,0x011024,0x1024,0x1198,0x011025,0x1025,0x119a,0x011025,0x1025,0x119d,0x011025,0x1025,0x115b,0x011025,0x1025,0x115e,0x011025,0x1025,0x1164,0x011025,0x1025)
USER32(NtUserGetProp                                     ,  NONE,  NONE,  NONE,  NONE,  NONE,0x1228,0x100e,0x100e,0x1232,0x100e,0x100e,0x124c,0x00100f,0x100f,0x124f,0x001010,0x1010,0x1255,0x001011,0x1011,0x1256,0x001011,0x1011,0x125c,0x001011,0x1011,0x121e,0x001011,0x1011,0x1224,0x001011,0x1011,0x122c,0x001011,0x1011)
USER32(NtUserGetQueueEventStatus                         ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x13d7,0x00133a,0x133a,0x1400,0x001365,0x1365,0x144d,0x0013ad,0x13ad,0x1451,0x0013b1,0x13b1,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
USER32(NtUserGetQueueStatusReadonly                      ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x144d,0x0013ac,0x13ac,0x1427,0x0013a1,0x13a1,0x1448,0x0013c2,0x13c2,0x1473,0x0013ea,0x13ea)
USER32(NtUserGetRawInputBuffer                           ,  NONE,0x11ab,0x11aa,0x1250,0x1250,0x11b9,0x12a8,0x12a8,0x11bf,0x12bd,0x12bd,0x1194,0x00133b,0x133b,0x1196,0x001366,0x1366,0x1197,0x0013ae,0x13ae,0x1199,0x0013b2,0x13b2,0x119b,0x0013ad,0x13ad,0x1159,0x0013a2,0x13a2,0x115a,0x0013c3,0x13c3,0x115f,0x0013eb,0x13eb)
USER32(NtUserGetRawInputData                             ,  NONE,0x11ac,0x11ab,0x1251,0x1251,0x11ba,0x12a9,0x12a9,0x11c0,0x12be,0x12be,0x1193,0x00133c,0x133c,0x1195,0x001367,0x1367,0x1196,0x0013af,0x13af,0x1198,0x0013b3,0x13b3,0x119a,0x0013ae,0x13ae,0x1158,0x0013a3,0x13a3,0x1159,0x0013c4,0x13c4,0x115e,0x0013ec,0x13ec)
USER32(NtUserGetRawInputDeviceInfo                       ,  NONE,0x11ad,0x11ac,0x1252,0x1252,0x11bb,0x12aa,0x12aa,0x11c1,0x12bf,0x12bf,0x1192,0x00133d,0x133d,0x1194,0x001368,0x1368,0x1195,0x0013b0,0x13b0,0x1197,0x0013b4,0x13b4,0x1199,0x0013af,0x13af,0x1157,0x0013a4,0x13a4,0x1158,0x0013c5,0x13c5,0x115d,0x0013ed,0x13ed)
USER32(NtUserGetRawInputDeviceList                       ,  NONE,0x11ae,0x11ad,0x1253,0x1253,0x11bc,0x12ab,0x12ab,0x11c2,0x12c0,0x12c0,0x1191,0x00133e,0x133e,0x1193,0x001369,0x1369,0x1194,0x0013b1,0x13b1,0x1196,0x0013b5,0x13b5,0x1198,0x0013b0,0x13b0,0x1156,0x0013a5,0x13a5,0x1157,0x0013c6,0x13c6,0x115c,0x0013ee,0x13ee)
USER32(NtUserGetRawPointerDeviceData                     ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x12c7,0x00133f,0x133f,0x12cc,0x00136a,0x136a,0x12d7,0x0013b2,0x13b2,0x12d8,0x0013b6,0x13b6,0x12de,0x0013b1,0x13b1,0x129f,0x0013a6,0x13a6,0x12a8,0x0013c7,0x13c7,0x12b0,0x0013ef,0x13ef)
USER32(NtUserGetRegisteredRawInputDevices                ,  NONE,0x11af,0x11ae,0x1254,0x1254,0x11bd,0x12ac,0x12ac,0x11c3,0x12c1,0x12c1,0x1190,0x001340,0x1340,0x1192,0x00136b,0x136b,0x1193,0x0013b3,0x13b3,0x1195,0x0013b7,0x13b7,0x1197,0x0013b2,0x13b2,0x1155,0x0013a7,0x13a7,0x1156,0x0013c8,0x13c8,0x115b,0x0013f0,0x13f0)
USER32(NtUserGetResizeDCompositionSynchronizationObject  ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1407,0x0013a8,0x13a8,0x1428,0x0013c9,0x13c9,0x144c,0x0013f1,0x13f1)
USER32(NtUserGetScrollBarInfo                            ,0x11a0,0x11b0,0x11af,0x1094,0x1094,0x11be,0x1095,0x1095,0x11c4,0x1093,0x1093,0x118f,0x0c1093,0x1093,0x1191,0x0c1094,0x1094,0x1192,0x0c1095,0x1095,0x1194,0x0c1095,0x1095,0x1196,0x0c1095,0x1095,0x1154,0x0c1093,0x1093,0x1155,0x0c1093,0x1093,0x115a,0x0c1093,0x1093)
USER32(NtUserGetSystemDpiForProcess                      ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1163,0x0013f2,0x13f2)
USER32(NtUserGetSystemMenu                               ,0x11a1,0x11b1,0x11b0,0x1060,0x1060,0x11bf,0x1061,0x1061,0x11c5,0x1061,0x1061,0x118e,0x071061,0x1061,0x1190,0x071062,0x1062,0x1191,0x071063,0x1063,0x1193,0x071063,0x1063,0x1195,0x071063,0x1063,0x1153,0x071063,0x1063,0x1154,0x071063,0x1063,0x1159,0x071063,0x1063)
/* i#487: this has a different sysnum on some platforms.  It must be before NtUserGetThreadDesktop (i#1418). */
USER32(GetThreadDesktop                                  ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
USER32(NtUserGetThreadDesktop                            ,  NONE,0x11b2,0x11b1,0x1084,0x1084,0x11c0,0x1085,0x1085,0x11c6,0x1083,0x1083,0x118d,0x051083,0x1083,0x118f,0x051084,0x1084,0x1190,0x051085,0x1085,0x1192,0x051085,0x1085,0x1194,0x051085,0x1085,0x1152,0x051083,0x1083,0x1153,0x051083,0x1083,0x1158,0x051083,0x1083)
USER32(NtUserGetThreadState                              ,0x11a3,0x11b3,0x11b2,0x1000,0x1000,0x11c1,0x1000,0x1000,0x11c7,0x1000,0x1000,0x118c,0x041001,0x1001,0x118e,0x041002,0x1002,0x118f,0x041003,0x1003,0x1191,0x041003,0x1003,0x1193,0x041003,0x1003,0x1151,0x041003,0x1003,0x1152,0x041003,0x1003,0x1157,0x041003,0x1003)
USER32(NtUserGetTitleBarInfo                             ,0x11a4,0x11b4,0x11b3,0x1090,0x1090,0x11c2,0x1091,0x1091,0x11c8,0x108f,0x108f,0x118b,0x07108f,0x108f,0x118d,0x071090,0x1090,0x118e,0x071091,0x1091,0x1190,0x071091,0x1091,0x1192,0x071091,0x1091,0x1150,0x07108f,0x108f,0x1151,0x07108f,0x108f,0x1156,0x07108f,0x108f)
USER32(NtUserGetTopLevelWindow                           ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x11c9,0x12c2,0x12c2,0x118a,0x001341,0x1341,0x118c,0x00136c,0x136c,0x118d,0x0013b4,0x13b4,0x118f,0x0013b8,0x13b8,0x1191,0x0013b3,0x13b3,0x114f,0x0013a9,0x13a9,0x1150,0x0013ca,0x13ca,0x1155,0x0013f3,0x13f3)
USER32(NtUserGetTouchInputInfo                           ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x1290,0x12c3,0x12c3,0x12a0,0x001342,0x1342,0x12a3,0x00136d,0x136d,0x12ad,0x0013b5,0x13b5,0x12ae,0x0013b9,0x13b9,0x12b4,0x0013b4,0x13b4,0x1275,0x0013aa,0x13aa,0x127e,0x0013cb,0x13cb,0x1286,0x0013f4,0x13f4)
USER32(NtUserGetTouchValidationStatus                    ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x12c1,0x001343,0x1343,0x12c4,0x00136e,0x136e,0x12cf,0x0013b6,0x13b6,0x12d0,0x0013ba,0x13ba,0x12d6,0x0013b5,0x13b5,0x1297,0x0013ab,0x13ab,0x12a0,0x0013cc,0x13cc,0x12a8,0x0013f5,0x13f5)
USER32(NtUserGetUpdateRect                               ,  NONE,0x11b5,0x11b4,0x1053,0x1053,0x11c4,0x1054,0x1054,0x11cb,0x1054,0x1054,0x1188,0x0c1054,0x1054,0x118a,0x0c1055,0x1055,0x118b,0x0c1056,0x1056,0x118d,0x0c1056,0x1056,0x118f,0x0c1056,0x1056,0x114d,0x0c1056,0x1056,0x114e,0x0c1056,0x1056,0x1153,0x0c1056,0x1056)
USER32(NtUserGetUpdateRgn                                ,  NONE,0x11b6,0x11b5,0x1087,0x1087,0x11c5,0x1088,0x1088,0x11cc,0x1086,0x1086,0x1187,0x0c1086,0x1086,0x1189,0x0c1087,0x1087,0x118a,0x0c1088,0x1088,0x118c,0x0c1088,0x1088,0x118e,0x0c1088,0x1088,0x114c,0x0c1086,0x1086,0x114d,0x0c1086,0x1086,0x1152,0x0c1086,0x1086)
USER32(NtUserGetUpdatedClipboardFormats                  ,  NONE,  NONE,  NONE,  NONE,  NONE,0x11c3,0x12ad,0x12ad,0x11ca,0x12c4,0x12c4,0x1189,0x001344,0x1344,0x118b,0x00136f,0x136f,0x118c,0x0013b7,0x13b7,0x118e,0x0013bb,0x13bb,0x1190,0x0013b6,0x13b6,0x114e,0x0013ac,0x13ac,0x114f,0x0013cd,0x13cd,0x1154,0x0013f6,0x13f6)
USER32(NtUserGetWOWClass                                 ,  NONE,0x11b9,0x11b8,0x1255,0x1255,0x11c8,0x12ae,0x12ae,0x11d2,0x12c5,0x12c5,0x1180,0x001345,0x1345,0x1182,0x001370,0x1370,0x1183,0x0013b8,0x13b8,0x1185,0x0013bc,0x13bc,0x1187,0x0013b7,0x13b7,0x1145,0x0013ad,0x13ad,0x1146,0x0013ce,0x13ce,0x114b,0x0013f7,0x13f7)
USER32(NtUserGetWindowBand                               ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x1186,0x001346,0x1346,0x1188,0x001371,0x1371,0x1189,0x0013b9,0x13b9,0x118b,0x0013bd,0x13bd,0x118d,0x0013b8,0x13b8,0x114b,0x0013ae,0x13ae,0x114c,0x0013cf,0x13cf,0x1151,0x0013f8,0x13f8)
USER32(NtUserGetWindowCompositionAttribute               ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x11ce,0x12c6,0x12c6,0x1184,0x001347,0x1347,0x1186,0x001372,0x1372,0x1187,0x0013ba,0x13ba,0x1189,0x0013be,0x13be,0x118b,0x0013b9,0x13b9,0x1149,0x0013af,0x13af,0x114a,0x0013d0,0x13d0,0x114f,0x0013f9,0x13f9)
USER32(NtUserGetWindowCompositionInfo                    ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x11cd,0x12c7,0x12c7,0x1185,0x001348,0x1348,0x1187,0x001373,0x1373,0x1188,0x0013bb,0x13bb,0x118a,0x0013bf,0x13bf,0x118c,0x0013ba,0x13ba,0x114a,0x0013b0,0x13b0,0x114b,0x0013d1,0x13d1,0x1150,0x0013fa,0x13fa)
USER32(NtUserGetWindowDC                                 ,0x11a7,0x11b7,0x11b6,0x1063,0x1063,0x11c6,0x1064,0x1064,0x11cf,0x1064,0x1064,0x1183,0x031064,0x1064,0x1185,0x031065,0x1065,0x1186,0x031066,0x1066,0x1188,0x031066,0x1066,0x118a,0x031066,0x1066,0x1148,0x031066,0x1066,0x1149,0x031066,0x1066,0x114e,0x031066,0x1066)
USER32(NtUserGetWindowDisplayAffinity                    ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x11d0,0x12c8,0x12c8,0x1182,0x001349,0x1349,0x1184,0x001374,0x1374,0x1185,0x0013bc,0x13bc,0x1187,0x0013c0,0x13c0,0x1189,0x0013bb,0x13bb,0x1147,0x0013b1,0x13b1,0x1148,0x0013d2,0x13d2,0x114d,0x0013fb,0x13fb)
USER32(NtUserGetWindowFeedbackSetting                    ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x12bf,0x00134a,0x134a,0x12c2,0x001375,0x1375,0x12cd,0x0013bd,0x13bd,0x12ce,0x0013c1,0x13c1,0x12d4,0x0013bc,0x13bc,0x1295,0x0013b2,0x13b2,0x129e,0x0013d3,0x13d3,0x12a6,0x0013fc,0x13fc)
USER32(NtUserGetWindowGroupId                            ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1468,0x0013fd,0x13fd)
USER32(NtUserGetWindowMinimizeRect                       ,  NONE,  NONE,  NONE,  NONE,  NONE,0x1277,0x12af,0x12af,0x1281,0x12c9,0x12c9,0x127a,0x00134b,0x134b,0x127d,0x001376,0x1376,0x1287,0x0013be,0x13be,0x1288,0x0013c2,0x13c2,0x128e,0x0013bd,0x13bd,0x1250,0x0013b3,0x13b3,0x1258,0x0013d4,0x13d4,0x1260,0x0013fe,0x13fe)
USER32(NtUserGetWindowPlacement                          ,0x11a8,0x11b8,0x11b7,0x10d9,0x10d9,0x11c7,0x10da,0x10da,0x11d1,0x10d5,0x10d5,0x1181,0x0710d5,0x10d5,0x1183,0x0710d6,0x10d6,0x1184,0x0710d7,0x10d7,0x1186,0x0710d7,0x10d7,0x1188,0x0710d6,0x10d6,0x1146,0x0710cb,0x10cb,0x1147,0x0710cb,0x10cb,0x114c,0x0710cb,0x10cb)
USER32(NtUserGetWindowProcessHandle                      ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x14c7,0x0013ff,0x13ff)
USER32(NtUserGetWindowRgnEx                              ,  NONE,  NONE,  NONE,  NONE,  NONE,0x123a,0x12b0,0x12b0,0x1246,0x12ca,0x12ca,0x1232,0x00134c,0x134c,0x1235,0x001377,0x1377,0x1238,0x0013bf,0x13bf,0x123a,0x0013c3,0x13c3,0x123e,0x0013be,0x13be,0x11fd,0x0013b4,0x13b4,0x1202,0x0013d5,0x13d5,0x120a,0x001400,0x1400)
USER32(NtUserGhostWindowFromHungWindow                   ,  NONE,  NONE,  NONE,  NONE,  NONE,0x11c9,0x12b1,0x12b1,0x11d3,0x12cb,0x12cb,0x117f,0x00134d,0x134d,0x1181,0x001378,0x1378,0x1182,0x0013c0,0x13c0,0x1184,0x0013c4,0x13c4,0x1186,0x0013bf,0x13bf,0x1144,0x0013b5,0x13b5,0x1145,0x0013d6,0x13d6,0x114a,0x001401,0x1401)
USER32(NtUserHandleDelegatedInput                        ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x11e0,0x00134e,0x134e,0x11e1,0x001379,0x1379,0x11e3,0x0013c1,0x13c1,0x11e5,0x0013c5,0x13c5,0x11e7,0x0013c0,0x13c0,0x11a5,0x0013b6,0x13b6,0x11a8,0x0013d7,0x13d7,0x11ae,0x001402,0x1402)
USER32(NtUserHardErrorControl                            ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x12b2,  NONE,  NONE,0x12cc,  NONE,    NONE,0x134f,  NONE,    NONE,0x137a,0x1218,0x0013c2,0x13c2,0x121a,0x0013c6,0x13c6,0x121d,0x0013c1,0x13c1,0x11dc,0x0013b7,0x13b7,0x11df,0x0013d8,0x13d8,0x11e6,0x001403,0x1403)
USER32(NtUserHideCaret                                   ,0x11ab,0x11bb,0x11ba,0x101e,0x101e,0x11cb,0x101f,0x101f,0x11d5,0x101f,0x101f,0x1212,0x031020,0x1020,0x1215,0x031021,0x1021,0x1217,0x031022,0x1022,0x1219,0x031022,0x1022,0x121c,0x031022,0x1022,0x11db,0x031022,0x1022,0x11de,0x031022,0x1022,0x11e5,0x031022,0x1022)
USER32(NtUserHidePointerContactVisualization             ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x12c0,0x001350,0x1350,0x12c3,0x00137b,0x137b,0x12ce,0x0013c3,0x13c3,0x12cf,0x0013c7,0x13c7,0x12d5,0x0013c2,0x13c2,0x1296,0x0013b8,0x13b8,0x129f,0x0013d9,0x13d9,0x12a7,0x001404,0x1404)
USER32(NtUserHiliteMenuItem                              ,0x11ac,0x11bc,0x11bb,0x1257,0x1257,0x11cc,0x12b3,0x12b3,0x11d6,0x12cd,0x12cd,0x1211,0x141351,0x1351,0x1214,0x14137c,0x137c,0x1216,0x1413c4,0x13c4,0x1218,0x1413c8,0x13c8,0x121b,0x1413c3,0x13c3,0x11da,0x1413b9,0x13b9,0x11dd,0x1413da,0x13da,0x11e4,0x141405,0x1405)
USER32(NtUserHungWindowFromGhostWindow                   ,  NONE,  NONE,  NONE,  NONE,  NONE,0x11cd,0x12b4,0x12b4,0x11d7,0x12ce,0x12ce,0x1210,0x001352,0x1352,0x1213,0x00137d,0x137d,0x1215,0x0013c5,0x13c5,0x1217,0x0013c9,0x13c9,0x121a,0x0013c4,0x13c4,0x11d9,0x0013ba,0x13ba,0x11dc,0x0013db,0x13db,0x11e3,0x001406,0x1406)
USER32(NtUserHwndQueryRedirectionInfo                    ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x1337,0x12cf,0x12cf,0x137f,0x001353,0x1353,0x1391,0x00137e,0x137e,0x13ca,0x0013c6,0x13c6,0x13cc,0x0013ca,0x13ca,0x13db,0x0013c5,0x13c5,0x13ad,0x0013bb,0x13bb,0x13c8,0x0013dc,0x13dc,0x13da,0x001407,0x1407)
USER32(NtUserHwndSetRedirectionInfo                      ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x1338,0x12d0,0x12d0,0x137e,0x001354,0x1354,0x1390,0x00137f,0x137f,0x13c9,0x0013c7,0x13c7,0x13cb,0x0013cb,0x13cb,0x13da,0x0013c6,0x13c6,0x13ac,0x0013bc,0x13bc,0x13c7,0x0013dd,0x13dd,0x13d9,0x001408,0x1408)
USER32(NtUserImpersonateDdeClientWindow                  ,0x11ad,0x11bd,0x11bc,0x1258,0x1258,0x11ce,0x12b5,0x12b5,0x11d8,0x12d1,0x12d1,0x120f,0x001355,0x1355,0x1212,0x001380,0x1380,0x1214,0x0013c8,0x13c8,0x1216,0x0013cc,0x13cc,0x1219,0x0013c7,0x13c7,0x11d8,0x0013bd,0x13bd,0x11db,0x0013de,0x13de,0x11e2,0x001409,0x1409)
USER32(NtUserInheritWindowMonitor                        ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x11e8,0x0013c8,0x13c8,0x11a6,0x0013be,0x13be,0x11a9,0x0013df,0x13df,0x11af,0x00140a,0x140a)
USER32(NtUserInitTask                                    ,  NONE,0x11c0,0x11bf,0x1259,0x1259,0x11d1,0x12b6,0x12b6,0x11db,0x12d2,0x12d2,0x120c,0x001356,0x1356,0x120f,0x001381,0x1381,0x1211,0x0013c9,0x13c9,0x1213,0x0013cd,0x13cd,0x1216,0x0013c9,0x13c9,0x11d5,0x0013bf,0x13bf,0x11d8,0x0013e0,0x13e0,0x11df,0x00140b,0x140b)
USER32(NtUserInitialize                                  ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x12b7,  NONE,  NONE,0x12d3,  NONE,    NONE,0x1357,  NONE,    NONE,0x1382,0x1213,0x0013ca,0x13ca,0x1215,0x0013ce,0x13ce,0x1218,0x0013ca,0x13ca,0x11d7,0x0013c0,0x13c0,0x11da,0x0013e1,0x13e1,0x11e1,0x00140c,0x140c)
USER32(NtUserInitializeClientPfnArrays                   ,  NONE,0x11bf,0x11be,0x125b,0x125b,0x11d0,0x12b8,0x12b8,0x11da,0x12d4,0x12d4,0x120d,0x001358,0x1358,0x1210,0x001383,0x1383,0x1212,0x0013cb,0x13cb,0x1214,0x0013cf,0x13cf,0x1217,0x0013cb,0x13cb,0x11d6,0x0013c1,0x13c1,0x11d9,0x0013e2,0x13e2,0x11e0,0x00140d,0x140d)
USER32(NtUserInitializeGenericHidInjection               ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1433,0x0013c2,0x13c2,0x1456,0x0013e3,0x13e3,0x1481,0x00140e,0x140e)
USER32(NtUserInitializeInputDeviceInjection              ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1458,0x0013cc,0x13cc,0x145c,0x0013d0,0x13d0,0x1457,0x0013cc,0x13cc,0x1432,0x0013c3,0x13c3,0x1455,0x0013e4,0x13e4,0x1480,0x00140f,0x140f)
USER32(NtUserInitializePointerDeviceInjection            ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1459,0x0013cd,0x13cd,0x145d,0x0013d1,0x13d1,0x1458,0x0013cd,0x13cd,0x1434,0x0013c4,0x13c4,0x1457,0x0013e5,0x13e5,0x1482,0x001410,0x1410)
USER32(NtUserInitializePointerDeviceInjectionEx          ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x143c,0x0013c5,0x13c5,0x145f,0x0013e6,0x13e6,0x148a,0x001411,0x1411)
USER32(NtUserInitializeTouchInjection                    ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x12c2,0x001359,0x1359,0x12c5,0x001384,0x1384,0x12d0,0x0013ce,0x13ce,0x12d1,0x0013d2,0x13d2,0x12d7,0x0013ce,0x13ce,0x1298,0x0013c6,0x13c6,0x12a1,0x0013e7,0x13e7,0x12a9,0x001412,0x1412)
USER32(NtUserInjectDeviceInput                           ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x145b,0x0013cf,0x13cf,0x1460,0x0013d3,0x13d3,0x145b,0x0013cf,0x13cf,0x1437,0x0013c7,0x13c7,0x145a,0x0013e8,0x13e8,0x1485,0x001413,0x1413)
USER32(NtUserInjectGenericHidInput                       ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x143b,0x0013c8,0x13c8,0x145e,0x0013e9,0x13e9,0x1489,0x001414,0x1414)
USER32(NtUserInjectGesture                               ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x1292,0x12d5,0x12d5,0x12d8,0x00135a,0x135a,0x12e1,0x001385,0x1385,0x12ed,0x0013d0,0x13d0,0x12ee,0x0013d4,0x13d4,0x12f4,0x0013d0,0x13d0,0x12b5,0x0013c9,0x13c9,0x12be,0x0013ea,0x13ea,0x12c6,0x001415,0x1415)
USER32(NtUserInjectKeyboardInput                         ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x145d,0x0013d1,0x13d1,0x1462,0x0013d5,0x13d5,0x145d,0x0013d1,0x13d1,0x1439,0x0013ca,0x13ca,0x145c,0x0013eb,0x13eb,0x1487,0x001416,0x1416)
USER32(NtUserInjectMouseInput                            ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x145c,0x0013d2,0x13d2,0x1461,0x0013d6,0x13d6,0x145c,0x0013d2,0x13d2,0x1438,0x0013cb,0x13cb,0x145b,0x0013ec,0x13ec,0x1486,0x001417,0x1417)
USER32(NtUserInjectPointerInput                          ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x145e,0x0013d3,0x13d3,0x1463,0x0013d7,0x13d7,0x145e,0x0013d3,0x13d3,0x143a,0x0013cc,0x13cc,0x145d,0x0013ed,0x13ed,0x1488,0x001418,0x1418)
USER32(NtUserInjectTouchInput                            ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x12c3,0x00135b,0x135b,0x12c6,0x001386,0x1386,0x12d1,0x0013d4,0x13d4,0x12d2,0x0013d8,0x13d8,0x12d8,0x0013d4,0x13d4,0x1299,0x0013cd,0x13cd,0x12a2,0x0013ee,0x13ee,0x12aa,0x001419,0x1419)
USER32(NtUserInteractiveControlQueryUsage                ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x146d,0x0013d5,0x13d5,0x1459,0x0013ce,0x13ce,0x147d,0x0013ef,0x13ef,0x14a8,0x00141a,0x141a)
USER32(NtUserInternalClipCursor                          ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x13d4,0x00135c,0x135c,0x13fd,0x001387,0x1387,0x144a,0x0013d5,0x13d5,0x144e,0x0013d9,0x13d9,0x144a,0x0013d6,0x13d6,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
USER32(NtUserInternalGetWindowIcon                       ,  NONE,  NONE,  NONE,  NONE,  NONE,0x11d3,0x12b9,0x12b9,0x11dd,0x12d6,0x12d6,0x120a,0x00135d,0x135d,0x120d,0x001388,0x1388,0x120f,0x0013d6,0x13d6,0x1211,0x0013da,0x13da,0x1214,0x0013d7,0x13d7,0x11d3,0x0013cf,0x13cf,0x11d6,0x0013f0,0x13f0,0x11dd,0x00141b,0x141b)
USER32(NtUserInternalGetWindowText                       ,  NONE,0x11c1,0x11c0,0x1062,0x1062,0x11d2,0x1063,0x1063,0x11dc,0x1063,0x1063,0x120b,0x0c1063,0x1063,0x120e,0x0c1064,0x1064,0x1210,0x0c1065,0x1065,0x1212,0x0c1065,0x1065,0x1215,0x0c1065,0x1065,0x11d4,0x0c1065,0x1065,0x11d7,0x0c1065,0x1065,0x11de,0x0c1065,0x1065)
USER32(NtUserInvalidateRect                              ,0x11b2,0x11c2,0x11c1,0x1004,0x1004,0x11d4,0x1004,0x1004,0x11de,0x1004,0x1004,0x1209,0x0c1005,0x1005,0x120c,0x0c1006,0x1006,0x120e,0x0c1007,0x1007,0x1210,0x0c1007,0x1007,0x1213,0x0c1007,0x1007,0x11d2,0x0c1007,0x1007,0x11d5,0x0c1007,0x1007,0x11dc,0x0c1007,0x1007)
USER32(NtUserInvalidateRgn                               ,0x11b3,0x11c3,0x11c2,0x10cc,0x10cc,0x11d5,0x10cd,0x10cd,0x11df,0x10c8,0x10c8,0x1208,0x0c10c8,0x10c8,0x120b,0x0c10c9,0x10c9,0x120d,0x0c10ca,0x10ca,0x120f,0x0c10ca,0x10ca,0x1212,0x0c10ca,0x10ca,0x11d1,0x0c10bf,0x10bf,0x11d4,0x0c10bf,0x10bf,0x11db,0x0c10bf,0x10bf)
USER32(NtUserIsChildWindowDpiMessageEnabled              ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1246,0x0013d7,0x13d7,0x1248,0x0013db,0x13db,0x124c,0x0013d8,0x13d8,0x120b,0x0013d0,0x13d0,0x1210,0x0013f1,0x13f1,0x1218,0x00141c,0x141c)
USER32(NtUserIsClipboardFormatAvailable                  ,0x11b4,0x11c4,0x11c3,0x102e,0x102e,0x11d6,0x102f,0x102f,0x11e0,0x102f,0x102f,0x1207,0x041030,0x1030,0x120a,0x041031,0x1031,0x120c,0x041032,0x1032,0x120e,0x041032,0x1032,0x1211,0x041032,0x1032,0x11d0,0x041032,0x1032,0x11d3,0x041032,0x1032,0x11da,0x041032,0x1032)
USER32(NtUserIsMouseInPointerEnabled                     ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x13cf,0x00135e,0x135e,0x13f8,0x001389,0x1389,0x1445,0x0013d8,0x13d8,0x1449,0x0013dc,0x13dc,0x1445,0x0013d9,0x13d9,0x1420,0x0013d1,0x13d1,0x1441,0x0013f2,0x13f2,0x146c,0x00141d,0x141d)
USER32(NtUserIsMouseInputEnabled                         ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x13d3,0x00135f,0x135f,0x13fc,0x00138a,0x138a,0x1449,0x0013d9,0x13d9,0x144d,0x0013dd,0x13dd,0x1449,0x0013da,0x13da,0x1424,0x0013d2,0x13d2,0x1445,0x0013f3,0x13f3,0x1470,0x00141e,0x141e)
USER32(NtUserIsNonClientDpiScalingEnabled                ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x124e,0x0013db,0x13db,0x120d,0x0013d3,0x13d3,0x1212,0x0013f4,0x13f4,0x121a,0x00141f,0x141f)
USER32(NtUserIsResizeLayoutSynchronizationEnabled        ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x140a,0x0013d4,0x13d4,0x142b,0x0013f5,0x13f5,0x144f,0x001420,0x1420)
USER32(NtUserIsTopLevelWindow                            ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x11e1,0x12d7,0x12d7,0x1206,0x001360,0x1360,0x1209,0x00138b,0x138b,0x120b,0x0013da,0x13da,0x120d,0x0013de,0x13de,0x1210,0x0013dc,0x13dc,0x11cf,0x0013d5,0x13d5,0x11d2,0x0013f6,0x13f6,0x11d9,0x001421,0x1421)
USER32(NtUserIsTouchWindow                               ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x128d,0x12d8,0x12d8,0x12a1,0x001361,0x1361,0x12a4,0x00138c,0x138c,0x12ae,0x0013db,0x13db,0x12af,0x0013df,0x13df,0x12b5,0x0013dd,0x13dd,0x1276,0x0013d6,0x13d6,0x127f,0x0013f7,0x13f7,0x1287,0x001422,0x1422)
USER32(NtUserIsWindowBroadcastingDpiToChildren           ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1247,0x0013dc,0x13dc,0x1249,0x0013e0,0x13e0,0x124f,0x0013de,0x13de,0x120f,0x0013d7,0x13d7,0x1214,0x0013f8,0x13f8,0x121c,0x001423,0x1423)
USER32(NtUserIsWindowGDIScaledDpiMessageEnabled          ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1211,0x0013d8,0x13d8,0x1216,0x0013f9,0x13f9,0x121e,0x001424,0x1424)
USER32(NtUserKillTimer                                   ,0x11b5,0x11c5,0x11c4,0x101a,0x101a,0x11d7,0x101b,0x101b,0x11e2,0x101b,0x101b,0x1205,0x07101c,0x101c,0x1208,0x07101d,0x101d,0x120a,0x07101e,0x101e,0x120c,0x07101e,0x101e,0x120f,0x07101e,0x101e,0x11ce,0x07101e,0x101e,0x11d1,0x07101e,0x101e,0x11d8,0x07101e,0x101e)
USER32(NtUserLayoutCompleted                             ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x11ff,0x001362,0x1362,0x1201,0x00138d,0x138d,0x1203,0x0013dd,0x13dd,0x1205,0x0013e1,0x13e1,0x1208,0x0013df,0x13df,0x11c6,0x0013d9,0x13d9,0x11c9,0x0013fa,0x13fa,0x11d0,0x001425,0x1425)
USER32(NtUserLinkDpiCursor                               ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,0x1409,0x00138e,0x138e,0x1456,0x0013de,0x13de,0x145a,0x0013e2,0x13e2,0x1455,0x0013e0,0x13e0,0x1430,0x0013da,0x13da,0x1452,0x0013fb,0x13fb,0x147d,0x001426,0x1426)
USER32(NtUserLoadKeyboardLayoutEx                        ,  NONE,0x11c6,0x11c5,0x125c,0x125c,0x11d8,0x12ba,0x12ba,0x11e3,0x12d9,0x12d9,0x1204,0x001363,0x1363,0x1207,0x00138f,0x138f,0x1209,0x0013df,0x13df,0x120b,0x0013e3,0x13e3,0x120e,0x0013e1,0x13e1,0x11cd,0x0013db,0x13db,0x11d0,0x0013fc,0x13fc,0x11d7,0x001427,0x1427)
USER32(NtUserLockCursor                                  ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x11c9,0x0013dc,0x13dc,0x11cc,0x0013fd,0x13fd,0x11d3,0x001428,0x1428)
USER32(NtUserLockWindowStation                           ,0x11b7,0x11c7,0x11c6,0x125d,0x125d,0x11d9,0x12bb,0x12bb,0x11e4,0x12da,0x12da,0x1203,0x041364,0x1364,0x1206,0x041390,0x1390,0x1208,0x0413e0,0x13e0,0x120a,0x0413e4,0x13e4,0x120d,0x0413e2,0x13e2,0x11cc,0x0413dd,0x13dd,0x11cf,0x0413fe,0x13fe,0x11d6,0x041429,0x1429)
USER32(NtUserLockWindowUpdate                            ,0x11b8,0x11c8,0x11c7,0x110c,0x110c,0x11da,0x110d,0x110d,0x11e5,0x1106,0x1106,0x1202,0x031106,0x1106,0x1205,0x031107,0x1107,0x1207,0x031108,0x1108,0x1209,0x031108,0x1108,0x120c,0x031107,0x1107,0x11cb,0x0310fc,0x10fc,0x11ce,0x0310fc,0x10fc,0x11d5,0x0310fc,0x10fc)
USER32(NtUserLockWorkStation                             ,0x11b9,0x11c9,0x11c8,0x125e,0x125e,0x11db,0x12bc,0x12bc,0x11e6,0x12db,0x12db,0x1201,0x011365,0x1365,0x1204,0x011391,0x1391,0x1206,0x0113e1,0x13e1,0x1208,0x0113e5,0x13e5,0x120b,0x0113e3,0x13e3,0x11ca,0x0113de,0x13de,0x11cd,0x0113ff,0x13ff,0x11d4,0x01142a,0x142a)
USER32(NtUserLogicalToPerMonitorDPIPhysicalPoint         ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,0x1202,0x001392,0x1392,0x1204,0x0013e2,0x13e2,0x1206,0x0013e6,0x13e6,0x1209,0x0013e4,0x13e4,0x11c7,0x0013df,0x13df,0x11ca,0x001400,0x1400,0x11d1,0x00142b,0x142b)
USER32(NtUserLogicalToPhysicalDpiPointForWindow          ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1451,0x00142c,0x142c)
USER32(NtUserLogicalToPhysicalPoint                      ,  NONE,  NONE,  NONE,  NONE,  NONE,0x11dc,0x12bd,0x12bd,0x11e7,0x12dc,0x12dc,0x1200,0x001366,0x1366,0x1203,0x001393,0x1393,0x1205,0x0013e3,0x13e3,0x1207,0x0013e7,0x13e7,0x120a,0x0013e5,0x13e5,0x11c8,0x0013e0,0x13e0,0x11cb,0x001401,0x1401,0x11d2,0x00142d,0x142d)
USER32(NtUserLogon                                       ,  NONE,  NONE,  NONE,0x2059,  NONE,  NONE,0x205f,  NONE,  NONE,0x205d,  NONE,  NONE,0x002002,  NONE,  NONE,0x002003,  NONE,  NONE,0x002003,  NONE,  NONE,0x002003,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
USER32(NtUserMNDragLeave                                 ,  NONE,0x11ce,0x11cd,0x125f,0x125f,0x11e1,0x12be,0x12be,0x11ec,0x12dd,0x12dd,0x11fa,0x011367,0x1367,0x11fc,0x011394,0x1394,0x11fe,0x0113e4,0x13e4,0x1200,0x0113e8,0x13e8,0x1203,0x0113e6,0x13e6,0x11c1,0x0113e1,0x13e1,0x11c4,0x011402,0x1402,0x11cb,0x01142e,0x142e)
USER32(NtUserMNDragOver                                  ,  NONE,0x11cf,0x11ce,0x1260,0x1260,0x11e2,0x12bf,0x12bf,0x11ed,0x12de,0x12de,0x11f9,0x001368,0x1368,0x11fb,0x001395,0x1395,0x11fd,0x0013e5,0x13e5,0x11ff,0x0013e9,0x13e9,0x1202,0x0013e7,0x13e7,0x11c0,0x0013e2,0x13e2,0x11c3,0x001403,0x1403,0x11ca,0x00142f,0x142f)
USER32(NtUserMagControl                                  ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x1334,0x12df,0x12df,0x137c,0x001369,0x1369,0x138e,0x001396,0x1396,0x13c7,0x0013e6,0x13e6,0x13c9,0x0013ea,0x13ea,0x13d8,0x0013e8,0x13e8,0x13aa,0x0013e3,0x13e3,0x13c5,0x001404,0x1404,0x13d7,0x001430,0x1430)
USER32(NtUserMagGetContextInformation                    ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x1336,0x12e0,0x12e0,0x137a,0x00136a,0x136a,0x138c,0x001397,0x1397,0x13c5,0x0013e7,0x13e7,0x13c7,0x0013eb,0x13eb,0x13d6,0x0013e9,0x13e9,0x13a8,0x0013e4,0x13e4,0x13c3,0x001405,0x1405,0x13d5,0x001431,0x1431)
USER32(NtUserMagSetContextInformation                    ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x1335,0x12e1,0x12e1,0x137b,0x00136b,0x136b,0x138d,0x001398,0x1398,0x13c6,0x0013e8,0x13e8,0x13c8,0x0013ec,0x13ec,0x13d7,0x0013ea,0x13ea,0x13a9,0x0013e5,0x13e5,0x13c4,0x001406,0x1406,0x13d6,0x001432,0x1432)
USER32(NtUserManageGestureHandlerWindow                  ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x1295,0x12e2,0x12e2,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
USER32(NtUserMapVirtualKeyEx                             ,  NONE,0x11ca,0x11c9,0x105a,0x105a,0x11dd,0x105b,0x105b,0x11e8,0x105b,0x105b,0x11fe,0x00105b,0x105b,0x1200,0x00105c,0x105c,0x1202,0x00105d,0x105d,0x1204,0x00105d,0x105d,0x1207,0x00105d,0x105d,0x11c5,0x00105d,0x105d,0x11c8,0x00105d,0x105d,0x11cf,0x00105d,0x105d)
USER32(NtUserMenuItemFromPoint                           ,0x11bb,0x11cb,0x11ca,0x1261,0x1261,0x11de,0x12c0,0x12c0,0x11e9,0x12e3,0x12e3,0x11fd,0x00136c,0x136c,0x11ff,0x001399,0x1399,0x1201,0x0013e9,0x13e9,0x1203,0x0013ed,0x13ed,0x1206,0x0013eb,0x13eb,0x11c4,0x0013e6,0x13e6,0x11c7,0x001407,0x1407,0x11ce,0x001433,0x1433)
USER32(NtUserMessageCall                                 ,  NONE,0x11cc,0x11cb,0x1007,0x1007,0x11df,0x1007,0x1007,0x11ea,0x1007,0x1007,0x11fc,0x001008,0x1008,0x11fe,0x001009,0x1009,0x1200,0x00100a,0x100a,0x1202,0x00100a,0x100a,0x1205,0x00100a,0x100a,0x11c3,0x00100a,0x100a,0x11c6,0x00100a,0x100a,0x11cd,0x00100a,0x100a)
USER32(NtUserMinMaximize                                 ,  NONE,0x11cd,0x11cc,0x1262,0x1262,0x11e0,0x12c1,0x12c1,0x11eb,0x12e4,0x12e4,0x11fb,0x0c136d,0x136d,0x11fd,0x0c139a,0x139a,0x11ff,0x0c13ea,0x13ea,0x1201,0x0c13ee,0x13ee,0x1204,0x0c13ec,0x13ec,0x11c2,0x0c13e7,0x13e7,0x11c5,0x0c1408,0x1408,0x11cc,0x0c1434,0x1434)
USER32(NtUserModifyUserStartupInfoFlags                  ,  NONE,0x11d0,0x11cf,0x1114,0x1114,0x11e3,0x1115,0x1115,0x11ee,0x110e,0x110e,0x11f8,0x05110e,0x110e,0x11fa,0x05110f,0x110f,0x11fc,0x051110,0x1110,0x11fe,0x051110,0x1110,0x1201,0x05110f,0x110f,0x11bf,0x051104,0x1104,0x11c2,0x051104,0x1104,0x11c9,0x051104,0x1104)
USER32(NtUserModifyWindowTouchCapability                 ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x128c,0x12e5,0x12e5,0x12a2,0x00136e,0x136e,0x12a5,0x00139b,0x139b,0x12af,0x0013eb,0x13eb,0x12b0,0x0013ef,0x13ef,0x12b6,0x0013ed,0x13ed,0x1277,0x0013e8,0x13e8,0x1280,0x001409,0x1409,0x1288,0x001435,0x1435)
USER32(NtUserMoveWindow                                  ,0x11c1,0x11d1,0x11d0,0x105d,0x105d,0x11e4,0x105e,0x105e,0x11ef,0x105e,0x105e,0x11f7,0x00105e,0x105e,0x11f9,0x00105f,0x105f,0x11fb,0x001060,0x1060,0x11fd,0x001060,0x1060,0x1200,0x001060,0x1060,0x11be,0x001060,0x1060,0x11c1,0x001060,0x1060,0x11c8,0x001060,0x1060)
USER32(NtUserMsgWaitForMultipleObjectsEx                 ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1495,0x00140a,0x140a,0x14c6,0x001436,0x1436)
USER32(NtUserNavigateFocus                               ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1466,0x0013ec,0x13ec,0x146b,0x0013f0,0x13f0,0x1466,0x0013ee,0x13ee,0x1452,0x0013e9,0x13e9,0x1476,0x00140b,0x140b,0x14a1,0x001437,0x1437)
USER32(NtUserNotifyIMEStatus                             ,0x11c2,0x11d2,0x11d1,0x1263,0x1263,0x11e5,0x12c2,0x12c2,0x11f0,0x12e6,0x12e6,0x11f6,0x0c136f,0x136f,0x11f8,0x0c139c,0x139c,0x11fa,0x0c13ed,0x13ed,0x11fc,0x0c13f1,0x13f1,0x11ff,0x0c13ef,0x13ef,0x11bd,0x0c13ea,0x13ea,0x11c0,0x0c140c,0x140c,0x11c7,0x0c1438,0x1438)
USER32(NtUserNotifyProcessCreate                         ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x108f,  NONE,  NONE,0x108d,  NONE,    NONE,0x108d,  NONE,    NONE,0x108e,0x11f9,0x00108f,0x108f,0x11fb,0x00108f,0x108f,0x11fe,0x00108f,0x108f,0x11bc,0x00108d,0x108d,0x11bf,0x00108d,0x108d,0x11c6,0x00108d,0x108d)
USER32(NtUserNotifyWinEvent                              ,  NONE,0x11d4,0x11d3,0x102c,0x102c,0x11e7,0x102d,0x102d,0x11f2,0x102d,0x102d,0x11f4,0x16102e,0x102e,0x11f6,0x16102f,0x102f,0x11f8,0x161030,0x1030,0x11fa,0x161030,0x1030,0x11fd,0x161030,0x1030,0x11bb,0x161030,0x1030,0x11be,0x161030,0x1030,0x11c5,0x161030,0x1030)
USER32(NtUserOpenClipboard                               ,  NONE,0x11d5,0x11d4,0x10d3,0x10d3,0x11e8,0x10d4,0x10d4,0x11f3,0x10cf,0x10cf,0x11f3,0x0710cf,0x10cf,0x11f5,0x0710d0,0x10d0,0x11f7,0x0710d1,0x10d1,0x11f9,0x0710d1,0x10d1,0x11fc,0x0710d1,0x10d1,0x11ba,0x0710c6,0x10c6,0x11bd,0x0710c6,0x10c6,0x11c4,0x0710c6,0x10c6)
USER32(NtUserOpenDesktop                                 ,  NONE,0x11d6,0x11d5,0x10ab,0x10ab,0x11e9,0x10ac,0x10ac,0x11f4,0x10a7,0x10a7,0x11f2,0x0010a7,0x10a7,0x11f4,0x0010a8,0x10a8,0x11f6,0x0010a9,0x10a9,0x11f8,0x0010a9,0x10a9,0x11fb,0x0010a9,0x10a9,0x11b9,0x0010a3,0x10a3,0x11bc,0x0010a3,0x10a3,0x11c3,0x0010a3,0x10a3)
USER32(NtUserOpenInputDesktop                            ,0x11c7,0x11d7,0x11d6,0x1264,0x1264,0x11ea,0x12c3,0x12c3,0x11f5,0x12e7,0x12e7,0x11f1,0x0a1370,0x1370,0x11f3,0x0a139d,0x139d,0x11f5,0x0a13ee,0x13ee,0x11f7,0x0a13f2,0x13f2,0x11fa,0x0a13f0,0x13f0,0x11b8,0x0a13eb,0x13eb,0x11bb,0x0a140d,0x140d,0x11c2,0x0a1439,0x1439)
USER32(NtUserOpenThreadDesktop                           ,  NONE,  NONE,  NONE,  NONE,  NONE,0x11eb,0x2063,0x12c4,0x11f6,0x12e8,0x12e8,0x11f0,0x001371,0x1371,0x11f2,0x00139e,0x139e,0x11f4,0x0013ef,0x13ef,0x11f6,0x0013f3,0x13f3,0x11f9,0x0013f1,0x13f1,0x11b7,0x0013ec,0x13ec,0x11ba,0x00140e,0x140e,0x11c1,0x00143a,0x143a)
USER32(NtUserOpenWindowStation                           ,  NONE,0x11d8,0x11d7,0x10a1,0x10a1,0x11ec,0x10a2,0x10a2,0x11f7,0x10a0,0x10a0,0x11ef,0x0010a0,0x10a0,0x11f1,0x0010a1,0x10a1,0x11f3,0x0010a2,0x10a2,0x11f5,0x0010a2,0x10a2,0x11f8,0x0010a2,0x10a2,0x11b6,0x0010a0,0x10a0,0x11b9,0x0010a0,0x10a0,0x11c0,0x0010a0,0x10a0)
USER32(NtUserPaintDesktop                                ,0x11c9,0x11d9,0x11d8,0x111b,0x111b,0x11ed,0x111c,0x111c,0x11f8,0x1115,0x1115,0x11ee,0x001115,0x1115,0x11f0,0x001116,0x1116,0x11f2,0x001117,0x1117,0x11f4,0x001117,0x1117,0x11f7,0x001116,0x1116,0x11b5,0x00110b,0x110b,0x11b8,0x00110b,0x110b,0x11bf,0x00110b,0x110b)
USER32(NtUserPaintMenuBar                                ,  NONE,0x1237,0x1233,0x10f2,0x10f2,0x124f,0x10f3,0x10f3,0x125c,0x10ee,0x10ee,0x1218,0x0010ee,0x10ee,0x121b,0x0010ef,0x10ef,0x121e,0x0010f0,0x10f0,0x1220,0x0010f0,0x10f0,0x1223,0x0010ef,0x10ef,0x11e2,0x0010e4,0x10e4,0x11e6,0x0010e4,0x10e4,0x11ed,0x0010e4,0x10e4)
USER32(NtUserPaintMonitor                                ,  NONE,  NONE,  NONE,  NONE,  NONE,0x11ee,0x12c5,0x12c5,0x11f9,0x12e9,0x12e9,0x11ed,0x001372,0x1372,0x11ef,0x00139f,0x139f,0x11f1,0x0013f0,0x13f0,0x11f3,0x0013f4,0x13f4,0x11f6,0x0013f2,0x13f2,0x11b4,0x0013ed,0x13ed,0x11b7,0x00140f,0x140f,0x11be,0x00143b,0x143b)
USER32(NtUserPeekMessage                                 ,  NONE,0x11da,0x11d9,0x1001,0x1001,0x11ef,0x1001,0x1001,0x11fa,0x1001,0x1001,0x11ec,0x001002,0x1002,0x11ee,0x001003,0x1003,0x11f0,0x001004,0x1004,0x11f2,0x001004,0x1004,0x11f5,0x001004,0x1004,0x11b3,0x001004,0x1004,0x11b6,0x001004,0x1004,0x11bd,0x001004,0x1004)
USER32(NtUserPerMonitorDPIPhysicalToLogicalPoint         ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,0x11e8,0x0013a0,0x13a0,0x11ea,0x0013f1,0x13f1,0x11ec,0x0013f5,0x13f5,0x11ef,0x0013f3,0x13f3,0x11ad,0x0013ee,0x13ee,0x11b0,0x001410,0x1410,0x11b7,0x00143c,0x143c)
USER32(NtUserPhysicalToLogicalDpiPointForWindow          ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1452,0x00143d,0x143d)
USER32(NtUserPhysicalToLogicalPoint                      ,  NONE,  NONE,  NONE,  NONE,  NONE,0x11f0,0x12c6,0x12c6,0x11fb,0x12ea,0x12ea,0x11eb,0x001373,0x1373,0x11ed,0x0013a1,0x13a1,0x11ef,0x0013f2,0x13f2,0x11f1,0x0013f6,0x13f6,0x11f4,0x0013f4,0x13f4,0x11b2,0x0013ef,0x13ef,0x11b5,0x001411,0x1411,0x11bc,0x00143e,0x143e)
USER32(NtUserPostMessage                                 ,  NONE,0x11db,0x11da,0x100e,0x100e,0x11f1,0x100f,0x100f,0x11fc,0x100f,0x100f,0x11ea,0x001010,0x1010,0x11ec,0x001011,0x1011,0x11ee,0x001012,0x1012,0x11f0,0x001012,0x1012,0x11f3,0x001012,0x1012,0x11b1,0x001012,0x1012,0x11b4,0x001012,0x1012,0x11bb,0x001012,0x1012)
USER32(NtUserPostThreadMessage                           ,  NONE,0x11dc,0x11db,0x105e,0x105e,0x11f2,0x105f,0x105f,0x11fd,0x105f,0x105f,0x11e9,0x00105f,0x105f,0x11eb,0x001060,0x1060,0x11ed,0x001061,0x1061,0x11ef,0x001061,0x1061,0x11f2,0x001061,0x1061,0x11b0,0x001061,0x1061,0x11b3,0x001061,0x1061,0x11ba,0x001061,0x1061)
USER32(NtUserPrintWindow                                 ,  NONE,0x11dd,0x11dc,0x1265,0x1265,0x11f3,0x12c7,0x12c7,0x11fe,0x12eb,0x12eb,0x11e8,0x0e1374,0x1374,0x11ea,0x0e13a2,0x13a2,0x11ec,0x0e13f3,0x13f3,0x11ee,0x0e13f7,0x13f7,0x11f1,0x0e13f5,0x13f5,0x11af,0x0e13f0,0x13f0,0x11b2,0x0e1412,0x1412,0x11b9,0x0e143f,0x143f)
USER32(NtUserProcessConnect                              ,  NONE,0x11de,0x11dd,0x10fa,0x10fa,0x11f4,0x10fb,0x10fb,0x11ff,0x10f6,0x10f6,0x11e7,0x0010f6,0x10f6,0x11e9,0x0010f7,0x10f7,0x11eb,0x0010f8,0x10f8,0x11ed,0x0010f8,0x10f8,0x11f0,0x0010f7,0x10f7,0x11ae,0x0010ec,0x10ec,0x11b1,0x0010ec,0x10ec,0x11b8,0x0010ec,0x10ec)
USER32(NtUserProcessInkFeedbackCommand                   ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x145c,0x0013f1,0x13f1,0x1480,0x001413,0x1413,0x14ab,0x001440,0x1440)
USER32(NtUserPromoteMouseInPointer                       ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,0x1375,  NONE,    NONE,0x13a3,0x1446,0x0013f4,0x13f4,0x144a,0x0013f8,0x13f8,0x1446,0x0013f6,0x13f6,0x1421,0x0013f2,0x13f2,0x1442,0x001414,0x1414,0x146d,0x001441,0x1441)
USER32(NtUserPromotePointer                              ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x12ce,0x001376,0x1376,0x12d6,0x0013a4,0x13a4,0x12e1,0x0013f5,0x13f5,0x12e2,0x0013f9,0x13f9,0x12e8,0x0013f7,0x13f7,0x12a9,0x0013f3,0x13f3,0x12b2,0x001415,0x1415,0x12ba,0x001442,0x1442)
USER32(NtUserQueryBSDRWindow                             ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x11e6,0x001377,0x1377,0x11e7,0x0013a5,0x13a5,0x11e9,0x0013f6,0x13f6,0x11eb,0x0013fa,0x13fa,0x11ee,0x0013f8,0x13f8,0x11ac,0x0013f4,0x13f4,0x11af,0x001416,0x1416,0x11b6,0x001443,0x1443)
USER32(NtUserQueryDisplayConfig                          ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x1154,0x12ec,0x12ec,0x1157,0x001378,0x1378,0x1159,0x0013a6,0x13a6,0x115a,0x0013f7,0x13f7,0x115c,0x0013fb,0x13fb,0x115d,0x0013f9,0x13f9,0x111a,0x0013f5,0x13f5,0x111b,0x001417,0x1417,0x111b,0x001444,0x1444)
USER32(NtUserQueryInformationThread                      ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x12c8,  NONE,  NONE,0x12ed,  NONE,    NONE,0x1379,  NONE,    NONE,0x13a7,0x11e8,0x0013f8,0x13f8,0x11ea,0x0013fc,0x13fc,0x11ed,0x0013fa,0x13fa,0x11ab,0x0013f6,0x13f6,0x11ae,0x001418,0x1418,0x11b5,0x001445,0x1445)
USER32(NtUserQuerySendMessage                            ,0x11d0,0x11e1,0x11e0,0x1268,0x1268,0x11f7,0x12ca,0x12ca,0x1202,0x12ef,0x12ef,0x11e3,0x00137b,0x137b,0x11e4,0x0013a9,0x13a9,0x11e6,0x0013fa,0x13fa,0x11e8,0x0013fe,0x13fe,0x11eb,0x0013fc,0x13fc,0x11a9,0x0013f8,0x13f8,0x11ac,0x00141a,0x141a,0x11b3,0x001447,0x1447)
USER32(NtUserQueryUserCounters                           ,0x11d1,0x11e2,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
USER32(NtUserQueryWindow                                 ,0x11d2,0x11e3,0x11e1,0x100f,0x100f,0x11f8,0x1010,0x1010,0x1203,0x1010,0x1010,0x11e2,0x071011,0x1011,0x11e3,0x071012,0x1012,0x11e5,0x071013,0x1013,0x11e7,0x071013,0x1013,0x11ea,0x071013,0x1013,0x11a8,0x071013,0x1013,0x11ab,0x071013,0x1013,0x11b2,0x071013,0x1013)
USER32(NtUserRealChildWindowFromPoint                    ,0x11d3,0x11e4,0x11e2,0x1269,0x1269,0x11f9,0x12cb,0x12cb,0x1204,0x12f0,0x12f0,0x11e1,0x00137c,0x137c,0x11e2,0x0013aa,0x13aa,0x11e4,0x0013fb,0x13fb,0x11e6,0x0013ff,0x13ff,0x11e9,0x0013fd,0x13fd,0x11a7,0x0013f9,0x13f9,0x11aa,0x00141b,0x141b,0x11b0,0x001448,0x1448)
USER32(NtUserRealInternalGetMessage                      ,  NONE,0x11e5,0x11e3,0x10ee,0x10ee,0x11fa,0x10ef,0x10ef,0x1205,0x10ea,0x10ea,0x1214,0x0010ea,0x10ea,0x1217,0x0010eb,0x10eb,0x1219,0x0010ec,0x10ec,0x121b,0x0010ec,0x10ec,0x121e,0x0010eb,0x10eb,0x11dd,0x0010e0,0x10e0,0x11e0,0x0010e0,0x10e0,0x11e7,0x0010e0,0x10e0)
USER32(NtUserRealWaitMessageEx                           ,  NONE,0x11e6,0x11e4,0x126a,0x126a,0x11fb,0x12cc,0x12cc,0x1206,0x12f1,0x12f1,0x1215,0x00137d,0x137d,0x1218,0x0013ab,0x13ab,0x121a,0x0013fc,0x13fc,0x121c,0x001400,0x1400,0x121f,0x0013fe,0x13fe,0x11de,0x0013fa,0x13fa,0x11e1,0x00141c,0x141c,0x11e8,0x001449,0x1449)
USER32(NtUserRedrawWindow                                ,0x11d4,0x11e7,0x11e5,0x1012,0x1012,0x11fc,0x1013,0x1013,0x1207,0x1013,0x1013,0x1278,0x141014,0x1014,0x127b,0x141015,0x1015,0x1285,0x141016,0x1016,0x1286,0x141016,0x1016,0x128c,0x141016,0x1016,0x124e,0x141016,0x1016,0x1256,0x141016,0x1016,0x125e,0x141016,0x1016)
USER32(NtUserRegisterBSDRWindow                          ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x1277,0x00137e,0x137e,0x127a,0x0013ac,0x13ac,0x1284,0x0013fd,0x13fd,0x1285,0x001401,0x1401,0x128b,0x0013ff,0x13ff,0x124d,0x0013fb,0x13fb,0x1255,0x00141d,0x141d,0x125d,0x00144a,0x144a)
USER32(NtUserRegisterClassExWOW                          ,  NONE,0x11e8,0x11e6,0x10b4,0x10b4,0x11fd,0x10b5,0x10b5,0x1208,0x10b0,0x10b0,0x1276,0x0010b0,0x10b0,0x1279,0x0010b1,0x10b1,0x1283,0x0010b2,0x10b2,0x1284,0x0010b2,0x10b2,0x128a,0x0010b2,0x10b2,0x124c,0x0010ab,0x10ab,0x1254,0x0010ab,0x10ab,0x125c,0x0010ab,0x10ab)
USER32(NtUserRegisterDManipHook                          ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x127d,0x0013fe,0x13fe,0x127e,0x001402,0x1402,0x1284,0x001400,0x1400,0x1246,0x0013fc,0x13fc,0x124d,0x00141e,0x141e,0x1255,0x00144b,0x144b)
USER32(NtUserRegisterEdgy                                ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x12be,0x00137f,0x137f,0x12c1,0x0013ad,0x13ad,0x12cb,0x0013ff,0x13ff,0x12cc,0x001403,0x1403,0x12d2,0x001401,0x1401,0x1293,0x0013fd,0x13fd,0x129c,0x00141f,0x141f,0x12a4,0x00144c,0x144c)
USER32(NtUserRegisterErrorReportingDialog                ,  NONE,  NONE,  NONE,  NONE,  NONE,0x11fe,0x12cd,0x12cd,0x1209,0x12f2,0x12f2,0x1275,0x001380,0x1380,0x1278,0x0013ae,0x13ae,0x1282,0x001400,0x1400,0x1283,0x001404,0x1404,0x1289,0x001402,0x1402,0x124b,0x0013fe,0x13fe,0x1253,0x001420,0x1420,0x125b,0x00144d,0x144d)
USER32(NtUserRegisterHotKey                              ,0x11d6,0x11ea,0x11e8,0x126b,0x126b,0x1200,0x12ce,0x12ce,0x120b,0x12f3,0x12f3,0x1273,0x141381,0x1381,0x1276,0x1413af,0x13af,0x127c,0x141401,0x1401,0x127d,0x141405,0x1405,0x1283,0x141403,0x1403,0x1245,0x1413ff,0x13ff,0x124c,0x141421,0x1421,0x1254,0x14144e,0x144e)
USER32(NtUserRegisterManipulationThread                  ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x127f,0x001402,0x1402,0x1280,0x001406,0x1406,0x1286,0x001404,0x1404,0x1248,0x001400,0x1400,0x124f,0x001422,0x1422,0x1257,0x00144f,0x144f)
USER32(NtUserRegisterPointerDeviceNotifications          ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x12ca,0x001382,0x1382,0x12cf,0x0013b0,0x13b0,0x12da,0x001403,0x1403,0x12db,0x001407,0x1407,0x12e1,0x001405,0x1405,0x12a2,0x001401,0x1401,0x12ab,0x001423,0x1423,0x12b3,0x001450,0x1450)
USER32(NtUserRegisterPointerInputTarget                  ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x12d0,0x001383,0x1383,0x12d8,0x0013b1,0x13b1,0x12e3,0x001404,0x1404,0x12e4,0x001408,0x1408,0x12ea,0x001406,0x1406,0x12ab,0x001402,0x1402,0x12b4,0x001424,0x1424,0x12bc,0x001451,0x1451)
USER32(NtUserRegisterRawInputDevices                     ,  NONE,0x11eb,0x11e9,0x126c,0x126c,0x1201,0x12cf,0x12cf,0x120c,0x12f4,0x12f4,0x1272,0x001384,0x1384,0x1275,0x0013b2,0x13b2,0x127b,0x001405,0x1405,0x127c,0x001409,0x1409,0x1282,0x001407,0x1407,0x1244,0x001403,0x1403,0x124b,0x001425,0x1425,0x1253,0x001452,0x1452)
USER32(NtUserRegisterServicesProcess                     ,  NONE,  NONE,  NONE,0x205b,  NONE,  NONE,0x2061,  NONE,0x120d,0x12f5,0x12f5,0x1271,0x001385,0x1385,0x1274,0x0013b3,0x13b3,0x127a,0x001406,0x1406,0x127b,0x00140a,0x140a,0x1281,0x001408,0x1408,0x1243,0x001404,0x1404,0x124a,0x001426,0x1426,0x1252,0x001453,0x1453)
USER32(NtUserRegisterSessionPort                         ,  NONE,  NONE,  NONE,  NONE,  NONE,0x1270,0x12d0,0x12d0,0x127c,0x12f6,0x12f6,0x1282,0x001386,0x1386,0x1285,0x0013b4,0x13b4,0x128f,0x001407,0x1407,0x1290,0x00140b,0x140b,0x1296,0x001409,0x1409,0x1257,0x001405,0x1405,0x125f,0x001427,0x1427,0x1267,0x001454,0x1454)
USER32(NtUserRegisterShellPTPListener                    ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x12cc,0x001408,0x1408,0x12cd,0x00140c,0x140c,0x12d3,0x00140a,0x140a,0x1294,0x001406,0x1406,0x129d,0x001428,0x1428,0x12a5,0x001455,0x1455)
USER32(NtUserRegisterTasklist                            ,0x11d7,0x11ec,0x11ea,0x126d,0x126d,0x1202,0x12d1,0x12d1,0x120e,0x12f7,0x12f7,0x1270,0x031387,0x1387,0x1273,0x0313b5,0x13b5,0x1279,0x031409,0x1409,0x127a,0x03140d,0x140d,0x1280,0x03140b,0x140b,0x1242,0x031407,0x1407,0x1249,0x031429,0x1429,0x1251,0x031456,0x1456)
USER32(NtUserRegisterTouchHitTestingWindow               ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x12c4,0x001388,0x1388,0x12c7,0x0013b6,0x13b6,0x12d2,0x00140a,0x140a,0x12d3,0x00140e,0x140e,0x12d9,0x00140c,0x140c,0x129a,0x001408,0x1408,0x12a3,0x00142a,0x142a,0x12ab,0x001457,0x1457)
USER32(NtUserRegisterTouchPadCapable                     ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,0x12cb,0x0013b7,0x13b7,0x12d6,0x00140b,0x140b,0x12d7,0x00140f,0x140f,0x12dd,0x00140d,0x140d,0x129e,0x001409,0x1409,0x12a7,0x00142b,0x142b,0x12af,0x001458,0x1458)
USER32(NtUserRegisterUserApiHook                         ,  NONE,0x11e9,0x11e7,0x126e,0x126e,0x11ff,0x12d2,0x12d2,0x120a,0x12f8,0x12f8,0x1274,0x001389,0x1389,0x1277,0x0013b8,0x13b8,0x1281,0x00140c,0x140c,0x1282,0x001410,0x1410,0x1288,0x00140e,0x140e,0x124a,0x00140a,0x140a,0x1252,0x00142c,0x142c,0x125a,0x001459,0x1459)
USER32(NtUserRegisterWindowMessage                       ,  NONE,0x11ed,0x11eb,0x1036,0x1036,0x1203,0x1037,0x1037,0x120f,0x1037,0x1037,0x126f,0x001038,0x1038,0x1272,0x001039,0x1039,0x1278,0x00103a,0x103a,0x1279,0x00103a,0x103a,0x127f,0x00103a,0x103a,0x1241,0x00103a,0x103a,0x1248,0x00103a,0x103a,0x1250,0x00103a,0x103a)
USER32(NtUserReleaseDC                                   ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x11e3,0x00142d,0x142d,0x11ea,0x00145a,0x145a)
USER32(NtUserReleaseDwmHitTestWaiters                    ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x121b,0x00140d,0x140d,0x121d,0x001411,0x1411,0x1220,0x00140f,0x140f,0x11df,0x00140b,0x140b,0x11e2,0x00142e,0x142e,0x11e9,0x00145b,0x145b)
USER32(NtUserRemoteConnect                               ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x12d3,  NONE,  NONE,0x12f9,  NONE,    NONE,0x138a,  NONE,    NONE,0x13b9,0x1294,0x00140e,0x140e,0x1295,0x001412,0x1412,0x129b,0x001410,0x1410,0x125c,0x00140c,0x140c,0x1264,0x00142f,0x142f,0x126c,0x00145c,0x145c)
USER32(NtUserRemoteRedrawRectangle                       ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x12d4,  NONE,  NONE,0x12fa,  NONE,    NONE,0x138b,  NONE,    NONE,0x13ba,0x1293,0x00140f,0x140f,0x1294,0x001413,0x1413,0x129a,0x001411,0x1411,0x125b,0x00140d,0x140d,0x1263,0x001430,0x1430,0x126b,0x00145d,0x145d)
USER32(NtUserRemoteRedrawScreen                          ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x12d5,  NONE,  NONE,0x12fb,  NONE,    NONE,0x138c,  NONE,    NONE,0x13bb,0x1292,0x001410,0x1410,0x1293,0x001414,0x1414,0x1299,0x001412,0x1412,0x125a,0x00140e,0x140e,0x1262,0x001431,0x1431,0x126a,0x00145e,0x145e)
USER32(NtUserRemoteStopScreenUpdates                     ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x12d6,  NONE,  NONE,0x12fc,  NONE,    NONE,0x138d,  NONE,    NONE,0x13bc,0x1291,0x001411,0x1411,0x1292,0x001415,0x1415,0x1298,0x001413,0x1413,0x1259,0x00140f,0x140f,0x1261,0x001432,0x1432,0x1269,0x00145f,0x145f)
USER32(NtUserRemoveClipboardFormatListener               ,  NONE,  NONE,  NONE,  NONE,  NONE,0x1204,0x12d7,0x12d7,0x1210,0x12fd,0x12fd,0x126e,0x00138e,0x138e,0x1271,0x0013bd,0x13bd,0x1277,0x001412,0x1412,0x1278,0x001416,0x1416,0x127e,0x001414,0x1414,0x1240,0x001410,0x1410,0x1247,0x001433,0x1433,0x124f,0x001460,0x1460)
USER32(NtUserRemoveInjectionDevice                       ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x145e,0x001417,0x1417,0x1459,0x001415,0x1415,0x1435,0x001411,0x1411,0x1458,0x001434,0x1434,0x1483,0x001461,0x1461)
USER32(NtUserRemoveMenu                                  ,0x11d9,0x11ee,0x11ec,0x10fe,0x10fe,0x1205,0x10ff,0x10ff,0x1211,0x10fa,0x10fa,0x126d,0x0a10fa,0x10fa,0x1270,0x0a10fb,0x10fb,0x1276,0x0a10fc,0x10fc,0x1277,0x0a10fc,0x10fc,0x127d,0x0a10fb,0x10fb,0x123f,0x0a10f0,0x10f0,0x1246,0x0a10f0,0x10f0,0x124e,0x0a10f0,0x10f0)
USER32(NtUserRemoveProp                                  ,  NONE,0x11ef,0x11ed,0x1045,0x1045,0x1206,0x1046,0x1046,0x1212,0x1046,0x1046,0x126c,0x071047,0x1047,0x126f,0x071048,0x1048,0x1275,0x071049,0x1049,0x1276,0x071049,0x1049,0x127c,0x071049,0x1049,0x123e,0x071049,0x1049,0x1245,0x071049,0x1049,0x124d,0x071049,0x1049)
USER32(NtUserReportInertia                               ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,0x1408,0x0013be,0x13be,0x1455,0x001413,0x1413,0x1459,0x001418,0x1418,0x1454,0x001416,0x1416,0x142f,0x001412,0x1412,0x1451,0x001435,0x1435,0x147c,0x001462,0x1462)
USER32(NtUserRequestMoveSizeOperation                    ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x11b1,0x001463,0x1463)
USER32(NtUserResolveDesktop                              ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x1121,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
USER32(NtUserResolveDesktopForWOW                        ,0x11dc,0x11f1,0x11ef,0x1273,0x1273,0x1208,0x12d8,0x12d8,0x1213,0x12fe,0x12fe,0x126b,0x00138f,0x138f,0x126e,0x0013bf,0x13bf,0x1274,0x001414,0x1414,0x1275,0x001419,0x1419,0x127b,0x001417,0x1417,0x123d,0x001413,0x1413,0x1244,0x001436,0x1436,0x124c,0x001464,0x1464)
USER32(NtUserSBGetParms                                  ,  NONE,0x11f2,0x11f0,0x104d,0x104d,0x1209,0x104e,0x104e,0x1214,0x104e,0x104e,0x126a,0x14104f,0x104f,0x126d,0x141050,0x1050,0x1273,0x141051,0x1051,0x1274,0x141051,0x1051,0x127a,0x141051,0x1051,0x123c,0x141051,0x1051,0x1243,0x141051,0x1051,0x124b,0x141051,0x1051)
USER32(NtUserScrollDC                                    ,  NONE,0x11f3,0x11f1,0x106a,0x106a,0x120a,0x106b,0x106b,0x1215,0x106b,0x106b,0x1269,0x00106b,0x106b,0x126c,0x00106c,0x106c,0x1272,0x00106d,0x106d,0x1273,0x00106d,0x106d,0x1279,0x00106d,0x106d,0x123b,0x00106c,0x106c,0x1242,0x00106c,0x106c,0x124a,0x00106c,0x106c)
USER32(NtUserScrollWindowEx                              ,  NONE,0x11f4,0x11f2,0x10c2,0x10c2,0x120b,0x10c3,0x10c3,0x1216,0x10be,0x10be,0x1268,0x0010be,0x10be,0x126b,0x0010bf,0x10bf,0x1271,0x0010c0,0x10c0,0x1272,0x0010c0,0x10c0,0x1278,0x0010c0,0x10c0,0x123a,0x0010b9,0x10b9,0x1241,0x0010b9,0x10b9,0x1249,0x0010b9,0x10b9)
USER32(NtUserSelectPalette                               ,  NONE,0x11f5,0x11f3,0x101c,0x101c,0x120c,0x101d,0x101d,0x1217,0x101d,0x101d,0x1267,0x0c101e,0x101e,0x126a,0x0c101f,0x101f,0x1270,0x0c1020,0x1020,0x1271,0x0c1020,0x1020,0x1277,0x0c1020,0x1020,0x1239,0x0c1020,0x1020,0x1240,0x0c1020,0x1020,0x1248,0x0c1020,0x1020)
USER32(NtUserSendEventMessage                            ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x1221,0x001390,0x1390,0x1224,0x0013c0,0x13c0,0x1227,0x001415,0x1415,0x1229,0x00141a,0x141a,0x122c,0x001418,0x1418,0x11eb,0x001414,0x1414,0x11ef,0x001437,0x1437,0x11f6,0x001465,0x1465)
USER32(NtUserSendInput                                   ,0x11e1,0x11f6,0x11f4,0x1083,0x1083,0x120d,0x1084,0x1084,0x1218,0x1082,0x1082,0x1266,0x001082,0x1082,0x1269,0x001083,0x1083,0x126f,0x001084,0x1084,0x1270,0x001084,0x1084,0x1276,0x001084,0x1084,0x1238,0x001082,0x1082,0x123f,0x001082,0x1082,0x1247,0x001082,0x1082)
USER32(NtUserSendInteractiveControlHapticsReport         ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x146b,0x001419,0x1419,0x1457,0x001415,0x1415,0x147b,0x001438,0x1438,0x14a6,0x001466,0x1466)
USER32(NtUserSendTouchInput                              ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x128e,0x12ff,0x12ff,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
USER32(NtUserSetActivationFilter                         ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,0x13ef,0x0013c1,0x13c1,0x143a,0x001416,0x1416,0x143d,0x00141b,0x141b,0x1439,0x00141a,0x141a,0x1414,0x001416,0x1416,0x1435,0x001439,0x1439,0x145b,0x001467,0x1467)
USER32(NtUserSetActiveProcess                            ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x13bf,0x001391,0x1391,0x13e7,0x0013c2,0x13c2,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
USER32(NtUserSetActiveProcessForMonitor                  ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1432,0x001417,0x1417,0x1435,0x00141c,0x141c,0x1431,0x00141b,0x141b,0x140c,0x001417,0x1417,0x142d,0x00143a,0x143a,0x1453,0x001468,0x1468)
USER32(NtUserSetActiveWindow                             ,0x11e4,0x11f7,0x11f5,0x10e4,0x10e4,0x120e,0x10e5,0x10e5,0x1219,0x10e0,0x10e0,0x1265,0x0310e0,0x10e0,0x1268,0x0310e1,0x10e1,0x126e,0x0310e2,0x10e2,0x126f,0x0310e2,0x10e2,0x1275,0x0310e1,0x10e1,0x1237,0x0310d6,0x10d6,0x123e,0x0310d6,0x10d6,0x1246,0x0310d6,0x10d6)
USER32(NtUserSetAutoRotation                             ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x13c3,0x001393,0x1393,0x13eb,0x0013c4,0x13c4,0x1436,0x001419,0x1419,0x1439,0x00141e,0x141e,0x1435,0x00141d,0x141d,0x1410,0x001419,0x1419,0x1431,0x00143c,0x143c,0x1457,0x00146a,0x146a)
USER32(NtUserSetBridgeWindowChild                        ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1147,0x00146b,0x146b)
USER32(NtUserSetBrokeredForeground                       ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x13c9,0x001394,0x1394,0x13f1,0x0013c5,0x13c5,0x143c,0x00141a,0x141a,0x143f,0x00141f,0x141f,0x143b,0x00141e,0x141e,0x1416,0x00141a,0x141a,0x1437,0x00143d,0x143d,0x145d,0x00146c,0x146c)
USER32(NtUserSetCalibrationData                          ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x12c6,0x001395,0x1395,0x12c9,0x0013c6,0x13c6,0x12d4,0x00141b,0x141b,0x12d5,0x001420,0x1420,0x12db,0x00141f,0x141f,0x129c,0x00141b,0x141b,0x12a5,0x00143e,0x143e,0x12ad,0x00146d,0x146d)
USER32(NtUserSetCapture                                  ,0x11e6,0x11f9,0x11f7,0x1048,0x1048,0x1210,0x1049,0x1049,0x121b,0x1049,0x1049,0x1263,0x03104a,0x104a,0x1266,0x03104b,0x104b,0x126c,0x03104c,0x104c,0x126d,0x03104c,0x104c,0x1273,0x03104c,0x104c,0x1235,0x03104c,0x104c,0x123c,0x03104c,0x104c,0x1244,0x03104c,0x104c)
USER32(NtUserSetChildWindowNoActivate                    ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x121c,0x1301,0x1301,0x1262,0x001396,0x1396,0x1265,0x0013c7,0x13c7,0x126b,0x00141c,0x141c,0x126c,0x001421,0x1421,0x1272,0x001420,0x1420,0x1234,0x00141c,0x141c,0x123b,0x00143f,0x143f,0x1243,0x00146e,0x146e)
USER32(NtUserSetClassLong                                ,  NONE,0x11fa,0x11f8,0x10c4,0x10c4,0x1211,0x10c5,0x10c5,0x121d,0x10c0,0x10c0,0x1261,0x0010c0,0x10c0,0x1264,0x0010c1,0x10c1,0x126a,0x0010c2,0x10c2,0x126b,0x0010c2,0x10c2,0x1271,0x0010c2,0x10c2,0x1233,0x0010bb,0x10bb,0x123a,0x0010bb,0x10bb,0x1242,0x0010bb,0x10bb)
USER32(NtUserSetClassLongPtr                             ,  NONE,  NONE,  NONE,  NONE,0x1297,  NONE,  NONE,0x1304,  NONE,  NONE,0x1339,  NONE,    NONE,0x13d8,  NONE,    NONE,0x140c,  NONE,    NONE,0x1468,  NONE,    NONE,0x146d,  NONE,    NONE,0x1470,  NONE,    NONE,0x146c,  NONE,    NONE,0x1496,  NONE,    NONE,0x14c8)
USER32(NtUserSetClassWord                                ,0x11e8,0x11fb,0x11f9,0x1275,0x1275,0x1212,0x12da,0x12da,0x121e,0x1302,0x1302,0x1260,0x0c1397,0x1397,0x1263,0x0c13c8,0x13c8,0x1269,0x0c141d,0x141d,0x126a,0x0c1422,0x1422,0x1270,0x0c1421,0x1421,0x1232,0x0c141d,0x141d,0x1239,0x0c1440,0x1440,0x1241,0x0c146f,0x146f)
USER32(NtUserSetClipboardData                            ,  NONE,0x11fc,0x11fa,0x10d5,0x10d5,0x1213,0x10d6,0x10d6,0x121f,0x10d1,0x10d1,0x125f,0x0f10d1,0x10d1,0x1262,0x0f10d2,0x10d2,0x1268,0x0f10d3,0x10d3,0x1269,0x0f10d3,0x10d3,0x126f,0x0f10d2,0x10d2,0x1231,0x0f10c7,0x10c7,0x1238,0x0f10c7,0x10c7,0x1240,0x0f10c7,0x10c7)
USER32(NtUserSetClipboardViewer                          ,0x11ea,0x11fd,0x11fb,0x1121,0x1121,0x1214,0x1122,0x1122,0x1220,0x111a,0x111a,0x125e,0x03111a,0x111a,0x1261,0x03111b,0x111b,0x1267,0x03111c,0x111c,0x1268,0x03111c,0x111c,0x126e,0x03111b,0x111b,0x1230,0x031110,0x1110,0x1237,0x031110,0x1110,0x123f,0x031110,0x1110)
USER32(NtUserSetConsoleReserveKeys                       ,0x11eb,0x11fe,0x11fc,0x1123,0x1123,0x1215,0x1124,0x1124,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
USER32(NtUserSetCoreWindow                               ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1464,0x00141e,0x141e,0x1469,0x001423,0x1423,0x1464,0x001422,0x1422,0x1450,0x00141e,0x141e,0x1474,0x001441,0x1441,0x149f,0x001470,0x1470)
USER32(NtUserSetCoreWindowPartner                        ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1465,0x00141f,0x141f,0x146a,0x001424,0x1424,0x1465,0x001423,0x1423,0x1451,0x00141f,0x141f,0x1475,0x001442,0x1442,0x14a0,0x001471,0x1471)
USER32(NtUserSetCursor                                   ,0x11ec,0x11ff,0x11fd,0x1019,0x1019,0x1216,0x101a,0x101a,0x1221,0x101a,0x101a,0x125d,0x04101b,0x101b,0x1260,0x04101c,0x101c,0x1266,0x04101d,0x101d,0x1267,0x04101d,0x101d,0x126d,0x04101d,0x101d,0x122f,0x04101d,0x101d,0x1235,0x04101d,0x101d,0x123d,0x04101d,0x101d)
USER32(NtUserSetCursorContents                           ,0x11ed,0x1200,0x11fe,0x1276,0x1276,0x1217,0x12db,0x12db,0x1222,0x1303,0x1303,0x125c,0x051398,0x1398,0x125f,0x0513c9,0x13c9,0x1265,0x051420,0x1420,0x1266,0x051425,0x1425,0x126c,0x051424,0x1424,0x122e,0x051420,0x1420,0x1234,0x051443,0x1443,0x123c,0x051472,0x1472)
USER32(NtUserSetCursorIconData                           ,  NONE,0x1201,0x11ff,0x10a8,0x10a8,0x1218,0x10a9,0x10a9,0x1223,0x10a4,0x10a4,0x125b,0x0010a4,0x10a4,0x125e,0x0010a5,0x10a5,0x1264,0x0010a6,0x10a6,0x1265,0x0010a6,0x10a6,0x126b,0x0010a6,0x10a6,0x122d,0x0010a1,0x10a1,0x1233,0x0010a1,0x10a1,0x123b,0x0010a1,0x10a1)
USER32(NtUserSetCursorPos                                ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1236,0x001444,0x1444,0x123e,0x001473,0x1473)
USER32(NtUserSetDbgTag                                   ,0x11ef,0x1202,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
USER32(NtUserSetDesktopColorTransform                    ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x13c2,0x001445,0x1445,0x13d4,0x001474,0x1474)
USER32(NtUserSetDialogControlDpiChangeBehavior           ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x120e,0x001421,0x1421,0x1213,0x001446,0x1446,0x121b,0x001475,0x1475)
USER32(NtUserSetDimUndimTransitionTime                   ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1122,0x001476,0x1476)
USER32(NtUserSetDisplayAutoRotationPreferences           ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x13c2,0x001399,0x1399,0x13ea,0x0013ca,0x13ca,0x1435,0x001421,0x1421,0x1438,0x001426,0x1426,0x1434,0x001425,0x1425,0x140f,0x001422,0x1422,0x1430,0x001447,0x1447,0x1456,0x001477,0x1477)
USER32(NtUserSetDisplayConfig                            ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x1153,0x1304,0x1304,0x1158,0x00139a,0x139a,0x115a,0x0013cb,0x13cb,0x115b,0x001422,0x1422,0x115d,0x001427,0x1427,0x115e,0x001426,0x1426,0x111b,0x001423,0x1423,0x111c,0x001448,0x1448,0x111c,0x001478,0x1478)
USER32(NtUserSetDisplayMapping                           ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x12c5,0x00139b,0x139b,0x12c8,0x0013cc,0x13cc,0x12d3,0x001423,0x1423,0x12d4,0x001428,0x1428,0x12da,0x001427,0x1427,0x129b,0x001424,0x1424,0x12a4,0x001449,0x1449,0x12ac,0x001479,0x1479)
USER32(NtUserSetFallbackForeground                       ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x13c8,0x00139c,0x139c,0x13f0,0x0013cd,0x13cd,0x143b,0x001424,0x1424,0x143e,0x001429,0x1429,0x143a,0x001428,0x1428,0x1415,0x001425,0x1425,0x1436,0x00144a,0x144a,0x145c,0x00147a,0x147a)
USER32(NtUserSetFeatureReportResponse                    ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x145a,0x001425,0x1425,0x145f,0x00142a,0x142a,0x145a,0x001429,0x1429,0x1436,0x001426,0x1426,0x1459,0x00144b,0x144b,0x1484,0x00147b,0x147b)
USER32(NtUserSetFocus                                    ,0x11f0,0x1203,0x1200,0x1050,0x1050,0x1219,0x1051,0x1051,0x1224,0x1051,0x1051,0x125a,0x031052,0x1052,0x125d,0x031053,0x1053,0x1263,0x031054,0x1054,0x1264,0x031054,0x1054,0x126a,0x031054,0x1054,0x122c,0x031054,0x1054,0x1232,0x031054,0x1054,0x123a,0x031054,0x1054)
USER32(NtUserSetGestureConfig                            ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x1296,0x1305,0x1305,0x12d5,0x00139d,0x139d,0x12de,0x0013ce,0x13ce,0x12ea,0x001426,0x1426,0x12eb,0x00142b,0x142b,0x12f1,0x00142a,0x142a,0x12b2,0x001427,0x1427,0x12bb,0x00144c,0x144c,0x12c3,0x00147c,0x147c)
USER32(NtUserSetImeHotKey                                ,  NONE,0x1204,0x1201,0x1277,0x1277,0x121a,0x12dc,0x12dc,0x1225,0x1306,0x1306,0x1259,0x00139e,0x139e,0x125c,0x0013cf,0x13cf,0x1262,0x001427,0x1427,0x1263,0x00142c,0x142c,0x1269,0x00142b,0x142b,0x122b,0x001428,0x1428,0x1231,0x00144d,0x144d,0x1239,0x00147d,0x147d)
USER32(NtUserSetImeOwnerWindow                           ,  NONE,0x1206,0x1203,0x1279,0x1279,0x121c,0x12de,0x12de,0x1227,0x1308,0x1308,0x1257,0x0813a0,0x13a0,0x125a,0x0813d1,0x13d1,0x1260,0x081429,0x1429,0x1261,0x08142e,0x142e,0x1267,0x08142d,0x142d,0x1229,0x08142a,0x142a,0x122f,0x08144f,0x144f,0x1237,0x08147f,0x147f)
USER32(NtUserSetImmersiveBackgroundWindow                ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x1249,0x0013a1,0x13a1,0x124c,0x0013d2,0x13d2,0x1252,0x00142a,0x142a,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
USER32(NtUserSetInformationProcess                       ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x1101,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
USER32(NtUserSetInformationThread                        ,  NONE,0x1208,0x1205,0x10e5,0x10e5,0x121e,0x10e6,0x10e6,0x1228,0x10e1,0x10e1,0x1256,0x0010e1,0x10e1,0x1259,0x0010e2,0x10e2,0x125f,0x0010e3,0x10e3,0x1260,0x0010e3,0x10e3,0x1266,0x0010e2,0x10e2,0x1228,0x0010d7,0x10d7,0x122e,0x0010d7,0x10d7,0x1236,0x0010d7,0x10d7)
USER32(NtUserSetInteractiveControlFocus                  ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x146c,0x00142e,0x142e,0x1458,0x00142b,0x142b,0x147c,0x001450,0x1450,0x14a7,0x001480,0x1480)
USER32(NtUserSetInteractiveCtrlRotationAngle             ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x146e,0x00142f,0x142f,0x145a,0x00142c,0x142c,0x147e,0x001451,0x1451,0x14a9,0x001481,0x1481)
USER32(NtUserSetInternalWindowPos                        ,0x11f6,0x1209,0x1206,0x127a,0x127a,0x121f,0x12df,0x12df,0x1229,0x1309,0x1309,0x1255,0x1413a2,0x13a2,0x1258,0x1413d3,0x13d3,0x125e,0x14142b,0x142b,0x125f,0x14142f,0x142f,0x1265,0x141430,0x1430,0x1227,0x14142d,0x142d,0x122d,0x141452,0x1452,0x1235,0x141482,0x1482)
USER32(NtUserSetKeyboardState                            ,0x11f7,0x120a,0x1207,0x10f3,0x10f3,0x1220,0x10f4,0x10f4,0x122a,0x10ef,0x10ef,0x1254,0x0410ef,0x10ef,0x1257,0x0410f0,0x10f0,0x125d,0x0410f1,0x10f1,0x125e,0x0410f1,0x10f1,0x1264,0x0410f0,0x10f0,0x1226,0x0410e5,0x10e5,0x122c,0x0410e5,0x10e5,0x1234,0x0410e5,0x10e5)
USER32(NtUserSetLayeredWindowAttributes                  ,0x122e,0x1245,0x1241,0x127b,0x127b,0x125d,0x12e0,0x12e0,0x126a,0x130a,0x130a,0x1294,0x1413a3,0x13a3,0x1297,0x1413d4,0x13d4,0x12a1,0x14142c,0x142c,0x12a2,0x141430,0x1430,0x12a8,0x141431,0x1431,0x1269,0x14142e,0x142e,0x1272,0x141453,0x1453,0x127a,0x141483,0x1483)
USER32(NtUserSetLogonNotifyWindow                        ,0x11f8,0x120b,0x1208,0x127c,0x127c,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
USER32(NtUserSetManipulationInputTarget                  ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1280,0x00142d,0x142d,0x1281,0x001431,0x1431,0x1287,0x001432,0x1432,0x1249,0x00142f,0x142f,0x1250,0x001454,0x1454,0x1258,0x001484,0x1484)
USER32(NtUserSetMenu                                     ,  NONE,0x120c,0x1209,0x127d,0x127d,0x1221,0x12e1,0x12e1,0x122b,0x130b,0x130b,0x1253,0x0c13a4,0x13a4,0x1256,0x0c13d5,0x13d5,0x125c,0x0c142e,0x142e,0x125d,0x0c1432,0x1432,0x1263,0x0c1433,0x1433,0x1225,0x0c1430,0x1430,0x122b,0x0c1455,0x1455,0x1233,0x0c1485,0x1485)
USER32(NtUserSetMenuContextHelpId                        ,0x11fa,0x120d,0x120a,0x127e,0x127e,0x1222,0x12e2,0x12e2,0x122c,0x130c,0x130c,0x1252,0x0513a5,0x13a5,0x1255,0x0513d6,0x13d6,0x125b,0x05142f,0x142f,0x125c,0x051433,0x1433,0x1262,0x051434,0x1434,0x1224,0x051431,0x1431,0x122a,0x051456,0x1456,0x1232,0x051486,0x1486)
USER32(NtUserSetMenuDefaultItem                          ,0x11fb,0x120e,0x120b,0x1107,0x1107,0x1223,0x1108,0x1108,0x122d,0x1102,0x1102,0x1251,0x0a1102,0x1102,0x1254,0x0a1103,0x1103,0x125a,0x0a1104,0x1104,0x125b,0x0a1104,0x1104,0x1261,0x0a1103,0x1103,0x1223,0x0a10f8,0x10f8,0x1229,0x0a10f8,0x10f8,0x1231,0x0a10f8,0x10f8)
USER32(NtUserSetMenuFlagRtoL                             ,  NONE,0x120f,0x120c,0x127f,0x127f,0x1224,0x12e3,0x12e3,0x122e,0x130d,0x130d,0x1250,0x0413a6,0x13a6,0x1253,0x0413d7,0x13d7,0x1259,0x041430,0x1430,0x125a,0x041434,0x1434,0x1260,0x041435,0x1435,0x1222,0x041432,0x1432,0x1228,0x041457,0x1457,0x1230,0x041487,0x1487)
USER32(NtUserSetMirrorRendering                          ,  NONE,  NONE,  NONE,  NONE,  NONE,0x1302,0x12e4,0x12e4,0x1332,0x130e,0x130e,0x1379,0x0013a7,0x13a7,0x138b,0x0013d8,0x13d8,0x13c4,0x001431,0x1431,0x13c6,0x001435,0x1435,0x13d5,0x001436,0x1436,0x13a7,0x001433,0x1433,0x13c1,0x001458,0x1458,0x13d3,0x001488,0x1488)
USER32(NtUserSetMonitorBrightness                        ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1121,0x001489,0x1489)
USER32(NtUserSetObjectInformation                        ,0x11fd,0x1210,0x120d,0x1280,0x1280,0x1225,0x12e5,0x12e5,0x122f,0x130f,0x130f,0x124f,0x0013a8,0x13a8,0x1252,0x0013d9,0x13d9,0x1258,0x001432,0x1432,0x1259,0x001436,0x1436,0x125f,0x001437,0x1437,0x1221,0x001434,0x1434,0x1227,0x001459,0x1459,0x122f,0x00148a,0x148a)
USER32(NtUserSetParent                                   ,0x11fe,0x1211,0x120e,0x1078,0x1078,0x1226,0x1079,0x1079,0x1230,0x1077,0x1077,0x124e,0x081077,0x1077,0x1251,0x081078,0x1078,0x1257,0x081079,0x1079,0x1258,0x081079,0x1079,0x125e,0x081079,0x1079,0x1220,0x081078,0x1078,0x1226,0x081078,0x1078,0x122e,0x081078,0x1078)
USER32(NtUserSetPrecisionTouchPadConfiguration           ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,0x12d5,0x0013da,0x13da,0x12e0,0x001433,0x1433,0x12e1,0x001437,0x1437,0x12e7,0x001438,0x1438,0x12a8,0x001435,0x1435,0x12b1,0x00145a,0x145a,0x12b9,0x00148b,0x148b)
USER32(NtUserSetProcessDPIAware                          ,  NONE,  NONE,  NONE,  NONE,  NONE,0x1234,0x12e6,0x12e6,0x123e,0x1310,0x1310,0x123e,0x0013a9,0x13a9,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
USER32(NtUserSetProcessDpiAwareness                      ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,0x1241,0x0013db,0x13db,0x1244,0x001434,0x1434,0x1246,0x001438,0x1438,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
USER32(NtUserSetProcessDpiAwarenessContext               ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x124a,0x001439,0x1439,0x1209,0x001436,0x1436,0x120e,0x00145b,0x145b,0x1216,0x00148c,0x148c)
USER32(NtUserSetProcessInteractionFlags                  ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x146f,0x00143a,0x143a,0x145d,0x001437,0x1437,0x1481,0x00145c,0x145c,0x14ac,0x00148d,0x148d)
USER32(NtUserSetProcessRestrictionExemption              ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x13cd,0x0013aa,0x13aa,0x13f6,0x0013dc,0x13dc,0x1441,0x001435,0x1435,0x1444,0x001439,0x1439,0x1440,0x00143b,0x143b,0x141b,0x001438,0x1438,0x143c,0x00145d,0x145d,0x1462,0x00148e,0x148e)
USER32(NtUserSetProcessUIAccessZorder                    ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x123d,0x0013ab,0x13ab,0x1240,0x0013dd,0x13dd,0x1243,0x001436,0x1436,0x1245,0x00143a,0x143a,0x1249,0x00143c,0x143c,0x1208,0x001439,0x1439,0x120d,0x00145e,0x145e,0x1215,0x00148f,0x148f)
USER32(NtUserSetProcessWindowStation                     ,0x11ff,0x1212,0x120f,0x10ac,0x10ac,0x1227,0x10ad,0x10ad,0x1231,0x10a8,0x10a8,0x124d,0x0410a8,0x10a8,0x1250,0x0410a9,0x10a9,0x1256,0x0410aa,0x10aa,0x1257,0x0410aa,0x10aa,0x125d,0x0410aa,0x10aa,0x121f,0x0410a4,0x10a4,0x1225,0x0410a4,0x10a4,0x122d,0x0410a4,0x10a4)
USER32(NtUserSetProp                                     ,  NONE,0x1213,0x1210,0x104b,0x104b,0x1229,0x104c,0x104c,0x1233,0x104c,0x104c,0x124b,0x10104d,0x104d,0x124e,0x10104e,0x104e,0x1254,0x10104f,0x104f,0x1255,0x10104f,0x104f,0x125b,0x10104f,0x104f,0x121d,0x10104f,0x104f,0x1223,0x10104f,0x104f,0x122b,0x10104f,0x104f)
USER32(NtUserSetRipFlags                                 ,0x1201,0x1214,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
USER32(NtUserSetScrollInfo                               ,0x1202,0x1215,0x1211,0x102f,0x102f,0x122a,0x1030,0x1030,0x1234,0x1030,0x1030,0x124a,0x141031,0x1031,0x124d,0x141032,0x1032,0x1253,0x141033,0x1033,0x1254,0x141033,0x1033,0x125a,0x141033,0x1033,0x121c,0x141033,0x1033,0x1222,0x141033,0x1033,0x122a,0x141033,0x1033)
USER32(NtUserSetSensorPresence                           ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,0x13ac,  NONE,    NONE,0x1001,0x1002,0x001002,0x1002,0x1002,0x001002,0x1002,0x1002,0x001002,0x1002,0x1002,0x001002,0x1002,0x1002,0x001002,0x1002,0x1002,0x001002,0x1002)
USER32(NtUserSetShellWindowEx                            ,0x1203,0x1216,0x1212,0x1281,0x1281,0x122b,0x12e7,0x12e7,0x1235,0x1311,0x1311,0x1248,0x0813ad,0x13ad,0x124b,0x0813de,0x13de,0x1251,0x081437,0x1437,0x1253,0x08143b,0x143b,0x1259,0x08143d,0x143d,0x121b,0x08143a,0x143a,0x1221,0x08145f,0x145f,0x1229,0x081490,0x1490)
USER32(NtUserSetSysColors                                ,  NONE,0x1217,0x1213,0x1282,0x1282,0x122c,0x12e8,0x12e8,0x1236,0x1312,0x1312,0x1247,0x1113ae,0x13ae,0x124a,0x1113df,0x13df,0x1250,0x111438,0x1438,0x1252,0x11143c,0x143c,0x1258,0x11143e,0x143e,0x121a,0x11143b,0x143b,0x1220,0x111460,0x1460,0x1228,0x111491,0x1491)
USER32(NtUserSetSystemCursor                             ,  NONE,0x1218,0x1214,0x1283,0x1283,0x122d,0x12e9,0x12e9,0x1237,0x1313,0x1313,0x1246,0x0513af,0x13af,0x1249,0x0513e0,0x13e0,0x124f,0x051439,0x1439,0x1251,0x05143d,0x143d,0x1257,0x05143f,0x143f,0x1219,0x05143c,0x143c,0x121f,0x051461,0x1461,0x1227,0x051492,0x1492)
USER32(NtUserSetSystemMenu                               ,0x1206,0x1219,0x1215,0x110d,0x110d,0x122e,0x110e,0x110e,0x1238,0x1107,0x1107,0x1245,0x081107,0x1107,0x1248,0x081108,0x1108,0x124e,0x081109,0x1109,0x1250,0x081109,0x1109,0x1256,0x081108,0x1108,0x1218,0x0810fd,0x10fd,0x121e,0x0810fd,0x10fd,0x1226,0x0810fd,0x10fd)
USER32(NtUserSetSystemTimer                              ,0x1207,0x121a,0x1216,0x1284,0x1284,0x122f,0x12ea,0x12ea,0x1239,0x1314,0x1314,0x1244,0x0013b0,0x13b0,0x1247,0x0013e1,0x13e1,0x124d,0x00143a,0x143a,0x124f,0x00143e,0x143e,0x1255,0x001440,0x1440,0x1217,0x00143d,0x143d,0x121d,0x001462,0x1462,0x1225,0x001493,0x1493)
USER32(NtUserSetTargetForResourceBrokering               ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1217,0x001463,0x1463,0x121f,0x001494,0x1494)
USER32(NtUserSetThreadDesktop                            ,0x1208,0x121b,0x1217,0x1092,0x1092,0x1230,0x1093,0x1093,0x123a,0x1091,0x1091,0x1242,0x041091,0x1091,0x1245,0x041092,0x1092,0x124b,0x041093,0x1093,0x124d,0x041093,0x1093,0x1253,0x041093,0x1093,0x1215,0x041091,0x1091,0x121b,0x041091,0x1091,0x1223,0x041091,0x1091)
USER32(NtUserSetThreadInputBlocked                       ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x1243,0x0013b1,0x13b1,0x1246,0x0013e2,0x13e2,0x124c,0x00143b,0x143b,0x124e,0x00143f,0x143f,0x1254,0x001441,0x1441,0x1216,0x00143e,0x143e,0x121c,0x001464,0x1464,0x1224,0x001495,0x1495)
USER32(NtUserSetThreadState                              ,  NONE,0x121d,0x1219,0x10dd,0x10dd,0x1232,0x10de,0x10de,0x123c,0x10d9,0x10d9,0x1240,0x0510d9,0x10d9,0x1243,0x0510da,0x10da,0x1249,0x0510db,0x10db,0x124b,0x0510db,0x10db,0x1251,0x0510da,0x10da,0x1213,0x0510cf,0x10cf,0x1219,0x0510cf,0x10cf,0x1221,0x0510cf,0x10cf)
USER32(NtUserSetTimer                                    ,0x120b,0x121e,0x121a,0x1017,0x1017,0x1233,0x1018,0x1018,0x123d,0x1018,0x1018,0x123f,0x001019,0x1019,0x1242,0x00101a,0x101a,0x1248,0x00101b,0x101b,0x124a,0x00101b,0x101b,0x1250,0x00101b,0x101b,0x1212,0x00101b,0x101b,0x1218,0x00101b,0x101b,0x1220,0x00101b,0x101b)
USER32(NtUserSetWinEventHook                             ,  NONE,0x1228,0x1224,0x1109,0x1109,0x1240,0x110a,0x110a,0x124c,0x1104,0x1104,0x122c,0x001104,0x1104,0x122f,0x001105,0x1105,0x1232,0x001106,0x1106,0x1234,0x001106,0x1106,0x1238,0x001105,0x1105,0x11f7,0x0010fa,0x10fa,0x11fc,0x0010fa,0x10fa,0x1204,0x0010fa,0x10fa)
USER32(NtUserSetWindowArrangement                        ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1442,0x00143d,0x143d,0x1445,0x001441,0x1441,0x1441,0x001443,0x1443,0x141c,0x001440,0x1440,0x143d,0x001466,0x1466,0x1463,0x001497,0x1497)
USER32(NtUserSetWindowBand                               ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x123c,0x0013b3,0x13b3,0x123f,0x0013e4,0x13e4,0x1242,0x00143e,0x143e,0x1244,0x001442,0x1442,0x1248,0x001444,0x1444,0x1207,0x001441,0x1441,0x120c,0x001467,0x1467,0x1214,0x001498,0x1498)
USER32(NtUserSetWindowCompositionAttribute               ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x123f,0x1316,0x1316,0x123b,0x0013b4,0x13b4,0x123e,0x0013e5,0x13e5,0x1241,0x00143f,0x143f,0x1243,0x001443,0x1443,0x1247,0x001445,0x1445,0x1206,0x001442,0x1442,0x120b,0x001468,0x1468,0x1213,0x001499,0x1499)
USER32(NtUserSetWindowCompositionTransition              ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x1239,0x0013b5,0x13b5,0x123c,0x0013e6,0x13e6,0x123f,0x001440,0x1440,0x1241,0x001444,0x1444,0x1245,0x001446,0x1446,0x1204,0x001443,0x1443,0x1209,0x001469,0x1469,0x1211,0x00149a,0x149a)
USER32(NtUserSetWindowDisplayAffinity                    ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x1240,0x1317,0x1317,0x1238,0x0013b6,0x13b6,0x123b,0x0013e7,0x13e7,0x123e,0x001441,0x1441,0x1240,0x001445,0x1445,0x1244,0x001447,0x1447,0x1203,0x001444,0x1444,0x1208,0x00146a,0x146a,0x1210,0x00149b,0x149b)
USER32(NtUserSetWindowFNID                               ,  NONE,0x121f,0x121b,0x1096,0x1096,0x1235,0x1097,0x1097,0x1241,0x1095,0x1095,0x1237,0x071095,0x1095,0x123a,0x071096,0x1096,0x123d,0x071097,0x1097,0x123f,0x071097,0x1097,0x1243,0x071097,0x1097,0x1202,0x071095,0x1095,0x1207,0x071095,0x1095,0x120f,0x071095,0x1095)
USER32(NtUserSetWindowFeedbackSetting                    ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x12bd,0x0013b7,0x13b7,0x12c0,0x0013e8,0x13e8,0x12ca,0x001442,0x1442,0x12cb,0x001446,0x1446,0x12d1,0x001448,0x1448,0x1292,0x001445,0x1445,0x129b,0x00146b,0x146b,0x12a3,0x00149c,0x149c)
USER32(NtUserSetWindowGroup                              ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x146a,0x00149d,0x149d)
USER32(NtUserSetWindowLong                               ,  NONE,0x1220,0x121c,0x105b,0x105b,0x1236,0x105c,0x105c,0x1242,0x105c,0x105c,0x1236,0x00105c,0x105c,0x1239,0x00105d,0x105d,0x123c,0x00105e,0x105e,0x123e,0x00105e,0x105e,0x1242,0x00105e,0x105e,0x1201,0x00105e,0x105e,0x1206,0x00105e,0x105e,0x120e,0x00105e,0x105e)
USER32(NtUserSetWindowLongPtr                            ,  NONE,  NONE,  NONE,  NONE,0x1298,  NONE,  NONE,0x1305,  NONE,  NONE,0x133a,  NONE,    NONE,0x13d9,  NONE,    NONE,0x140d,  NONE,    NONE,0x1469,  NONE,    NONE,0x146e,  NONE,    NONE,0x1471,  NONE,    NONE,0x146d,  NONE,    NONE,0x1497,  NONE,    NONE,0x14c9)
USER32(NtUserSetWindowPlacement                          ,0x120e,0x1221,0x121d,0x10e6,0x10e6,0x1237,0x10e7,0x10e7,0x1243,0x10e2,0x10e2,0x1235,0x0710e2,0x10e2,0x1238,0x0710e3,0x10e3,0x123b,0x0710e4,0x10e4,0x123d,0x0710e4,0x10e4,0x1241,0x0710e3,0x10e3,0x1200,0x0710d8,0x10d8,0x1205,0x0710d8,0x10d8,0x120d,0x0710d8,0x10d8)
USER32(NtUserSetWindowPos                                ,0x120f,0x1222,0x121e,0x1023,0x1023,0x1238,0x1024,0x1024,0x1244,0x1024,0x1024,0x1234,0x001025,0x1025,0x1237,0x001026,0x1026,0x123a,0x001027,0x1027,0x123c,0x001027,0x1027,0x1240,0x001027,0x1027,0x11ff,0x001027,0x1027,0x1204,0x001027,0x1027,0x120c,0x001027,0x1027)
USER32(NtUserSetWindowRgn                                ,  NONE,0x1223,0x121f,0x10ce,0x10ce,0x1239,0x10cf,0x10cf,0x1245,0x10ca,0x10ca,0x1233,0x0c10ca,0x10ca,0x1236,0x0c10cb,0x10cb,0x1239,0x0c10cc,0x10cc,0x123b,0x0c10cc,0x10cc,0x123f,0x0c10cc,0x10cc,0x11fe,0x0c10c1,0x10c1,0x1203,0x0c10c1,0x10c1,0x120b,0x0c10c1,0x10c1)
USER32(NtUserSetWindowRgnEx                              ,  NONE,  NONE,  NONE,  NONE,  NONE,0x123b,0x12ec,0x12ec,0x1247,0x1318,0x1318,0x1231,0x0013b8,0x13b8,0x1234,0x0013e9,0x13e9,0x1237,0x001443,0x1443,0x1239,0x001447,0x1447,0x123d,0x001449,0x1449,0x11fc,0x001446,0x1446,0x1201,0x00146c,0x146c,0x1209,0x00149e,0x149e)
USER32(NtUserSetWindowShowState                          ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1443,0x001444,0x1444,0x1446,0x001448,0x1448,0x1442,0x00144a,0x144a,0x141d,0x001447,0x1447,0x143e,0x00146d,0x146d,0x1464,0x00149f,0x149f)
USER32(NtUserSetWindowStationUser                        ,  NONE,0x1226,0x1222,0x1286,0x1286,0x123e,0x12ed,0x12ed,0x124a,0x1319,0x1319,0x122e,0x0013b9,0x13b9,0x1231,0x0013ea,0x13ea,0x1234,0x001445,0x1445,0x1236,0x001449,0x1449,0x123a,0x00144b,0x144b,0x11f9,0x001448,0x1448,0x11fe,0x00146e,0x146e,0x1206,0x0014a0,0x14a0)
USER32(NtUserSetWindowWord                               ,0x1214,0x1227,0x1223,0x10ec,0x10ec,0x123f,0x10ed,0x10ed,0x124b,0x10e8,0x10e8,0x122d,0x0010e8,0x10e8,0x1230,0x0010e9,0x10e9,0x1233,0x0010ea,0x10ea,0x1235,0x0010ea,0x10ea,0x1239,0x0010e9,0x10e9,0x11f8,0x0010de,0x10de,0x11fd,0x0010de,0x10de,0x1205,0x0010de,0x10de)
USER32(NtUserSetWindowsHookAW                            ,  NONE,0x1224,0x1220,0x1106,0x1106,0x123c,0x1107,0x1107,0x1248,0x1101,0x1101,0x1230,0x001101,0x1101,0x1233,0x001102,0x1102,0x1236,0x001103,0x1103,0x1238,0x001103,0x1103,0x123c,0x001102,0x1102,0x11fb,0x0010f7,0x10f7,0x1200,0x0010f7,0x10f7,0x1208,0x0010f7,0x10f7)
USER32(NtUserSetWindowsHookEx                            ,  NONE,0x1225,0x1221,0x108d,0x108d,0x123d,0x108e,0x108e,0x1249,0x108c,0x108c,0x122f,0x00108c,0x108c,0x1232,0x00108d,0x108d,0x1235,0x00108e,0x108e,0x1237,0x00108e,0x108e,0x123b,0x00108e,0x108e,0x11fa,0x00108c,0x108c,0x11ff,0x00108c,0x108c,0x1207,0x00108c,0x108c)
USER32(NtUserSfmDestroyLogicalSurfaceBinding             ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x128b,0x131a,0x131a,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
USER32(NtUserSfmDxBindSwapChain                          ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x1282,0x131b,0x131b,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
USER32(NtUserSfmDxGetSwapChainStats                      ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x1288,0x131c,0x131c,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
USER32(NtUserSfmDxOpenSwapChain                          ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x1283,0x131d,0x131d,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
USER32(NtUserSfmDxQuerySwapChainBindingStatus            ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x1286,0x131e,0x131e,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
USER32(NtUserSfmDxReleaseSwapChain                       ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x1284,0x131f,0x131f,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
USER32(NtUserSfmDxReportPendingBindingsToDwm             ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x1287,0x1320,0x1320,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
USER32(NtUserSfmDxSetSwapChainBindingStatus              ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x1285,0x1321,0x1321,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
USER32(NtUserSfmDxSetSwapChainStats                      ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x1289,0x1322,0x1322,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
USER32(NtUserSfmGetLogicalSurfaceBinding                 ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x128a,0x1323,0x1323,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
USER32(NtUserShowCaret                                   ,0x1216,0x1229,0x1225,0x1024,0x1024,0x1241,0x1025,0x1025,0x124d,0x1025,0x1025,0x122b,0x031026,0x1026,0x122e,0x031027,0x1027,0x1231,0x031028,0x1028,0x1233,0x031028,0x1028,0x1237,0x031028,0x1028,0x11f6,0x031028,0x1028,0x11fa,0x031028,0x1028,0x1201,0x031028,0x1028)
USER32(NtUserShowCursor                                  ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x11fb,0x00146f,0x146f,0x1202,0x0014a1,0x14a1)
USER32(NtUserShowScrollBar                               ,0x1217,0x122a,0x1226,0x103c,0x103c,0x1242,0x103d,0x103d,0x124e,0x103d,0x103d,0x122a,0x0c103e,0x103e,0x122d,0x0c103f,0x103f,0x1230,0x0c1040,0x1040,0x1232,0x0c1040,0x1040,0x1236,0x0c1040,0x1040,0x11f5,0x0c1040,0x1040,0x11f9,0x0c1040,0x1040,0x1200,0x0c1040,0x1040)
USER32(NtUserShowSystemCursor                            ,  NONE,  NONE,  NONE,  NONE,  NONE,0x1303,0x12ee,0x12ee,0x1333,0x1324,0x1324,0x1378,0x0013ba,0x13ba,0x138a,0x0013eb,0x13eb,0x13c3,0x001446,0x1446,0x13c5,0x00144a,0x144a,0x13d4,0x00144c,0x144c,0x13a6,0x001449,0x1449,0x13c0,0x001470,0x1470,0x13d2,0x0014a2,0x14a2)
USER32(NtUserShowWindow                                  ,0x1218,0x122b,0x1227,0x1057,0x1057,0x1243,0x1058,0x1058,0x124f,0x1058,0x1058,0x1229,0x081058,0x1058,0x122c,0x081059,0x1059,0x122f,0x08105a,0x105a,0x1231,0x08105a,0x105a,0x1235,0x08105a,0x105a,0x11f4,0x08105a,0x105a,0x11f8,0x08105a,0x105a,0x11ff,0x08105a,0x105a)
USER32(NtUserShowWindowAsync                             ,0x1219,0x122c,0x1228,0x1122,0x1122,0x1244,0x1123,0x1123,0x1250,0x111b,0x111b,0x1228,0x08111b,0x111b,0x122b,0x08111c,0x111c,0x122e,0x08111d,0x111d,0x1230,0x08111d,0x111d,0x1234,0x08111c,0x111c,0x11f3,0x081111,0x1111,0x11f7,0x081111,0x1111,0x11fe,0x081111,0x1111)
USER32(NtUserShutdownBlockReasonCreate                   ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x2067,  NONE,  NONE,0x2060,  NONE,0x1227,0x0013bb,0x13bb,0x122a,0x0013ec,0x13ec,0x122d,0x001447,0x1447,0x122f,0x00144b,0x144b,0x1233,0x00144d,0x144d,0x11f2,0x00144a,0x144a,0x11f6,0x001471,0x1471,0x11fd,0x0014a3,0x14a3)
USER32(NtUserShutdownBlockReasonDestroy                  ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x2069,  NONE,  NONE,0x2062,  NONE,0x1225,0x0013bd,0x13bd,0x1228,0x0013ee,0x13ee,0x122b,0x001449,0x1449,0x122d,0x00144d,0x144d,0x1231,0x00144f,0x144f,0x11f0,0x00144c,0x144c,0x11f4,0x001473,0x1473,0x11fb,0x0014a5,0x14a5)
USER32(NtUserShutdownBlockReasonQuery                    ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x2068,  NONE,  NONE,0x2061,  NONE,0x1226,0x0013bc,0x13bc,0x1229,0x0013ed,0x13ed,0x122c,0x001448,0x1448,0x122e,0x00144c,0x144c,0x1232,0x00144e,0x144e,0x11f1,0x00144b,0x144b,0x11f5,0x001472,0x1472,0x11fc,0x0014a4,0x14a4)
USER32(NtUserSignalRedirectionStartComplete              ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x13be,0x0013be,0x13be,0x13e6,0x0013ef,0x13ef,0x1431,0x00144a,0x144a,0x1434,0x00144e,0x144e,0x1430,0x001450,0x1450,0x1405,0x00144d,0x144d,0x1426,0x001474,0x1474,0x144a,0x0014a6,0x14a6)
USER32(NtUserSlicerControl                               ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x137d,0x0013bf,0x13bf,0x138f,0x0013f0,0x13f0,0x13c8,0x00144b,0x144b,0x13ca,0x00144f,0x144f,0x13d9,0x001451,0x1451,0x13ab,0x00144e,0x144e,0x13c6,0x001475,0x1475,0x13d8,0x0014a7,0x14a7)
USER32(NtUserSoundSentry                                 ,  NONE,  NONE,  NONE,  NONE,  NONE,0x1245,0x12ef,0x12ef,0x1251,0x1325,0x1325,0x1224,0x0013c0,0x13c0,0x1227,0x0013f1,0x13f1,0x122a,0x00144c,0x144c,0x122c,0x001450,0x1450,0x1230,0x001452,0x1452,0x11ef,0x00144f,0x144f,0x11f3,0x001476,0x1476,0x11fa,0x0014a8,0x14a8)
USER32(NtUserStopAndEndInertia                           ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1251,0x001477,0x1477,0x1259,0x0014a9,0x14a9)
USER32(NtUserSwitchDesktop                               ,0x121b,0x122e,0x122a,0x1288,0x1288,0x1246,0x12f0,0x12f0,0x1252,0x1326,0x1326,0x1223,0x0513c1,0x13c1,0x1226,0x0513f2,0x13f2,0x1229,0x05144d,0x144d,0x122b,0x051451,0x1451,0x122f,0x051453,0x1453,0x11ee,0x051450,0x1450,0x11f2,0x051478,0x1478,0x11f9,0x0514aa,0x14aa)
USER32(NtUserSystemParametersInfo                        ,  NONE,0x122f,0x122b,0x1041,0x1041,0x1247,0x1042,0x1042,0x1253,0x1042,0x1042,0x1222,0x001043,0x1043,0x1225,0x001044,0x1044,0x1228,0x001045,0x1045,0x122a,0x001045,0x1045,0x122d,0x001045,0x1045,0x11ec,0x001045,0x1045,0x11f0,0x001045,0x1045,0x11f7,0x001045,0x1045)
USER32(NtUserSystemParametersInfoForDpi                  ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x122e,0x001454,0x1454,0x11ed,0x001451,0x1451,0x11f1,0x001479,0x1479,0x11f8,0x0014ab,0x14ab)
USER32(NtUserTestForInteractiveUser                      ,  NONE,  NONE,  NONE,0x1289,0x1289,0x1248,0x12f1,0x12f1,0x1254,0x1327,0x1327,0x1220,0x0413c2,0x13c2,0x1223,0x0413f3,0x13f3,0x1226,0x04144e,0x144e,0x1228,0x041452,0x1452,0x122b,0x041455,0x1455,0x11ea,0x041452,0x1452,0x11ee,0x04147a,0x147a,0x11f5,0x0414ac,0x14ac)
USER32(NtUserThunkedMenuInfo                             ,  NONE,0x1231,0x122d,0x110e,0x110e,0x1249,0x110f,0x110f,0x1255,0x1108,0x1108,0x121f,0x001108,0x1108,0x1222,0x001109,0x1109,0x1225,0x00110a,0x110a,0x1227,0x00110a,0x110a,0x122a,0x001109,0x1109,0x11e9,0x0010fe,0x10fe,0x11ed,0x0010fe,0x10fe,0x11f4,0x0010fe,0x10fe)
USER32(NtUserThunkedMenuItemInfo                         ,  NONE,0x1232,0x122e,0x1099,0x1099,0x124a,0x109a,0x109a,0x1256,0x1098,0x1098,0x121e,0x001098,0x1098,0x1221,0x001099,0x1099,0x1224,0x00109a,0x109a,0x1226,0x00109a,0x109a,0x1229,0x00109a,0x109a,0x11e8,0x001098,0x1098,0x11ec,0x001098,0x1098,0x11f3,0x001098,0x1098)
USER32(NtUserToUnicodeEx                                 ,  NONE,0x1233,0x122f,0x107a,0x107a,0x124b,0x107b,0x107b,0x1257,0x1079,0x1079,0x121d,0x001079,0x1079,0x1220,0x00107a,0x107a,0x1223,0x00107b,0x107b,0x1225,0x00107b,0x107b,0x1228,0x00107b,0x107b,0x11e7,0x00107a,0x107a,0x11eb,0x00107a,0x107a,0x11f2,0x00107a,0x107a)
USER32(NtUserTrackMouseEvent                             ,0x1221,0x1234,0x1230,0x10df,0x10df,0x124c,0x10e0,0x10e0,0x1258,0x10db,0x10db,0x121c,0x0010db,0x10db,0x121f,0x0010dc,0x10dc,0x1222,0x0010dd,0x10dd,0x1224,0x0010dd,0x10dd,0x1227,0x0010dc,0x10dc,0x11e6,0x0010d1,0x10d1,0x11ea,0x0010d1,0x10d1,0x11f1,0x0010d1,0x10d1)
USER32(NtUserTrackPopupMenuEx                            ,0x1222,0x1235,0x1231,0x128a,0x128a,0x124d,0x12f2,0x12f2,0x1259,0x1328,0x1328,0x121b,0x0013c3,0x13c3,  NONE,0x0013f4,0x13f4,0x1221,0x00144f,0x144f,0x1223,0x001453,0x1453,0x1226,0x001456,0x1456,0x11e5,0x001453,0x1453,0x11e9,0x00147b,0x147b,0x11f0,0x0014ad,0x14ad)
USER32(NtUserTransformPoint                              ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,0x1403,0x0013f5,0x13f5,0x1450,0x001450,0x1450,0x1454,0x001454,0x1454,0x1450,0x001457,0x1457,0x142b,0x001454,0x1454,0x144c,0x00147c,0x147c,0x1477,0x0014ae,0x14ae)
USER32(NtUserTransformRect                               ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,0x1404,0x0013f6,0x13f6,0x1451,0x001451,0x1451,0x1455,0x001455,0x1455,0x1451,0x001458,0x1458,0x142c,0x001455,0x1455,0x144d,0x00147d,0x147d,0x1478,0x0014af,0x14af)
USER32(NtUserTranslateAccelerator                        ,  NONE,0x1238,0x1234,0x1010,0x1010,0x1250,0x1011,0x1011,0x125d,0x1011,0x1011,0x1217,0x001012,0x1012,0x121a,0x001013,0x1013,0x121d,0x001014,0x1014,0x121f,0x001014,0x1014,0x1222,0x001014,0x1014,0x11e1,0x001014,0x1014,0x11e5,0x001014,0x1014,0x11ec,0x001014,0x1014)
USER32(NtUserTranslateMessage                            ,  NONE,0x1239,0x1235,0x100d,0x100d,0x1251,0x100d,0x100d,0x125e,0x100d,0x100d,0x1216,0x00100e,0x100e,0x1219,0x00100f,0x100f,0x121c,0x001010,0x1010,0x121e,0x001010,0x1010,0x1221,0x001010,0x1010,0x11e0,0x001010,0x1010,0x11e4,0x001010,0x1010,0x11eb,0x001010,0x1010)
USER32(NtUserUndelegateInput                             ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x1279,0x0013c4,0x13c4,0x127c,0x0013f7,0x13f7,0x1286,0x001452,0x1452,0x1287,0x001456,0x1456,0x128d,0x001459,0x1459,0x124f,0x001456,0x1456,0x1257,0x00147e,0x147e,0x125f,0x0014b0,0x14b0)
USER32(NtUserUnhookWinEvent                              ,0x1226,0x123b,0x1237,0x110a,0x110a,0x1253,0x110b,0x110b,0x1260,0x1105,0x1105,0x129e,0x001105,0x1105,0x12a1,0x001106,0x1106,0x12ab,0x001107,0x1107,0x12ac,0x001107,0x1107,0x12b2,0x001106,0x1106,0x1273,0x0010fb,0x10fb,0x127c,0x0010fb,0x10fb,0x1284,0x0010fb,0x10fb)
USER32(NtUserUnhookWindowsHookEx                         ,0x1225,0x123a,0x1236,0x1070,0x1070,0x1252,0x1071,0x1071,0x125f,0x1070,0x1070,0x129f,0x001070,0x1070,0x12a2,0x001071,0x1071,0x12ac,0x001072,0x1072,0x12ad,0x001072,0x1072,0x12b3,0x001072,0x1072,0x1274,0x001071,0x1071,0x127d,0x001071,0x1071,0x1285,0x001071,0x1071)
USER32(NtUserUnloadKeyboardLayout                        ,  NONE,0x123c,0x1238,0x128b,0x128b,0x1254,0x12f3,0x12f3,0x1261,0x1329,0x1329,0x129d,0x0313c5,0x13c5,0x12a0,0x0313f8,0x13f8,0x12aa,0x031453,0x1453,0x12ab,0x031457,0x1457,0x12b1,0x03145a,0x145a,0x1272,0x031457,0x1457,0x127b,0x03147f,0x147f,0x1283,0x0314b1,0x14b1)
USER32(NtUserUnlockWindowStation                         ,0x1228,0x123d,0x1239,0x128c,0x128c,0x1255,0x12f4,0x12f4,0x1262,0x132a,0x132a,0x129c,0x0413c6,0x13c6,0x129f,0x0413f9,0x13f9,0x12a9,0x041454,0x1454,0x12aa,0x041458,0x1458,0x12b0,0x04145b,0x145b,0x1271,0x041458,0x1458,0x127a,0x041480,0x1480,0x1282,0x0414b2,0x14b2)
USER32(NtUserUnregisterClass                             ,  NONE,0x123e,0x123a,0x10bf,0x10bf,0x1256,0x10c0,0x10c0,0x1263,0x10bb,0x10bb,0x129b,0x0010bb,0x10bb,0x129e,0x0010bc,0x10bc,0x12a8,0x0010bd,0x10bd,0x12a9,0x0010bd,0x10bd,0x12af,0x0010bd,0x10bd,0x1270,0x0010b6,0x10b6,0x1279,0x0010b6,0x10b6,0x1281,0x0010b6,0x10b6)
USER32(NtUserUnregisterHotKey                            ,0x122a,0x1240,0x123c,0x128d,0x128d,0x1258,0x12f5,0x12f5,0x1265,0x132b,0x132b,0x1299,0x0813c7,0x13c7,0x129c,0x0813fa,0x13fa,0x12a6,0x081455,0x1455,0x12a7,0x081459,0x1459,0x12ad,0x08145c,0x145c,0x126e,0x081459,0x1459,0x1277,0x081481,0x1481,0x127f,0x0814b3,0x14b3)
USER32(NtUserUnregisterSessionPort                       ,  NONE,  NONE,  NONE,  NONE,  NONE,0x1271,0x12f6,0x12f6,0x127d,0x132c,0x132c,0x1281,0x0013c8,0x13c8,0x1284,0x0013fb,0x13fb,0x128e,0x001456,0x1456,0x128f,0x00145a,0x145a,0x1295,0x00145d,0x145d,0x1256,0x00145a,0x145a,0x125e,0x001482,0x1482,0x1266,0x0014b4,0x14b4)
USER32(NtUserUnregisterUserApiHook                       ,  NONE,0x123f,0x123b,0x128e,0x128e,0x1257,0x12f7,0x12f7,0x1264,0x132d,0x132d,0x129a,0x0013c9,0x13c9,0x129d,0x0013fc,0x13fc,0x12a7,0x001457,0x1457,0x12a8,0x00145b,0x145b,0x12ae,0x00145e,0x145e,0x126f,0x00145b,0x145b,0x1278,0x001483,0x1483,0x1280,0x0014b5,0x14b5)
USER32(NtUserUpdateDefaultDesktopThumbnail               ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x123a,0x0013ca,0x13ca,0x123d,0x0013fd,0x13fd,0x1240,0x001458,0x1458,0x1242,0x00145c,0x145c,0x1246,0x00145f,0x145f,0x1205,0x00145c,0x145c,0x120a,0x001484,0x1484,0x1212,0x0014b6,0x14b6)
USER32(NtUserUpdateInputContext                          ,0x122b,0x1241,0x123d,0x128f,0x128f,0x1259,0x12f8,0x12f8,0x1266,0x132e,0x132e,0x1298,0x0013cb,0x13cb,0x129b,0x0013fe,0x13fe,0x12a5,0x001459,0x1459,0x12a6,0x00145d,0x145d,0x12ac,0x001460,0x1460,0x126d,0x00145d,0x145d,0x1276,0x001485,0x1485,0x127e,0x0014b7,0x14b7)
USER32(NtUserUpdateInstance                              ,  NONE,0x1242,0x123e,0x1290,0x1290,0x125a,0x12f9,0x12f9,0x1267,0x132f,0x132f,0x1297,0x0c13cc,0x13cc,0x129a,0x0c13ff,0x13ff,0x12a4,0x0c145a,0x145a,0x12a5,0x0c145e,0x145e,0x12ab,0x0c1461,0x1461,0x126c,0x0c145e,0x145e,0x1275,0x0c1486,0x1486,0x127d,0x0c14b8,0x14b8)
USER32(NtUserUpdateLayeredWindow                         ,0x122d,0x1243,0x123f,0x1291,0x1291,0x125b,0x12fa,0x12fa,0x1268,0x1330,0x1330,0x1296,0x0013cd,0x13cd,0x1299,0x001400,0x1400,0x12a3,0x00145b,0x145b,0x12a4,0x00145f,0x145f,0x12aa,0x001462,0x1462,0x126b,0x00145f,0x145f,0x1274,0x001487,0x1487,0x127c,0x0014b9,0x14b9)
USER32(NtUserUpdatePerUserSystemParameters               ,  NONE,0x1246,0x1242,0x1292,0x1292,0x125e,0x12fb,0x12fb,0x126b,0x1331,0x1331,0x1293,0x0713ce,0x13ce,0x1296,0x071401,0x1401,0x12a0,0x07145c,0x145c,0x12a1,0x071460,0x1460,0x12a7,0x071463,0x1463,0x1268,0x071460,0x1460,0x1271,0x071488,0x1488,0x1279,0x0714ba,0x14ba)
USER32(NtUserUpdateWindowInputSinkHints                  ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,0x1402,0x001402,0x1402,0x144f,0x00145d,0x145d,0x1453,0x001461,0x1461,0x144f,0x001464,0x1464,0x142a,0x001461,0x1461,0x144b,0x001489,0x1489,0x1476,0x0014bb,0x14bb)
USER32(NtUserUpdateWindowTrackingInfo                    ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1447,0x001462,0x1462,0x1443,0x001465,0x1465,0x141e,0x001462,0x1462,0x143f,0x00148a,0x148a,0x1465,0x0014bc,0x14bc)
USER32(NtUserUpdateWindowTransform                       ,  NONE,  NONE,  NONE,  NONE,  NONE,0x1272,0x12fc,0x12fc,0x127e,0x1332,0x1332,0x127f,0x0013cf,0x13cf,0x1282,0x001403,0x1403,0x128c,0x00145e,0x145e,0x128d,0x001463,0x1463,0x1293,0x001466,0x1466,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
USER32(NtUserUserConnectToServer                         ,  NONE,  NONE,  NONE,0x205e,  NONE,  NONE,0x2066,  NONE,  NONE,0x205f,  NONE,  NONE,0x002004,  NONE,  NONE,0x002000,  NONE,  NONE,0x002000,  NONE,  NONE,0x002000,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
USER32(NtUserUserHandleGrantAccess                       ,0x1230,0x1247,0x1243,0x1293,0x1293,0x125f,0x12fd,0x12fd,0x126c,0x1333,0x1333,0x1292,0x0e13d0,0x13d0,0x1295,0x0e1404,0x1404,0x129f,0x0e145f,0x145f,0x12a0,0x0e1464,0x1464,0x12a6,0x0e1467,0x1467,0x1267,0x0e1463,0x1463,0x1270,0x0e148b,0x148b,0x1278,0x0e14bd,0x14bd)
USER32(NtUserValidateHandleSecure                        ,0x1231,0x1248,0x1244,0x1294,0x1294,0x1260,0x12fe,0x12fe,0x126d,0x1334,0x1334,0x1291,0x0313d1,0x13d1,0x1294,0x031405,0x1405,0x129e,0x031460,0x1460,0x129f,0x031465,0x1465,0x12a5,0x031468,0x1468,0x1266,0x031464,0x1464,0x126f,0x03148c,0x148c,0x1277,0x0314be,0x14be)
USER32(NtUserValidateRect                                ,0x1232,0x1249,0x1245,0x10d1,0x10d1,0x1261,0x10d2,0x10d2,0x126e,0x10cd,0x10cd,0x1290,0x0710cd,0x10cd,0x1293,0x0710ce,0x10ce,0x129d,0x0710cf,0x10cf,0x129e,0x0710cf,0x10cf,0x12a4,0x0710cf,0x10cf,0x1265,0x0710c4,0x10c4,0x126e,0x0710c4,0x10c4,0x1276,0x0710c4,0x10c4)
USER32(NtUserValidateTimerCallback                       ,  NONE,0x124a,0x1246,0x1015,0x1015,0x1262,0x1016,0x1016,0x126f,0x1016,0x1016,0x128f,0x041017,0x1017,0x1292,0x041018,0x1018,0x129c,0x041019,0x1019,0x129d,0x041019,0x1019,0x12a3,0x041019,0x1019,0x1264,0x041019,0x1019,0x126d,0x041019,0x1019,0x1275,0x041019,0x1019)
USER32(NtUserVkKeyScanEx                                 ,  NONE,0x124b,0x1247,0x1027,0x1027,0x1263,0x1028,0x1028,0x1270,0x1028,0x1028,0x128e,0x0f1029,0x1029,0x1291,0x0f102a,0x102a,0x129b,0x0f102b,0x102b,0x129c,0x0f102b,0x102b,0x12a2,0x0f102b,0x102b,0x1263,0x0f102b,0x102b,0x126c,0x0f102b,0x102b,0x1274,0x0f102b,0x102b)
USER32(NtUserWOWCleanup                                  ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1425,0x00148d,0x148d,0x1449,0x0014bf,0x14bf)
USER32(NtUserWaitAvailableMessageEx                      ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x1288,0x0013d2,0x13d2,0x128b,0x001406,0x1406,0x1295,0x001461,0x1461,0x1296,0x001466,0x1466,0x129c,0x001469,0x1469,0x125d,0x001465,0x1465,0x1265,0x00148e,0x148e,0x126d,0x0014c0,0x14c0)
USER32(NtUserWaitForInputIdle                            ,  NONE,0x124c,0x1248,0x1295,0x1295,0x1264,0x12ff,0x12ff,0x1271,0x1335,0x1335,0x128d,0x0c13d3,0x13d3,0x1290,0x0c1407,0x1407,0x129a,0x0c1462,0x1462,0x129b,0x0c1467,0x1467,0x12a1,0x0c146a,0x146a,0x1262,0x0c1466,0x1466,0x126b,0x0c148f,0x148f,0x1273,0x0c14c1,0x14c1)
USER32(NtUserWaitForMsgAndEvent                          ,  NONE,0x124d,0x1249,0x1296,0x1296,0x1265,0x1300,0x1300,0x1272,0x1336,0x1336,0x128c,0x0313d4,0x13d4,0x128f,0x031408,0x1408,0x1299,0x031463,0x1463,0x129a,0x031468,0x1468,0x12a0,0x03146b,0x146b,0x1261,0x031467,0x1467,0x126a,0x031490,0x1490,0x1272,0x0314c2,0x14c2)
USER32(NtUserWaitForRedirectionStartComplete             ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x13bd,0x0013d5,0x13d5,0x13e5,0x001409,0x1409,0x1430,0x001464,0x1464,0x1433,0x001469,0x1469,0x142f,0x00146c,0x146c,0x1404,0x001468,0x1468,0x1424,0x001491,0x1491,0x1448,0x0014c3,0x14c3)
USER32(NtUserWaitMessage                                 ,0x1236,0x124e,0x124a,0x100c,0x100c,0x1266,0x100c,0x100c,0x1273,0x100c,0x100c,0x128b,0x01100d,0x100d,0x128e,0x01100e,0x100e,0x1298,0x01100f,0x100f,0x1299,0x01100f,0x100f,0x129f,0x01100f,0x100f,0x1260,0x01100f,0x100f,0x1269,0x01100f,0x100f,0x1271,0x01100f,0x100f)
USER32(NtUserWin32PoolAllocationStats                    ,0x1237,0x124f,0x124b,0x1297,0x1299,0x1267,0x1301,0x1301,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
USER32(NtUserWindowFromDC                                ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1268,0x001492,0x1492,0x1270,0x0014c4,0x14c4)
USER32(NtUserWindowFromPhysicalPoint                     ,  NONE,  NONE,  NONE,  NONE,  NONE,0x1268,0x1302,0x1302,0x1274,0x1337,0x1337,0x128a,0x0013d6,0x13d6,0x128d,0x00140a,0x140a,0x1297,0x001465,0x1465,0x1298,0x00146a,0x146a,0x129e,0x00146d,0x146d,0x125f,0x001469,0x1469,0x1267,0x001493,0x1493,0x126f,0x0014c5,0x14c5)
USER32(NtUserWindowFromPoint                             ,0x1238,0x1250,0x124c,0x1013,0x1013,0x1269,0x1014,0x1014,0x1275,0x1014,0x1014,0x1289,0x001015,0x1015,0x128c,0x001016,0x1016,0x1296,0x001017,0x1017,0x1297,0x001017,0x1017,0x129d,0x001017,0x1017,0x125e,0x001017,0x1017,0x1266,0x001017,0x1017,0x126e,0x001017,0x1017)
USER32(NtUserYieldTask                                   ,  NONE,0x1251,0x124d,0x1298,0x129a,0x126a,0x1303,0x1303,0x1276,0x1338,0x1338,0x1000,0x011000,0x1000,0x1000,0x011000,0x1000,0x1001,0x011001,0x1001,0x1001,0x011001,0x1001,0x1001,0x011001,0x1001,0x1001,0x011001,0x1001,0x1001,0x011001,0x1001,0x1001,0x011001,0x1001)
USER32(NtValidateCompositionSurfaceHandle                ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,0x13d7,  NONE,    NONE,0x140b,0x13cc,0x001466,0x1466,0x13ce,0x00146b,0x146b,0x13dd,0x00146e,0x146e,0x13af,0x00146a,0x146a,0x13ca,0x001494,0x1494,0x13dc,0x0014c6,0x14c6)
USER32(NtVisualCaptureBits                               ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x13dc,0x001467,0x1467,0x13de,0x00146c,0x146c,0x13ef,0x00146f,0x146f,0x13c1,0x00146b,0x146b,0x13dc,0x001495,0x1495,0x13ed,0x0014c7,0x14c7)

/* a number of these duplicate user32 wrappers (see comments at top of file) so not
 * listed here, even though on win2k the imm32 wrapper is not present in user32
 */
IMM32(NtUserAssociateInputContext                       ,0x1128,0x1132,0x1131,0x1225,0x1225,0x113c,0x1275,0x1275,0x113d,0x1282,0x1282,0x116e,0x012dd,0x12dd,0x1170,0x012fe,0x12fe,0x1171,0x001343,0x1343,0x1173,0x001347,0x1347,0x1175,0x00133f,0x133f,0x1133,0x00132c,0x132c,0x1134,0x001348,0x1348,0x1138,0x001368,0x1368)
IMM32(NtUserBuildHimcList                               ,0x112d,0x1137,0x1136,0x1227,0x1227,0x1141,0x1277,0x1277,0x1142,0x1284,0x1284,0x1169,0x012e1,0x12e1,0x116b,0x01302,0x1302,0x116c,0x001347,0x1347,0x116e,0x00134b,0x134b,0x1170,0x001344,0x1344,0x112e,0x001332,0x1332,0x112f,0x00134e,0x134e,0x1132,0x00136e,0x136e)
IMM32(NtUserCreateInputContext                          ,0x114b,0x1155,0x1154,0x122e,0x122e,0x1162,0x1281,0x1281,0x1167,0x1290,0x1290,0x1144,0x012f0,0x12f0,0x1146,0x01313,0x1313,0x1147,0x001358,0x1358,0x1149,0x00135c,0x135c,0x114a,0x001355,0x1355,0x1107,0x001345,0x1345,0x1108,0x001362,0x1362,0x1108,0x001382,0x1382)
IMM32(NtUserDestroyInputContext                         ,0x1157,0x1161,0x1160,0x1233,0x1233,0x116c,0x1284,0x1284,0x1171,0x1293,0x1293,0x1179,0x012f7,0x12f7,0x117b,0x0131a,0x131a,0x117c,0x00135f,0x135f,0x117e,0x001363,0x1363,0x1180,0x00135c,0x135c,0x113e,0x00134c,0x134c,0x113f,0x001369,0x1369,0x1143,0x00138c,0x138c)
IMM32(NtUserGetAppImeLevel                              ,0x1174,0x117e,0x117d,0x123e,0x123e,0x118c,0x1296,0x1296,0x1191,0x12a6,0x12a6,0x11c6,0x31312,0x1312,0x11c7,0x31336,0x1336,0x11c9,0x03137c,0x137c,0x11cb,0x031380,0x1380,0x11cd,0x03137a,0x137a,0x118b,0x03136e,0x136e,0x118e,0x03138c,0x138c,0x1194,0x0313b2,0x13b2)
IMM32(NtUserGetImeInfoEx                                ,0x118f,0x119a,0x1199,0x1246,0x1246,0x11a8,0x129e,0x129e,0x11ad,0x12b2,0x12b2,0x11a6,0x01326,0x1326,0x11a8,0x0134c,0x134c,0x11a9,0x001394,0x1394,0x11ab,0x001398,0x1398,0x11ae,0x001391,0x1391,0x116c,0x001385,0x1385,0x116f,0x0013a6,0x13a6,0x1175,0x0013cc,0x13cc)
IMM32(NtUserQueryInputContext                           ,0x11cf,0x11e0,0x11df,0x1267,0x1267,0x11f6,0x12c9,0x12c9,0x1201,0x12ee,0x12ee,0x11e4,0x0137a,0x137a,0x11e5,0x013a8,0x13a8,0x11e7,0x0013f9,0x13f9,0x11e9,0x0013fd,0x13fd,0x11ec,0x0013fb,0x13fb,0x11aa,0x0013f7,0x13f7,0x11ad,0x001419,0x1419,0x11b4,0x001446,0x1446)
IMM32(NtUserSetAppImeLevel                              ,0x11e5,0x11f8,0x11f6,0x1274,0x1274,0x120f,0x12d9,0x12d9,0x121a,0x1300,0x1300,0x1264,0x71392,0x1392,0x1267,0x713c3,0x13c3,0x126d,0x071418,0x1418,0x126e,0x07141d,0x141d,0x1274,0x07141c,0x141c,0x1236,0x071418,0x1418,0x123d,0x07143b,0x143b,0x1245,0x071469,0x1469)
IMM32(NtUserSetImeInfoEx                                ,0x11f2,0x1205,0x1202,0x1278,0x1278,0x121b,0x12dd,0x12dd,0x1226,0x1307,0x1307,0x1258,0x0139f,0x139f,0x125b,0x013d0,0x13d0,0x1261,0x001428,0x1428,0x1262,0x00142d,0x142d,0x1268,0x00142c,0x142c,0x122a,0x001429,0x1429,0x1230,0x00144e,0x144e,0x1238,0x00147e,0x147e)
IMM32(NtUserSetThreadLayoutHandles                      ,0x1209,0x121c,0x1218,0x1285,0x1285,0x1231,0x12eb,0x12eb,0x123b,0x1315,0x1315,0x1241,0x013b2,0x13b2,0x1244,0x013e3,0x13e3,0x124a,0x00143c,0x143c,0x124c,0x001440,0x1440,0x1252,0x001442,0x1442,0x1214,0x00143f,0x143f,0x121a,0x001465,0x1465,0x1222,0x001496,0x1496)

GDI32(NtGdiAbortDoc                                     ,  NONE,0x1000,0x1000,0x1127,0x1127,0x1000,0x1128,0x1128,0x1000,0x111f,0x111f,0x113c,0x03114c,0x114c,0x113e,0x03115b,0x115b,0x113f,0x031167,0x1167,0x1141,0x031168,0x1168,0x1142,0x031154,0x1154,0x10ff,0x03114c,0x114c,0x1100,0x031154,0x1154,0x1100,0x03116c,0x116c)
GDI32(NtGdiAbortPath                                    ,  NONE,0x1001,0x1001,0x1128,0x1128,0x1001,0x1129,0x1129,0x1001,0x1120,0x1120,0x113b,0x03114d,0x114d,0x113d,0x03115c,0x115c,0x113e,0x031168,0x1168,0x1140,0x031169,0x1169,0x1141,0x031155,0x1155,0x10fe,0x03114d,0x114d,0x10ff,0x031155,0x1155,0x10ff,0x03116d,0x116d)
GDI32(NtGdiAddEmbFontToDC                               ,  NONE,0x10d6,0x10d5,0x1129,0x1129,0x10de,0x112a,0x112a,0x10e0,0x1121,0x1121,0x105a,0x00114e,0x114e,0x105b,0x00115d,0x115d,0x105c,0x001169,0x1169,0x105c,0x00116a,0x116a,0x105c,0x001156,0x1156,0x1063,0x00114e,0x114e,0x1064,0x001156,0x1156,0x1064,0x00116e,0x116e)
GDI32(NtGdiAddFontMemResourceEx                         ,  NONE,0x1004,0x1004,0x1116,0x1116,0x1004,0x1117,0x1117,0x1004,0x1110,0x1110,0x1138,0x001110,0x1110,0x113a,0x001111,0x1111,0x113b,0x001112,0x1112,0x113d,0x001112,0x1112,0x113e,0x001111,0x1111,0x10fb,0x001106,0x1106,0x10fc,0x001106,0x1106,0x10fc,0x001106,0x1106)
GDI32(NtGdiAddFontResourceW                             ,  NONE,0x1002,0x1002,0x112a,0x112a,0x1002,0x112b,0x112b,0x1002,0x1122,0x1122,0x113a,0x00114f,0x114f,0x113c,0x00115e,0x115e,0x113d,0x00116a,0x116a,0x113f,0x00116b,0x116b,0x1140,0x001157,0x1157,0x10fd,0x00114f,0x114f,0x10fe,0x001157,0x1157,0x10fe,0x00116f,0x116f)
GDI32(NtGdiAddInitialFonts                              ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x13a5,0x001150,0x1150,0x13bf,0x001158,0x1158,0x13d1,0x001170,0x1170)
GDI32(NtGdiAddRemoteFontToDC                            ,  NONE,0x1003,0x1003,0x112b,0x112b,0x1003,0x112c,0x112c,0x1003,0x1123,0x1123,0x1139,0x001150,0x1150,0x113b,0x00115f,0x115f,0x113c,0x00116b,0x116b,0x113e,0x00116c,0x116c,0x113f,0x001158,0x1158,0x10fc,0x001151,0x1151,0x10fd,0x001159,0x1159,0x10fd,0x001171,0x1171)
GDI32(NtGdiAddRemoteMMInstanceToDC                      ,  NONE,0x1006,0x1006,0x112c,0x112c,0x1006,0x112d,0x112d,0x1006,0x1124,0x1124,0x1136,0x0c1151,0x1151,0x1138,0x0c1160,0x1160,0x1139,0x0c116c,0x116c,0x113b,0x0c116d,0x116d,0x113c,0x0c1159,0x1159,0x10f9,0x0c1152,0x1152,0x10fa,0x0c115a,0x115a,0x10fa,0x0c1172,0x1172)
GDI32(NtGdiAlphaBlend                                   ,  NONE,0x1007,0x1007,0x107d,0x107d,0x1007,0x107e,0x107e,0x1007,0x107c,0x107c,0x1135,0x00107c,0x107c,0x1137,0x00107d,0x107d,0x1138,0x00107e,0x107e,0x113a,0x00107e,0x107e,0x113b,0x00107e,0x107e,0x10f8,0x00107d,0x107d,0x10f9,0x00107d,0x107d,0x10f9,0x00107d,0x107d)
GDI32(NtGdiAngleArc                                     ,  NONE,0x1008,0x1008,0x112d,0x112d,0x1008,0x112e,0x112e,0x1008,0x1125,0x1125,0x1134,0x001152,0x1152,0x1136,0x001161,0x1161,0x1137,0x00116d,0x116d,0x1139,0x00116e,0x116e,0x113a,0x00115a,0x115a,0x10f7,0x001153,0x1153,0x10f8,0x00115b,0x115b,0x10f8,0x001173,0x1173)
GDI32(NtGdiAnyLinkedFonts                               ,0x1009,0x1009,0x1009,0x112e,0x112e,0x1009,0x112f,0x112f,0x1009,0x1126,0x1126,0x1133,0x011153,0x1153,0x1135,0x011162,0x1162,0x1136,0x01116e,0x116e,0x1138,0x01116f,0x116f,0x1139,0x01115b,0x115b,0x10f6,0x011154,0x1154,0x10f7,0x01115c,0x115c,0x10f7,0x011174,0x1174)
GDI32(NtGdiArcInternal                                  ,  NONE,0x100b,0x100b,0x112f,0x112f,0x100b,0x1130,0x1130,0x100b,0x1127,0x1127,0x1131,0x001154,0x1154,0x1133,0x001163,0x1163,0x1134,0x00116f,0x116f,0x1136,0x001170,0x1170,0x1137,0x00115c,0x115c,0x10f4,0x001155,0x1155,0x10f5,0x00115d,0x115d,0x10f5,0x001175,0x1175)
GDI32(NtGdiBRUSHOBJ_DeleteRbrush                        ,  NONE,0x1298,0x1294,0x1130,0x1130,0x12b9,0x1131,0x1131,0x12d9,0x1128,0x1128,0x1301,0x001155,0x1155,0x130a,0x001164,0x1164,0x1316,0x001170,0x1170,0x1317,0x001171,0x1171,0x131d,0x00115d,0x115d,0x12de,0x001156,0x1156,0x12e7,0x00115e,0x115e,0x12f1,0x001176,0x1176)
GDI32(NtGdiBRUSHOBJ_hGetColorTransform                  ,0x1265,0x127d,0x1279,0x1131,0x1131,0x129e,0x1132,0x1132,0x12be,0x1129,0x1129,0x12e3,0x001156,0x1156,0x12ec,0x001165,0x1165,0x12f8,0x001171,0x1171,0x12f9,0x001172,0x1172,0x12ff,0x00115e,0x115e,0x12c0,0x001157,0x1157,0x12c9,0x00115f,0x115f,0x12d3,0x001177,0x1177)
GDI32(NtGdiBRUSHOBJ_pvAllocRbrush                       ,0x1263,0x127b,0x1277,0x1132,0x1132,0x129c,0x1133,0x1133,0x12bc,0x112a,0x112a,0x12e5,0x001157,0x1157,0x12ee,0x001166,0x1166,0x12fa,0x001172,0x1172,0x12fb,0x001173,0x1173,0x1301,0x00115f,0x115f,0x12c2,0x001158,0x1158,0x12cb,0x001160,0x1160,0x12d5,0x001178,0x1178)
GDI32(NtGdiBRUSHOBJ_pvGetRbrush                         ,0x1264,0x127c,0x1278,0x1133,0x1133,0x129d,0x1134,0x1134,0x12bd,0x112b,0x112b,0x12e4,0x001158,0x1158,0x12ed,0x001167,0x1167,0x12f9,0x001173,0x1173,0x12fa,0x001174,0x1174,0x1300,0x001160,0x1160,0x12c1,0x001159,0x1159,0x12ca,0x001161,0x1161,0x12d4,0x001179,0x1179)
GDI32(NtGdiBRUSHOBJ_ulGetBrushColor                     ,0x1262,0x127a,0x1276,0x1134,0x1134,0x129b,0x1135,0x1135,0x12bb,0x112c,0x112c,0x12e6,0x001159,0x1159,0x12ef,0x001168,0x1168,0x12fb,0x001174,0x1174,0x12fc,0x001175,0x1175,0x1302,0x001161,0x1161,0x12c3,0x00115a,0x115a,0x12cc,0x001162,0x1162,0x12d6,0x00117a,0x117a)
GDI32(NtGdiBeginGdiRendering                            ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x100c,0x112d,0x112d,0x1130,0x00115a,0x115a,0x1132,0x001169,0x1169,0x1133,0x001175,0x1175,0x1135,0x001176,0x1176,0x1136,0x001162,0x1162,0x10f3,0x00115b,0x115b,0x10f4,0x001163,0x1163,0x10f4,0x00117b,0x117b)
GDI32(NtGdiBeginPath                                    ,  NONE,0x100c,0x100c,0x110f,0x110f,0x100c,0x1110,0x1110,0x100d,0x1109,0x1109,0x112f,0x031109,0x1109,0x1131,0x03110a,0x110a,0x1132,0x03110b,0x110b,0x1134,0x03110b,0x110b,0x1135,0x03110a,0x110a,0x10f2,0x0310ff,0x10ff,0x10f3,0x0310ff,0x10ff,0x10f3,0x0310ff,0x10ff)
GDI32(NtGdiBitBlt                                       ,  NONE,0x100d,0x100d,0x1008,0x1008,0x100d,0x1008,0x1008,0x100e,0x1008,0x1008,0x112e,0x001009,0x1009,0x1130,0x00100a,0x100a,0x1131,0x00100b,0x100b,0x1133,0x00100b,0x100b,0x1134,0x00100b,0x100b,0x10f1,0x00100b,0x100b,0x10f2,0x00100b,0x100b,0x10f2,0x00100b,0x100b)
GDI32(NtGdiCLIPOBJ_bEnum                                ,0x125c,0x1274,0x1270,0x1135,0x1135,0x1295,0x1136,0x1136,0x12b5,0x112e,0x112e,0x12e2,0x00115b,0x115b,0x12eb,0x00116a,0x116a,0x12f7,0x001176,0x1176,0x12f8,0x001177,0x1177,0x12fe,0x001163,0x1163,0x12bf,0x00115c,0x115c,0x12c8,0x001164,0x1164,0x12d2,0x00117c,0x117c)
GDI32(NtGdiCLIPOBJ_cEnumStart                           ,0x125d,0x1275,0x1271,0x1136,0x1136,0x1296,0x1137,0x1137,0x12b6,0x112f,0x112f,0x12e1,0x00115c,0x115c,0x12ea,0x00116b,0x116b,0x12f6,0x001177,0x1177,0x12f7,0x001178,0x1178,0x12fd,0x001164,0x1164,0x12be,0x00115d,0x115d,0x12c7,0x001165,0x1165,0x12d1,0x00117d,0x117d)
GDI32(NtGdiCLIPOBJ_ppoGetPath                           ,0x125e,0x1276,0x1272,0x1137,0x1137,0x1297,0x1138,0x1138,0x12b7,0x1130,0x1130,0x12e0,0x00115d,0x115d,0x12e9,0x00116c,0x116c,0x12f5,0x001178,0x1178,0x12f6,0x001179,0x1179,0x12fc,0x001165,0x1165,0x12bd,0x00115e,0x115e,0x12c6,0x001166,0x1166,0x12d0,0x00117e,0x117e)
GDI32(NtGdiCancelDC                                     ,  NONE,0x100e,0x100e,0x1138,0x1138,0x100e,0x1139,0x1139,0x100f,0x1131,0x1131,0x112d,0x03115e,0x115e,0x112f,0x03116d,0x116d,0x1130,0x031179,0x1179,0x1132,0x03117a,0x117a,0x1133,0x031166,0x1166,0x10f0,0x03115f,0x115f,0x10f1,0x031167,0x1167,0x10f1,0x03117f,0x117f)
GDI32(NtGdiChangeGhostFont                              ,  NONE,0x10d5,0x10d4,0x1139,0x1139,0x10dd,0x113a,0x113a,0x10df,0x1132,0x1132,0x105b,0x00115f,0x115f,0x105c,0x00116e,0x116e,0x105d,0x00117a,0x117a,0x105d,0x00117b,0x117b,0x105d,0x001167,0x1167,0x1064,0x001160,0x1160,0x1065,0x001168,0x1168,0x1065,0x001180,0x1180)
GDI32(NtGdiCheckBitmapBits                              ,  NONE,0x100f,0x100f,0x113a,0x113a,0x100f,0x113b,0x113b,0x1010,0x1133,0x1133,0x112c,0x001160,0x1160,0x112e,0x00116f,0x116f,0x112f,0x00117b,0x117b,0x1131,0x00117c,0x117c,0x1132,0x001168,0x1168,0x10ef,0x001161,0x1161,0x10f0,0x001169,0x1169,0x10f0,0x001181,0x1181)
GDI32(NtGdiClearBitmapAttributes                        ,  NONE,0x1011,0x1011,0x113b,0x113b,0x1011,0x113c,0x113c,0x1012,0x1134,0x1134,0x112a,0x051161,0x1161,0x112c,0x051170,0x1170,0x112d,0x05117c,0x117c,0x112f,0x05117d,0x117d,0x1130,0x051169,0x1169,0x10ed,0x051162,0x1162,0x10ee,0x05116a,0x116a,0x10ee,0x051182,0x1182)
GDI32(NtGdiClearBrushAttributes                         ,  NONE,0x1012,0x1012,0x113c,0x113c,0x1012,0x113d,0x113d,0x1013,0x1135,0x1135,0x1129,0x051162,0x1162,0x112b,0x051171,0x1171,0x112c,0x05117d,0x117d,0x112e,0x05117e,0x117e,0x112f,0x05116a,0x116a,0x10ec,0x051163,0x1163,0x10ed,0x05116b,0x116b,0x10ed,0x051183,0x1183)
GDI32(NtGdiCloseFigure                                  ,  NONE,0x1010,0x1010,0x1105,0x1105,0x1010,0x1106,0x1106,0x1011,0x1100,0x1100,0x112b,0x031100,0x1100,0x112d,0x031101,0x1101,0x112e,0x031102,0x1102,0x1130,0x031102,0x1102,0x1131,0x031101,0x1101,0x10ee,0x0310f6,0x10f6,0x10ef,0x0310f6,0x10f6,0x10ef,0x0310f6,0x10f6)
GDI32(NtGdiColorCorrectPalette                          ,  NONE,0x1013,0x1013,0x113d,0x113d,0x1013,0x113e,0x113e,0x1014,0x1136,0x1136,0x1128,0x001163,0x1163,0x112a,0x001172,0x1172,0x112b,0x00117e,0x117e,0x112d,0x00117f,0x117f,0x112e,0x00116b,0x116b,0x10eb,0x001164,0x1164,0x10ec,0x00116c,0x116c,0x10ec,0x001184,0x1184)
GDI32(NtGdiCombineRgn                                   ,  NONE,0x1014,0x1014,0x1033,0x1033,0x1014,0x1034,0x1034,0x1015,0x1034,0x1034,0x1127,0x111035,0x1035,0x1129,0x111036,0x1036,0x112a,0x111037,0x1037,0x112c,0x111037,0x1037,0x112d,0x111037,0x1037,0x10ea,0x111037,0x1037,0x10eb,0x111037,0x1037,0x10eb,0x111037,0x1037)
GDI32(NtGdiCombineTransform                             ,  NONE,0x1015,0x1015,0x10f4,0x10f4,0x1015,0x10f5,0x10f5,0x1016,0x10f0,0x10f0,0x1126,0x0a10f0,0x10f0,0x1128,0x0a10f1,0x10f1,0x1129,0x0a10f2,0x10f2,0x112b,0x0a10f2,0x10f2,0x112c,0x0a10f1,0x10f1,0x10e9,0x0a10e6,0x10e6,0x10ea,0x0a10e6,0x10e6,0x10ea,0x0a10e6,0x10e6)
GDI32(NtGdiComputeXformCoefficients                     ,  NONE,0x1016,0x1016,0x108c,0x108c,0x1016,0x108d,0x108d,0x1017,0x108b,0x108b,0x1125,0x03108b,0x108b,0x1127,0x03108c,0x108c,0x1128,0x03108d,0x108d,0x112a,0x03108d,0x108d,0x112b,0x03108d,0x108d,0x10e8,0x03108b,0x108b,0x10e9,0x03108b,0x108b,0x10e9,0x03108b,0x108b)
GDI32(NtGdiConfigureOPMProtectedOutput                  ,  NONE,  NONE,  NONE,  NONE,  NONE,0x1017,0x113f,0x113f,0x1018,0x1137,0x1137,0x1124,0x001164,0x1164,0x1126,0x001173,0x1173,0x1127,0x00117f,0x117f,0x1129,0x001180,0x1180,0x112a,0x00116c,0x116c,0x10e7,0x001165,0x1165,0x10e8,0x00116d,0x116d,0x10e8,0x001185,0x1185)
GDI32(NtGdiConsoleTextOut                               ,0x1015,0x1017,0x1017,0x106d,0x106d,0x1018,0x106e,0x106e,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
GDI32(NtGdiConvertMetafileRect                          ,  NONE,0x1018,0x1018,0x113e,0x113e,0x1019,0x1140,0x1140,0x1019,0x1138,0x1138,0x1123,0x071165,0x1165,0x1125,0x071174,0x1174,0x1126,0x071180,0x1180,0x1128,0x071181,0x1181,0x1129,0x07116d,0x116d,0x10e6,0x071166,0x1166,0x10e7,0x07116e,0x116e,0x10e7,0x071186,0x1186)
GDI32(NtGdiCreateBitmap                                 ,  NONE,0x1019,0x1019,0x106c,0x106c,0x101a,0x106d,0x106d,0x101a,0x106d,0x106d,0x1122,0x00106d,0x106d,0x1124,0x00106e,0x106e,0x1125,0x00106f,0x106f,0x1127,0x00106f,0x106f,0x1128,0x00106f,0x106f,0x10e5,0x00106e,0x106e,0x10e6,0x00106e,0x106e,0x10e6,0x00106e,0x106e)
GDI32(NtGdiCreateBitmapFromDxSurface                    ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x101b,0x1139,0x1139,0x1121,0x001166,0x1166,0x1123,0x001175,0x1175,0x1124,0x001181,0x1181,0x1126,0x001182,0x1182,0x1127,0x00116e,0x116e,0x10e4,0x001167,0x1167,0x10e5,0x00116f,0x116f,0x10e5,0x001187,0x1187)
GDI32(NtGdiCreateBitmapFromDxSurface2                   ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x1120,0x001167,0x1167,0x1122,0x001176,0x1176,0x1123,0x001182,0x1182,0x1125,0x001183,0x1183,0x1126,0x00116f,0x116f,0x10e3,0x001168,0x1168,0x10e4,0x001170,0x1170,0x10e4,0x001188,0x1188)
GDI32(NtGdiCreateClientObj                              ,  NONE,0x101a,0x101a,0x10be,0x10be,0x101b,0x10bf,0x10bf,0x101c,0x10ba,0x10ba,0x111f,0x0410ba,0x10ba,0x1121,0x0410bb,0x10bb,0x1122,0x0410bc,0x10bc,0x1124,0x0410bc,0x10bc,0x1125,0x0410bc,0x10bc,0x10e2,0x0410b5,0x10b5,0x10e3,0x0410b5,0x10b5,0x10e3,0x0410b5,0x10b5)
GDI32(NtGdiCreateColorSpace                             ,  NONE,0x101b,0x101b,0x1124,0x1124,0x101c,0x1125,0x1125,0x101d,0x111c,0x111c,0x111e,0x04111c,0x111c,0x1120,0x04111d,0x111d,0x1121,0x04111e,0x111e,0x1123,0x04111e,0x111e,0x1124,0x04111d,0x111d,0x10e1,0x041112,0x1112,0x10e2,0x041112,0x1112,0x10e2,0x041112,0x1112)
GDI32(NtGdiCreateColorTransform                         ,  NONE,0x101c,0x101c,0x113f,0x113f,0x101d,0x1141,0x1141,0x101e,0x113a,0x113a,0x111d,0x001168,0x1168,0x111f,0x001177,0x1177,0x1120,0x001183,0x1183,0x1122,0x001184,0x1184,0x1123,0x001170,0x1170,0x10e0,0x001169,0x1169,0x10e1,0x001171,0x1171,0x10e1,0x001189,0x1189)
GDI32(NtGdiCreateCompatibleBitmap                       ,  NONE,0x101d,0x101d,0x104a,0x104a,0x101e,0x104b,0x104b,0x101f,0x104b,0x104b,0x111c,0x0b104c,0x104c,0x111e,0x0b104d,0x104d,0x111f,0x0b104e,0x104e,0x1121,0x0b104e,0x104e,0x1122,0x0b104e,0x104e,0x10df,0x0b104e,0x104e,0x10e0,0x0b104e,0x104e,0x10e0,0x0b104e,0x104e)
GDI32(NtGdiCreateCompatibleDC                           ,  NONE,0x101e,0x101e,0x1054,0x1054,0x101f,0x1055,0x1055,0x1020,0x1055,0x1055,0x111b,0x031055,0x1055,0x111d,0x031056,0x1056,0x111e,0x031057,0x1057,0x1120,0x031057,0x1057,0x1121,0x031057,0x1057,0x10de,0x031057,0x1057,0x10df,0x031057,0x1057,0x10df,0x031057,0x1057)
GDI32(NtGdiCreateDIBBrush                               ,  NONE,0x101f,0x101f,0x1075,0x1075,0x1020,0x1076,0x1076,0x1021,0x1074,0x1074,0x111a,0x001074,0x1074,0x111c,0x001075,0x1075,0x111d,0x001076,0x1076,0x111f,0x001076,0x1076,0x1120,0x001076,0x1076,0x10dd,0x001075,0x1075,0x10de,0x001075,0x1075,0x10de,0x001075,0x1075)
GDI32(NtGdiCreateDIBSection                             ,  NONE,0x1021,0x1021,0x109b,0x109b,0x1022,0x109c,0x109c,0x1023,0x109a,0x109a,0x1118,0x00109a,0x109a,0x111a,0x00109b,0x109b,0x111b,0x00109c,0x109c,0x111d,0x00109c,0x109c,0x111e,0x00109c,0x109c,0x10db,0x00109a,0x109a,0x10dc,0x00109a,0x109a,0x10dc,0x00109a,0x109a)
GDI32(NtGdiCreateDIBitmapInternal                       ,  NONE,0x1020,0x1020,0x10a0,0x10a0,0x1021,0x10a1,0x10a1,0x1022,0x109f,0x109f,0x1119,0x00109f,0x109f,0x111b,0x0010a0,0x10a0,0x111c,0x0010a1,0x10a1,0x111e,0x0010a1,0x10a1,0x111f,0x0010a1,0x10a1,0x10dc,0x00109f,0x109f,0x10dd,0x00109f,0x109f,0x10dd,0x00109f,0x109f)
GDI32(NtGdiCreateEllipticRgn                            ,  NONE,0x1022,0x1022,0x1140,0x1140,0x1023,0x1142,0x1142,0x1024,0x113b,0x113b,0x1116,0x111169,0x1169,0x1118,0x111178,0x1178,0x1119,0x111184,0x1184,0x111b,0x111185,0x1185,0x111c,0x111171,0x1171,0x10d9,0x11116a,0x116a,0x10da,0x111172,0x1172,0x10da,0x11118a,0x118a)
GDI32(NtGdiCreateHalftonePalette                        ,0x1021,0x1023,0x1023,0x10f1,0x10f1,0x1024,0x10f2,0x10f2,0x1025,0x10ed,0x10ed,0x1115,0x0310ed,0x10ed,0x1117,0x0310ee,0x10ee,0x1118,0x0310ef,0x10ef,0x111a,0x0310ef,0x10ef,0x111b,0x0310ee,0x10ee,0x10d8,0x0310e3,0x10e3,0x10d9,0x0310e3,0x10e3,0x10d9,0x0310e3,0x10e3)
GDI32(NtGdiCreateHatchBrushInternal                     ,  NONE,0x1024,0x1024,0x1141,0x1141,0x1025,0x1143,0x1143,0x1026,0x113c,0x113c,0x1114,0x0a116a,0x116a,0x1116,0x0a1179,0x1179,0x1117,0x0a1185,0x1185,0x1119,0x0a1186,0x1186,0x111a,0x0a1172,0x1172,0x10d7,0x0a116b,0x116b,0x10d8,0x0a1173,0x1173,0x10d8,0x0a118b,0x118b)
GDI32(NtGdiCreateMetafileDC                             ,  NONE,0x1025,0x1025,0x1142,0x1142,0x1026,0x1144,0x1144,0x1027,0x113d,0x113d,0x1113,0x03116b,0x116b,0x1115,0x03117a,0x117a,0x1116,0x031186,0x1186,0x1118,0x031187,0x1187,0x1119,0x031173,0x1173,0x10d6,0x03116c,0x116c,0x10d7,0x031174,0x1174,0x10d7,0x03118c,0x118c)
GDI32(NtGdiCreateOPMProtectedOutput                     ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1118,0x001174,0x1174,0x10d5,0x00116d,0x116d,0x10d6,0x001175,0x1175,0x10d6,0x00118d,0x118d)
GDI32(NtGdiCreateOPMProtectedOutputs                    ,  NONE,  NONE,  NONE,  NONE,  NONE,0x1027,0x1145,0x1145,0x1028,0x113e,0x113e,0x1112,0x00116c,0x116c,0x1114,0x00117b,0x117b,0x1115,0x001187,0x1187,0x1117,0x001188,0x1188,0x1117,0x001175,0x1175,0x10d4,0x00116e,0x116e,0x10d5,0x001176,0x1176,0x10d5,0x00118e,0x118e)
GDI32(NtGdiCreatePaletteInternal                        ,  NONE,0x1026,0x1026,0x10b0,0x10b0,0x1028,0x10b1,0x10b1,0x1029,0x10ac,0x10ac,0x1111,0x0510ac,0x10ac,0x1113,0x0510ad,0x10ad,0x1114,0x0510ae,0x10ae,0x1116,0x0510ae,0x10ae,0x1116,0x0510ae,0x10ae,0x10d3,0x0510a7,0x10a7,0x10d4,0x0510a7,0x10a7,0x10d4,0x0510a7,0x10a7)
GDI32(NtGdiCreatePatternBrushInternal                   ,  NONE,0x1027,0x1027,0x10b5,0x10b5,0x1029,0x10b6,0x10b6,0x102a,0x10b1,0x10b1,0x1110,0x0a10b1,0x10b1,0x1112,0x0a10b2,0x10b2,0x1113,0x0a10b3,0x10b3,0x1115,0x0a10b3,0x10b3,0x1115,0x0a10b3,0x10b3,0x10d2,0x0a10ac,0x10ac,0x10d3,0x0a10ac,0x10ac,0x10d3,0x0a10ac,0x10ac)
GDI32(NtGdiCreatePen                                    ,  NONE,0x1028,0x1028,0x1056,0x1056,0x102a,0x1057,0x1057,0x102b,0x1057,0x1057,0x110f,0x111057,0x1057,0x1111,0x111058,0x1058,0x1112,0x111059,0x1059,0x1114,0x111059,0x1059,0x1114,0x111059,0x1059,0x10d1,0x111059,0x1059,0x10d2,0x111059,0x1059,0x10d2,0x111059,0x1059)
GDI32(NtGdiCreateRectRgn                                ,  NONE,0x1029,0x1029,0x1085,0x1085,0x102b,0x1086,0x1086,0x102c,0x1084,0x1084,0x110e,0x111084,0x1084,0x1110,0x111085,0x1085,0x1111,0x111086,0x1086,0x1113,0x111086,0x1086,0x1113,0x111086,0x1086,0x10d0,0x111084,0x1084,0x10d1,0x111084,0x1084,0x10d1,0x111084,0x1084)
GDI32(NtGdiCreateRoundRectRgn                           ,  NONE,0x102a,0x102a,0x1143,0x1143,0x102c,0x1146,0x1146,0x102d,0x113f,0x113f,0x110d,0x00116d,0x116d,0x110f,0x00117c,0x117c,0x1110,0x001188,0x1188,0x1112,0x001189,0x1189,0x1112,0x001176,0x1176,0x10cf,0x00116f,0x116f,0x10d0,0x001177,0x1177,0x10d0,0x00118f,0x118f)
GDI32(NtGdiCreateServerMetaFile                         ,  NONE,0x102b,0x102b,0x1144,0x1144,0x102d,0x1147,0x1147,0x102e,0x1140,0x1140,0x110c,0x00116e,0x116e,0x110e,0x00117d,0x117d,0x110f,0x001189,0x1189,0x1111,0x00118a,0x118a,0x1111,0x001177,0x1177,0x10ce,0x001170,0x1170,0x10cf,0x001178,0x1178,0x10cf,0x001190,0x1190)
GDI32(NtGdiCreateSessionMappedDIBSection                ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x1117,0x00116f,0x116f,0x1119,0x00117e,0x117e,0x111a,0x00118a,0x118a,0x111c,0x00118b,0x118b,0x111d,0x001178,0x1178,0x10da,0x001171,0x1171,0x10db,0x001179,0x1179,0x10db,0x001191,0x1191)
GDI32(NtGdiCreateSolidBrush                             ,  NONE,0x102c,0x102c,0x10bc,0x10bc,0x102e,0x10bd,0x10bd,0x102f,0x10b8,0x10b8,0x110b,0x0510b8,0x10b8,0x110d,0x0510b9,0x10b9,0x110e,0x0510ba,0x10ba,0x1110,0x0510ba,0x10ba,0x1110,0x0510ba,0x10ba,0x10cd,0x0510b3,0x10b3,0x10ce,0x0510b3,0x10b3,0x10ce,0x0510b3,0x10b3)
GDI32(NtGdiD3dContextCreate                             ,  NONE,0x102d,0x102d,0x1145,0x1145,0x102f,0x1148,0x1148,0x1030,0x1141,0x1141,0x110a,0x001170,0x1170,0x110c,0x00117f,0x117f,0x110d,0x00118b,0x118b,0x110f,0x00118c,0x118c,0x110f,0x001179,0x1179,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
GDI32(NtGdiD3dContextDestroy                            ,  NONE,0x102e,0x102e,0x1146,0x1146,0x1030,0x1149,0x1149,0x1031,0x1142,0x1142,0x1109,0x001171,0x1171,0x110b,0x001180,0x1180,0x110c,0x00118c,0x118c,0x110e,0x00118d,0x118d,0x110e,0x00117a,0x117a,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
GDI32(NtGdiD3dContextDestroyAll                         ,  NONE,0x102f,0x102f,0x1147,0x1147,0x1031,0x114a,0x114a,0x1032,0x1143,0x1143,0x1108,0x041172,0x1172,0x110a,0x041181,0x1181,0x110b,0x04118d,0x118d,0x110d,0x04118e,0x118e,0x110d,0x04117b,0x117b,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
GDI32(NtGdiD3dDrawPrimitives2                           ,  NONE,0x1031,0x1031,0x1064,0x1064,0x1033,0x1065,0x1065,0x1034,0x1065,0x1065,0x1106,0x001065,0x1065,0x1108,0x001066,0x1066,0x1109,0x001067,0x1067,0x110b,0x001067,0x1067,0x110b,0x001067,0x1067,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
GDI32(NtGdiD3dValidateTextureStageState                 ,  NONE,0x1030,0x1030,0x1148,0x1148,0x1032,0x114b,0x114b,0x1033,0x1144,0x1144,0x1107,0x001173,0x1173,0x1109,0x001182,0x1182,0x110a,0x00118e,0x118e,0x110c,0x00118f,0x118f,0x110c,0x00117c,0x117c,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
GDI32(NtGdiDDCCIGetCapabilitiesString                   ,  NONE,  NONE,  NONE,  NONE,  NONE,0x1300,0x114c,0x114c,0x132c,0x1145,0x1145,0x136f,0x001174,0x1174,0x1381,0x001183,0x1183,0x13ba,0x00118f,0x118f,0x13bc,0x001190,0x1190,0x13c9,0x00117d,0x117d,0x139a,0x001172,0x1172,0x13b4,0x00117a,0x117a,0x13c6,0x001192,0x1192)
GDI32(NtGdiDDCCIGetCapabilitiesStringLength             ,  NONE,  NONE,  NONE,  NONE,  NONE,0x12ff,0x114d,0x114d,0x132b,0x1146,0x1146,0x1370,0x001175,0x1175,0x1382,0x001184,0x1184,0x13bb,0x001190,0x1190,0x13bd,0x001191,0x1191,0x13ca,0x00117e,0x117e,0x139b,0x001173,0x1173,0x13b5,0x00117b,0x117b,0x13c7,0x001193,0x1193)
GDI32(NtGdiDDCCIGetTimingReport                         ,  NONE,  NONE,  NONE,  NONE,  NONE,0x1301,0x114e,0x114e,0x132d,0x1147,0x1147,0x136e,0x001176,0x1176,0x1380,0x001185,0x1185,0x13b9,0x001191,0x1191,0x13bb,0x001192,0x1192,0x13c8,0x00117f,0x117f,0x1399,0x001174,0x1174,0x13b3,0x00117c,0x117c,0x13c5,0x001194,0x1194)
GDI32(NtGdiDDCCIGetVCPFeature                           ,  NONE,  NONE,  NONE,  NONE,  NONE,0x12fc,0x114f,0x114f,0x1328,0x1148,0x1148,0x1373,0x001177,0x1177,0x1385,0x001186,0x1186,0x13be,0x001192,0x1192,0x13c0,0x001193,0x1193,0x13cd,0x001180,0x1180,0x139e,0x001175,0x1175,0x13b8,0x00117d,0x117d,0x13ca,0x001195,0x1195)
GDI32(NtGdiDDCCISaveCurrentSettings                     ,  NONE,  NONE,  NONE,  NONE,  NONE,0x12fe,0x1150,0x1150,0x132a,0x1149,0x1149,0x1371,0x001178,0x1178,0x1383,0x001187,0x1187,0x13bc,0x001193,0x1193,0x13be,0x001194,0x1194,0x13cb,0x001181,0x1181,0x139c,0x001176,0x1176,0x13b6,0x00117e,0x117e,0x13c8,0x001196,0x1196)
GDI32(NtGdiDDCCISetVCPFeature                           ,  NONE,  NONE,  NONE,  NONE,  NONE,0x12fd,0x1151,0x1151,0x1329,0x114a,0x114a,0x1372,0x001179,0x1179,0x1384,0x001188,0x1188,0x13bd,0x001194,0x1194,0x13bf,0x001195,0x1195,0x13cc,0x001182,0x1182,0x139d,0x001177,0x1177,0x13b7,0x00117f,0x117f,0x13c9,0x001197,0x1197)
GDI32(NtGdiDdAddAttachedSurface                         ,  NONE,0x1033,0x1033,0x1149,0x1149,0x1035,0x1152,0x1152,0x1036,0x114b,0x114b,0x1104,0x00117a,0x117a,0x1106,0x001189,0x1189,0x1107,0x001195,0x1195,0x1109,0x001196,0x1196,0x1109,0x001183,0x1183,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
GDI32(NtGdiDdAlphaBlt                                   ,  NONE,0x1034,0x1034,0x114a,0x114a,0x1036,0x1153,0x1153,0x1037,0x114c,0x114c,0x1103,0x00117b,0x117b,0x1105,0x00118a,0x118a,0x1106,0x001196,0x1196,0x1108,0x001197,0x1197,0x1108,0x001184,0x1184,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
GDI32(NtGdiDdAttachSurface                              ,  NONE,0x1035,0x1035,0x114b,0x114b,0x1037,0x1154,0x1154,0x1038,0x114d,0x114d,0x1102,0x05117c,0x117c,0x1104,0x05118b,0x118b,0x1105,0x051197,0x1197,0x1107,0x051198,0x1198,0x1107,0x051185,0x1185,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
GDI32(NtGdiDdBeginMoCompFrame                           ,  NONE,0x1036,0x1036,0x114c,0x114c,0x1038,0x1155,0x1155,0x1039,0x114e,0x114e,0x1101,0x00117d,0x117d,0x1103,0x00118c,0x118c,0x1104,0x001198,0x1198,0x1106,0x001199,0x1199,0x1106,0x001186,0x1186,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
GDI32(NtGdiDdBlt                                        ,  NONE,0x1037,0x1037,0x107e,0x107e,0x1039,0x107f,0x107f,0x103a,0x107d,0x107d,0x1100,0x00107d,0x107d,0x1102,0x00107e,0x107e,0x1103,0x00107f,0x107f,0x1105,0x00107f,0x107f,0x1105,0x00107f,0x107f,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
GDI32(NtGdiDdCanCreateD3DBuffer                         ,  NONE,0x1039,0x1039,0x114d,0x114d,0x103b,0x1156,0x1156,0x103c,0x114f,0x114f,0x10fe,0x00117e,0x117e,0x1100,0x00118d,0x118d,0x1101,0x001199,0x1199,0x1103,0x00119a,0x119a,0x1103,0x001187,0x1187,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
GDI32(NtGdiDdCanCreateSurface                           ,  NONE,0x1038,0x1038,0x10a6,0x10a6,0x103a,0x10a7,0x10a7,0x103b,0x10a2,0x10a2,0x10ff,0x0010a2,0x10a2,0x1101,0x0010a3,0x10a3,0x1102,0x0010a4,0x10a4,0x1104,0x0010a4,0x10a4,0x1104,0x0010a4,0x10a4,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
GDI32(NtGdiDdColorControl                               ,  NONE,0x103a,0x103a,0x114e,0x114e,0x103c,0x1157,0x1157,0x103d,0x1150,0x1150,0x10fd,0x00117f,0x117f,0x10ff,0x00118e,0x118e,0x1100,0x00119a,0x119a,0x1102,0x00119b,0x119b,0x1102,0x001188,0x1188,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
GDI32(NtGdiDdCreateD3DBuffer                            ,  NONE,0x103d,0x103d,0x114f,0x114f,0x103f,0x1158,0x1158,0x1040,0x1151,0x1151,0x10fa,0x001180,0x1180,0x10fc,0x00118f,0x118f,0x10fd,0x00119b,0x119b,0x10ff,0x00119c,0x119c,0x10ff,0x001189,0x1189,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
GDI32(NtGdiDdCreateDirectDrawObject                     ,  NONE,0x103b,0x103b,0x1150,0x1150,0x103d,0x1159,0x1159,0x103e,0x1152,0x1152,0x10fc,0x031181,0x1181,0x10fe,0x031190,0x1190,0x10ff,0x03119c,0x119c,0x1101,0x03119d,0x119d,0x1101,0x03118a,0x118a,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
GDI32(NtGdiDdCreateFullscreenSprite                     ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x132e,0x1153,0x1153,0x1377,0x001182,0x1182,0x1389,0x001191,0x1191,0x13c2,0x00119d,0x119d,0x13c4,0x00119e,0x119e,0x13d1,0x00118b,0x118b,0x13a2,0x001178,0x1178,0x13bc,0x001180,0x1180,0x13ce,0x001198,0x1198)
GDI32(NtGdiDdCreateMoComp                               ,  NONE,0x103e,0x103e,0x1151,0x1151,0x1040,0x115a,0x115a,0x1041,0x1154,0x1154,0x10f9,0x001183,0x1183,0x10fb,0x001192,0x1192,0x10fc,0x00119e,0x119e,0x10fe,0x00119f,0x119f,0x10fe,0x00118c,0x118c,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
GDI32(NtGdiDdCreateSurface                              ,  NONE,0x103c,0x103c,0x10a7,0x10a7,0x103e,0x10a8,0x10a8,0x103f,0x10a3,0x10a3,0x10fb,0x0010a3,0x10a3,0x10fd,0x0010a4,0x10a4,0x10fe,0x0010a5,0x10a5,0x1100,0x0010a5,0x10a5,0x1100,0x0010a5,0x10a5,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
GDI32(NtGdiDdCreateSurfaceEx                            ,  NONE,0x105e,0x105e,0x10c6,0x10c6,0x1060,0x10c7,0x10c7,0x1061,0x10c2,0x10c2,0x10d9,0x0a10c2,0x10c2,0x10db,0x0a10c3,0x10c3,0x10dc,0x0a10c4,0x10c4,0x10de,0x0a10c4,0x10c4,0x10de,0x0a10c4,0x10c4,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
GDI32(NtGdiDdCreateSurfaceObject                        ,  NONE,0x103f,0x103f,0x10c7,0x10c7,0x1041,0x10c8,0x10c8,0x1042,0x10c3,0x10c3,0x10f8,0x0010c3,0x10c3,0x10fa,0x0010c4,0x10c4,0x10fb,0x0010c5,0x10c5,0x10fd,0x0010c5,0x10c5,0x10fd,0x0010c5,0x10c5,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
GDI32(NtGdiDdDDIAbandonSwapChain                        ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x13a8,0x00119f,0x119f,0x13a8,0x0011a0,0x11a0,0x13af,0x00118d,0x118d,0x1370,0x001179,0x1179,0x137b,0x001181,0x1181,0x1385,0x001199,0x1199)
GDI32(NtGdiDdDDIAcquireKeyedMutex                       ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x131c,0x1155,0x1155,0x1353,0x001184,0x1184,0x135c,0x001193,0x1193,0x136c,0x0011a0,0x11a0,0x136d,0x0011a1,0x11a1,0x1373,0x00118e,0x118e,0x1334,0x00117a,0x117a,0x133c,0x001182,0x1182,0x1346,0x00119a,0x119a)
GDI32(NtGdiDdDDIAcquireKeyedMutex2                      ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x1344,0x001185,0x1185,0x134d,0x001194,0x1194,0x135d,0x0011a1,0x11a1,0x135e,0x0011a2,0x11a2,0x1364,0x00118f,0x118f,0x1325,0x00117b,0x117b,0x132d,0x001183,0x1183,0x1337,0x00119b,0x119b)
GDI32(NtGdiDdDDIAcquireSwapChain                        ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x13a6,0x0011a2,0x11a2,0x13a6,0x0011a3,0x11a3,0x13ad,0x001190,0x1190,0x136e,0x00117c,0x117c,0x1379,0x001184,0x1184,0x1383,0x00119c,0x119c)
GDI32(NtGdiDdDDIAddSurfaceToSwapChain                   ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1375,0x001185,0x1185,0x137f,0x00119d,0x119d)
GDI32(NtGdiDdDDIAdjustFullscreenGamma                   ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x13b0,0x0011a3,0x11a3,0x13b0,0x0011a4,0x11a4,0x13b7,0x001191,0x1191,0x1378,0x00117d,0x117d,0x1383,0x001186,0x1186,0x138d,0x00119e,0x119e)
GDI32(NtGdiDdDDICacheHybridQueryValue                   ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,0x1374,0x001195,0x1195,0x1399,0x0011a4,0x11a4,0x139a,0x0011a5,0x11a5,0x13a1,0x001192,0x1192,0x1362,0x00117e,0x117e,0x136a,0x001187,0x1187,0x1374,0x00119f,0x119f)
GDI32(NtGdiDdDDIChangeVideoMemoryReservation            ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x13a2,0x0011a5,0x11a5,0x13a2,0x0011a6,0x11a6,0x13a9,0x001193,0x1193,0x136a,0x00117f,0x117f,0x1372,0x001188,0x1188,0x137c,0x0011a0,0x11a0)
GDI32(NtGdiDdDDICheckExclusiveOwnership                 ,  NONE,  NONE,  NONE,  NONE,  NONE,0x12f2,0x115b,0x115b,0x1315,0x1156,0x1156,0x135a,0x001186,0x1186,0x1363,0x001196,0x1196,0x1373,0x0011a6,0x11a6,0x1374,0x0011a7,0x11a7,0x137a,0x001194,0x1194,0x133b,0x001180,0x1180,0x1343,0x001189,0x1189,0x134d,0x0011a1,0x11a1)
GDI32(NtGdiDdDDICheckMonitorPowerState                  ,  NONE,  NONE,  NONE,  NONE,  NONE,0x12f1,0x115c,0x115c,0x1314,0x1157,0x1157,0x135b,0x001187,0x1187,0x1364,0x001197,0x1197,0x1374,0x0011a7,0x11a7,0x1375,0x0011a8,0x11a8,0x137b,0x001195,0x1195,0x133c,0x001181,0x1181,0x1344,0x00118a,0x118a,0x134e,0x0011a2,0x11a2)
GDI32(NtGdiDdDDICheckMultiPlaneOverlaySupport           ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,0x1372,0x001198,0x1198,0x1384,0x0011a8,0x11a8,0x1385,0x0011a9,0x11a9,0x138b,0x001196,0x1196,0x134c,0x001182,0x1182,0x1354,0x00118b,0x118b,0x135e,0x0011a3,0x11a3)
GDI32(NtGdiDdDDICheckMultiPlaneOverlaySupport2          ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x13aa,0x0011a9,0x11a9,0x13aa,0x0011aa,0x11aa,0x13b1,0x001197,0x1197,0x1372,0x001183,0x1183,0x137d,0x00118c,0x118c,0x1387,0x0011a4,0x11a4)
GDI32(NtGdiDdDDICheckMultiPlaneOverlaySupport3          ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x13be,0x001198,0x1198,0x137f,0x001184,0x1184,0x138a,0x00118d,0x118d,0x1394,0x0011a5,0x11a5)
GDI32(NtGdiDdDDICheckOcclusion                          ,  NONE,  NONE,  NONE,  NONE,  NONE,0x12ef,0x115d,0x115d,0x1312,0x1158,0x1158,0x135d,0x001188,0x1188,0x1366,0x001199,0x1199,0x1376,0x0011aa,0x11aa,0x1377,0x0011ab,0x11ab,0x137d,0x001199,0x1199,0x133e,0x001185,0x1185,0x1346,0x00118e,0x118e,0x1350,0x0011a6,0x11a6)
GDI32(NtGdiDdDDICheckSharedResourceAccess               ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x1321,0x1159,0x1159,0x134e,0x001189,0x1189,0x1357,0x00119a,0x119a,0x1367,0x0011ab,0x11ab,0x1368,0x0011ac,0x11ac,0x136e,0x00119a,0x119a,0x132f,0x001186,0x1186,0x1337,0x00118f,0x118f,0x1341,0x0011a7,0x11a7)
GDI32(NtGdiDdDDICheckVidPnExclusiveOwnership            ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x1320,0x115a,0x115a,0x134f,0x00118a,0x118a,0x1358,0x00119b,0x119b,0x1368,0x0011ac,0x11ac,0x1369,0x0011ad,0x11ad,0x136f,0x00119b,0x119b,0x1330,0x001187,0x1187,0x1338,0x001190,0x1190,0x1342,0x0011a8,0x11a8)
GDI32(NtGdiDdDDICloseAdapter                            ,  NONE,  NONE,  NONE,  NONE,  NONE,0x12d7,0x115e,0x115e,0x12f9,0x115b,0x115b,0x131e,0x00118b,0x118b,0x1327,0x00119c,0x119c,0x1333,0x0011ad,0x11ad,0x1334,0x0011ae,0x11ae,0x133a,0x00119c,0x119c,0x12fb,0x001188,0x1188,0x1303,0x001191,0x1191,0x130d,0x0011a9,0x11a9)
GDI32(NtGdiDdDDIConfigureSharedResource                 ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x131e,0x115c,0x115c,0x1351,0x00118c,0x118c,0x135a,0x00119d,0x119d,0x136a,0x0011ae,0x11ae,0x136b,0x0011af,0x11af,0x1371,0x00119d,0x119d,0x1332,0x001189,0x1189,0x133a,0x001192,0x1192,0x1344,0x0011aa,0x11aa)
GDI32(NtGdiDdDDICreateAllocation                        ,  NONE,  NONE,  NONE,  NONE,  NONE,0x12be,0x115f,0x115f,0x12df,0x115d,0x115d,0x133f,0x00118d,0x118d,0x1348,0x00119e,0x119e,0x1358,0x0011af,0x11af,0x1359,0x0011b0,0x11b0,0x135f,0x00119e,0x119e,0x1320,0x00118a,0x118a,0x1328,0x001193,0x1193,0x1332,0x0011ab,0x11ab)
GDI32(NtGdiDdDDICreateBundleObject                      ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x13a9,0x001194,0x1194,0x13b4,0x0011ac,0x11ac)
GDI32(NtGdiDdDDICreateContext                           ,  NONE,  NONE,  NONE,  NONE,  NONE,0x12c6,0x1160,0x1160,0x12e7,0x115e,0x115e,0x1332,0x00118e,0x118e,0x133b,0x00119f,0x119f,0x1348,0x0011b0,0x11b0,0x1349,0x0011b1,0x11b1,0x134f,0x00119f,0x119f,0x1310,0x00118b,0x118b,0x1318,0x001195,0x1195,0x1322,0x0011ad,0x11ad)
GDI32(NtGdiDdDDICreateContextVirtual                    ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1396,0x0011b1,0x11b1,0x1397,0x0011b2,0x11b2,0x139e,0x0011a0,0x11a0,0x135f,0x00118c,0x118c,0x1367,0x001196,0x1196,0x1371,0x0011ae,0x11ae)
GDI32(NtGdiDdDDICreateDCFromMemory                      ,  NONE,  NONE,  NONE,  NONE,  NONE,0x12e4,0x1161,0x1161,0x1307,0x115f,0x115f,0x130f,0x00118f,0x118f,0x1318,0x0011a0,0x11a0,0x1324,0x0011b2,0x11b2,0x1325,0x0011b3,0x11b3,0x132b,0x0011a1,0x11a1,0x12ec,0x00118d,0x118d,0x12f5,0x001197,0x1197,0x12ff,0x0011af,0x11af)
GDI32(NtGdiDdDDICreateDevice                            ,  NONE,  NONE,  NONE,  NONE,  NONE,0x12c4,0x1162,0x1162,0x12e5,0x1160,0x1160,0x1334,0x001190,0x1190,0x133d,0x0011a1,0x11a1,0x134a,0x0011b3,0x11b3,0x134b,0x0011b4,0x11b4,0x1351,0x0011a2,0x11a2,0x1312,0x00118e,0x118e,0x131a,0x001198,0x1198,0x1324,0x0011b0,0x11b0)
GDI32(NtGdiDdDDICreateHwContext                         ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1383,0x00118f,0x118f,0x138e,0x001199,0x1199,0x1398,0x0011b1,0x11b1)
GDI32(NtGdiDdDDICreateHwQueue                           ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1385,0x001190,0x1190,0x1390,0x00119a,0x119a,0x139a,0x0011b2,0x11b2)
GDI32(NtGdiDdDDICreateKeyedMutex                        ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x1319,0x1161,0x1161,0x1356,0x001191,0x1191,0x135f,0x0011a2,0x11a2,0x136f,0x0011b4,0x11b4,0x1370,0x0011b5,0x11b5,0x1376,0x0011a3,0x11a3,0x1337,0x001191,0x1191,0x133f,0x00119b,0x119b,0x1349,0x0011b3,0x11b3)
GDI32(NtGdiDdDDICreateKeyedMutex2                       ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x1346,0x001192,0x1192,0x134f,0x0011a3,0x11a3,0x135f,0x0011b5,0x11b5,0x1360,0x0011b6,0x11b6,0x1366,0x0011a4,0x11a4,0x1327,0x001192,0x1192,0x132f,0x00119c,0x119c,0x1339,0x0011b4,0x11b4)
GDI32(NtGdiDdDDICreateOutputDupl                        ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x134b,0x001193,0x1193,0x1354,0x0011a4,0x11a4,0x1364,0x0011b6,0x11b6,0x1365,0x0011b7,0x11b7,0x136b,0x0011a5,0x11a5,0x132c,0x001193,0x1193,0x1334,0x00119d,0x119d,0x133e,0x0011b5,0x11b5)
GDI32(NtGdiDdDDICreateOverlay                           ,  NONE,  NONE,  NONE,  NONE,  NONE,0x12dd,0x1163,0x1163,0x1300,0x1162,0x1162,0x1316,0x001194,0x1194,0x131f,0x0011a5,0x11a5,0x132b,0x0011b7,0x11b7,0x132c,0x0011b8,0x11b8,0x1332,0x0011a6,0x11a6,0x12f3,0x001194,0x1194,0x12fc,0x00119e,0x119e,0x1306,0x0011b6,0x11b6)
GDI32(NtGdiDdDDICreatePagingQueue                       ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x138c,0x0011b8,0x11b8,0x138d,0x0011b9,0x11b9,0x1394,0x0011a7,0x11a7,0x1355,0x001195,0x1195,0x135d,0x00119f,0x119f,0x1367,0x0011b7,0x11b7)
GDI32(NtGdiDdDDICreateProtectedSession                  ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x13a2,0x0011a0,0x11a0,0x13ad,0x0011b8,0x11b8)
GDI32(NtGdiDdDDICreateSwapChain                         ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x13a3,0x0011b9,0x11b9,0x13a3,0x0011ba,0x11ba,0x13aa,0x0011a8,0x11a8,0x136b,0x001196,0x1196,0x1373,0x0011a1,0x11a1,0x137d,0x0011b9,0x11b9)
GDI32(NtGdiDdDDICreateSynchronizationObject             ,  NONE,  NONE,  NONE,  NONE,  NONE,0x12c8,0x1164,0x1164,0x12e9,0x1163,0x1163,0x1330,0x001195,0x1195,0x1339,0x0011a6,0x11a6,0x1346,0x0011ba,0x11ba,0x1347,0x0011bb,0x11bb,0x134d,0x0011a9,0x11a9,0x130e,0x001197,0x1197,0x1316,0x0011a2,0x11a2,0x1320,0x0011ba,0x11ba)
GDI32(NtGdiDdDDIDDisplayEnum                            ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x139e,0x0011a3,0x11a3,0x13a9,0x0011bb,0x11bb)
GDI32(NtGdiDdDDIDestroyAllocation                       ,  NONE,  NONE,  NONE,  NONE,  NONE,0x12c1,0x1165,0x1165,0x12e2,0x1164,0x1164,0x1337,0x001196,0x1196,0x1340,0x0011a7,0x11a7,0x134d,0x0011bb,0x11bb,0x134e,0x0011bc,0x11bc,0x1354,0x0011aa,0x11aa,0x1315,0x001198,0x1198,0x131d,0x0011a4,0x11a4,0x1327,0x0011bc,0x11bc)
GDI32(NtGdiDdDDIDestroyAllocation2                      ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x134e,0x0011bc,0x11bc,0x134f,0x0011bd,0x11bd,0x1355,0x0011ab,0x11ab,0x1316,0x001199,0x1199,0x131e,0x0011a5,0x11a5,0x1328,0x0011bd,0x11bd)
GDI32(NtGdiDdDDIDestroyContext                          ,  NONE,  NONE,  NONE,  NONE,  NONE,0x12c7,0x1166,0x1166,0x12e8,0x1165,0x1165,0x1331,0x001197,0x1197,0x133a,0x0011a8,0x11a8,0x1347,0x0011bd,0x11bd,0x1348,0x0011be,0x11be,0x134e,0x0011ac,0x11ac,0x130f,0x00119a,0x119a,0x1317,0x0011a6,0x11a6,0x1321,0x0011be,0x11be)
GDI32(NtGdiDdDDIDestroyDCFromMemory                     ,  NONE,  NONE,  NONE,  NONE,  NONE,0x12e5,0x1167,0x1167,0x1308,0x1166,0x1166,0x130e,0x001198,0x1198,0x1317,0x0011a9,0x11a9,0x1323,0x0011be,0x11be,0x1324,0x0011bf,0x11bf,0x132a,0x0011ad,0x11ad,0x12eb,0x00119b,0x119b,0x12f4,0x0011a7,0x11a7,0x12fe,0x0011bf,0x11bf)
GDI32(NtGdiDdDDIDestroyDevice                           ,  NONE,  NONE,  NONE,  NONE,  NONE,0x12c5,0x1168,0x1168,0x12e6,0x1167,0x1167,0x1333,0x001199,0x1199,0x133c,0x0011aa,0x11aa,0x1349,0x0011bf,0x11bf,0x134a,0x0011c0,0x11c0,0x1350,0x0011ae,0x11ae,0x1311,0x00119c,0x119c,0x1319,0x0011a8,0x11a8,0x1323,0x0011c0,0x11c0)
GDI32(NtGdiDdDDIDestroyHwContext                        ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1384,0x00119d,0x119d,0x138f,0x0011a9,0x11a9,0x1399,0x0011c1,0x11c1)
GDI32(NtGdiDdDDIDestroyHwQueue                          ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1386,0x00119e,0x119e,0x1391,0x0011aa,0x11aa,0x139b,0x0011c2,0x11c2)
GDI32(NtGdiDdDDIDestroyKeyedMutex                       ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x131b,0x1168,0x1168,0x1354,0x00119a,0x119a,0x135d,0x0011ab,0x11ab,0x136d,0x0011c0,0x11c0,0x136e,0x0011c1,0x11c1,0x1374,0x0011af,0x11af,0x1335,0x00119f,0x119f,0x133d,0x0011ab,0x11ab,0x1347,0x0011c3,0x11c3)
GDI32(NtGdiDdDDIDestroyOutputDupl                       ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x134a,0x00119b,0x119b,0x1353,0x0011ac,0x11ac,0x1363,0x0011c1,0x11c1,0x1364,0x0011c2,0x11c2,0x136a,0x0011b0,0x11b0,0x132b,0x0011a0,0x11a0,0x1333,0x0011ac,0x11ac,0x133d,0x0011c4,0x11c4)
GDI32(NtGdiDdDDIDestroyOverlay                          ,  NONE,  NONE,  NONE,  NONE,  NONE,0x12e0,0x1169,0x1169,0x1303,0x1169,0x1169,0x1313,0x00119c,0x119c,0x131c,0x0011ad,0x11ad,0x1328,0x0011c2,0x11c2,0x1329,0x0011c3,0x11c3,0x132f,0x0011b1,0x11b1,0x12f0,0x0011a1,0x11a1,0x12f9,0x0011ad,0x11ad,0x1303,0x0011c5,0x11c5)
GDI32(NtGdiDdDDIDestroyPagingQueue                      ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x138d,0x0011c3,0x11c3,0x138e,0x0011c4,0x11c4,0x1395,0x0011b2,0x11b2,0x1356,0x0011a2,0x11a2,0x135e,0x0011ae,0x11ae,0x1368,0x0011c6,0x11c6)
GDI32(NtGdiDdDDIDestroyProtectedSession                 ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x13a3,0x0011af,0x11af,0x13ae,0x0011c7,0x11c7)
GDI32(NtGdiDdDDIDestroySynchronizationObject            ,  NONE,  NONE,  NONE,  NONE,  NONE,0x12c9,0x116a,0x116a,0x12eb,0x116a,0x116a,0x132e,0x00119d,0x119d,0x1337,0x0011ae,0x11ae,0x1344,0x0011c4,0x11c4,0x1345,0x0011c5,0x11c5,0x134b,0x0011b3,0x11b3,0x130c,0x0011a3,0x11a3,0x1314,0x0011b0,0x11b0,0x131e,0x0011c8,0x11c8)
GDI32(NtGdiDdDDIDispMgrCreate                           ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x139f,0x0011b1,0x11b1,0x13aa,0x0011c9,0x11c9)
GDI32(NtGdiDdDDIDispMgrSourceOperation                  ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x13a1,0x0011b2,0x11b2,0x13ac,0x0011ca,0x11ca)
GDI32(NtGdiDdDDIDispMgrTargetOperation                  ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x13a0,0x0011b3,0x11b3,0x13ab,0x0011cb,0x11cb)
GDI32(NtGdiDdDDIEnumAdapters                            ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x1320,0x00119e,0x119e,0x1329,0x0011af,0x11af,0x1335,0x0011c5,0x11c5,0x1336,0x0011c6,0x11c6,0x133c,0x0011b4,0x11b4,0x12fd,0x0011a4,0x11a4,0x1305,0x0011b4,0x11b4,0x130f,0x0011cc,0x11cc)
GDI32(NtGdiDdDDIEnumAdapters2                           ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1336,0x0011c6,0x11c6,0x1337,0x0011c7,0x11c7,0x133d,0x0011b5,0x11b5,0x12fe,0x0011a5,0x11a5,0x1306,0x0011b5,0x11b5,0x1310,0x0011cd,0x11cd)
GDI32(NtGdiDdDDIEscape                                  ,  NONE,  NONE,  NONE,  NONE,  NONE,0x12d9,0x116b,0x116b,0x12fb,0x116b,0x116b,0x131c,0x00119f,0x119f,0x1325,0x0011b0,0x11b0,0x1331,0x0011c7,0x11c7,0x1332,0x0011c8,0x11c8,0x1338,0x0011b6,0x11b6,0x12f9,0x0011a6,0x11a6,0x1301,0x0011b6,0x11b6,0x130b,0x0011ce,0x11ce)
GDI32(NtGdiDdDDIEvict                                   ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1386,0x0011c8,0x11c8,0x1387,0x0011c9,0x11c9,0x138d,0x0011b7,0x11b7,0x134e,0x0011a7,0x11a7,0x1356,0x0011b7,0x11b7,0x1360,0x0011cf,0x11cf)
GDI32(NtGdiDdDDIExtractBundleObject                     ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x13aa,0x0011b8,0x11b8,0x13b5,0x0011d0,0x11d0)
GDI32(NtGdiDdDDIFlipOverlay                             ,  NONE,  NONE,  NONE,  NONE,  NONE,0x12df,0x116c,0x116c,0x1302,0x116c,0x116c,0x1314,0x0011a0,0x11a0,0x131d,0x0011b1,0x11b1,0x1329,0x0011c9,0x11c9,0x132a,0x0011ca,0x11ca,0x1330,0x0011b8,0x11b8,0x12f1,0x0011a8,0x11a8,0x12fa,0x0011b9,0x11b9,0x1304,0x0011d1,0x11d1)
GDI32(NtGdiDdDDIFlushHeapTransitions                    ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x13b3,0x0011cb,0x11cb,0x13ba,0x0011b9,0x11b9,0x137b,0x0011a9,0x11a9,0x1386,0x0011ba,0x11ba,0x1390,0x0011d2,0x11d2)
GDI32(NtGdiDdDDIFreeGpuVirtualAddress                   ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1394,0x0011ca,0x11ca,0x1395,0x0011cc,0x11cc,0x139c,0x0011ba,0x11ba,0x135d,0x0011aa,0x11aa,0x1365,0x0011bb,0x11bb,0x136f,0x0011d3,0x11d3)
GDI32(NtGdiDdDDIGetAllocationPriority                   ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x138a,0x0011ab,0x11ab,0x1395,0x0011bc,0x11bc,0x13a0,0x0011d4,0x11d4)
GDI32(NtGdiDdDDIGetCachedHybridQueryValue               ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,0x1373,0x0011b2,0x11b2,0x1398,0x0011cb,0x11cb,0x1399,0x0011cd,0x11cd,0x13a0,0x0011bb,0x11bb,0x1361,0x0011ac,0x11ac,0x1369,0x0011bd,0x11bd,0x1373,0x0011d5,0x11d5)
GDI32(NtGdiDdDDIGetContextInProcessSchedulingPriority   ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x1365,0x0011a1,0x11a1,0x136e,0x0011b3,0x11b3,0x1380,0x0011cc,0x11cc,0x1381,0x0011ce,0x11ce,0x1387,0x0011bc,0x11bc,0x1348,0x0011ad,0x11ad,0x1350,0x0011be,0x11be,0x135a,0x0011d6,0x11d6)
GDI32(NtGdiDdDDIGetContextSchedulingPriority            ,  NONE,  NONE,  NONE,  NONE,  NONE,0x12e7,0x116d,0x116d,0x130a,0x116d,0x116d,0x130c,0x0011a2,0x11a2,0x1315,0x0011b4,0x11b4,0x1321,0x0011cd,0x11cd,0x1322,0x0011cf,0x11cf,0x1328,0x0011bd,0x11bd,0x12e9,0x0011ae,0x11ae,0x12f2,0x0011bf,0x11bf,0x12fc,0x0011d7,0x11d7)
GDI32(NtGdiDdDDIGetDWMVerticalBlankEvent                ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x137d,0x0011ce,0x11ce,0x137e,0x0011d0,0x11d0,0x1384,0x0011be,0x11be,0x1345,0x0011af,0x11af,0x134d,0x0011c0,0x11c0,0x1357,0x0011d8,0x11d8)
GDI32(NtGdiDdDDIGetDeviceState                          ,  NONE,  NONE,  NONE,  NONE,  NONE,0x12e3,0x116e,0x116e,0x1306,0x116e,0x116e,0x1310,0x0011a3,0x11a3,0x1319,0x0011b5,0x11b5,0x1325,0x0011cf,0x11cf,0x1326,0x0011d1,0x11d1,0x132c,0x0011bf,0x11bf,0x12ed,0x0011b0,0x11b0,0x12f6,0x0011c1,0x11c1,0x1300,0x0011d9,0x11d9)
GDI32(NtGdiDdDDIGetDisplayModeList                      ,  NONE,  NONE,  NONE,  NONE,  NONE,0x12d0,0x116f,0x116f,0x12f2,0x116f,0x116f,0x1327,0x0011a4,0x11a4,0x1330,0x0011b6,0x11b6,0x133d,0x0011d0,0x11d0,0x133e,0x0011d2,0x11d2,0x1344,0x0011c0,0x11c0,0x1305,0x0011b1,0x11b1,0x130d,0x0011c2,0x11c2,0x1317,0x0011da,0x11da)
GDI32(NtGdiDdDDIGetMemoryBudgetTarget                   ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1392,0x0011b2,0x11b2,0x139d,0x0011c3,0x11c3,0x13a8,0x0011db,0x11db)
GDI32(NtGdiDdDDIGetMultiPlaneOverlayCaps                ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x138b,0x0011b3,0x11b3,0x1396,0x0011c4,0x11c4,0x13a1,0x0011dc,0x11dc)
GDI32(NtGdiDdDDIGetMultisampleMethodList                ,  NONE,  NONE,  NONE,  NONE,  NONE,0x12d2,0x1170,0x1170,0x12f4,0x1170,0x1170,0x1325,0x0011a5,0x11a5,0x132e,0x0011b7,0x11b7,0x133b,0x0011d1,0x11d1,0x133c,0x0011d3,0x11d3,0x1342,0x0011c1,0x11c1,0x1303,0x0011b4,0x11b4,0x130b,0x0011c5,0x11c5,0x1315,0x0011dd,0x11dd)
GDI32(NtGdiDdDDIGetOverlayState                         ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x131f,0x1171,0x1171,0x1350,0x0011a6,0x11a6,0x1359,0x0011b8,0x11b8,0x1369,0x0011d2,0x11d2,0x136a,0x0011d4,0x11d4,0x1370,0x0011c2,0x11c2,0x1331,0x0011b5,0x11b5,0x1339,0x0011c6,0x11c6,0x1343,0x0011de,0x11de)
GDI32(NtGdiDdDDIGetPostCompositionCaps                  ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x138c,0x0011b6,0x11b6,0x1397,0x0011c7,0x11c7,0x13a2,0x0011df,0x11df)
GDI32(NtGdiDdDDIGetPresentHistory                       ,  NONE,  NONE,  NONE,  NONE,  NONE,0x12dc,0x1171,0x1171,0x12fe,0x1172,0x1172,0x1318,0x0011a7,0x11a7,0x1321,0x0011b9,0x11b9,0x132d,0x0011d3,0x11d3,0x132e,0x0011d5,0x11d5,0x1334,0x0011c3,0x11c3,0x12f5,0x0011b7,0x11b7,0x12fe,0x0011c8,0x11c8,0x1308,0x0011e0,0x11e0)
GDI32(NtGdiDdDDIGetPresentQueueEvent                    ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x12ff,0x1173,0x1173,0x1317,0x0011a8,0x11a8,0x1320,0x0011ba,0x11ba,0x132c,0x0011d4,0x11d4,0x132d,0x0011d6,0x11d6,0x1333,0x0011c4,0x11c4,0x12f4,0x0011b8,0x11b8,0x12fd,0x0011c9,0x11c9,0x1307,0x0011e1,0x11e1)
GDI32(NtGdiDdDDIGetProcessDeviceLostSupport             ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x13ac,0x0011ca,0x11ca,  NONE,    NONE,  NONE)
GDI32(NtGdiDdDDIGetProcessDeviceRemovalSupport          ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x13b8,0x0011e2,0x11e2)
GDI32(NtGdiDdDDIGetProcessSchedulingPriorityBand        ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1391,0x0011b9,0x11b9,0x139c,0x0011cb,0x11cb,0x13a7,0x0011e3,0x11e3)
GDI32(NtGdiDdDDIGetProcessSchedulingPriorityClass       ,  NONE,  NONE,  NONE,  NONE,  NONE,0x12e9,0x1172,0x1172,0x130c,0x1174,0x1174,0x130a,0x0011a9,0x11a9,0x1313,0x0011bb,0x11bb,0x131f,0x0011d5,0x11d5,0x1320,0x0011d7,0x11d7,0x1326,0x0011c5,0x11c5,0x12e7,0x0011ba,0x11ba,0x12f0,0x0011cc,0x11cc,0x12fa,0x0011e4,0x11e4)
GDI32(NtGdiDdDDIGetResourcePresentPrivateDriverData     ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1391,0x0011d6,0x11d6,0x1392,0x0011d8,0x11d8,0x1399,0x0011c6,0x11c6,0x135a,0x0011bb,0x11bb,0x1362,0x0011cd,0x11cd,0x136c,0x0011e5,0x11e5)
GDI32(NtGdiDdDDIGetRuntimeData                          ,  NONE,  NONE,  NONE,  NONE,  NONE,0x12cc,0x1173,0x1173,0x12ee,0x1175,0x1175,0x132b,0x0011aa,0x11aa,0x1334,0x0011bc,0x11bc,0x1341,0x0011d7,0x11d7,0x1342,0x0011d9,0x11d9,0x1348,0x0011c7,0x11c7,0x1309,0x0011bc,0x11bc,0x1311,0x0011ce,0x11ce,0x131b,0x0011e6,0x11e6)
GDI32(NtGdiDdDDIGetScanLine                             ,  NONE,  NONE,  NONE,  NONE,  NONE,0x12eb,0x1174,0x1174,0x130e,0x1176,0x1176,0x1308,0x0011ab,0x11ab,0x1311,0x0011bd,0x11bd,0x131d,0x0011d8,0x11d8,0x131e,0x0011da,0x11da,0x1324,0x0011c8,0x11c8,0x12e5,0x0011bd,0x11bd,0x12ee,0x0011cf,0x11cf,0x12f8,0x0011e7,0x11e7)
GDI32(NtGdiDdDDIGetSetSwapChainMetadata                 ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x13a5,0x0011d9,0x11d9,0x13a5,0x0011db,0x11db,0x13ac,0x0011c9,0x11c9,0x136d,0x0011be,0x11be,0x1378,0x0011d0,0x11d0,0x1382,0x0011e8,0x11e8)
GDI32(NtGdiDdDDIGetSharedPrimaryHandle                  ,  NONE,  NONE,  NONE,  NONE,  NONE,0x12d8,0x1175,0x1175,0x12fa,0x1177,0x1177,0x131d,0x0011ac,0x11ac,0x1326,0x0011be,0x11be,0x1332,0x0011da,0x11da,0x1333,0x0011dc,0x11dc,0x1339,0x0011ca,0x11ca,0x12fa,0x0011bf,0x11bf,0x1302,0x0011d1,0x11d1,0x130c,0x0011e9,0x11e9)
GDI32(NtGdiDdDDIGetSharedResourceAdapterLuid            ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x1366,0x0011ad,0x11ad,0x136f,0x0011bf,0x11bf,0x1381,0x0011db,0x11db,0x1382,0x0011dd,0x11dd,0x1388,0x0011cb,0x11cb,0x1349,0x0011c0,0x11c0,0x1351,0x0011d2,0x11d2,0x135b,0x0011ea,0x11ea)
GDI32(NtGdiDdDDIGetYieldPercentage                      ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1390,0x0011c1,0x11c1,0x139b,0x0011d3,0x11d3,0x13a6,0x0011eb,0x11eb)
GDI32(NtGdiDdDDIInvalidateActiveVidPn                   ,  NONE,  NONE,  NONE,  NONE,  NONE,0x12ee,0x1176,0x1176,0x1311,0x1178,0x1178,0x135e,0x0011ae,0x11ae,0x1367,0x0011c0,0x11c0,0x1377,0x0011dc,0x11dc,0x1378,0x0011de,0x11de,0x137e,0x0011cc,0x11cc,0x133f,0x0011c2,0x11c2,0x1347,0x0011d4,0x11d4,0x1351,0x0011ec,0x11ec)
GDI32(NtGdiDdDDIInvalidateCache                         ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1390,0x0011dd,0x11dd,0x1391,0x0011df,0x11df,0x1398,0x0011cd,0x11cd,0x1359,0x0011c3,0x11c3,0x1361,0x0011d5,0x11d5,0x136b,0x0011ed,0x11ed)
GDI32(NtGdiDdDDILock                                    ,  NONE,  NONE,  NONE,  NONE,  NONE,0x12ce,0x1177,0x1177,0x12f0,0x1179,0x1179,0x1329,0x0011af,0x11af,0x1332,0x0011c1,0x11c1,0x133f,0x0011de,0x11de,0x1340,0x0011e0,0x11e0,0x1346,0x0011ce,0x11ce,0x1307,0x0011c4,0x11c4,0x130f,0x0011d6,0x11d6,0x1319,0x0011ee,0x11ee)
GDI32(NtGdiDdDDILock2                                   ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x138e,0x0011df,0x11df,0x138f,0x0011e1,0x11e1,0x1396,0x0011cf,0x11cf,0x1357,0x0011c5,0x11c5,0x135f,0x0011d7,0x11d7,0x1369,0x0011ef,0x11ef)
GDI32(NtGdiDdDDIMakeResident                            ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1385,0x0011e0,0x11e0,0x1386,0x0011e2,0x11e2,0x138c,0x0011d0,0x11d0,0x134d,0x0011c6,0x11c6,0x1355,0x0011d8,0x11d8,0x135f,0x0011f0,0x11f0)
GDI32(NtGdiDdDDIMapGpuVirtualAddress                    ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1392,0x0011e1,0x11e1,0x1393,0x0011e3,0x11e3,0x139a,0x0011d1,0x11d1,0x135b,0x0011c7,0x11c7,0x1363,0x0011d9,0x11d9,0x136d,0x0011f1,0x11f1)
GDI32(NtGdiDdDDIMarkDeviceAsError                       ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x13b2,0x0011e2,0x11e2,0x13b2,0x0011e4,0x11e4,0x13b9,0x0011d2,0x11d2,0x137a,0x0011c8,0x11c8,0x1385,0x0011da,0x11da,0x138f,0x0011f2,0x11f2)
GDI32(NtGdiDdDDINetDispGetNextChunkInfo                 ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,0x1375,0x0011c2,0x11c2,0x139a,0x0011e3,0x11e3,0x139b,0x0011e5,0x11e5,0x13a2,0x0011d3,0x11d3,0x1363,0x0011c9,0x11c9,0x136b,0x0011db,0x11db,0x1375,0x0011f3,0x11f3)
GDI32(NtGdiDdDDINetDispQueryMiracastDisplayDeviceStatus ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,0x1379,0x0011c3,0x11c3,0x139f,0x0011e4,0x11e4,0x139f,0x0011e6,0x11e6,0x13a6,0x0011d4,0x11d4,0x1367,0x0011ca,0x11ca,0x136f,0x0011dc,0x11dc,0x1379,0x0011f4,0x11f4)
GDI32(NtGdiDdDDINetDispQueryMiracastDisplayDeviceSupport,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,0x1376,0x0011c4,0x11c4,0x139b,0x0011e5,0x11e5,0x139c,0x0011e7,0x11e7,0x13a3,0x0011d5,0x11d5,0x1364,0x0011cb,0x11cb,0x136c,0x0011dd,0x11dd,0x1376,0x0011f5,0x11f5)
GDI32(NtGdiDdDDINetDispStartMiracastDisplayDevice       ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,0x1377,0x0011c5,0x11c5,0x139c,0x0011e6,0x11e6,0x139d,0x0011e8,0x11e8,0x13a4,0x0011d6,0x11d6,0x1365,0x0011cc,0x11cc,0x136d,0x0011de,0x11de,0x1377,0x0011f6,0x11f6)
GDI32(NtGdiDdDDINetDispStartMiracastDisplayDeviceEx     ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x139d,0x0011e7,0x11e7,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
GDI32(NtGdiDdDDINetDispStopMiracastDisplayDevice        ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,0x1378,0x0011c6,0x11c6,0x139e,0x0011e8,0x11e8,0x139e,0x0011e9,0x11e9,0x13a5,0x0011d7,0x11d7,0x1366,0x0011cd,0x11cd,0x136e,0x0011df,0x11df,0x1378,0x0011f7,0x11f7)
GDI32(NtGdiDdDDINetDispStopSessions                     ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x13a0,0x0011e9,0x11e9,0x13a0,0x0011ea,0x11ea,0x13a7,0x0011d8,0x11d8,0x1368,0x0011ce,0x11ce,0x1370,0x0011e0,0x11e0,0x137a,0x0011f8,0x11f8)
GDI32(NtGdiDdDDIOfferAllocations                        ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x134d,0x0011b0,0x11b0,0x1356,0x0011c7,0x11c7,0x1366,0x0011ea,0x11ea,0x1367,0x0011eb,0x11eb,0x136d,0x0011d9,0x11d9,0x132e,0x0011cf,0x11cf,0x1336,0x0011e1,0x11e1,0x1340,0x0011f9,0x11f9)
GDI32(NtGdiDdDDIOpenAdapterFromDeviceName               ,  NONE,  NONE,  NONE,  NONE,  NONE,0x12d5,0x1178,0x1178,0x12f7,0x117a,0x117a,0x1322,0x0011b1,0x11b1,0x132b,0x0011c8,0x11c8,0x1338,0x0011eb,0x11eb,0x1339,0x0011ec,0x11ec,0x133f,0x0011da,0x11da,0x1300,0x0011d0,0x11d0,0x1308,0x0011e2,0x11e2,0x1312,0x0011fa,0x11fa)
GDI32(NtGdiDdDDIOpenAdapterFromHdc                      ,  NONE,  NONE,  NONE,  NONE,  NONE,0x12d6,0x1179,0x1179,0x12f8,0x117b,0x117b,0x1321,0x0011b2,0x11b2,0x132a,0x0011c9,0x11c9,0x1337,0x0011ec,0x11ec,0x1338,0x0011ed,0x11ed,0x133e,0x0011db,0x11db,0x12ff,0x0011d1,0x11d1,0x1307,0x0011e3,0x11e3,0x1311,0x0011fb,0x11fb)
GDI32(NtGdiDdDDIOpenAdapterFromLuid                     ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x131f,0x0011b3,0x11b3,0x1328,0x0011ca,0x11ca,0x1334,0x0011ed,0x11ed,0x1335,0x0011ee,0x11ee,0x133b,0x0011dc,0x11dc,0x12fc,0x0011d2,0x11d2,0x1304,0x0011e4,0x11e4,0x130e,0x0011fc,0x11fc)
GDI32(NtGdiDdDDIOpenBundleObjectNtHandleFromName        ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x13b6,0x0011fd,0x11fd)
GDI32(NtGdiDdDDIOpenKeyedMutex                          ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x131a,0x117c,0x117c,0x1355,0x0011b4,0x11b4,0x135e,0x0011cb,0x11cb,0x136e,0x0011ee,0x11ee,0x136f,0x0011ef,0x11ef,0x1375,0x0011dd,0x11dd,0x1336,0x0011d3,0x11d3,0x133e,0x0011e5,0x11e5,0x1348,0x0011fe,0x11fe)
GDI32(NtGdiDdDDIOpenKeyedMutex2                         ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x1345,0x0011b5,0x11b5,0x134e,0x0011cc,0x11cc,0x135e,0x0011ef,0x11ef,0x135f,0x0011f0,0x11f0,0x1365,0x0011de,0x11de,0x1326,0x0011d4,0x11d4,0x132e,0x0011e6,0x11e6,0x1338,0x0011ff,0x11ff)
GDI32(NtGdiDdDDIOpenKeyedMutexFromNtHandle              ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x13b9,0x001200,0x1200)
GDI32(NtGdiDdDDIOpenNtHandleFromName                    ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x133b,0x0011b6,0x11b6,0x1344,0x0011cd,0x11cd,0x1353,0x0011f0,0x11f0,0x1354,0x0011f1,0x11f1,0x135a,0x0011df,0x11df,0x131b,0x0011d5,0x11d5,0x1323,0x0011e7,0x11e7,0x132d,0x001201,0x1201)
GDI32(NtGdiDdDDIOpenProtectedSessionFromNtHandle        ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x13a6,0x0011e8,0x11e8,0x13b1,0x001202,0x1202)
GDI32(NtGdiDdDDIOpenResource                            ,  NONE,  NONE,  NONE,  NONE,  NONE,0x12c0,0x117a,0x117a,0x12e1,0x117d,0x117d,0x133a,0x0011b7,0x11b7,0x1343,0x0011ce,0x11ce,0x1352,0x0011f1,0x11f1,0x1353,0x0011f2,0x11f2,0x1359,0x0011e0,0x11e0,0x131a,0x0011d6,0x11d6,0x1322,0x0011e9,0x11e9,0x132c,0x001203,0x1203)
GDI32(NtGdiDdDDIOpenResourceFromNtHandle                ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x1338,0x0011b8,0x11b8,0x1341,0x0011cf,0x11cf,0x134f,0x0011f2,0x11f2,0x1350,0x0011f3,0x11f3,0x1356,0x0011e1,0x11e1,0x1317,0x0011d7,0x11d7,0x131f,0x0011ea,0x11ea,0x1329,0x001204,0x1204)
GDI32(NtGdiDdDDIOpenSwapChain                           ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x13a4,0x0011f3,0x11f3,0x13a4,0x0011f4,0x11f4,0x13ab,0x0011e2,0x11e2,0x136c,0x0011d8,0x11d8,0x1374,0x0011eb,0x11eb,0x137e,0x001205,0x1205)
GDI32(NtGdiDdDDIOpenSyncObjectFromNtHandle              ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x1339,0x0011b9,0x11b9,0x1342,0x0011d0,0x11d0,0x1350,0x0011f4,0x11f4,0x1351,0x0011f5,0x11f5,0x1357,0x0011e3,0x11e3,0x1318,0x0011d9,0x11d9,0x1320,0x0011ec,0x11ec,0x132a,0x001206,0x1206)
GDI32(NtGdiDdDDIOpenSyncObjectFromNtHandle2             ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1351,0x0011f5,0x11f5,0x1352,0x0011f6,0x11f6,0x1358,0x0011e4,0x11e4,0x1319,0x0011da,0x11da,0x1321,0x0011ed,0x11ed,0x132b,0x001207,0x1207)
GDI32(NtGdiDdDDIOpenSyncObjectNtHandleFromName          ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1354,0x0011f6,0x11f6,0x1355,0x0011f7,0x11f7,0x135b,0x0011e5,0x11e5,0x131c,0x0011db,0x11db,0x1324,0x0011ee,0x11ee,0x132e,0x001208,0x1208)
GDI32(NtGdiDdDDIOpenSynchronizationObject               ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x12ea,0x117e,0x117e,0x132f,0x0011ba,0x11ba,0x1338,0x0011d1,0x11d1,0x1345,0x0011f7,0x11f7,0x1346,0x0011f8,0x11f8,0x134c,0x0011e6,0x11e6,0x130d,0x0011dc,0x11dc,0x1315,0x0011ef,0x11ef,0x131f,0x001209,0x1209)
GDI32(NtGdiDdDDIOutputDuplGetFrameInfo                  ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x1349,0x0011bb,0x11bb,0x1352,0x0011d2,0x11d2,0x1362,0x0011f8,0x11f8,0x1363,0x0011f9,0x11f9,0x1369,0x0011e7,0x11e7,0x132a,0x0011dd,0x11dd,0x1332,0x0011f0,0x11f0,0x133c,0x00120a,0x120a)
GDI32(NtGdiDdDDIOutputDuplGetMetaData                   ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x1348,0x0011bc,0x11bc,0x1351,0x0011d3,0x11d3,0x1361,0x0011f9,0x11f9,0x1362,0x0011fa,0x11fa,0x1368,0x0011e8,0x11e8,0x1329,0x0011de,0x11de,0x1331,0x0011f1,0x11f1,0x133b,0x00120b,0x120b)
GDI32(NtGdiDdDDIOutputDuplGetPointerShapeData           ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x1347,0x0011bd,0x11bd,0x1350,0x0011d4,0x11d4,0x1360,0x0011fa,0x11fa,0x1361,0x0011fb,0x11fb,0x1367,0x0011e9,0x11e9,0x1328,0x0011df,0x11df,0x1330,0x0011f2,0x11f2,0x133a,0x00120c,0x120c)
GDI32(NtGdiDdDDIOutputDuplPresent                       ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x1342,0x0011be,0x11be,0x134b,0x0011d5,0x11d5,0x135b,0x0011fb,0x11fb,0x135c,0x0011fc,0x11fc,0x1362,0x0011ea,0x11ea,0x1323,0x0011e0,0x11e0,0x132b,0x0011f3,0x11f3,0x1335,0x00120d,0x120d)
GDI32(NtGdiDdDDIOutputDuplReleaseFrame                  ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x1340,0x0011bf,0x11bf,0x1349,0x0011d6,0x11d6,0x1359,0x0011fc,0x11fc,0x135a,0x0011fd,0x11fd,0x1360,0x0011eb,0x11eb,0x1321,0x0011e1,0x11e1,0x1329,0x0011f4,0x11f4,0x1333,0x00120e,0x120e)
GDI32(NtGdiDdDDIPinDirectFlipResources                  ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x1361,0x0011c0,0x11c0,0x136a,0x0011d7,0x11d7,0x137a,0x0011fd,0x11fd,0x137b,0x0011fe,0x11fe,0x1381,0x0011ec,0x11ec,0x1342,0x0011e2,0x11e2,0x134a,0x0011f5,0x11f5,0x1354,0x00120f,0x120f)
GDI32(NtGdiDdDDIPollDisplayChildren                     ,  NONE,  NONE,  NONE,  NONE,  NONE,0x12ed,0x117b,0x117b,0x1310,0x117f,0x117f,0x135f,0x0011c1,0x11c1,0x1368,0x0011d8,0x11d8,0x1378,0x0011fe,0x11fe,0x1379,0x0011ff,0x11ff,0x137f,0x0011ed,0x11ed,0x1340,0x0011e3,0x11e3,0x1348,0x0011f6,0x11f6,0x1352,0x001210,0x1210)
GDI32(NtGdiDdDDIPresent                                 ,  NONE,  NONE,  NONE,  NONE,  NONE,0x12d3,0x117c,0x117c,0x12f5,0x1180,0x1180,0x1324,0x0011c2,0x11c2,0x132d,0x0011d9,0x11d9,0x133a,0x0011ff,0x11ff,0x133b,0x001200,0x1200,0x1341,0x0011ee,0x11ee,0x1302,0x0011e4,0x11e4,0x130a,0x0011f7,0x11f7,0x1314,0x001211,0x1211)
GDI32(NtGdiDdDDIPresentMultiPlaneOverlay                ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,0x1371,0x0011da,0x11da,0x1383,0x001200,0x1200,0x1384,0x001201,0x1201,0x138a,0x0011ef,0x11ef,0x134b,0x0011e5,0x11e5,0x1353,0x0011f8,0x11f8,0x135d,0x001212,0x1212)
GDI32(NtGdiDdDDIPresentMultiPlaneOverlay2               ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x13ab,0x001201,0x1201,0x13ab,0x001202,0x1202,0x13b2,0x0011f0,0x11f0,0x1373,0x0011e6,0x11e6,0x137e,0x0011f9,0x11f9,0x1388,0x001213,0x1213)
GDI32(NtGdiDdDDIPresentMultiPlaneOverlay3               ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x13bf,0x0011f1,0x11f1,0x1380,0x0011e7,0x11e7,0x138b,0x0011fa,0x11fa,0x1395,0x001214,0x1214)
GDI32(NtGdiDdDDIPresentRedirected                       ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x13a8,0x0011fb,0x11fb,0x13b3,0x001215,0x1215)
GDI32(NtGdiDdDDIQueryAdapterInfo                        ,  NONE,  NONE,  NONE,  NONE,  NONE,0x12cd,0x117d,0x117d,0x12ef,0x1181,0x1181,0x132a,0x0011c3,0x11c3,0x1333,0x0011db,0x11db,0x1340,0x001202,0x1202,0x1341,0x001203,0x1203,0x1347,0x0011f2,0x11f2,0x1308,0x0011e8,0x11e8,0x1310,0x0011fc,0x11fc,0x131a,0x001216,0x1216)
GDI32(NtGdiDdDDIQueryAllocationResidency                ,  NONE,  NONE,  NONE,  NONE,  NONE,0x12c3,0x117e,0x117e,0x12e4,0x1182,0x1182,0x1335,0x0011c4,0x11c4,0x133e,0x0011dc,0x11dc,0x134b,0x001203,0x1203,0x134c,0x001204,0x1204,0x1352,0x0011f3,0x11f3,0x1313,0x0011e9,0x11e9,0x131b,0x0011fd,0x11fd,0x1325,0x001217,0x1217)
GDI32(NtGdiDdDDIQueryClockCalibration                   ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x13ae,0x001204,0x1204,0x13ae,0x001205,0x1205,0x13b5,0x0011f4,0x11f4,0x1376,0x0011ea,0x11ea,0x1381,0x0011fe,0x11fe,0x138b,0x001218,0x1218)
GDI32(NtGdiDdDDIQueryFSEBlock                           ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x13c1,0x0011f5,0x11f5,0x1382,0x0011eb,0x11eb,0x138d,0x0011ff,0x11ff,0x1397,0x001219,0x1219)
GDI32(NtGdiDdDDIQueryProcessOfferInfo                   ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x13bc,0x0011f6,0x11f6,0x137d,0x0011ec,0x11ec,0x1388,0x001200,0x1200,0x1392,0x00121a,0x121a)
GDI32(NtGdiDdDDIQueryProtectedSessionInfoFromNtHandle   ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x13a5,0x001201,0x1201,0x13b0,0x00121b,0x121b)
GDI32(NtGdiDdDDIQueryProtectedSessionStatus             ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x13a4,0x001202,0x1202,0x13af,0x00121c,0x121c)
GDI32(NtGdiDdDDIQueryRemoteVidPnSourceFromGdiDisplayName,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x1341,0x0011c5,0x11c5,0x134a,0x0011dd,0x11dd,0x135a,0x001205,0x1205,0x135b,0x001206,0x1206,0x1361,0x0011f7,0x11f7,0x1322,0x0011ed,0x11ed,0x132a,0x001203,0x1203,0x1334,0x00121d,0x121d)
GDI32(NtGdiDdDDIQueryResourceInfo                       ,  NONE,  NONE,  NONE,  NONE,  NONE,0x12bf,0x117f,0x117f,0x12e0,0x1183,0x1183,0x133e,0x0011c6,0x11c6,0x1347,0x0011de,0x11de,0x1357,0x001206,0x1206,0x1358,0x001207,0x1207,0x135e,0x0011f8,0x11f8,0x131f,0x0011ee,0x11ee,0x1327,0x001204,0x1204,0x1331,0x00121e,0x121e)
GDI32(NtGdiDdDDIQueryResourceInfoFromNtHandle           ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x133d,0x0011c7,0x11c7,0x1346,0x0011df,0x11df,0x1356,0x001207,0x1207,0x1357,0x001208,0x1208,0x135d,0x0011f9,0x11f9,0x131e,0x0011ef,0x11ef,0x1326,0x001205,0x1205,0x1330,0x00121f,0x121f)
GDI32(NtGdiDdDDIQueryStatistics                         ,  NONE,  NONE,  NONE,  NONE,  NONE,0x12da,0x1180,0x1180,0x12fc,0x1184,0x1184,0x131b,0x0011c8,0x11c8,0x1324,0x0011e0,0x11e0,0x1330,0x001208,0x1208,0x1331,0x001209,0x1209,0x1337,0x0011fa,0x11fa,0x12f8,0x0011f0,0x11f0,0x1300,0x001206,0x1206,0x130a,0x001220,0x1220)
GDI32(NtGdiDdDDIQueryVidPnExclusiveOwnership            ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x13af,0x001209,0x1209,0x13af,0x00120a,0x120a,0x13b6,0x0011fb,0x11fb,0x1377,0x0011f1,0x11f1,0x1382,0x001207,0x1207,0x138c,0x001221,0x1221)
GDI32(NtGdiDdDDIQueryVideoMemoryInfo                    ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x13a1,0x00120a,0x120a,0x13a1,0x00120b,0x120b,0x13a8,0x0011fc,0x11fc,0x1369,0x0011f2,0x11f2,0x1371,0x001208,0x1208,0x137b,0x001222,0x1222)
GDI32(NtGdiDdDDIReclaimAllocations                      ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x134c,0x0011c9,0x11c9,0x1355,0x0011e1,0x11e1,0x1365,0x00120b,0x120b,0x1366,0x00120c,0x120c,0x136c,0x0011fd,0x11fd,0x132d,0x0011f3,0x11f3,0x1335,0x001209,0x1209,0x133f,0x001223,0x1223)
GDI32(NtGdiDdDDIReclaimAllocations2                     ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x13ac,0x00120c,0x120c,0x13ac,0x00120d,0x120d,0x13b3,0x0011fe,0x11fe,0x1374,0x0011f4,0x11f4,0x137f,0x00120a,0x120a,0x1389,0x001224,0x1224)
GDI32(NtGdiDdDDIReleaseKeyedMutex                       ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x131d,0x1185,0x1185,0x1352,0x0011ca,0x11ca,0x135b,0x0011e2,0x11e2,0x136b,0x00120d,0x120d,0x136c,0x00120e,0x120e,0x1372,0x0011ff,0x11ff,0x1333,0x0011f5,0x11f5,0x133b,0x00120b,0x120b,0x1345,0x001225,0x1225)
GDI32(NtGdiDdDDIReleaseKeyedMutex2                      ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x1343,0x0011cb,0x11cb,0x134c,0x0011e3,0x11e3,0x135c,0x00120e,0x120e,0x135d,0x00120f,0x120f,0x1363,0x001200,0x1200,0x1324,0x0011f6,0x11f6,0x132c,0x00120c,0x120c,0x1336,0x001226,0x1226)
GDI32(NtGdiDdDDIReleaseProcessVidPnSourceOwners         ,  NONE,  NONE,  NONE,  NONE,  NONE,0x12ea,0x1181,0x1181,0x130d,0x1186,0x1186,0x1309,0x0011cc,0x11cc,0x1312,0x0011e4,0x11e4,0x131e,0x00120f,0x120f,0x131f,0x001210,0x1210,0x1325,0x001201,0x1201,0x12e6,0x0011f7,0x11f7,0x12ef,0x00120d,0x120d,0x12f9,0x001227,0x1227)
GDI32(NtGdiDdDDIReleaseSwapChain                        ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x13a7,0x001210,0x1210,0x13a7,0x001211,0x1211,0x13ae,0x001202,0x1202,0x136f,0x0011f8,0x11f8,0x137a,0x00120e,0x120e,0x1384,0x001228,0x1228)
GDI32(NtGdiDdDDIRemoveSurfaceFromSwapChain              ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1376,0x00120f,0x120f,0x1380,0x001229,0x1229)
GDI32(NtGdiDdDDIRender                                  ,  NONE,  NONE,  NONE,  NONE,  NONE,0x12d4,0x1182,0x1182,0x12f6,0x1187,0x1187,0x1323,0x0011cd,0x11cd,0x132c,0x0011e5,0x11e5,0x1339,0x001211,0x1211,0x133a,0x001212,0x1212,0x1340,0x001203,0x1203,0x1301,0x0011f9,0x11f9,0x1309,0x001210,0x1210,0x1313,0x00122a,0x122a)
GDI32(NtGdiDdDDIReserveGpuVirtualAddress                ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1393,0x001212,0x1212,0x1394,0x001213,0x1213,0x139b,0x001204,0x1204,0x135c,0x0011fa,0x11fa,0x1364,0x001211,0x1211,0x136e,0x00122b,0x122b)
GDI32(NtGdiDdDDISetAllocationPriority                   ,  NONE,  NONE,  NONE,  NONE,  NONE,0x12c2,0x1183,0x1183,0x12e3,0x1188,0x1188,0x1336,0x0011ce,0x11ce,0x133f,0x0011e6,0x11e6,0x134c,0x001213,0x1213,0x134d,0x001214,0x1214,0x1353,0x001205,0x1205,0x1314,0x0011fb,0x11fb,0x131c,0x001212,0x1212,0x1326,0x00122c,0x122c)
GDI32(NtGdiDdDDISetContextInProcessSchedulingPriority   ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x1364,0x0011cf,0x11cf,0x136d,0x0011e7,0x11e7,0x137f,0x001214,0x1214,0x1380,0x001215,0x1215,0x1386,0x001206,0x1206,0x1347,0x0011fc,0x11fc,0x134f,0x001213,0x1213,0x1359,0x00122d,0x122d)
GDI32(NtGdiDdDDISetContextSchedulingPriority            ,  NONE,  NONE,  NONE,  NONE,  NONE,0x12e6,0x1184,0x1184,0x1309,0x1189,0x1189,0x130d,0x0011d0,0x11d0,0x1316,0x0011e8,0x11e8,0x1322,0x001215,0x1215,0x1323,0x001216,0x1216,0x1329,0x001207,0x1207,0x12ea,0x0011fd,0x11fd,0x12f3,0x001214,0x1214,0x12fd,0x00122e,0x122e)
GDI32(NtGdiDdDDISetDeviceLostSupport                    ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x13ab,0x001215,0x1215,  NONE,    NONE,  NONE)
GDI32(NtGdiDdDDISetDisplayMode                          ,  NONE,  NONE,  NONE,  NONE,  NONE,0x12d1,0x1185,0x1185,0x12f3,0x118a,0x118a,0x1326,0x0011d1,0x11d1,0x132f,0x0011e9,0x11e9,0x133c,0x001216,0x1216,0x133d,0x001217,0x1217,0x1343,0x001208,0x1208,0x1304,0x0011fe,0x11fe,0x130c,0x001216,0x1216,0x1316,0x00122f,0x122f)
GDI32(NtGdiDdDDISetDisplayPrivateDriverFormat           ,  NONE,  NONE,  NONE,  NONE,  NONE,0x12f3,0x1186,0x1186,0x1316,0x118b,0x118b,0x1359,0x0011d2,0x11d2,0x1362,0x0011ea,0x11ea,0x1372,0x001217,0x1217,0x1373,0x001218,0x1218,0x1379,0x001209,0x1209,0x133a,0x0011ff,0x11ff,0x1342,0x001217,0x1217,0x134c,0x001230,0x1230)
GDI32(NtGdiDdDDISetDodIndirectSwapchain                 ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x13a9,0x001218,0x1218,0x13a9,0x001219,0x1219,0x13b0,0x00120a,0x120a,0x1371,0x001200,0x1200,0x137c,0x001218,0x1218,0x1386,0x001231,0x1231)
GDI32(NtGdiDdDDISetFSEBlock                             ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x13c0,0x00120b,0x120b,0x1381,0x001201,0x1201,0x138c,0x001219,0x1219,0x1396,0x001232,0x1232)
GDI32(NtGdiDdDDISetGammaRamp                            ,  NONE,  NONE,  NONE,  NONE,  NONE,0x12e2,0x1187,0x1187,0x1305,0x118c,0x118c,0x1311,0x0011d3,0x11d3,0x131a,0x0011eb,0x11eb,0x1326,0x001219,0x1219,0x1327,0x00121a,0x121a,0x132d,0x00120c,0x120c,0x12ee,0x001202,0x1202,0x12f7,0x00121a,0x121a,0x1301,0x001233,0x1233)
GDI32(NtGdiDdDDISetHwProtectionTeardownRecovery         ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x13b4,0x00121b,0x121b,0x13bb,0x00120d,0x120d,0x137c,0x001203,0x1203,0x1387,0x00121b,0x121b,0x1391,0x001234,0x1234)
GDI32(NtGdiDdDDISetMemoryBudgetTarget                   ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x138f,0x001204,0x1204,0x139a,0x00121c,0x121c,0x13a5,0x001235,0x1235)
GDI32(NtGdiDdDDISetMonitorColorSpaceTransform           ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x13a7,0x00121d,0x121d,0x13b2,0x001236,0x1236)
GDI32(NtGdiDdDDISetProcessDeviceRemovalSupport          ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x13b7,0x001237,0x1237)
GDI32(NtGdiDdDDISetProcessSchedulingPriorityBand        ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x138e,0x001205,0x1205,0x1399,0x00121e,0x121e,0x13a4,0x001238,0x1238)
GDI32(NtGdiDdDDISetProcessSchedulingPriorityClass       ,  NONE,  NONE,  NONE,  NONE,  NONE,0x12e8,0x1188,0x1188,0x130b,0x118d,0x118d,0x130b,0x0011d4,0x11d4,0x1314,0x0011ec,0x11ec,0x1320,0x00121a,0x121a,0x1321,0x00121c,0x121c,0x1327,0x00120e,0x120e,0x12e8,0x001206,0x1206,0x12f1,0x00121f,0x121f,0x12fb,0x001239,0x1239)
GDI32(NtGdiDdDDISetQueuedLimit                          ,  NONE,  NONE,  NONE,  NONE,  NONE,0x12ec,0x1189,0x1189,0x130f,0x118e,0x118e,0x1360,0x0011d5,0x11d5,0x1369,0x0011ed,0x11ed,0x1379,0x00121b,0x121b,0x137a,0x00121d,0x121d,0x1380,0x00120f,0x120f,0x1341,0x001207,0x1207,0x1349,0x001220,0x1220,0x1353,0x00123a,0x123a)
GDI32(NtGdiDdDDISetStablePowerState                     ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x13ad,0x00121c,0x121c,0x13ad,0x00121e,0x121e,0x13b4,0x001210,0x1210,0x1375,0x001208,0x1208,0x1380,0x001221,0x1221,0x138a,0x00123b,0x123b)
GDI32(NtGdiDdDDISetStereoEnabled                        ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x1367,0x0011d6,0x11d6,0x1370,0x0011ee,0x11ee,0x1382,0x00121d,0x121d,0x1383,0x00121f,0x121f,0x1389,0x001211,0x1211,0x134a,0x001209,0x1209,0x1352,0x001222,0x1222,0x135c,0x00123c,0x123c)
GDI32(NtGdiDdDDISetSyncRefreshCountWaitTarget           ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x137e,0x00121e,0x121e,0x137f,0x001220,0x1220,0x1385,0x001212,0x1212,0x1346,0x00120a,0x120a,0x134e,0x001223,0x1223,0x1358,0x00123d,0x123d)
GDI32(NtGdiDdDDISetVidPnSourceHwProtection              ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x13b1,0x00121f,0x121f,0x13b1,0x001221,0x1221,0x13b8,0x001213,0x1213,0x1379,0x00120b,0x120b,0x1384,0x001224,0x1224,0x138e,0x00123e,0x123e)
GDI32(NtGdiDdDDISetVidPnSourceOwner                     ,  NONE,  NONE,  NONE,  NONE,  NONE,0x12db,0x118a,0x118a,0x12fd,0x118f,0x118f,0x131a,0x0011d7,0x11d7,0x1323,0x0011ef,0x11ef,0x132f,0x001220,0x1220,0x1330,0x001222,0x1222,0x1336,0x001214,0x1214,0x12f7,0x00120c,0x120c,0x12ff,0x001225,0x1225,0x1309,0x00123f,0x123f)
GDI32(NtGdiDdDDISetVidPnSourceOwner1                    ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x1319,0x0011d8,0x11d8,0x1322,0x0011f0,0x11f0,0x132e,0x001221,0x1221,0x132f,0x001223,0x1223,0x1335,0x001215,0x1215,0x12f6,0x00120d,0x120d,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
GDI32(NtGdiDdDDISetYieldPercentage                      ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x138d,0x00120e,0x120e,0x1398,0x001226,0x1226,0x13a3,0x001240,0x1240)
GDI32(NtGdiDdDDIShareObjects                            ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x133c,0x0011d9,0x11d9,0x1345,0x0011f1,0x11f1,0x1355,0x001222,0x1222,0x1356,0x001224,0x1224,0x135c,0x001216,0x1216,0x131d,0x00120f,0x120f,0x1325,0x001227,0x1227,0x132f,0x001241,0x1241)
GDI32(NtGdiDdDDISharedPrimaryLockNotification           ,  NONE,  NONE,  NONE,  NONE,  NONE,0x12f4,0x118b,0x118b,0x1317,0x1190,0x1190,0x1358,0x0011da,0x11da,0x1361,0x0011f2,0x11f2,0x1371,0x001223,0x1223,0x1372,0x001225,0x1225,0x1378,0x001217,0x1217,0x1339,0x001210,0x1210,0x1341,0x001228,0x1228,0x134b,0x001242,0x1242)
GDI32(NtGdiDdDDISharedPrimaryUnLockNotification         ,  NONE,  NONE,  NONE,  NONE,  NONE,0x12f5,0x118c,0x118c,0x1318,0x1191,0x1191,0x1357,0x0011db,0x11db,0x1360,0x0011f3,0x11f3,0x1370,0x001224,0x1224,0x1371,0x001226,0x1226,0x1377,0x001218,0x1218,0x1338,0x001211,0x1211,0x1340,0x001229,0x1229,0x134a,0x001243,0x1243)
GDI32(NtGdiDdDDISignalSynchronizationObject             ,  NONE,  NONE,  NONE,  NONE,  NONE,0x12cb,0x118d,0x118d,0x12ed,0x1192,0x1192,0x132c,0x0011dc,0x11dc,0x1335,0x0011f4,0x11f4,0x1342,0x001225,0x1225,0x1343,0x001227,0x1227,0x1349,0x001219,0x1219,0x130a,0x001212,0x1212,0x1312,0x00122a,0x122a,0x131c,0x001244,0x1244)
GDI32(NtGdiDdDDISignalSynchronizationObjectFromCpu      ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1388,0x001226,0x1226,0x1389,0x001228,0x1228,0x1390,0x00121a,0x121a,0x1351,0x001213,0x1213,0x1359,0x00122b,0x122b,0x1363,0x001245,0x1245)
GDI32(NtGdiDdDDISignalSynchronizationObjectFromGpu      ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x138a,0x001227,0x1227,0x138b,0x001229,0x1229,0x1392,0x00121b,0x121b,0x1353,0x001214,0x1214,0x135b,0x00122c,0x122c,0x1365,0x001246,0x1246)
GDI32(NtGdiDdDDISignalSynchronizationObjectFromGpu2     ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x138b,0x001228,0x1228,0x138c,0x00122a,0x122a,0x1393,0x00121c,0x121c,0x1354,0x001215,0x1215,0x135c,0x00122d,0x122d,0x1366,0x001247,0x1247)
GDI32(NtGdiDdDDISubmitCommand                           ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1397,0x001229,0x1229,0x1398,0x00122b,0x122b,0x139f,0x00121d,0x121d,0x1360,0x001216,0x1216,0x1368,0x00122e,0x122e,0x1372,0x001248,0x1248)
GDI32(NtGdiDdDDISubmitCommandToHwQueue                  ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1387,0x001217,0x1217,0x1392,0x00122f,0x122f,0x139c,0x001249,0x1249)
GDI32(NtGdiDdDDISubmitSignalSyncObjectsToHwQueue        ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1389,0x001218,0x1218,0x1394,0x001230,0x1230,0x139e,0x00124a,0x124a)
GDI32(NtGdiDdDDISubmitWaitForSyncObjectsToHwQueue       ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1388,0x001219,0x1219,0x1393,0x001231,0x1231,0x139d,0x00124b,0x124b)
GDI32(NtGdiDdDDITrimProcessCommitment                   ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x13bd,0x00121e,0x121e,0x137e,0x00121a,0x121a,0x1389,0x001232,0x1232,0x1393,0x00124c,0x124c)
GDI32(NtGdiDdDDIUnOrderedPresentSwapChain               ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1377,0x001233,0x1233,0x1381,0x00124d,0x124d)
GDI32(NtGdiDdDDIUnlock                                  ,  NONE,  NONE,  NONE,  NONE,  NONE,0x12cf,0x118e,0x118e,0x12f1,0x1193,0x1193,0x1328,0x0011dd,0x11dd,0x1331,0x0011f5,0x11f5,0x133e,0x00122a,0x122a,0x133f,0x00122c,0x122c,0x1345,0x00121f,0x121f,0x1306,0x00121b,0x121b,0x130e,0x001234,0x1234,0x1318,0x00124e,0x124e)
GDI32(NtGdiDdDDIUnlock2                                 ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x138f,0x00122b,0x122b,0x1390,0x00122d,0x122d,0x1397,0x001220,0x1220,0x1358,0x00121c,0x121c,0x1360,0x001235,0x1235,0x136a,0x00124f,0x124f)
GDI32(NtGdiDdDDIUnpinDirectFlipResources                ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x1362,0x0011de,0x11de,0x136b,0x0011f6,0x11f6,0x137b,0x00122c,0x122c,0x137c,0x00122e,0x122e,0x1382,0x001221,0x1221,0x1343,0x00121d,0x121d,0x134b,0x001236,0x1236,0x1355,0x001250,0x1250)
GDI32(NtGdiDdDDIUpdateAllocationProperty                ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x138e,0x001222,0x1222,0x134f,0x00121e,0x121e,0x1357,0x001237,0x1237,0x1361,0x001251,0x1251)
GDI32(NtGdiDdDDIUpdateGpuVirtualAddress                 ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1395,0x00122d,0x122d,0x1396,0x00122f,0x122f,0x139d,0x001223,0x1223,0x135e,0x00121f,0x121f,0x1366,0x001238,0x1238,0x1370,0x001252,0x1252)
GDI32(NtGdiDdDDIUpdateOverlay                           ,  NONE,  NONE,  NONE,  NONE,  NONE,0x12de,0x118f,0x118f,0x1301,0x1194,0x1194,0x1315,0x0011df,0x11df,0x131e,0x0011f7,0x11f7,0x132a,0x00122e,0x122e,0x132b,0x001230,0x1230,0x1331,0x001224,0x1224,0x12f2,0x001220,0x1220,0x12fb,0x001239,0x1239,0x1305,0x001253,0x1253)
GDI32(NtGdiDdDDIWaitForIdle                             ,  NONE,  NONE,  NONE,  NONE,  NONE,0x12f0,0x1190,0x1190,0x1313,0x1195,0x1195,0x135c,0x0011e0,0x11e0,0x1365,0x0011f8,0x11f8,0x1375,0x00122f,0x122f,0x1376,0x001231,0x1231,0x137c,0x001225,0x1225,0x133d,0x001221,0x1221,0x1345,0x00123a,0x123a,0x134f,0x001254,0x1254)
GDI32(NtGdiDdDDIWaitForSynchronizationObject            ,  NONE,  NONE,  NONE,  NONE,  NONE,0x12ca,0x1191,0x1191,0x12ec,0x1196,0x1196,0x132d,0x0011e1,0x11e1,0x1336,0x0011f9,0x11f9,0x1343,0x001230,0x1230,0x1344,0x001232,0x1232,0x134a,0x001226,0x1226,0x130b,0x001222,0x1222,0x1313,0x00123b,0x123b,0x131d,0x001255,0x1255)
GDI32(NtGdiDdDDIWaitForSynchronizationObjectFromCpu     ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1387,0x001231,0x1231,0x1388,0x001233,0x1233,0x138f,0x001227,0x1227,0x1350,0x001223,0x1223,0x1358,0x00123c,0x123c,0x1362,0x001256,0x1256)
GDI32(NtGdiDdDDIWaitForSynchronizationObjectFromGpu     ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1389,0x001232,0x1232,0x138a,0x001234,0x1234,0x1391,0x001228,0x1228,0x1352,0x001224,0x1224,0x135a,0x00123d,0x123d,0x1364,0x001257,0x1257)
GDI32(NtGdiDdDDIWaitForVerticalBlankEvent               ,  NONE,  NONE,  NONE,  NONE,  NONE,0x12e1,0x1192,0x1192,0x1304,0x1197,0x1197,0x1312,0x0011e2,0x11e2,0x131b,0x0011fa,0x11fa,0x1327,0x001233,0x1233,0x1328,0x001235,0x1235,0x132e,0x001229,0x1229,0x12ef,0x001225,0x1225,0x12f8,0x00123e,0x123e,0x1302,0x001258,0x1258)
GDI32(NtGdiDdDDIWaitForVerticalBlankEvent2              ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x1363,0x0011e3,0x11e3,0x136c,0x0011fb,0x11fb,0x137c,0x001234,0x1234,0x137d,0x001236,0x1236,0x1383,0x00122a,0x122a,0x1344,0x001226,0x1226,0x134c,0x00123f,0x123f,0x1356,0x001259,0x1259)
GDI32(NtGdiDdDeleteDirectDrawObject                     ,  NONE,0x1040,0x1040,0x1152,0x1152,0x1042,0x1193,0x1193,0x1043,0x1198,0x1198,0x10f7,0x0411e4,0x11e4,0x10f9,0x0411fc,0x11fc,0x10fa,0x041235,0x1235,0x10fc,0x041237,0x1237,0x10fc,0x04122b,0x122b,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
GDI32(NtGdiDdDeleteSurfaceObject                        ,  NONE,0x1041,0x1041,0x10a2,0x10a2,0x1043,0x10a3,0x10a3,0x1044,0x10a1,0x10a1,0x10f6,0x0410a1,0x10a1,0x10f8,0x0410a2,0x10a2,0x10f9,0x0410a3,0x10a3,0x10fb,0x0410a3,0x10a3,0x10fb,0x0410a3,0x10a3,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
GDI32(NtGdiDdDestroyD3DBuffer                           ,  NONE,0x1044,0x1044,0x1153,0x1153,0x1046,0x1194,0x1194,0x1047,0x1199,0x1199,0x10f3,0x0411e5,0x11e5,0x10f5,0x0411fd,0x11fd,0x10f6,0x041236,0x1236,0x10f8,0x041238,0x1238,0x10f8,0x04122c,0x122c,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
GDI32(NtGdiDdDestroyFullscreenSprite                    ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x1330,0x119a,0x119a,0x1375,0x0011e6,0x11e6,0x1387,0x0011fe,0x11fe,0x13c0,0x001237,0x1237,0x13c2,0x001239,0x1239,0x13cf,0x00122d,0x122d,0x13a0,0x001227,0x1227,0x13ba,0x001240,0x1240,0x13cc,0x00125a,0x125a)
GDI32(NtGdiDdDestroyMoComp                              ,  NONE,0x1042,0x1042,0x1154,0x1154,0x1044,0x1195,0x1195,0x1045,0x119b,0x119b,0x10f5,0x0011e7,0x11e7,0x10f7,0x0011ff,0x11ff,0x10f8,0x001238,0x1238,0x10fa,0x00123a,0x123a,0x10fa,0x00122e,0x122e,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
GDI32(NtGdiDdDestroySurface                             ,  NONE,0x1043,0x1043,0x10a9,0x10a9,0x1045,0x10aa,0x10aa,0x1046,0x10a5,0x10a5,0x10f4,0x0510a5,0x10a5,0x10f6,0x0510a6,0x10a6,0x10f7,0x0510a7,0x10a7,0x10f9,0x0510a7,0x10a7,0x10f9,0x0510a7,0x10a7,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
GDI32(NtGdiDdEndMoCompFrame                             ,  NONE,0x1045,0x1045,0x1155,0x1155,0x1047,0x1196,0x1196,0x1048,0x119c,0x119c,0x10f2,0x0011e8,0x11e8,0x10f4,0x001200,0x1200,0x10f5,0x001239,0x1239,0x10f7,0x00123b,0x123b,0x10f7,0x00122f,0x122f,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
GDI32(NtGdiDdFlip                                       ,  NONE,0x1046,0x1046,0x1156,0x1156,0x1048,0x1197,0x1197,0x1049,0x119d,0x119d,0x10f1,0x0011e9,0x11e9,0x10f3,0x001201,0x1201,0x10f4,0x00123a,0x123a,0x10f6,0x00123c,0x123c,0x10f6,0x001230,0x1230,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
GDI32(NtGdiDdFlipToGDISurface                           ,  NONE,0x1047,0x1047,0x1157,0x1157,0x1049,0x1198,0x1198,0x104a,0x119e,0x119e,0x10f0,0x0011ea,0x11ea,0x10f2,0x001202,0x1202,0x10f3,0x00123b,0x123b,0x10f5,0x00123d,0x123d,0x10f5,0x001231,0x1231,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
GDI32(NtGdiDdGetAvailDriverMemory                       ,  NONE,0x1048,0x1048,0x1158,0x1158,0x104a,0x1199,0x1199,0x104b,0x119f,0x119f,0x10ef,0x0011eb,0x11eb,0x10f1,0x001203,0x1203,0x10f2,0x00123c,0x123c,0x10f4,0x00123e,0x123e,0x10f4,0x001232,0x1232,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
GDI32(NtGdiDdGetBltStatus                               ,  NONE,0x1049,0x1049,0x1159,0x1159,0x104b,0x119a,0x119a,0x104c,0x11a0,0x11a0,0x10ee,0x0011ec,0x11ec,0x10f0,0x001204,0x1204,0x10f1,0x00123d,0x123d,0x10f3,0x00123f,0x123f,0x10f3,0x001233,0x1233,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
GDI32(NtGdiDdGetDC                                      ,  NONE,0x104a,0x104a,0x115a,0x115a,0x104c,0x119b,0x119b,0x104d,0x11a1,0x11a1,0x10ed,0x0511ed,0x11ed,0x10ef,0x051205,0x1205,0x10f0,0x05123e,0x123e,0x10f2,0x051240,0x1240,0x10f2,0x051234,0x1234,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
GDI32(NtGdiDdGetDriverInfo                              ,  NONE,0x104b,0x104b,0x115b,0x115b,0x104d,0x119c,0x119c,0x104e,0x11a2,0x11a2,0x10ec,0x0011ee,0x11ee,0x10ee,0x001206,0x1206,0x10ef,0x00123f,0x123f,0x10f1,0x001241,0x1241,0x10f1,0x001235,0x1235,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
GDI32(NtGdiDdGetDriverState                             ,  NONE,0x1032,0x1032,0x115c,0x115c,0x1034,0x119d,0x119d,0x1035,0x11a3,0x11a3,0x1105,0x0011ef,0x11ef,0x1107,0x001207,0x1207,0x1108,0x001240,0x1240,0x110a,0x001242,0x1242,0x110a,0x001236,0x1236,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
GDI32(NtGdiDdGetDxHandle                                ,  NONE,0x104c,0x104c,0x115d,0x115d,0x104e,0x119e,0x119e,0x104f,0x11a4,0x11a4,0x10eb,0x0a11f0,0x11f0,0x10ed,0x0a1208,0x1208,0x10ee,0x0a1241,0x1241,0x10f0,0x0a1243,0x1243,0x10f0,0x0a1237,0x1237,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
GDI32(NtGdiDdGetFlipStatus                              ,  NONE,0x104d,0x104d,0x115e,0x115e,0x104f,0x119f,0x119f,0x1050,0x11a5,0x11a5,0x10ea,0x0011f1,0x11f1,0x10ec,0x001209,0x1209,0x10ed,0x001242,0x1242,0x10ef,0x001244,0x1244,0x10ef,0x001238,0x1238,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
GDI32(NtGdiDdGetInternalMoCompInfo                      ,  NONE,0x104e,0x104e,0x115f,0x115f,0x1050,0x11a0,0x11a0,0x1051,0x11a6,0x11a6,0x10e9,0x0011f2,0x11f2,0x10eb,0x00120a,0x120a,0x10ec,0x001243,0x1243,0x10ee,0x001245,0x1245,0x10ee,0x001239,0x1239,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
GDI32(NtGdiDdGetMoCompBuffInfo                          ,  NONE,0x104f,0x104f,0x1160,0x1160,0x1051,0x11a1,0x11a1,0x1052,0x11a7,0x11a7,0x10e8,0x0011f3,0x11f3,0x10ea,0x00120b,0x120b,0x10eb,0x001244,0x1244,0x10ed,0x001246,0x1246,0x10ed,0x00123a,0x123a,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
GDI32(NtGdiDdGetMoCompFormats                           ,  NONE,0x1051,0x1051,0x1161,0x1161,0x1053,0x11a2,0x11a2,0x1054,0x11a8,0x11a8,0x10e6,0x0011f4,0x11f4,0x10e8,0x00120c,0x120c,0x10e9,0x001245,0x1245,0x10eb,0x001247,0x1247,0x10eb,0x00123b,0x123b,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
GDI32(NtGdiDdGetMoCompGuids                             ,  NONE,0x1050,0x1050,0x1162,0x1162,0x1052,0x11a3,0x11a3,0x1053,0x11a9,0x11a9,0x10e7,0x0011f5,0x11f5,0x10e9,0x00120d,0x120d,0x10ea,0x001246,0x1246,0x10ec,0x001248,0x1248,0x10ec,0x00123c,0x123c,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
GDI32(NtGdiDdGetScanLine                                ,  NONE,0x1052,0x1052,0x1163,0x1163,0x1054,0x11a4,0x11a4,0x1055,0x11aa,0x11aa,0x10e5,0x0011f6,0x11f6,0x10e7,0x00120e,0x120e,0x10e8,0x001247,0x1247,0x10ea,0x001249,0x1249,0x10ea,0x00123d,0x123d,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
GDI32(NtGdiDdLock                                       ,  NONE,0x1053,0x1053,0x1164,0x1164,0x1055,0x11a5,0x11a5,0x1056,0x11ab,0x11ab,0x10e4,0x0011f7,0x11f7,0x10e6,0x00120f,0x120f,0x10e7,0x001248,0x1248,0x10e9,0x00124a,0x124a,0x10e9,0x00123e,0x123e,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
GDI32(NtGdiDdLockD3D                                    ,  NONE,0x1054,0x1054,0x10c9,0x10c9,0x1056,0x10ca,0x10ca,0x1057,0x10c5,0x10c5,0x10e3,0x0010c5,0x10c5,0x10e5,0x0010c6,0x10c6,0x10e6,0x0010c7,0x10c7,0x10e8,0x0010c7,0x10c7,0x10e8,0x0010c7,0x10c7,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
GDI32(NtGdiDdNotifyFullscreenSpriteUpdate               ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x132f,0x11ac,0x11ac,0x1376,0x0011f8,0x11f8,0x1388,0x001210,0x1210,0x13c1,0x001249,0x1249,0x13c3,0x00124b,0x124b,0x13d0,0x00123f,0x123f,0x13a1,0x001228,0x1228,0x13bb,0x001241,0x1241,0x13cd,0x00125b,0x125b)
GDI32(NtGdiDdQueryDirectDrawObject                      ,  NONE,0x1055,0x1055,0x1165,0x1165,0x1057,0x11a6,0x11a6,0x1058,0x11ad,0x11ad,0x10e2,0x0011f9,0x11f9,0x10e4,0x001211,0x1211,0x10e5,0x00124a,0x124a,0x10e7,0x00124c,0x124c,0x10e7,0x001240,0x1240,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
GDI32(NtGdiDdQueryMoCompStatus                          ,  NONE,0x1056,0x1056,0x1166,0x1166,0x1058,0x11a7,0x11a7,0x1059,0x11ae,0x11ae,0x10e1,0x0011fa,0x11fa,0x10e3,0x001212,0x1212,0x10e4,0x00124b,0x124b,0x10e6,0x00124d,0x124d,0x10e6,0x001241,0x1241,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
GDI32(NtGdiDdQueryVisRgnUniqueness                      ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x1331,0x11af,0x11af,0x1374,0x0011fb,0x11fb,0x1386,0x001213,0x1213,0x13bf,0x00124c,0x124c,0x13c1,0x00124e,0x124e,0x13ce,0x001242,0x1242,0x139f,0x001229,0x1229,0x13b9,0x001242,0x1242,0x13cb,0x00125c,0x125c)
GDI32(NtGdiDdReenableDirectDrawObject                   ,  NONE,0x1057,0x1057,0x1167,0x1167,0x1059,0x11a8,0x11a8,0x105a,0x11b0,0x11b0,0x10e0,0x0511fc,0x11fc,0x10e2,0x051214,0x1214,0x10e3,0x05124d,0x124d,0x10e5,0x05124f,0x124f,0x10e5,0x051243,0x1243,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
GDI32(NtGdiDdReleaseDC                                  ,  NONE,0x1058,0x1058,0x1168,0x1168,0x105a,0x11a9,0x11a9,0x105b,0x11b1,0x11b1,0x10df,0x0411fd,0x11fd,0x10e1,0x041215,0x1215,0x10e2,0x04124e,0x124e,0x10e4,0x041250,0x1250,0x10e4,0x041244,0x1244,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
GDI32(NtGdiDdRenderMoComp                               ,  NONE,0x1059,0x1059,0x1169,0x1169,0x105b,0x11aa,0x11aa,0x105c,0x11b2,0x11b2,0x10de,0x0011fe,0x11fe,0x10e0,0x001216,0x1216,0x10e1,0x00124f,0x124f,0x10e3,0x001251,0x1251,0x10e3,0x001245,0x1245,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
GDI32(NtGdiDdResetVisrgn                                ,  NONE,0x105a,0x105a,0x10ae,0x10ae,0x105c,0x10af,0x10af,0x105d,0x10aa,0x10aa,0x10dd,0x0910aa,0x10aa,0x10df,0x0910ab,0x10ab,0x10e0,0x0910ac,0x10ac,0x10e2,0x0910ac,0x10ac,0x10e2,0x0910ac,0x10ac,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
GDI32(NtGdiDdSetColorKey                                ,  NONE,0x105b,0x105b,0x116a,0x116a,0x105d,0x11ab,0x11ab,0x105e,0x11b3,0x11b3,0x10dc,0x0011ff,0x11ff,0x10de,0x001217,0x1217,0x10df,0x001250,0x1250,0x10e1,0x001252,0x1252,0x10e1,0x001246,0x1246,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
GDI32(NtGdiDdSetExclusiveMode                           ,  NONE,0x105c,0x105c,0x116b,0x116b,0x105e,0x11ac,0x11ac,0x105f,0x11b4,0x11b4,0x10db,0x001200,0x1200,0x10dd,0x001218,0x1218,0x10de,0x001251,0x1251,0x10e0,0x001253,0x1253,0x10e0,0x001247,0x1247,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
GDI32(NtGdiDdSetGammaRamp                               ,  NONE,0x105d,0x105d,0x116c,0x116c,0x105f,0x11ad,0x11ad,0x1060,0x11b5,0x11b5,0x10da,0x0f1201,0x1201,0x10dc,0x0f1219,0x1219,0x10dd,0x0f1252,0x1252,0x10df,0x0f1254,0x1254,0x10df,0x0f1248,0x1248,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
GDI32(NtGdiDdSetOverlayPosition                         ,  NONE,0x105f,0x105f,0x116d,0x116d,0x1061,0x11ae,0x11ae,0x1062,0x11b6,0x11b6,0x10d8,0x001202,0x1202,0x10da,0x00121a,0x121a,0x10db,0x001253,0x1253,0x10dd,0x001255,0x1255,0x10dd,0x001249,0x1249,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
GDI32(NtGdiDdUnattachSurface                            ,  NONE,0x1060,0x1060,0x116e,0x116e,0x1062,0x11af,0x11af,0x1063,0x11b7,0x11b7,0x10d7,0x051203,0x1203,0x10d9,0x05121b,0x121b,0x10da,0x051254,0x1254,0x10dc,0x051256,0x1256,0x10dc,0x05124a,0x124a,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
GDI32(NtGdiDdUnlock                                     ,  NONE,0x1061,0x1061,0x116f,0x116f,0x1063,0x11b0,0x11b0,0x1064,0x11b8,0x11b8,0x10d6,0x001204,0x1204,0x10d8,0x00121c,0x121c,0x10d9,0x001255,0x1255,0x10db,0x001257,0x1257,0x10db,0x00124b,0x124b,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
GDI32(NtGdiDdUnlockD3D                                  ,  NONE,0x1062,0x1062,0x10ca,0x10ca,0x1064,0x10cb,0x10cb,0x1065,0x10c6,0x10c6,0x10d5,0x0010c6,0x10c6,0x10d7,0x0010c7,0x10c7,0x10d8,0x0010c8,0x10c8,0x10da,0x0010c8,0x10c8,0x10da,0x0010c8,0x10c8,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
GDI32(NtGdiDdUpdateOverlay                              ,  NONE,0x1063,0x1063,0x1170,0x1170,0x1065,0x11b1,0x11b1,0x1066,0x11b9,0x11b9,0x10d4,0x001205,0x1205,0x10d6,0x00121d,0x121d,0x10d7,0x001256,0x1256,0x10d9,0x001258,0x1258,0x10d9,0x00124c,0x124c,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
GDI32(NtGdiDdWaitForVerticalBlank                       ,  NONE,0x1064,0x1064,0x1171,0x1171,0x1066,0x11b2,0x11b2,0x1067,0x11ba,0x11ba,0x10d3,0x001206,0x1206,0x10d5,0x00121e,0x121e,0x10d6,0x001257,0x1257,0x10d8,0x001259,0x1259,0x10d8,0x00124d,0x124d,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
GDI32(NtGdiDeleteClientObj                              ,  NONE,0x1077,0x1077,0x1088,0x1088,0x1079,0x1089,0x1089,0x107a,0x1087,0x1087,0x10c0,0x031087,0x1087,0x10c2,0x031088,0x1088,0x10c3,0x031089,0x1089,0x10c5,0x031089,0x1089,0x10c5,0x031089,0x1089,0x10cc,0x031087,0x1087,0x10cd,0x031087,0x1087,0x10cd,0x031087,0x1087)
GDI32(NtGdiDeleteColorSpace                             ,  NONE,0x1078,0x1078,0x1125,0x1125,0x107a,0x1126,0x1126,0x107b,0x111d,0x111d,0x10bf,0x03111d,0x111d,0x10c1,0x03111e,0x111e,0x10c2,0x03111f,0x111f,0x10c4,0x03111f,0x111f,0x10c4,0x03111e,0x111e,0x10cb,0x031113,0x1113,0x10cc,0x031113,0x1113,0x10cc,0x031113,0x1113)
GDI32(NtGdiDeleteColorTransform                         ,  NONE,0x1079,0x1079,0x1172,0x1172,0x107b,0x11b3,0x11b3,0x107c,0x11bb,0x11bb,0x10be,0x081207,0x1207,0x10c0,0x08121f,0x121f,0x10c1,0x081258,0x1258,0x10c3,0x08125a,0x125a,0x10c3,0x08124e,0x124e,0x10ca,0x08122a,0x122a,0x10cb,0x081243,0x1243,0x10cb,0x08125d,0x125d)
GDI32(NtGdiDeleteObjectApp                              ,  NONE,0x107a,0x107a,0x1022,0x1022,0x107c,0x1023,0x1023,0x107d,0x1023,0x1023,0x10bd,0x031024,0x1024,0x10bf,0x031025,0x1025,0x10c0,0x031026,0x1026,0x10c2,0x031026,0x1026,0x10c2,0x031026,0x1026,0x10c9,0x031026,0x1026,0x10ca,0x031026,0x1026,0x10ca,0x031026,0x1026)
GDI32(NtGdiDescribePixelFormat                          ,  NONE,0x107b,0x107b,0x1173,0x1173,0x107d,0x11b4,0x11b4,0x107e,0x11bc,0x11bc,0x10bc,0x141208,0x1208,0x10be,0x141220,0x1220,0x10bf,0x141259,0x1259,0x10c1,0x14125b,0x125b,0x10c1,0x14124f,0x124f,0x10c8,0x14122b,0x122b,0x10c9,0x141244,0x1244,0x10c9,0x14125e,0x125e)
GDI32(NtGdiDestroyOPMProtectedOutput                    ,  NONE,  NONE,  NONE,  NONE,  NONE,0x107e,0x11b5,0x11b5,0x107f,0x11bd,0x11bd,0x10bb,0x001209,0x1209,0x10bd,0x001221,0x1221,0x10be,0x00125a,0x125a,0x10c0,0x00125c,0x125c,0x10c0,0x001250,0x1250,0x10c7,0x00122c,0x122c,0x10c8,0x001245,0x1245,0x10c8,0x00125f,0x125f)
GDI32(NtGdiDestroyPhysicalMonitor                       ,  NONE,  NONE,  NONE,  NONE,  NONE,0x12fb,0x11b6,0x11b6,0x1327,0x11be,0x11be,0x136a,0x00120a,0x120a,0x137c,0x001222,0x1222,0x13b5,0x00125b,0x125b,0x13b7,0x00125d,0x125d,0x13c4,0x001251,0x1251,0x1395,0x00122d,0x122d,0x13af,0x001246,0x1246,0x13c1,0x001260,0x1260)
GDI32(NtGdiDoBanding                                    ,  NONE,0x107d,0x107d,0x1174,0x1174,0x1080,0x11b7,0x11b7,0x1081,0x11bf,0x11bf,  NONE,0x14120b,0x120b,0x10bb,0x141223,0x1223,0x10bc,0x14125c,0x125c,0x10be,0x14125e,0x125e,0x10be,0x141252,0x1252,0x10c5,0x14122e,0x122e,0x10c6,0x141247,0x1247,0x10c6,0x141261,0x1261)
GDI32(NtGdiDoPalette                                    ,  NONE,0x107e,0x107e,0x1046,0x1046,0x1081,0x1047,0x1047,0x1082,0x1047,0x1047,0x10b8,0x001048,0x1048,0x10ba,0x001049,0x1049,0x10bb,0x00104a,0x104a,0x10bd,0x00104a,0x104a,0x10bd,0x00104a,0x104a,0x10c4,0x00104a,0x104a,0x10c5,0x00104a,0x104a,0x10c5,0x00104a,0x104a)
GDI32(NtGdiDrawEscape                                   ,  NONE,0x107f,0x107f,0x1175,0x1175,0x1082,0x11b8,0x11b8,0x1083,0x11c0,0x11c0,0x10b7,0x14120c,0x120c,0x10b9,0x141224,0x1224,0x10ba,0x14125d,0x125d,0x10bc,0x14125f,0x125f,0x10bc,0x141253,0x1253,0x10c3,0x14122f,0x122f,0x10c4,0x141248,0x1248,0x10c4,0x141262,0x1262)
GDI32(NtGdiDrawStream                                   ,  NONE,0x129a,0x1296,0x1061,0x1061,0x12bb,0x1062,0x1062,0x12db,0x1062,0x1062,0x1304,0x001062,0x1062,0x130d,0x001063,0x1063,0x1319,0x001064,0x1064,0x131a,0x001064,0x1064,0x1320,0x001064,0x1064,0x12e1,0x001064,0x1064,0x12ea,0x001064,0x1064,0x12f4,0x001064,0x1064)
GDI32(NtGdiDvpAcquireNotification                       ,  NONE,0x1074,0x1074,0x1176,0x1176,0x1076,0x11b9,0x11b9,0x1077,0x11c1,0x11c1,0x10c3,0x00120d,0x120d,0x10c5,0x001225,0x1225,0x10c6,0x00125e,0x125e,0x10c8,0x001260,0x1260,0x10c8,0x001254,0x1254,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
GDI32(NtGdiDvpCanCreateVideoPort                        ,  NONE,0x1065,0x1065,0x1177,0x1177,0x1067,0x11ba,0x11ba,0x1068,0x11c2,0x11c2,0x10d2,0x00120e,0x120e,0x10d4,0x001226,0x1226,0x10d5,0x00125f,0x125f,0x10d7,0x001261,0x1261,0x10d7,0x001255,0x1255,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
GDI32(NtGdiDvpColorControl                              ,  NONE,0x1066,0x1066,0x1178,0x1178,0x1068,0x11bb,0x11bb,0x1069,0x11c3,0x11c3,0x10d1,0x00120f,0x120f,0x10d3,0x001227,0x1227,0x10d4,0x001260,0x1260,0x10d6,0x001262,0x1262,0x10d6,0x001256,0x1256,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
GDI32(NtGdiDvpCreateVideoPort                           ,  NONE,0x1067,0x1067,0x1179,0x1179,0x1069,0x11bc,0x11bc,0x106a,0x11c4,0x11c4,0x10d0,0x001210,0x1210,0x10d2,0x001228,0x1228,0x10d3,0x001261,0x1261,0x10d5,0x001263,0x1263,0x10d5,0x001257,0x1257,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
GDI32(NtGdiDvpDestroyVideoPort                          ,  NONE,0x1068,0x1068,0x117a,0x117a,0x106a,0x11bd,0x11bd,0x106b,0x11c5,0x11c5,0x10cf,0x001211,0x1211,0x10d1,0x001229,0x1229,0x10d2,0x001262,0x1262,0x10d4,0x001264,0x1264,0x10d4,0x001258,0x1258,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
GDI32(NtGdiDvpFlipVideoPort                             ,  NONE,0x1069,0x1069,0x117b,0x117b,0x106b,0x11be,0x11be,0x106c,0x11c6,0x11c6,0x10ce,0x001212,0x1212,0x10d0,0x00122a,0x122a,0x10d1,0x001263,0x1263,0x10d3,0x001265,0x1265,0x10d3,0x001259,0x1259,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
GDI32(NtGdiDvpGetVideoPortBandwidth                     ,  NONE,0x106a,0x106a,0x117c,0x117c,0x106c,0x11bf,0x11bf,0x106d,0x11c7,0x11c7,0x10cd,0x001213,0x1213,0x10cf,0x00122b,0x122b,0x10d0,0x001264,0x1264,0x10d2,0x001266,0x1266,0x10d2,0x00125a,0x125a,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
GDI32(NtGdiDvpGetVideoPortConnectInfo                   ,  NONE,0x1070,0x1070,0x117d,0x117d,0x1072,0x11c0,0x11c0,0x1073,0x11c8,0x11c8,0x10c7,0x001214,0x1214,0x10c9,0x00122c,0x122c,0x10ca,0x001265,0x1265,0x10cc,0x001267,0x1267,0x10cc,0x00125b,0x125b,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
GDI32(NtGdiDvpGetVideoPortField                         ,  NONE,0x106b,0x106b,0x117e,0x117e,0x106d,0x11c1,0x11c1,0x106e,0x11c9,0x11c9,0x10cc,0x001215,0x1215,0x10ce,0x00122d,0x122d,0x10cf,0x001266,0x1266,0x10d1,0x001268,0x1268,0x10d1,0x00125c,0x125c,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
GDI32(NtGdiDvpGetVideoPortFlipStatus                    ,  NONE,0x106c,0x106c,0x117f,0x117f,0x106e,0x11c2,0x11c2,0x106f,0x11ca,0x11ca,0x10cb,0x001216,0x1216,0x10cd,0x00122e,0x122e,0x10ce,0x001267,0x1267,0x10d0,0x001269,0x1269,0x10d0,0x00125d,0x125d,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
GDI32(NtGdiDvpGetVideoPortInputFormats                  ,  NONE,0x106d,0x106d,0x1180,0x1180,0x106f,0x11c3,0x11c3,0x1070,0x11cb,0x11cb,0x10ca,0x001217,0x1217,0x10cc,0x00122f,0x122f,0x10cd,0x001268,0x1268,0x10cf,0x00126a,0x126a,0x10cf,0x00125e,0x125e,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
GDI32(NtGdiDvpGetVideoPortLine                          ,  NONE,0x106e,0x106e,0x1181,0x1181,0x1070,0x11c4,0x11c4,0x1071,0x11cc,0x11cc,0x10c9,0x001218,0x1218,0x10cb,0x001230,0x1230,0x10cc,0x001269,0x1269,0x10ce,0x00126b,0x126b,0x10ce,0x00125f,0x125f,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
GDI32(NtGdiDvpGetVideoPortOutputFormats                 ,  NONE,0x106f,0x106f,0x1182,0x1182,0x1071,0x11c5,0x11c5,0x1072,0x11cd,0x11cd,0x10c8,0x001219,0x1219,0x10ca,0x001231,0x1231,0x10cb,0x00126a,0x126a,0x10cd,0x00126c,0x126c,0x10cd,0x001260,0x1260,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
GDI32(NtGdiDvpGetVideoSignalStatus                      ,  NONE,0x1071,0x1071,0x1183,0x1183,0x1073,0x11c6,0x11c6,0x1074,0x11ce,0x11ce,0x10c6,0x00121a,0x121a,0x10c8,0x001232,0x1232,0x10c9,0x00126b,0x126b,0x10cb,0x00126d,0x126d,0x10cb,0x001261,0x1261,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
GDI32(NtGdiDvpReleaseNotification                       ,  NONE,0x1075,0x1075,0x1184,0x1184,0x1077,0x11c7,0x11c7,0x1078,0x11cf,0x11cf,0x10c2,0x00121b,0x121b,0x10c4,0x001233,0x1233,0x10c5,0x00126c,0x126c,0x10c7,0x00126e,0x126e,0x10c7,0x001262,0x1262,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
GDI32(NtGdiDvpUpdateVideoPort                           ,  NONE,0x1072,0x1072,0x1185,0x1185,0x1074,0x11c8,0x11c8,0x1075,0x11d0,0x11d0,0x10c5,0x00121c,0x121c,0x10c7,0x001234,0x1234,0x10c8,0x00126d,0x126d,0x10ca,0x00126f,0x126f,0x10ca,0x001263,0x1263,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
GDI32(NtGdiDvpWaitForVideoPortSync                      ,  NONE,0x1073,0x1073,0x1186,0x1186,0x1075,0x11c9,0x11c9,0x1076,0x11d1,0x11d1,0x10c4,0x00121d,0x121d,0x10c6,0x001235,0x1235,0x10c7,0x00126e,0x126e,0x10c9,0x001270,0x1270,0x10c9,0x001264,0x1264,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
GDI32(NtGdiDwmCreatedBitmapRemotingOutput               ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x1307,0x00121e,0x121e,0x1310,0x001236,0x1236,0x131c,0x00126f,0x126f,0x131d,0x001271,0x1271,0x1323,0x001265,0x1265,0x12e4,0x001230,0x1230,0x12ed,0x001249,0x1249,0x12f7,0x001263,0x1263)
GDI32(NtGdiDwmGetDirtyRgn                               ,  NONE,  NONE,  NONE,  NONE,  NONE,0x12bc,0x11ca,0x11ca,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
GDI32(NtGdiDwmGetSurfaceData                            ,  NONE,  NONE,  NONE,  NONE,  NONE,0x12bd,0x11cb,0x11cb,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
GDI32(NtGdiDxgGenericThunk                              ,  NONE,0x1076,0x1076,0x1187,0x1187,0x1078,0x11cc,0x11cc,0x1079,0x11d2,0x11d2,0x10c1,0x00121f,0x121f,0x10c3,0x001237,0x1237,0x10c4,0x001270,0x1270,0x10c6,0x001272,0x1272,0x10c6,0x001266,0x1266,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
GDI32(NtGdiEllipse                                      ,0x11e0,0x1080,0x1080,0x1188,0x1188,0x1083,0x11cd,0x11cd,0x1084,0x11d3,0x11d3,0x10b6,0x001220,0x1220,0x10b8,0x001238,0x1238,0x10b9,0x001271,0x1271,0x10bb,0x001273,0x1273,0x10bb,0x001267,0x1267,0x10c2,0x001231,0x1231,0x10c3,0x00124a,0x124a,0x10c3,0x001264,0x1264)
GDI32(NtGdiEnableEudc                                   ,0x107c,0x1081,0x1081,0x1189,0x1189,0x1084,0x11ce,0x11ce,0x1085,0x11d4,0x11d4,0x10b5,0x041221,0x1221,0x10b7,0x041239,0x1239,0x10b8,0x041272,0x1272,0x10ba,0x041274,0x1274,0x10ba,0x041268,0x1268,0x10c1,0x041232,0x1232,0x10c2,0x04124b,0x124b,0x10c2,0x041265,0x1265)
GDI32(NtGdiEndDoc                                       ,  NONE,0x1082,0x1082,0x118a,0x118a,0x1085,0x11cf,0x11cf,0x1086,0x11d5,0x11d5,0x10b4,0x031222,0x1222,0x10b6,0x03123a,0x123a,0x10b7,0x031273,0x1273,0x10b9,0x031275,0x1275,0x10b9,0x031269,0x1269,0x10c0,0x031233,0x1233,0x10c1,0x03124c,0x124c,0x10c1,0x031266,0x1266)
GDI32(NtGdiEndGdiRendering                              ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x1087,0x11d6,0x11d6,0x10b3,0x001223,0x1223,0x10b5,0x00123b,0x123b,0x10b6,0x001274,0x1274,0x10b8,0x001276,0x1276,0x10b8,0x00126a,0x126a,0x10bf,0x001234,0x1234,0x10c0,0x00124d,0x124d,0x10c0,0x001267,0x1267)
GDI32(NtGdiEndPage                                      ,  NONE,0x1083,0x1083,0x118b,0x118b,0x1086,0x11d0,0x11d0,0x1088,0x11d7,0x11d7,0x10b2,0x031224,0x1224,0x10b4,0x03123c,0x123c,0x10b5,0x031275,0x1275,0x10b7,0x031277,0x1277,0x10b7,0x03126b,0x126b,0x10be,0x031235,0x1235,0x10bf,0x03124e,0x124e,0x10bf,0x031268,0x1268)
GDI32(NtGdiEndPath                                      ,  NONE,0x1084,0x1084,0x1110,0x1110,0x1087,0x1111,0x1111,0x1089,0x110a,0x110a,0x10b1,0x03110a,0x110a,0x10b3,0x03110b,0x110b,0x10b4,0x03110c,0x110c,0x10b6,0x03110c,0x110c,0x10b6,0x03110b,0x110b,0x10bd,0x031100,0x1100,0x10be,0x031100,0x1100,0x10be,0x031100,0x1100)
GDI32(NtGdiEngAlphaBlend                                ,0x1254,0x126c,0x1268,0x118c,0x118c,0x128d,0x11d1,0x11d1,0x12ad,0x11d8,0x11d8,0x12a7,0x001225,0x1225,0x12aa,0x00123d,0x123d,0x12b4,0x001276,0x1276,0x12b5,0x001278,0x1278,0x12bb,0x00126c,0x126c,0x127c,0x001236,0x1236,0x1285,0x00124f,0x124f,0x128d,0x001269,0x1269)
GDI32(NtGdiEngAssociateSurface                          ,0x123f,0x1257,0x1253,0x118d,0x118d,0x1278,0x11d2,0x11d2,0x1298,0x11d9,0x11d9,0x12bc,0x001226,0x1226,0x12bf,0x00123e,0x123e,0x12c9,0x001277,0x1277,0x12ca,0x001279,0x1279,0x12d0,0x00126d,0x126d,0x1291,0x001237,0x1237,0x129a,0x001250,0x1250,0x12a2,0x00126a,0x126a)
GDI32(NtGdiEngBitBlt                                    ,0x124b,0x1263,0x125f,0x118e,0x118e,0x1284,0x11d3,0x11d3,0x12a4,0x11da,0x11da,0x12b0,0x001227,0x1227,0x12b3,0x00123f,0x123f,0x12bd,0x001278,0x1278,0x12be,0x00127a,0x127a,0x12c4,0x00126e,0x126e,0x1285,0x001238,0x1238,0x128e,0x001251,0x1251,0x1296,0x00126b,0x126b)
GDI32(NtGdiEngCheckAbort                                ,0x127b,0x1293,0x128f,0x118f,0x118f,0x12b4,0x11d4,0x11d4,0x12d4,0x11db,0x11db,0x12fb,0x001228,0x1228,0x1304,0x001240,0x1240,0x1310,0x001279,0x1279,0x1311,0x00127b,0x127b,0x1317,0x00126f,0x126f,0x12d8,0x001239,0x1239,0x12e1,0x001252,0x1252,0x12eb,0x00126c,0x126c)
GDI32(NtGdiEngComputeGlyphSet                           ,  NONE,0x125c,0x1258,0x1190,0x1190,0x127d,0x11d5,0x11d5,0x129d,0x11dc,0x11dc,0x12b7,0x001229,0x1229,0x12ba,0x001241,0x1241,0x12c4,0x00127a,0x127a,0x12c5,0x00127c,0x127c,0x12cb,0x001270,0x1270,0x128c,0x00123a,0x123a,0x1295,0x001253,0x1253,0x129d,0x00126d,0x126d)
GDI32(NtGdiEngCopyBits                                  ,0x1245,0x125d,0x1259,0x1191,0x1191,0x127e,0x11d6,0x11d6,0x129e,0x11dd,0x11dd,0x12b6,0x00122a,0x122a,0x12b9,0x001242,0x1242,0x12c3,0x00127b,0x127b,0x12c4,0x00127d,0x127d,0x12ca,0x001271,0x1271,0x128b,0x00123b,0x123b,0x1294,0x001254,0x1254,0x129c,0x00126e,0x126e)
GDI32(NtGdiEngCreateBitmap                              ,0x1240,0x1258,0x1254,0x1192,0x1192,0x1279,0x11d7,0x11d7,0x1299,0x11de,0x11de,0x12bb,0x00122b,0x122b,0x12be,0x001243,0x1243,0x12c8,0x00127c,0x127c,0x12c9,0x00127e,0x127e,0x12cf,0x001272,0x1272,0x1290,0x00123c,0x123c,0x1299,0x001255,0x1255,0x12a1,0x00126f,0x126f)
GDI32(NtGdiEngCreateClip                                ,0x1260,0x1278,0x1274,0x1193,0x1193,0x1299,0x11d8,0x11d8,0x12b9,0x11df,0x11df,0x12de,0x00122c,0x122c,0x12e7,0x001244,0x1244,0x12f3,0x00127d,0x127d,0x12f4,0x00127f,0x127f,0x12fa,0x001273,0x1273,0x12bb,0x00123d,0x123d,0x12c4,0x001256,0x1256,0x12ce,0x001270,0x1270)
GDI32(NtGdiEngCreateDeviceBitmap                        ,0x1242,0x125a,0x1256,0x1194,0x1194,0x127b,0x11d9,0x11d9,0x129b,0x11e0,0x11e0,0x12b9,0x00122d,0x122d,0x12bc,0x001245,0x1245,0x12c6,0x00127e,0x127e,0x12c7,0x001280,0x1280,0x12cd,0x001274,0x1274,0x128e,0x00123e,0x123e,0x1297,0x001257,0x1257,0x129f,0x001271,0x1271)
GDI32(NtGdiEngCreateDeviceSurface                       ,0x1241,0x1259,0x1255,0x1195,0x1195,0x127a,0x11da,0x11da,0x129a,0x11e1,0x11e1,0x12ba,0x00122e,0x122e,0x12bd,0x001246,0x1246,0x12c7,0x00127f,0x127f,0x12c8,0x001281,0x1281,0x12ce,0x001275,0x1275,0x128f,0x00123f,0x123f,0x1298,0x001258,0x1258,0x12a0,0x001272,0x1272)
GDI32(NtGdiEngCreatePalette                             ,0x1243,0x125b,0x1257,0x1196,0x1196,0x127c,0x11db,0x11db,0x129c,0x11e2,0x11e2,0x12b8,0x00122f,0x122f,0x12bb,0x001247,0x1247,0x12c5,0x001280,0x1280,0x12c6,0x001282,0x1282,0x12cc,0x001276,0x1276,0x128d,0x001240,0x1240,0x1296,0x001259,0x1259,0x129e,0x001273,0x1273)
GDI32(NtGdiEngDeleteClip                                ,0x1261,0x1279,0x1275,0x1197,0x1197,0x129a,0x11dc,0x11dc,0x12ba,0x11e3,0x11e3,0x12dd,0x001230,0x1230,0x12e6,0x001248,0x1248,0x12f2,0x001281,0x1281,0x12f3,0x001283,0x1283,0x12f9,0x001277,0x1277,0x12ba,0x001241,0x1241,0x12c3,0x00125a,0x125a,0x12cd,0x001274,0x1274)
GDI32(NtGdiEngDeletePalette                             ,0x1246,0x125e,0x125a,0x1198,0x1198,0x127f,0x11dd,0x11dd,0x129f,0x11e4,0x11e4,0x12b5,0x001231,0x1231,0x12b8,0x001249,0x1249,0x12c2,0x001282,0x1282,0x12c3,0x001284,0x1284,0x12c9,0x001278,0x1278,0x128a,0x001242,0x1242,0x1293,0x00125b,0x125b,0x129b,0x001275,0x1275)
GDI32(NtGdiEngDeletePath                                ,0x125f,0x1277,0x1273,0x1199,0x1199,0x1298,0x11de,0x11de,0x12b8,0x11e5,0x11e5,0x12df,0x001232,0x1232,0x12e8,0x00124a,0x124a,0x12f4,0x001283,0x1283,0x12f5,0x001285,0x1285,0x12fb,0x001279,0x1279,0x12bc,0x001243,0x1243,0x12c5,0x00125c,0x125c,0x12cf,0x001276,0x1276)
GDI32(NtGdiEngDeleteSurface                             ,0x1247,0x125f,0x125b,0x119a,0x119a,0x1280,0x11df,0x11df,0x12a0,0x11e6,0x11e6,0x12b4,0x001233,0x1233,0x12b7,0x00124b,0x124b,0x12c1,0x001284,0x1284,0x12c2,0x001286,0x1286,0x12c8,0x00127a,0x127a,0x1289,0x001244,0x1244,0x1292,0x00125d,0x125d,0x129a,0x001277,0x1277)
GDI32(NtGdiEngEraseSurface                              ,0x1248,0x1260,0x125c,0x119b,0x119b,0x1281,0x11e0,0x11e0,0x12a1,0x11e7,0x11e7,0x12b3,0x001234,0x1234,0x12b6,0x00124c,0x124c,0x12c0,0x001285,0x1285,0x12c1,0x001287,0x1287,0x12c7,0x00127b,0x127b,0x1288,0x001245,0x1245,0x1291,0x00125e,0x125e,0x1299,0x001278,0x1278)
GDI32(NtGdiEngFillPath                                  ,0x1250,0x1268,0x1264,0x119c,0x119c,0x1289,0x11e1,0x11e1,0x12a9,0x11e8,0x11e8,0x12ab,0x001235,0x1235,0x12ae,0x00124d,0x124d,0x12b8,0x001286,0x1286,0x12b9,0x001288,0x1288,0x12bf,0x00127c,0x127c,0x1280,0x001246,0x1246,0x1289,0x00125f,0x125f,0x1291,0x001279,0x1279)
GDI32(NtGdiEngGradientFill                              ,0x1255,0x126d,0x1269,0x119d,0x119d,0x128e,0x11e2,0x11e2,0x12ae,0x11e9,0x11e9,0x12a6,0x001236,0x1236,0x12a9,0x00124e,0x124e,0x12b3,0x001287,0x1287,0x12b4,0x001289,0x1289,0x12ba,0x00127d,0x127d,0x127b,0x001247,0x1247,0x1284,0x001260,0x1260,0x128c,0x00127a,0x127a)
GDI32(NtGdiEngLineTo                                    ,0x1253,0x126b,0x1267,0x119e,0x119e,0x128c,0x11e3,0x11e3,0x12ac,0x11ea,0x11ea,0x12a8,0x001237,0x1237,0x12ab,0x00124f,0x124f,0x12b5,0x001288,0x1288,0x12b6,0x00128a,0x128a,0x12bc,0x00127e,0x127e,0x127d,0x001248,0x1248,0x1286,0x001261,0x1261,0x128e,0x00127b,0x127b)
GDI32(NtGdiEngLockSurface                               ,0x124a,0x1262,0x125e,0x119f,0x119f,0x1283,0x11e4,0x11e4,0x12a3,0x11eb,0x11eb,0x12b1,0x001238,0x1238,0x12b4,0x001250,0x1250,0x12be,0x001289,0x1289,0x12bf,0x00128b,0x128b,0x12c5,0x00127f,0x127f,0x1286,0x001249,0x1249,0x128f,0x001262,0x1262,0x1297,0x00127c,0x127c)
GDI32(NtGdiEngMarkBandingSurface                        ,0x124e,0x1266,0x1262,0x11a0,0x11a0,0x1287,0x11e5,0x11e5,0x12a7,0x11ec,0x11ec,0x12ad,0x001239,0x1239,0x12b0,0x001251,0x1251,0x12ba,0x00128a,0x128a,0x12bb,0x00128c,0x128c,0x12c1,0x001280,0x1280,0x1282,0x00124a,0x124a,0x128b,0x001263,0x1263,0x1293,0x00127d,0x127d)
GDI32(NtGdiEngPaint                                     ,0x1252,0x126a,0x1266,0x11a1,0x11a1,0x128b,0x11e6,0x11e6,0x12ab,0x11ed,0x11ed,0x12a9,0x00123a,0x123a,0x12ac,0x001252,0x1252,0x12b6,0x00128b,0x128b,0x12b7,0x00128d,0x128d,0x12bd,0x001281,0x1281,0x127e,0x00124b,0x124b,0x1287,0x001264,0x1264,0x128f,0x00127e,0x127e)
GDI32(NtGdiEngPlgBlt                                    ,0x124d,0x1265,0x1261,0x11a2,0x11a2,0x1286,0x11e7,0x11e7,0x12a6,0x11ee,0x11ee,0x12ae,0x00123b,0x123b,0x12b1,0x001253,0x1253,0x12bb,0x00128c,0x128c,0x12bc,0x00128e,0x128e,0x12c2,0x001282,0x1282,0x1283,0x00124c,0x124c,0x128c,0x001265,0x1265,0x1294,0x00127f,0x127f)
GDI32(NtGdiEngStretchBlt                                ,0x124c,0x1264,0x1260,0x11a3,0x11a3,0x1285,0x11e8,0x11e8,0x12a5,0x11ef,0x11ef,0x12af,0x00123c,0x123c,0x12b2,0x001254,0x1254,0x12bc,0x00128d,0x128d,0x12bd,0x00128f,0x128f,0x12c3,0x001283,0x1283,0x1284,0x00124d,0x124d,0x128d,0x001266,0x1266,0x1295,0x001280,0x1280)
GDI32(NtGdiEngStretchBltROP                             ,0x1258,0x1270,0x126c,0x11a4,0x11a4,0x1291,0x11e9,0x11e9,0x12b1,0x11f0,0x11f0,0x12a3,0x00123d,0x123d,0x12a6,0x001255,0x1255,0x12b0,0x00128e,0x128e,0x12b1,0x001290,0x1290,0x12b7,0x001284,0x1284,0x1278,0x00124e,0x124e,0x1281,0x001267,0x1267,0x1289,0x001281,0x1281)
GDI32(NtGdiEngStrokeAndFillPath                         ,0x1251,0x1269,0x1265,0x11a5,0x11a5,0x128a,0x11ea,0x11ea,0x12aa,0x11f1,0x11f1,0x12aa,0x00123e,0x123e,0x12ad,0x001256,0x1256,0x12b7,0x00128f,0x128f,0x12b8,0x001291,0x1291,0x12be,0x001285,0x1285,0x127f,0x00124f,0x124f,0x1288,0x001268,0x1268,0x1290,0x001282,0x1282)
GDI32(NtGdiEngStrokePath                                ,0x124f,0x1267,0x1263,0x11a6,0x11a6,0x1288,0x11eb,0x11eb,0x12a8,0x11f2,0x11f2,0x12ac,0x00123f,0x123f,0x12af,0x001257,0x1257,0x12b9,0x001290,0x1290,0x12ba,0x001292,0x1292,0x12c0,0x001286,0x1286,0x1281,0x001250,0x1250,0x128a,0x001269,0x1269,0x1292,0x001283,0x1283)
GDI32(NtGdiEngTextOut                                   ,0x1257,0x126f,0x126b,0x11a7,0x11a7,0x1290,0x11ec,0x11ec,0x12b0,0x11f3,0x11f3,0x12a4,0x001240,0x1240,0x12a7,0x001258,0x1258,0x12b1,0x001291,0x1291,0x12b2,0x001293,0x1293,0x12b8,0x001287,0x1287,0x1279,0x001251,0x1251,0x1282,0x00126a,0x126a,0x128a,0x001284,0x1284)
GDI32(NtGdiEngTransparentBlt                            ,0x1256,0x126e,0x126a,0x11a8,0x11a8,0x128f,0x11ed,0x11ed,0x12af,0x11f4,0x11f4,0x12a5,0x001241,0x1241,0x12a8,0x001259,0x1259,0x12b2,0x001292,0x1292,0x12b3,0x001294,0x1294,0x12b9,0x001288,0x1288,0x127a,0x001252,0x1252,0x1283,0x00126b,0x126b,0x128b,0x001285,0x1285)
GDI32(NtGdiEngUnlockSurface                             ,0x1249,0x1261,0x125d,0x11a9,0x11a9,0x1282,0x11ee,0x11ee,0x12a2,0x11f5,0x11f5,0x12b2,0x001242,0x1242,0x12b5,0x00125a,0x125a,0x12bf,0x001293,0x1293,0x12c0,0x001295,0x1295,0x12c6,0x001289,0x1289,0x1287,0x001253,0x1253,0x1290,0x00126c,0x126c,0x1298,0x001286,0x1286)
GDI32(NtGdiEnsureDpiDepDefaultGuiFontForPlateau         ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x100e,0x00126d,0x126d,0x100e,0x001287,0x1287)
GDI32(NtGdiEnumFontChunk                                ,  NONE,0x1085,0x1085,0x10a5,0x10a5,0x1088,0x10a6,0x10a6,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
GDI32(NtGdiEnumFontClose                                ,  NONE,0x1086,0x1086,0x10a3,0x10a3,0x1089,0x10a4,0x10a4,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
GDI32(NtGdiEnumFontOpen                                 ,  NONE,0x1087,0x1087,0x10a4,0x10a4,0x108a,0x10a5,0x10a5,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
GDI32(NtGdiEnumFonts                                    ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x108a,0x11f6,0x11f6,0x10b0,0x001243,0x1243,0x10b2,0x00125b,0x125b,0x10b3,0x001294,0x1294,0x10b5,0x001296,0x1296,0x10b5,0x00128a,0x128a,0x10bc,0x001254,0x1254,0x10bd,0x00126e,0x126e,0x10bd,0x001288,0x1288)
GDI32(NtGdiEnumObjects                                  ,  NONE,0x1088,0x1088,0x11aa,0x11aa,0x108b,0x11ef,0x11ef,0x108b,0x11f7,0x11f7,0x10af,0x001244,0x1244,0x10b1,0x00125c,0x125c,0x10b2,0x001295,0x1295,0x10b4,0x001297,0x1297,0x10b4,0x00128b,0x128b,0x10bb,0x001255,0x1255,0x10bc,0x00126f,0x126f,0x10bc,0x001289,0x1289)
GDI32(NtGdiEqualRgn                                     ,  NONE,0x1089,0x1089,0x1117,0x1117,0x108c,0x1118,0x1118,0x108c,0x1111,0x1111,0x10ae,0x051111,0x1111,0x10b0,0x051112,0x1112,0x10b1,0x051113,0x1113,0x10b3,0x051113,0x1113,0x10b3,0x051112,0x1112,0x10ba,0x051107,0x1107,0x10bb,0x051107,0x1107,0x10bb,0x051107,0x1107)
GDI32(NtGdiEudcLoadUnloadLink                           ,  NONE,0x108a,0x108a,0x11ab,0x11ab,0x108d,0x11f0,0x11f0,0x108d,0x11f8,0x11f8,0x10ad,0x001245,0x1245,0x10af,0x00125d,0x125d,0x10b0,0x001296,0x1296,0x10b2,0x001298,0x1298,0x10b2,0x00128c,0x128c,0x10b9,0x001256,0x1256,0x10ba,0x001270,0x1270,0x10ba,0x00128a,0x128a)
GDI32(NtGdiExcludeClipRect                              ,  NONE,0x108b,0x108b,0x109a,0x109a,0x108e,0x109b,0x109b,0x108e,0x1099,0x1099,0x10ac,0x001099,0x1099,0x10ae,0x00109a,0x109a,0x10af,0x00109b,0x109b,0x10b1,0x00109b,0x109b,0x10b1,0x00109b,0x109b,0x10b8,0x001099,0x1099,0x10b9,0x001099,0x1099,0x10b9,0x001099,0x1099)
GDI32(NtGdiExtCreatePen                                 ,  NONE,0x108c,0x108c,0x10af,0x10af,0x108f,0x10b0,0x10b0,0x108f,0x10ab,0x10ab,0x10ab,0x0010ab,0x10ab,0x10ad,0x0010ac,0x10ac,0x10ae,0x0010ad,0x10ad,0x10b0,0x0010ad,0x10ad,0x10b0,0x0010ad,0x10ad,0x10b7,0x0010a6,0x10a6,0x10b8,0x0010a6,0x10a6,0x10b8,0x0010a6,0x10a6)
GDI32(NtGdiExtCreateRegion                              ,  NONE,0x108d,0x108d,0x108b,0x108b,0x1090,0x108c,0x108c,0x1090,0x108a,0x108a,0x10aa,0x0a108a,0x108a,0x10ac,0x0a108b,0x108b,0x10ad,0x0a108c,0x108c,0x10af,0x0a108c,0x108c,0x10af,0x0a108c,0x108c,0x10b6,0x0a108a,0x108a,0x10b7,0x0a108a,0x108a,0x10b7,0x0a108a,0x108a)
GDI32(NtGdiExtEscape                                    ,  NONE,0x108e,0x108e,0x111c,0x111c,0x1091,0x111d,0x111d,0x1091,0x1116,0x1116,0x10a9,0x001116,0x1116,0x10ab,0x001117,0x1117,0x10ac,0x001118,0x1118,0x10ae,0x001118,0x1118,0x10ae,0x001117,0x1117,0x10b5,0x00110c,0x110c,0x10b6,0x00110c,0x110c,0x10b6,0x00110c,0x110c)
GDI32(NtGdiExtFloodFill                                 ,  NONE,0x108f,0x108f,0x11ac,0x11ac,0x1092,0x11f1,0x11f1,0x1092,0x11f9,0x11f9,0x10a8,0x001246,0x1246,0x10aa,0x00125e,0x125e,0x10ab,0x001297,0x1297,0x10ad,0x001299,0x1299,0x10ad,0x00128d,0x128d,0x10b4,0x001257,0x1257,0x10b5,0x001271,0x1271,0x10b5,0x00128b,0x128b)
GDI32(NtGdiExtGetObjectW                                ,  NONE,0x1090,0x1090,0x1051,0x1051,0x1093,0x1052,0x1052,0x1093,0x1052,0x1052,0x10a7,0x001053,0x1053,0x10a9,0x001054,0x1054,0x10aa,0x001055,0x1055,0x10ac,0x001055,0x1055,0x10ac,0x001055,0x1055,0x10b3,0x001055,0x1055,0x10b4,0x001055,0x1055,0x10b4,0x001055,0x1055)
GDI32(NtGdiExtSelectClipRgn                             ,  NONE,0x1091,0x1091,0x102d,0x102d,0x1094,0x102e,0x102e,0x1094,0x102e,0x102e,0x10a6,0x10102f,0x102f,0x10a8,0x101030,0x1030,0x10a9,0x101031,0x1031,0x10ab,0x101031,0x1031,0x10ab,0x101031,0x1031,0x10b2,0x101031,0x1031,0x10b3,0x101031,0x1031,0x10b3,0x101031,0x1031)
GDI32(NtGdiExtTextOutW                                  ,  NONE,0x1092,0x1092,0x1037,0x1037,0x1095,0x1038,0x1038,0x1095,0x1038,0x1038,0x10a5,0x001039,0x1039,0x10a7,0x00103a,0x103a,0x10a8,0x00103b,0x103b,0x10aa,0x00103b,0x103b,0x10aa,0x00103b,0x103b,0x10b1,0x00103b,0x103b,0x10b2,0x00103b,0x103b,0x10b2,0x00103b,0x103b)
GDI32(NtGdiFONTOBJ_cGetAllGlyphHandles                  ,0x126f,0x1287,0x1283,0x11ad,0x11ad,0x12a8,0x11f2,0x11f2,0x12c8,0x11fa,0x11fa,0x12ef,0x001247,0x1247,0x12f8,0x00125f,0x125f,0x1304,0x001298,0x1298,0x1305,0x00129a,0x129a,0x130b,0x00128e,0x128e,0x12cc,0x001258,0x1258,0x12d5,0x001272,0x1272,0x12df,0x00128c,0x128c)
GDI32(NtGdiFONTOBJ_cGetGlyphs                           ,0x126a,0x1282,0x127e,0x11ae,0x11ae,0x12a3,0x11f3,0x11f3,0x12c3,0x11fb,0x11fb,0x12ec,0x001248,0x1248,0x12f5,0x001260,0x1260,0x1301,0x001299,0x1299,0x1302,0x00129b,0x129b,0x1308,0x00128f,0x128f,0x12c9,0x001259,0x1259,0x12d2,0x001273,0x1273,0x12dc,0x00128d,0x128d)
GDI32(NtGdiFONTOBJ_pQueryGlyphAttrs                     ,0x126d,0x1285,0x1281,0x11af,0x11af,0x12a6,0x11f4,0x11f4,0x12c6,0x11fc,0x11fc,0x12e9,0x001249,0x1249,0x12f2,0x001261,0x1261,0x12fe,0x00129a,0x129a,0x12ff,0x00129c,0x129c,0x1305,0x001290,0x1290,0x12c6,0x00125a,0x125a,0x12cf,0x001274,0x1274,0x12d9,0x00128e,0x128e)
GDI32(NtGdiFONTOBJ_pfdg                                 ,0x126c,0x1284,0x1280,0x11b0,0x11b0,0x12a5,0x11f5,0x11f5,0x12c5,0x11fd,0x11fd,0x12ea,0x00124a,0x124a,0x12f3,0x001262,0x1262,0x12ff,0x00129b,0x129b,0x1300,0x00129d,0x129d,0x1306,0x001291,0x1291,0x12c7,0x00125b,0x125b,0x12d0,0x001275,0x1275,0x12da,0x00128f,0x128f)
GDI32(NtGdiFONTOBJ_pifi                                 ,0x126b,0x1283,0x127f,0x11b1,0x11b1,0x12a4,0x11f6,0x11f6,0x12c4,0x11fe,0x11fe,0x12eb,0x00124b,0x124b,0x12f4,0x001263,0x1263,0x1300,0x00129c,0x129c,0x1301,0x00129e,0x129e,0x1307,0x001292,0x1292,0x12c8,0x00125c,0x125c,0x12d1,0x001276,0x1276,0x12db,0x001290,0x1290)
GDI32(NtGdiFONTOBJ_pvTrueTypeFontFile                   ,0x126e,0x1286,0x1282,0x11b2,0x11b2,0x12a7,0x11f7,0x11f7,0x12c7,0x11ff,0x11ff,0x12f0,0x00124c,0x124c,0x12f9,0x001264,0x1264,0x1305,0x00129d,0x129d,0x1306,0x00129f,0x129f,0x130c,0x001293,0x1293,0x12cd,0x00125d,0x125d,0x12d6,0x001277,0x1277,0x12e0,0x001291,0x1291)
GDI32(NtGdiFONTOBJ_pxoGetXform                          ,0x1269,0x1281,0x127d,0x11b3,0x11b3,0x12a2,0x11f8,0x11f8,0x12c2,0x1200,0x1200,0x12ed,0x00124d,0x124d,0x12f6,0x001265,0x1265,0x1302,0x00129e,0x129e,0x1303,0x0012a0,0x12a0,0x1309,0x001294,0x1294,0x12ca,0x00125e,0x125e,0x12d3,0x001278,0x1278,0x12dd,0x001292,0x1292)
GDI32(NtGdiFONTOBJ_vGetInfo                             ,0x1268,0x1280,0x127c,0x11b4,0x11b4,0x12a1,0x11f9,0x11f9,0x12c1,0x1201,0x1201,0x12ee,0x00124e,0x124e,0x12f7,0x001266,0x1266,0x1303,0x00129f,0x129f,0x1304,0x0012a1,0x12a1,0x130a,0x001295,0x1295,0x12cb,0x00125f,0x125f,0x12d4,0x001279,0x1279,0x12de,0x001293,0x1293)
GDI32(NtGdiFillPath                                     ,  NONE,0x1093,0x1093,0x1111,0x1111,0x1096,0x1112,0x1112,0x1096,0x110b,0x110b,0x10a4,0x03110b,0x110b,0x10a6,0x03110c,0x110c,0x10a7,0x03110d,0x110d,0x10a9,0x03110d,0x110d,0x10a9,0x03110c,0x110c,0x10b0,0x031101,0x1101,0x10b1,0x031101,0x1101,0x10b1,0x031101,0x1101)
GDI32(NtGdiFillRgn                                      ,  NONE,0x1094,0x1094,0x10d8,0x10d8,0x1097,0x10d9,0x10d9,0x1097,0x10d4,0x10d4,0x10a3,0x0010d4,0x10d4,0x10a5,0x0010d5,0x10d5,0x10a6,0x0010d6,0x10d6,0x10a8,0x0010d6,0x10d6,0x10a8,0x0010d5,0x10d5,0x10af,0x0010ca,0x10ca,0x10b0,0x0010ca,0x10ca,0x10b0,0x0010ca,0x10ca)
GDI32(NtGdiFlattenPath                                  ,  NONE,0x1095,0x1095,0x11b5,0x11b5,0x1098,0x11fa,0x11fa,0x1098,0x1202,0x1202,0x10a2,0x03124f,0x124f,0x10a4,0x031267,0x1267,0x10a5,0x0312a0,0x12a0,0x10a7,0x0312a2,0x12a2,0x10a7,0x031296,0x1296,0x10ae,0x031260,0x1260,0x10af,0x03127a,0x127a,0x10af,0x031294,0x1294)
GDI32(NtGdiFlush                                        ,  NONE,0x1097,0x1096,0x1011,0x1011,0x1099,0x1012,0x1012,0x1099,0x1012,0x1012,0x10a1,0x011013,0x1013,0x10a3,0x011014,0x1014,0x10a4,0x011015,0x1015,0x10a6,0x011015,0x1015,0x10a6,0x011015,0x1015,0x10ad,0x011015,0x1015,0x10ae,0x011015,0x1015,0x10ae,0x011015,0x1015)
GDI32(NtGdiFontIsLinked                                 ,0x100a,0x100a,0x100a,0x11b6,0x11b6,0x100a,0x11fb,0x11fb,0x100a,0x1203,0x1203,0x1132,0x031250,0x1250,0x1134,0x031268,0x1268,0x1135,0x0312a1,0x12a1,0x1137,0x0312a3,0x12a3,0x1138,0x031297,0x1297,0x10f5,0x031261,0x1261,0x10f6,0x03127b,0x127b,0x10f6,0x031295,0x1295)
GDI32(NtGdiForceUFIMapping                              ,  NONE,0x1098,0x1097,0x11b7,0x11b7,0x109a,0x11fc,0x11fc,0x109a,0x1204,0x1204,0x10a0,0x071251,0x1251,0x10a2,0x071269,0x1269,0x10a3,0x0712a2,0x12a2,0x10a5,0x0712a4,0x12a4,0x10a5,0x071298,0x1298,0x10ac,0x071262,0x1262,0x10ad,0x07127c,0x127c,0x10ad,0x071296,0x1296)
GDI32(NtGdiFrameRgn                                     ,  NONE,0x1099,0x1098,0x11b8,0x11b8,0x109b,0x11fd,0x11fd,0x109b,0x1205,0x1205,0x109f,0x001252,0x1252,0x10a1,0x00126a,0x126a,0x10a2,0x0012a3,0x12a3,0x10a4,0x0012a5,0x12a5,0x10a4,0x001299,0x1299,0x10ab,0x001263,0x1263,0x10ac,0x00127d,0x127d,0x10ac,0x001297,0x1297)
GDI32(NtGdiFullscreenControl                            ,0x1096,0x109a,0x1099,0x11b9,0x11b9,0x109c,0x11fe,0x11fe,0x109c,0x1206,0x1206,0x109e,0x001253,0x1253,0x10a0,0x00126b,0x126b,0x10a1,0x0012a4,0x12a4,0x10a3,0x0012a6,0x12a6,0x10a3,0x00129a,0x129a,0x10aa,0x001264,0x1264,0x10ab,0x00127e,0x127e,0x10ab,0x001298,0x1298)
GDI32(NtGdiGetAndSetDCDword                             ,  NONE,0x109b,0x109a,0x1067,0x1067,0x109d,0x1068,0x1068,0x109d,0x1068,0x1068,0x109d,0x141068,0x1068,0x109f,0x141069,0x1069,0x10a0,0x14106a,0x106a,0x10a2,0x14106a,0x106a,0x10a2,0x14106a,0x106a,0x10a9,0x141069,0x1069,0x10aa,0x141069,0x1069,0x10aa,0x141069,0x1069)
GDI32(NtGdiGetAppClipBox                                ,  NONE,0x109c,0x109b,0x1042,0x1042,0x109e,0x1043,0x1043,0x109e,0x1043,0x1043,0x109c,0x071044,0x1044,0x109e,0x071045,0x1045,0x109f,0x071046,0x1046,0x10a1,0x071046,0x1046,0x10a1,0x071046,0x1046,0x10a8,0x071046,0x1046,0x10a9,0x071046,0x1046,0x10a9,0x071046,0x1046)
GDI32(NtGdiGetAppliedDeviceGammaRamp                    ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x102b,0x071265,0x1265,0x102c,0x07127f,0x127f,0x102c,0x071299,0x1299)
GDI32(NtGdiGetBitmapBits                                ,  NONE,0x109d,0x109c,0x10e2,0x10e2,0x109f,0x10e3,0x10e3,0x109f,0x10de,0x10de,0x109b,0x0a10de,0x10de,0x109d,0x0a10df,0x10df,0x109e,0x0a10e0,0x10e0,0x10a0,0x0a10e0,0x10e0,0x10a0,0x0a10df,0x10df,0x10a7,0x0a10d4,0x10d4,0x10a8,0x0a10d4,0x10d4,0x10a8,0x0a10d4,0x10d4)
GDI32(NtGdiGetBitmapDimension                           ,  NONE,0x109e,0x109d,0x1101,0x1101,0x10a0,0x1102,0x1102,0x10a0,0x10fc,0x10fc,0x109a,0x0510fc,0x10fc,0x109c,0x0510fd,0x10fd,0x109d,0x0510fe,0x10fe,0x109f,0x0510fe,0x10fe,0x109f,0x0510fd,0x10fd,0x10a6,0x0510f2,0x10f2,0x10a7,0x0510f2,0x10f2,0x10a7,0x0510f2,0x10f2)
GDI32(NtGdiGetBitmapDpiScaleValue                       ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x100d,0x001266,0x1266,0x100d,0x001280,0x1280,0x100d,0x00129a,0x129a)
GDI32(NtGdiGetBoundsRect                                ,  NONE,0x109f,0x109e,0x11ba,0x11ba,0x10a1,0x11ff,0x11ff,0x10a1,0x1207,0x1207,0x1099,0x0c1254,0x1254,0x109b,0x0c126c,0x126c,0x109c,0x0c12a5,0x12a5,0x109e,0x0c12a7,0x12a7,0x109e,0x0c129b,0x129b,0x10a5,0x0c1267,0x1267,0x10a6,0x0c1281,0x1281,0x10a6,0x0c129b,0x129b)
GDI32(NtGdiGetCOPPCompatibleOPMInformation              ,  NONE,  NONE,  NONE,  NONE,  NONE,0x10ab,0x1200,0x1200,0x10ab,0x1208,0x1208,0x108f,0x001255,0x1255,0x1091,0x00126d,0x126d,0x1092,0x0012a6,0x12a6,0x1092,0x0012a8,0x12a8,0x1092,0x00129c,0x129c,0x1099,0x001268,0x1268,0x109a,0x001282,0x1282,0x109a,0x00129c,0x129c)
GDI32(NtGdiGetCertificate                               ,  NONE,  NONE,  NONE,  NONE,  NONE,0x10a2,0x1201,0x1201,0x10a2,0x1209,0x1209,0x1098,0x001256,0x1256,0x109a,0x00126e,0x126e,0x109b,0x0012a7,0x12a7,0x109b,0x0012a9,0x12a9,0x109b,0x00129d,0x129d,0x10a2,0x001269,0x1269,0x10a3,0x001283,0x1283,0x10a3,0x00129d,0x129d)
GDI32(NtGdiGetCertificateByHandle                       ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x109d,0x0012aa,0x12aa,0x109d,0x00129e,0x129e,0x10a4,0x00126a,0x126a,0x10a5,0x001284,0x1284,0x10a5,0x00129e,0x129e)
GDI32(NtGdiGetCertificateSize                           ,  NONE,  NONE,  NONE,  NONE,  NONE,0x10a3,0x1202,0x1202,0x10a3,0x120a,0x120a,0x1097,0x001257,0x1257,0x1099,0x00126f,0x126f,0x109a,0x0012a8,0x12a8,0x109a,0x0012ab,0x12ab,0x109a,0x00129f,0x129f,0x10a1,0x00126b,0x126b,0x10a2,0x001285,0x1285,0x10a2,0x00129f,0x129f)
GDI32(NtGdiGetCertificateSizeByHandle                   ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x109c,0x0012ac,0x12ac,0x109c,0x0012a0,0x12a0,0x10a3,0x00126c,0x126c,0x10a4,0x001286,0x1286,0x10a4,0x0012a0,0x12a0)
GDI32(NtGdiGetCharABCWidthsW                            ,  NONE,0x10a0,0x109f,0x11bb,0x11bb,0x10a4,0x1203,0x1203,0x10a4,0x120b,0x120b,0x1096,0x001258,0x1258,0x1098,0x001270,0x1270,0x1099,0x0012a9,0x12a9,0x1099,0x0012ad,0x12ad,0x1099,0x0012a1,0x12a1,0x10a0,0x00126d,0x126d,0x10a1,0x001287,0x1287,0x10a1,0x0012a1,0x12a1)
GDI32(NtGdiGetCharSet                                   ,  NONE,0x10a2,0x10a1,0x1009,0x1009,0x10a6,0x1009,0x1009,0x10a6,0x1009,0x1009,0x1094,0x03100a,0x100a,0x1096,0x03100b,0x100b,0x1097,0x03100c,0x100c,0x1097,0x03100c,0x100c,0x1097,0x03100c,0x100c,0x109e,0x03100c,0x100c,0x109f,0x03100c,0x100c,0x109f,0x03100c,0x100c)
GDI32(NtGdiGetCharWidthInfo                             ,0x10a0,0x10a4,0x10a3,0x10d0,0x10d0,0x10a8,0x10d1,0x10d1,0x10a8,0x10cc,0x10cc,0x1092,0x0710cc,0x10cc,0x1094,0x0710cd,0x10cd,0x1095,0x0710ce,0x10ce,0x1095,0x0710ce,0x10ce,0x1095,0x0710ce,0x10ce,0x109c,0x0710c3,0x10c3,0x109d,0x0710c3,0x10c3,0x109d,0x0710c3,0x10c3)
GDI32(NtGdiGetCharWidthW                                ,  NONE,0x10a3,0x10a2,0x10cb,0x10cb,0x10a7,0x10cc,0x10cc,0x10a7,0x10c7,0x10c7,0x1093,0x0010c7,0x10c7,0x1095,0x0010c8,0x10c8,0x1096,0x0010c9,0x10c9,0x1096,0x0010c9,0x10c9,0x1096,0x0010c9,0x10c9,0x109d,0x0010be,0x10be,0x109e,0x0010be,0x10be,0x109e,0x0010be,0x10be)
GDI32(NtGdiGetCharacterPlacementW                       ,  NONE,0x10a1,0x10a0,0x11bc,0x11bc,0x10a5,0x1204,0x1204,0x10a5,0x120c,0x120c,0x1095,0x001259,0x1259,0x1097,0x001271,0x1271,0x1098,0x0012aa,0x12aa,0x1098,0x0012ae,0x12ae,0x1098,0x0012a2,0x12a2,0x109f,0x00126e,0x126e,0x10a0,0x001288,0x1288,0x10a0,0x0012a2,0x12a2)
GDI32(NtGdiGetCodePage                                  ,0x10b6,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
GDI32(NtGdiGetColorAdjustment                           ,  NONE,0x10a5,0x10a4,0x11bd,0x11bd,0x10a9,0x1205,0x1205,0x10a9,0x120d,0x120d,0x1091,0x07125a,0x125a,0x1093,0x071272,0x1272,0x1094,0x0712ab,0x12ab,0x1094,0x0712af,0x12af,0x1094,0x0712a3,0x12a3,0x109b,0x07126f,0x126f,0x109c,0x071289,0x1289,0x109c,0x0712a3,0x12a3)
GDI32(NtGdiGetColorSpaceforBitmap                       ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x1206,  NONE,  NONE,0x120e,  NONE,    NONE,0x125b,  NONE,    NONE,0x1273,0x1093,0x0412ac,0x12ac,0x1093,0x0412b0,0x12b0,0x1093,0x0412a4,0x12a4,0x109a,0x041270,0x1270,0x109b,0x04128a,0x128a,0x109b,0x0412a4,0x12a4)
GDI32(NtGdiGetCurrentDpiInfo                            ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,0x1090,0x001274,0x1274,0x1091,0x0012ad,0x12ad,0x1091,0x0012b1,0x12b1,0x1091,0x0012a5,0x12a5,0x1098,0x001271,0x1271,0x1099,0x00128b,0x128b,0x1099,0x0012a5,0x12a5)
GDI32(NtGdiGetDCDpiScaleValue                           ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x100c,0x001272,0x1272,0x100c,0x00128c,0x128c,0x100c,0x0012a6,0x12a6)
GDI32(NtGdiGetDCDword                                   ,  NONE,0x10a7,0x10a6,0x103e,0x103e,0x10ac,0x103f,0x103f,0x10ac,0x103f,0x103f,0x108e,0x0c1040,0x1040,0x108f,0x0c1041,0x1041,0x1090,0x0c1042,0x1042,0x1090,0x0c1042,0x1042,0x1090,0x0c1042,0x1042,0x1097,0x0c1042,0x1042,0x1098,0x0c1042,0x1042,0x1098,0x0c1042,0x1042)
GDI32(NtGdiGetDCObject                                  ,  NONE,0x10a9,0x10a8,0x1034,0x1034,0x10ae,0x1035,0x1035,0x10ae,0x1035,0x1035,0x108c,0x081036,0x1036,0x108d,0x081037,0x1037,0x108e,0x081038,0x1038,0x108e,0x081038,0x1038,0x108e,0x081038,0x1038,0x1095,0x081038,0x1038,0x1096,0x081038,0x1038,0x1096,0x081038,0x1038)
GDI32(NtGdiGetDCPoint                                   ,  NONE,0x10aa,0x10a9,0x1073,0x1073,0x10af,0x1074,0x1074,0x10af,0x1073,0x1073,0x108b,0x0c1073,0x1073,0x108c,0x0c1074,0x1074,0x108d,0x0c1075,0x1075,0x108d,0x0c1075,0x1075,0x108d,0x0c1075,0x1075,0x1094,0x0c1074,0x1074,0x1095,0x0c1074,0x1074,0x1095,0x0c1074,0x1074)
GDI32(NtGdiGetDCforBitmap                               ,  NONE,0x10a8,0x10a7,0x109c,0x109c,0x10ad,0x109d,0x109d,0x10ad,0x109b,0x109b,0x108d,0x04109b,0x109b,0x108e,0x04109c,0x109c,0x108f,0x04109d,0x109d,0x108f,0x04109d,0x109d,0x108f,0x04109d,0x109d,0x1096,0x04109b,0x109b,0x1097,0x04109b,0x109b,0x1097,0x04109b,0x109b)
GDI32(NtGdiGetDIBitsInternal                            ,  NONE,0x10ae,0x10ad,0x1086,0x1086,0x10b3,0x1087,0x1087,0x10b3,0x1085,0x1085,0x1087,0x001085,0x1085,0x1088,0x001086,0x1086,0x1089,0x001087,0x1087,0x1089,0x001087,0x1087,0x1089,0x001087,0x1087,0x1090,0x001085,0x1085,0x1091,0x001085,0x1085,0x1091,0x001085,0x1085)
GDI32(NtGdiGetDeviceCaps                                ,  NONE,0x10ab,0x10aa,0x11bf,0x11bf,0x10b0,0x1207,0x1207,0x10b0,0x120f,0x120f,0x108a,0x08125c,0x125c,0x108b,0x081275,0x1275,0x108c,0x0812ae,0x12ae,0x108c,0x0812b2,0x12b2,0x108c,0x0812a6,0x12a6,0x1093,0x081273,0x1273,0x1094,0x08128d,0x128d,0x1094,0x0812a7,0x12a7)
GDI32(NtGdiGetDeviceCapsAll                             ,  NONE,0x10ad,0x10ac,0x11c0,0x11c0,0x10b2,0x1208,0x1208,0x10b2,0x1210,0x1210,0x1088,0x07125d,0x125d,0x1089,0x071276,0x1276,0x108a,0x0712af,0x12af,0x108a,0x0712b3,0x12b3,0x108a,0x0712a7,0x12a7,0x1091,0x071274,0x1274,0x1092,0x07128e,0x128e,0x1092,0x0712a8,0x12a8)
GDI32(NtGdiGetDeviceGammaRamp                           ,  NONE,0x10ac,0x10ab,0x11c1,0x11c1,0x10b1,0x1209,0x1209,0x10b1,0x1211,0x1211,0x1089,0x00125e,0x125e,0x108a,0x001277,0x1277,0x108b,0x0012b0,0x12b0,0x108b,0x0012b4,0x12b4,0x108b,0x0012a8,0x12a8,0x1092,0x001275,0x1275,0x1093,0x00128f,0x128f,0x1093,0x0012a9,0x12a9)
GDI32(NtGdiGetDeviceWidth                               ,  NONE,0x1117,0x1116,0x11c2,0x11c2,0x111f,0x120a,0x120a,0x1121,0x1212,0x1212,0x1019,0x03125f,0x125f,0x101a,0x031278,0x1278,0x101b,0x0312b1,0x12b1,0x101b,0x0312b5,0x12b5,0x101b,0x0312a9,0x12a9,0x101f,0x031276,0x1276,0x1020,0x031290,0x1290,0x1020,0x0312aa,0x12aa)
GDI32(NtGdiGetDhpdev                                    ,  NONE,0x1292,0x128e,0x11c3,0x11c3,0x12b3,0x120b,0x120b,0x12d3,0x1213,0x1213,0x12fc,0x001260,0x1260,0x1305,0x001279,0x1279,0x1311,0x0012b2,0x12b2,0x1312,0x0012b6,0x12b6,0x1318,0x0012aa,0x12aa,0x12d9,0x001277,0x1277,0x12e2,0x001291,0x1291,0x12ec,0x0012ab,0x12ab)
GDI32(NtGdiGetETM                                       ,  NONE,0x10af,0x10ae,0x11c4,0x11c4,0x10b4,0x120c,0x120c,0x10b4,0x1214,0x1214,0x1086,0x071261,0x1261,0x1087,0x07127a,0x127a,0x1088,0x0712b3,0x12b3,0x1088,0x0712b7,0x12b7,0x1087,0x0712ab,0x12ab,0x108e,0x071278,0x1278,0x108f,0x071292,0x1292,0x108f,0x0712ac,0x12ac)
GDI32(NtGdiGetEmbUFI                                    ,  NONE,0x10d2,0x10d1,0x11c5,0x11c5,0x10da,0x120d,0x120d,0x10dc,0x1215,0x1215,0x105e,0x001262,0x1262,0x105f,0x00127b,0x127b,0x1060,0x0012b4,0x12b4,0x1060,0x0012b8,0x12b8,0x1060,0x0012ac,0x12ac,0x1067,0x001279,0x1279,0x1068,0x001293,0x1293,0x1068,0x0012ad,0x12ad)
GDI32(NtGdiGetEmbedFonts                                ,  NONE,0x10d4,0x10d3,0x11c6,0x11c6,0x10dc,0x120e,0x120e,0x10de,0x1216,0x1216,0x105c,0x011263,0x1263,0x105d,0x01127c,0x127c,0x105e,0x0112b5,0x12b5,0x105e,0x0112b9,0x12b9,0x105e,0x0112ad,0x12ad,0x1065,0x01127a,0x127a,0x1066,0x011294,0x1294,0x1066,0x0112ae,0x12ae)
GDI32(NtGdiGetEntry                                     ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1088,0x0012ae,0x12ae,0x108f,0x00127b,0x127b,0x1090,0x001295,0x1295,0x1090,0x0012af,0x12af)
GDI32(NtGdiGetEudcTimeStampEx                           ,  NONE,0x10b0,0x10af,0x11c7,0x11c7,0x10b5,0x120f,0x120f,0x10b5,0x1217,0x1217,0x1085,0x0a1264,0x1264,0x1086,0x0a127d,0x127d,0x1087,0x0a12b6,0x12b6,0x1087,0x0a12ba,0x12ba,0x1086,0x0a12af,0x12af,0x108d,0x0a127c,0x127c,0x108e,0x0a1296,0x1296,0x108e,0x0a12b0,0x12b0)
GDI32(NtGdiGetFontData                                  ,  NONE,0x10b1,0x10b0,0x10db,0x10db,0x10b6,0x10dc,0x10dc,0x10b6,0x10d7,0x10d7,0x1084,0x0010d7,0x10d7,0x1085,0x0010d8,0x10d8,0x1086,0x0010d9,0x10d9,0x1086,0x0010d9,0x10d9,0x1085,0x0010d8,0x10d8,0x108c,0x0010cd,0x10cd,0x108d,0x0010cd,0x10cd,0x108d,0x0010cd,0x10cd)
GDI32(NtGdiGetFontFileData                              ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x10b7,0x1218,0x1218,0x1083,0x001265,0x1265,0x1084,0x00127e,0x127e,0x1085,0x0012b7,0x12b7,0x1085,0x0012bb,0x12bb,0x1084,0x0012b0,0x12b0,0x108b,0x00127d,0x127d,0x108c,0x001297,0x1297,0x108c,0x0012b1,0x12b1)
GDI32(NtGdiGetFontFileInfo                              ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x10b8,0x1219,0x1219,0x1082,0x001266,0x1266,0x1083,0x00127f,0x127f,0x1084,0x0012b8,0x12b8,0x1084,0x0012bc,0x12bc,0x1083,0x0012b1,0x12b1,0x108a,0x00127e,0x127e,0x108b,0x001298,0x1298,0x108b,0x0012b2,0x12b2)
GDI32(NtGdiGetFontResourceInfoInternalW                 ,  NONE,0x10b2,0x10b1,0x11c8,0x11c8,0x10b7,0x1210,0x1210,0x10b9,0x121a,0x121a,0x1081,0x001267,0x1267,0x1082,0x001280,0x1280,0x1083,0x0012b9,0x12b9,0x1083,0x0012bd,0x12bd,0x1082,0x0012b2,0x12b2,0x1089,0x00127f,0x127f,0x108a,0x001299,0x1299,0x108a,0x0012b3,0x12b3)
GDI32(NtGdiGetFontUnicodeRanges                         ,0x10cf,0x10d7,0x10d6,0x11c9,0x11c9,0x10df,0x1211,0x1211,0x10e1,0x121b,0x121b,0x1059,0x071268,0x1268,0x105a,0x071281,0x1281,0x105b,0x0712ba,0x12ba,0x105b,0x0712be,0x12be,0x105b,0x0712b3,0x12b3,0x1062,0x071280,0x1280,0x1063,0x07129a,0x129a,0x1063,0x0712b4,0x12b4)
GDI32(NtGdiGetGammaRampCapability                       ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x1029,0x031281,0x1281,0x102a,0x03129b,0x129b,0x102a,0x0312b5,0x12b5)
GDI32(NtGdiGetGlyphIndicesW                             ,0x10af,0x10b3,0x10b2,0x11ca,0x11ca,0x10b8,0x1212,0x1212,0x10ba,0x121c,0x121c,0x1080,0x001269,0x1269,0x1081,0x001282,0x1282,0x1082,0x0012bb,0x12bb,0x1082,0x0012bf,0x12bf,0x1081,0x0012b4,0x12b4,0x1088,0x001282,0x1282,0x1089,0x00129c,0x129c,0x1089,0x0012b6,0x12b6)
GDI32(NtGdiGetGlyphIndicesWInternal                     ,  NONE,0x10b4,0x10b3,0x11cb,0x11cb,0x10b9,0x1213,0x1213,0x10bb,0x121d,0x121d,0x107f,0x00126a,0x126a,0x1080,0x001283,0x1283,0x1081,0x0012bc,0x12bc,0x1081,0x0012c0,0x12c0,0x1080,0x0012b5,0x12b5,0x1087,0x001283,0x1283,0x1088,0x00129d,0x129d,0x1088,0x0012b7,0x12b7)
GDI32(NtGdiGetGlyphOutline                              ,  NONE,0x10b5,0x10b4,0x11cc,0x11cc,0x10ba,0x1214,0x1214,0x10bc,0x121e,0x121e,0x107e,0x00126b,0x126b,0x107f,0x001284,0x1284,0x1080,0x0012bd,0x12bd,0x1080,0x0012c1,0x12c1,0x107f,0x0012b6,0x12b6,0x1086,0x001284,0x1284,0x1087,0x00129e,0x129e,0x1087,0x0012b8,0x12b8)
GDI32(NtGdiGetKerningPairs                              ,  NONE,0x10b6,0x10b5,0x11cd,0x11cd,0x10bc,0x1215,0x1215,0x10be,0x121f,0x121f,0x107c,0x0c126c,0x126c,0x107d,0x0c1285,0x1285,0x107e,0x0c12be,0x12be,0x107e,0x0c12c2,0x12c2,0x107d,0x0c12b7,0x12b7,0x1084,0x0c1285,0x1285,0x1085,0x0c129f,0x129f,0x1085,0x0c12b9,0x12b9)
GDI32(NtGdiGetLinkedUFIs                                ,  NONE,0x10b7,0x10b6,0x11ce,0x11ce,0x10bd,0x1216,0x1216,0x10bf,0x1220,0x1220,0x107b,0x0c126d,0x126d,0x107c,0x0c1286,0x1286,0x107d,0x0c12bf,0x12bf,0x107d,0x0c12c3,0x12c3,0x107c,0x0c12b8,0x12b8,0x1083,0x0c1286,0x1286,0x1084,0x0c12a0,0x12a0,0x1084,0x0c12ba,0x12ba)
GDI32(NtGdiGetMiterLimit                                ,  NONE,0x10b8,0x10b7,0x11cf,0x11cf,0x10be,0x1217,0x1217,0x10c0,0x1221,0x1221,0x107a,0x07126e,0x126e,0x107b,0x071287,0x1287,0x107c,0x0712c0,0x12c0,0x107c,0x0712c4,0x12c4,0x107b,0x0712b9,0x12b9,0x1082,0x071287,0x1287,0x1083,0x0712a1,0x12a1,0x1083,0x0712bb,0x12bb)
GDI32(NtGdiGetMonitorID                                 ,  NONE,0x10b9,0x10b8,0x11d0,0x11d0,0x10bf,0x1218,0x1218,0x10c1,0x1222,0x1222,0x1079,0x0c126f,0x126f,0x107a,0x0c1288,0x1288,0x107b,0x0c12c1,0x12c1,0x107b,0x0c12c5,0x12c5,0x107a,0x0c12ba,0x12ba,0x1081,0x0c1288,0x1288,0x1082,0x0c12a2,0x12a2,0x1082,0x0c12bc,0x12bc)
GDI32(NtGdiGetNearestColor                              ,  NONE,0x10ba,0x10b9,0x1071,0x1071,0x10c0,0x1072,0x1072,0x10c2,0x1071,0x1071,0x1078,0x071071,0x1071,0x1079,0x071072,0x1072,0x107a,0x071073,0x1073,0x107a,0x071073,0x1073,0x1079,0x071073,0x1073,0x1080,0x071072,0x1072,0x1081,0x071072,0x1072,0x1081,0x071072,0x1072)
GDI32(NtGdiGetNearestPaletteIndex                       ,  NONE,0x10bb,0x10ba,0x10c8,0x10c8,0x10c1,0x10c9,0x10c9,0x10c3,0x10c4,0x10c4,0x1077,0x0510c4,0x10c4,0x1078,0x0510c5,0x10c5,0x1079,0x0510c6,0x10c6,0x1079,0x0510c6,0x10c6,0x1078,0x0510c6,0x10c6,0x107f,0x0510bd,0x10bd,0x1080,0x0510bd,0x10bd,0x1080,0x0510bd,0x10bd)
GDI32(NtGdiGetNumberOfPhysicalMonitors                  ,  NONE,  NONE,  NONE,  NONE,  NONE,0x12f8,0x1219,0x1219,0x1324,0x1223,0x1223,0x136d,0x001270,0x1270,0x137f,0x001289,0x1289,0x13b8,0x0012c2,0x12c2,0x13ba,0x0012c6,0x12c6,0x13c7,0x0012bb,0x12bb,0x1398,0x001289,0x1289,0x13b2,0x0012a3,0x12a3,0x13c4,0x0012bd,0x12bd)
GDI32(NtGdiGetOPMInformation                            ,  NONE,  NONE,  NONE,  NONE,  NONE,0x10bb,0x121a,0x121a,0x10bd,0x1224,0x1224,0x107d,0x001271,0x1271,0x107e,0x00128a,0x128a,0x107f,0x0012c3,0x12c3,0x107f,0x0012c7,0x12c7,0x107e,0x0012bc,0x12bc,0x1085,0x00128a,0x128a,0x1086,0x0012a4,0x12a4,0x1086,0x0012be,0x12be)
GDI32(NtGdiGetOPMRandomNumber                           ,  NONE,  NONE,  NONE,  NONE,  NONE,0x10c3,0x121b,0x121b,0x10c5,0x1225,0x1225,0x1075,0x001272,0x1272,0x1076,0x00128b,0x128b,0x1077,0x0012c4,0x12c4,0x1077,0x0012c8,0x12c8,0x1076,0x0012bd,0x12bd,0x107d,0x00128b,0x128b,0x107e,0x0012a5,0x12a5,0x107e,0x0012bf,0x12bf)
GDI32(NtGdiGetObjectBitmapHandle                        ,  NONE,0x10bc,0x10bb,0x11d1,0x11d1,0x10c2,0x121c,0x121c,0x10c4,0x1226,0x1226,0x1076,0x051273,0x1273,0x1077,0x05128c,0x128c,0x1078,0x0512c5,0x12c5,0x1078,0x0512c9,0x12c9,0x1077,0x0512be,0x12be,0x107e,0x05128c,0x128c,0x107f,0x0512a6,0x12a6,0x107f,0x0512c0,0x12c0)
GDI32(NtGdiGetOutlineTextMetricsInternalW               ,  NONE,0x10bd,0x10bc,0x10b7,0x10b7,0x10c4,0x10b8,0x10b8,0x10c6,0x10b3,0x10b3,0x1074,0x0010b3,0x10b3,0x1075,0x0010b4,0x10b4,0x1076,0x0010b5,0x10b5,0x1076,0x0010b5,0x10b5,0x1075,0x0010b5,0x10b5,0x107c,0x0010ae,0x10ae,0x107d,0x0010ae,0x10ae,0x107d,0x0010ae,0x10ae)
GDI32(NtGdiGetPath                                      ,  NONE,0x10be,0x10bd,0x11d2,0x11d2,0x10c5,0x121d,0x121d,0x10c7,0x1227,0x1227,0x1073,0x141274,0x1274,0x1074,0x14128d,0x128d,0x1075,0x1412c6,0x12c6,0x1075,0x1412ca,0x12ca,0x1074,0x1412bf,0x12bf,0x107b,0x14128d,0x128d,0x107c,0x1412a7,0x12a7,0x107c,0x1412c1,0x12c1)
GDI32(NtGdiGetPerBandInfo                               ,  NONE,0x107c,0x107c,0x11d3,0x11d3,0x107f,0x121e,0x121e,0x1080,0x1228,0x1228,0x10ba,0x071275,0x1275,0x10bc,0x07128e,0x128e,0x10bd,0x0712c7,0x12c7,0x10bf,0x0712cb,0x12cb,0x10bf,0x0712c0,0x12c0,0x10c6,0x07128e,0x128e,0x10c7,0x0712a8,0x12a8,0x10c7,0x0712c2,0x12c2)
GDI32(NtGdiGetPhysicalMonitorDescription                ,  NONE,  NONE,  NONE,  NONE,  NONE,0x12fa,0x121f,0x121f,0x1326,0x1229,0x1229,0x136b,0x001276,0x1276,0x137d,0x00128f,0x128f,0x13b6,0x0012c8,0x12c8,0x13b8,0x0012cc,0x12cc,0x13c5,0x0012c1,0x12c1,0x1396,0x00128f,0x128f,0x13b0,0x0012a9,0x12a9,0x13c2,0x0012c3,0x12c3)
GDI32(NtGdiGetPhysicalMonitors                          ,  NONE,  NONE,  NONE,  NONE,  NONE,0x12f9,0x1220,0x1220,0x1325,0x122a,0x122a,0x136c,0x001277,0x1277,0x137e,0x001290,0x1290,0x13b7,0x0012c9,0x12c9,0x13b9,0x0012cd,0x12cd,0x13c6,0x0012c2,0x12c2,0x1397,0x001290,0x1290,0x13b1,0x0012aa,0x12aa,0x13c3,0x0012c4,0x12c4)
GDI32(NtGdiGetPixel                                     ,  NONE,0x10bf,0x10be,0x10c3,0x10c3,0x10c6,0x10c4,0x10c4,0x10c8,0x10bf,0x10bf,0x1072,0x0b10bf,0x10bf,0x1073,0x0b10c0,0x10c0,0x1074,0x0b10c1,0x10c1,0x1074,0x0b10c1,0x10c1,0x1073,0x0b10c1,0x10c1,0x107a,0x0b10ba,0x10ba,0x107b,0x0b10ba,0x10ba,0x107b,0x0b10ba,0x10ba)
GDI32(NtGdiGetProcessSessionFonts                       ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x13d2,0x0012c3,0x12c3,0x13a3,0x001291,0x1291,0x13bd,0x0012ab,0x12ab,0x13cf,0x0012c5,0x12c5)
GDI32(NtGdiGetPublicFontTableChangeCookie               ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x13d3,0x0012c4,0x12c4,0x13a4,0x001292,0x1292,0x13be,0x0012ac,0x12ac,0x13d0,0x0012c6,0x12c6)
GDI32(NtGdiGetRandomRgn                                 ,  NONE,0x10c0,0x10bf,0x102a,0x102a,0x10c7,0x102b,0x102b,0x10c9,0x102b,0x102b,0x1071,0x10102c,0x102c,0x1072,0x10102d,0x102d,0x1073,0x10102e,0x102e,0x1073,0x10102e,0x102e,0x1072,0x10102e,0x102e,0x1079,0x10102e,0x102e,0x107a,0x10102e,0x102e,0x107a,0x10102e,0x102e)
GDI32(NtGdiGetRasterizerCaps                            ,  NONE,0x10c1,0x10c0,0x10eb,0x10eb,0x10c8,0x10ec,0x10ec,0x10ca,0x10e7,0x10e7,0x1070,0x0510e7,0x10e7,0x1071,0x0510e8,0x10e8,0x1072,0x0510e9,0x10e9,0x1072,0x0510e9,0x10e9,0x1071,0x0510e8,0x10e8,0x1078,0x0510dd,0x10dd,0x1079,0x0510dd,0x10dd,0x1079,0x0510dd,0x10dd)
GDI32(NtGdiGetRealizationInfo                           ,  NONE,0x10c2,0x10c1,0x11d4,0x11d4,0x10c9,0x1221,0x1221,0x10cb,0x122b,0x122b,0x106f,0x071278,0x1278,0x1070,0x071291,0x1291,0x1071,0x0712ca,0x12ca,0x1071,0x0712ce,0x12ce,0x1070,0x0712c5,0x12c5,0x1077,0x071293,0x1293,0x1078,0x0712ad,0x12ad,0x1078,0x0712c7,0x12c7)
GDI32(NtGdiGetRegionData                                ,  NONE,0x10c3,0x10c2,0x103f,0x103f,0x10ca,0x1040,0x1040,0x10cc,0x1040,0x1040,0x106e,0x0a1041,0x1041,0x106f,0x0a1042,0x1042,0x1070,0x0a1043,0x1043,0x1070,0x0a1043,0x1043,0x106f,0x0a1043,0x1043,0x1076,0x0a1043,0x1043,0x1077,0x0a1043,0x1043,0x1077,0x0a1043,0x1043)
GDI32(NtGdiGetRgnBox                                    ,  NONE,0x10c4,0x10c3,0x1066,0x1066,0x10cb,0x1067,0x1067,0x10cd,0x1067,0x1067,0x106d,0x051067,0x1067,0x106e,0x051068,0x1068,0x106f,0x051069,0x1069,0x106f,0x051069,0x1069,0x106e,0x051069,0x1069,0x1075,0x051068,0x1068,0x1076,0x051068,0x1068,0x1076,0x051068,0x1068)
GDI32(NtGdiGetServerMetaFileBits                        ,  NONE,0x10c5,0x10c4,0x11d5,0x11d5,0x10cc,0x1222,0x1222,0x10ce,0x122c,0x122c,0x106c,0x001279,0x1279,0x106d,0x001292,0x1292,0x106e,0x0012cb,0x12cb,0x106e,0x0012cf,0x12cf,0x106d,0x0012c6,0x12c6,0x1074,0x001294,0x1294,0x1075,0x0012ae,0x12ae,0x1075,0x0012c8,0x12c8)
GDI32(NtGdiGetSpoolMessage                              ,0x10c2,0x10c6,0x10c5,0x11d6,0x11d6,0x10cd,0x1223,0x1223,0x10cf,0x122d,0x122d,0x106b,0x00127a,0x127a,0x106c,0x001293,0x1293,0x106d,0x0012cc,0x12cc,0x106d,0x0012d0,0x12d0,0x106c,0x0012c7,0x12c7,0x1073,0x001295,0x1295,0x1074,0x0012af,0x12af,0x1074,0x0012c9,0x12c9)
GDI32(NtGdiGetStats                                     ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x1224,  NONE,  NONE,0x122e,  NONE,    NONE,0x127b,  NONE,    NONE,0x1294,0x106c,0x0012cd,0x12cd,0x106c,0x0012d1,0x12d1,0x106b,0x0012c8,0x12c8,0x1072,0x001296,0x1296,0x1073,0x0012b0,0x12b0,0x1073,0x0012ca,0x12ca)
GDI32(NtGdiGetStockObject                               ,  NONE,0x10c8,0x10c7,0x10d4,0x10d4,0x10cf,0x10d5,0x10d5,0x10d1,0x10d0,0x10d0,0x1069,0x0310d0,0x10d0,0x106a,0x0310d1,0x10d1,0x106b,0x0310d2,0x10d2,0x106b,0x0310d2,0x10d2,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
GDI32(NtGdiGetStringBitmapW                             ,  NONE,0x10c9,0x10c8,0x11d8,0x11d8,0x10d0,0x1225,0x1225,0x10d2,0x122f,0x122f,0x1068,0x00127c,0x127c,0x1069,0x001295,0x1295,0x106a,0x0012ce,0x12ce,0x106a,0x0012d2,0x12d2,0x106a,0x0012c9,0x12c9,0x1071,0x001297,0x1297,0x1072,0x0012b1,0x12b1,0x1072,0x0012cb,0x12cb)
GDI32(NtGdiGetSuggestedOPMProtectedOutputArraySize      ,  NONE,  NONE,  NONE,  NONE,  NONE,0x10d1,0x1226,0x1226,0x10d3,0x1230,0x1230,0x1067,0x00127d,0x127d,0x1068,0x001296,0x1296,0x1069,0x0012cf,0x12cf,0x1069,0x0012d3,0x12d3,0x1069,0x0012ca,0x12ca,0x1070,0x001298,0x1298,0x1071,0x0012b2,0x12b2,0x1071,0x0012cc,0x12cc)
GDI32(NtGdiGetSystemPaletteUse                          ,  NONE,0x10ca,0x10c9,0x1118,0x1118,0x10d2,0x1119,0x1119,0x10d4,0x1112,0x1112,0x1066,0x031112,0x1112,0x1067,0x031113,0x1113,0x1068,0x031114,0x1114,0x1068,0x031114,0x1114,0x1068,0x031113,0x1113,0x106f,0x031108,0x1108,0x1070,0x031108,0x1108,0x1070,0x031108,0x1108)
GDI32(NtGdiGetTextCharsetInfo                           ,0x10c7,0x10cb,0x10ca,0x104c,0x104c,0x10d3,0x104d,0x104d,0x10d5,0x104d,0x104d,0x1065,0x0c104e,0x104e,0x1066,0x0c104f,0x104f,0x1067,0x0c1050,0x1050,0x1067,0x0c1050,0x1050,0x1067,0x0c1050,0x1050,0x106e,0x0c1050,0x1050,0x106f,0x0c1050,0x1050,0x106f,0x0c1050,0x1050)
GDI32(NtGdiGetTextExtent                                ,  NONE,0x10cc,0x10cb,0x1095,0x1095,0x10d4,0x1096,0x1096,0x10d6,0x1094,0x1094,0x1064,0x001094,0x1094,0x1065,0x001095,0x1095,0x1066,0x001096,0x1096,0x1066,0x001096,0x1096,0x1066,0x001096,0x1096,0x106d,0x001094,0x1094,0x106e,0x001094,0x1094,0x106e,0x001094,0x1094)
GDI32(NtGdiGetTextExtentExW                             ,  NONE,0x10cd,0x10cc,0x11d9,0x11d9,0x10d5,0x1227,0x1227,0x10d7,0x1231,0x1231,0x1063,0x00127e,0x127e,0x1064,0x001297,0x1297,0x1065,0x0012d0,0x12d0,0x1065,0x0012d4,0x12d4,0x1065,0x0012cb,0x12cb,0x106c,0x001299,0x1299,0x106d,0x0012b3,0x12b3,0x106d,0x0012cd,0x12cd)
GDI32(NtGdiGetTextFaceW                                 ,  NONE,0x10ce,0x10cd,0x1081,0x1081,0x10d6,0x1082,0x1082,0x10d8,0x1080,0x1080,0x1062,0x141080,0x1080,0x1063,0x141081,0x1081,0x1064,0x141082,0x1082,0x1064,0x141082,0x1082,0x1064,0x141082,0x1082,0x106b,0x141080,0x1080,0x106c,0x141080,0x1080,0x106c,0x141080,0x1080)
GDI32(NtGdiGetTextMetricsW                              ,  NONE,0x10cf,0x10ce,0x1076,0x1076,0x10d7,0x1077,0x1077,0x10d9,0x1075,0x1075,0x1061,0x0c1075,0x1075,0x1062,0x0c1076,0x1076,0x1063,0x0c1077,0x1077,0x1063,0x0c1077,0x1077,0x1063,0x0c1077,0x1077,0x106a,0x0c1076,0x1076,0x106b,0x0c1076,0x1076,0x106b,0x0c1076,0x1076)
GDI32(NtGdiGetTransform                                 ,  NONE,0x10d0,0x10cf,0x10e0,0x10e0,0x10d8,0x10e1,0x10e1,0x10da,0x10dc,0x10dc,0x1060,0x0c10dc,0x10dc,0x1061,0x0c10dd,0x10dd,0x1062,0x0c10de,0x10de,0x1062,0x0c10de,0x10de,0x1062,0x0c10dd,0x10dd,0x1069,0x0c10d2,0x10d2,0x106a,0x0c10d2,0x10d2,0x106a,0x0c10d2,0x10d2)
GDI32(NtGdiGetUFI                                       ,  NONE,0x10d1,0x10d0,0x11da,0x11da,0x10d9,0x1228,0x1228,0x10db,0x1232,0x1232,0x105f,0x00127f,0x127f,0x1060,0x001298,0x1298,0x1061,0x0012d1,0x12d1,0x1061,0x0012d5,0x12d5,0x1061,0x0012cc,0x12cc,0x1068,0x00129a,0x129a,0x1069,0x0012b4,0x12b4,0x1069,0x0012ce,0x12ce)
GDI32(NtGdiGetUFIPathname                               ,  NONE,0x10d3,0x10d2,0x11db,0x11db,0x10db,0x1229,0x1229,0x10dd,0x1233,0x1233,0x105d,0x001280,0x1280,0x105e,0x001299,0x1299,0x105f,0x0012d2,0x12d2,0x105f,0x0012d6,0x12d6,0x105f,0x0012cd,0x12cd,0x1066,0x00129b,0x129b,0x1067,0x0012b5,0x12b5,0x1067,0x0012cf,0x12cf)
GDI32(NtGdiGetWidthTable                                ,  NONE,0x10d8,0x10d7,0x1069,0x1069,0x10e0,0x106a,0x106a,0x10e2,0x106a,0x106a,0x1058,0x00106a,0x106a,0x1059,0x00106b,0x106b,0x105a,0x00106c,0x106c,0x105a,0x00106c,0x106c,0x105a,0x00106c,0x106c,0x1061,0x00106b,0x106b,0x1062,0x00106b,0x106b,0x1062,0x00106b,0x106b)
GDI32(NtGdiGradientFill                                 ,  NONE,0x10d9,0x10d8,0x11dc,0x11dc,0x10e1,0x122a,0x122a,0x10e3,0x1234,0x1234,0x1057,0x001281,0x1281,0x1058,0x00129a,0x129a,0x1059,0x0012d3,0x12d3,0x1059,0x0012d7,0x12d7,0x1059,0x0012ce,0x12ce,0x1060,0x00129c,0x129c,0x1061,0x0012b6,0x12b6,0x1061,0x0012d0,0x12d0)
GDI32(NtGdiHLSurfGetInformation                         ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x12dd,0x1235,0x1235,0x1306,0x001282,0x1282,0x130f,0x00129b,0x129b,0x131b,0x0012d4,0x12d4,0x131c,0x0012d8,0x12d8,0x1322,0x0012cf,0x12cf,0x12e3,0x00129d,0x129d,0x12ec,0x0012b7,0x12b7,0x12f6,0x0012d1,0x12d1)
GDI32(NtGdiHLSurfSetInformation                         ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x12de,0x1236,0x1236,0x1305,0x001283,0x1283,0x130e,0x00129c,0x129c,0x131a,0x0012d5,0x12d5,0x131b,0x0012d9,0x12d9,0x1321,0x0012d0,0x12d0,0x12e2,0x00129e,0x129e,0x12eb,0x0012b8,0x12b8,0x12f5,0x0012d2,0x12d2)
GDI32(NtGdiHT_Get8BPPFormatPalette                      ,0x127c,0x1294,0x1290,0x11dd,0x11dd,0x12b5,0x122b,0x122b,0x12d5,0x1237,0x1237,0x12fe,0x001284,0x1284,0x1307,0x00129d,0x129d,0x1313,0x0012d6,0x12d6,0x1314,0x0012da,0x12da,0x131a,0x0012d1,0x12d1,0x12db,0x00129f,0x129f,0x12e4,0x0012b9,0x12b9,0x12ee,0x0012d3,0x12d3)
GDI32(NtGdiHT_Get8BPPMaskPalette                        ,0x127d,0x1295,0x1291,0x11de,0x11de,0x12b6,0x122c,0x122c,0x12d6,0x1238,0x1238,0x12fd,0x001285,0x1285,0x1306,0x00129e,0x129e,0x1312,0x0012d7,0x12d7,0x1313,0x0012db,0x12db,0x1319,0x0012d2,0x12d2,0x12da,0x0012a0,0x12a0,0x12e3,0x0012ba,0x12ba,0x12ed,0x0012d4,0x12d4)
GDI32(NtGdiHfontCreate                                  ,  NONE,0x10da,0x10d9,0x105c,0x105c,0x10e2,0x105d,0x105d,0x10e4,0x105d,0x105d,0x1056,0x00105d,0x105d,0x1057,0x00105e,0x105e,0x1058,0x00105f,0x105f,0x1058,0x00105f,0x105f,0x1058,0x00105f,0x105f,0x105f,0x00105f,0x105f,0x1060,0x00105f,0x105f,0x1060,0x00105f,0x105f)
GDI32(NtGdiIcmBrushInfo                                 ,  NONE,0x10db,0x10da,0x11df,0x11df,0x10e3,0x122d,0x122d,0x10e5,0x1239,0x1239,0x1055,0x001286,0x1286,0x1056,0x00129f,0x129f,0x1057,0x0012d8,0x12d8,0x1057,0x0012dc,0x12dc,0x1057,0x0012d3,0x12d3,0x105e,0x0012a1,0x12a1,0x105f,0x0012bb,0x12bb,0x105f,0x0012d5,0x12d5)
GDI32(NtGdiInit                                         ,  NONE,0x10dc,0x10db,0x11e0,0x11e0,0x10e4,0x122e,0x122e,0x10e6,0x123a,0x123a,0x1054,0x011287,0x1287,0x1055,0x0112a0,0x12a0,0x1056,0x0112d9,0x12d9,0x1056,0x0112dd,0x12dd,0x1056,0x0112d4,0x12d4,0x105d,0x0112a2,0x12a2,0x105e,0x0112bc,0x12bc,0x105e,0x0112d6,0x12d6)
GDI32(NtGdiInitSpool                                    ,0x10d5,0x10dd,0x10dc,0x11e1,0x11e1,0x10e5,0x122f,0x122f,0x10e7,0x123b,0x123b,0x1053,0x011288,0x1288,0x1054,0x0112a1,0x12a1,0x1055,0x0112da,0x12da,0x1055,0x0112de,0x12de,0x1055,0x0112d5,0x12d5,0x105c,0x0112a3,0x12a3,0x105d,0x0112bd,0x12bd,0x105d,0x0112d7,0x12d7)
GDI32(NtGdiIntersectClipRect                            ,  NONE,0x10de,0x10dd,0x101f,0x101f,0x10e6,0x1020,0x1020,0x10e8,0x1020,0x1020,0x1052,0x001021,0x1021,0x1053,0x001022,0x1022,0x1054,0x001023,0x1023,0x1054,0x001023,0x1023,0x1054,0x001023,0x1023,0x105b,0x001023,0x1023,0x105c,0x001023,0x1023,0x105c,0x001023,0x1023)
GDI32(NtGdiInvertRgn                                    ,  NONE,0x10df,0x10de,0x1065,0x1065,0x10e7,0x1066,0x1066,0x10e9,0x1066,0x1066,0x1051,0x071066,0x1066,0x1052,0x071067,0x1067,0x1053,0x071068,0x1068,0x1053,0x071068,0x1068,0x1053,0x071068,0x1068,0x105a,0x071067,0x1067,0x105b,0x071067,0x1067,0x105b,0x071067,0x1067)
GDI32(NtGdiLineTo                                       ,  NONE,0x10e0,0x10df,0x1040,0x1040,0x10e8,0x1041,0x1041,0x10ea,0x1041,0x1041,0x1050,0x0c1042,0x1042,0x1051,0x0c1043,0x1043,0x1052,0x0c1044,0x1044,0x1052,0x0c1044,0x1044,0x1052,0x0c1044,0x1044,0x1059,0x0c1044,0x1044,0x105a,0x0c1044,0x1044,0x105a,0x0c1044,0x1044)
GDI32(NtGdiMakeFontDir                                  ,  NONE,0x10e1,0x10e0,0x11e2,0x11e2,0x10e9,0x1230,0x1230,0x10eb,0x123c,0x123c,0x104f,0x001289,0x1289,0x1050,0x0012a2,0x12a2,0x1051,0x0012db,0x12db,0x1051,0x0012df,0x12df,0x1051,0x0012d6,0x12d6,0x1058,0x0012a4,0x12a4,0x1059,0x0012be,0x12be,0x1059,0x0012d8,0x12d8)
GDI32(NtGdiMakeInfoDC                                   ,  NONE,0x10e2,0x10e1,0x11e3,0x11e3,0x10ea,0x1231,0x1231,0x10ec,0x123d,0x123d,0x104e,0x07128a,0x128a,0x104f,0x0712a3,0x12a3,0x1050,0x0712dc,0x12dc,0x1050,0x0712e0,0x12e0,0x1050,0x0712d7,0x12d7,0x1057,0x0712a5,0x12a5,0x1058,0x0712bf,0x12bf,0x1058,0x0712d9,0x12d9)
GDI32(NtGdiMakeObjectUnXferable                         ,  NONE,  NONE,  NONE,0x11e4,0x11e4,  NONE,0x1232,0x1232,  NONE,0x123e,0x123e,0x1368,0x00128b,0x128b,0x137a,0x0012a4,0x12a4,0x13b3,0x0012dd,0x12dd,0x13b5,0x0012e1,0x12e1,0x13c2,0x0012d8,0x12d8,0x1393,0x0012a6,0x12a6,0x13ad,0x0012c0,0x12c0,0x13bf,0x0012da,0x12da)
GDI32(NtGdiMakeObjectXferable                           ,  NONE,  NONE,  NONE,0x11e5,0x11e5,  NONE,0x1233,0x1233,  NONE,0x123f,0x123f,0x1369,0x00128c,0x128c,0x137b,0x0012a5,0x12a5,0x13b4,0x0012de,0x12de,0x13b6,0x0012e2,0x12e2,0x13c3,0x0012d9,0x12d9,0x1394,0x0012a7,0x12a7,0x13ae,0x0012c1,0x12c1,0x13c0,0x0012db,0x12db)
GDI32(NtGdiMaskBlt                                      ,  NONE,0x10e3,0x10e2,0x1068,0x1068,0x10eb,0x1069,0x1069,0x10ed,0x1069,0x1069,0x104d,0x001069,0x1069,0x104e,0x00106a,0x106a,0x104f,0x00106b,0x106b,0x104f,0x00106b,0x106b,0x104f,0x00106b,0x106b,0x1056,0x00106a,0x106a,0x1057,0x00106a,0x106a,0x1057,0x00106a,0x106a)
GDI32(NtGdiMirrorWindowOrg                              ,  NONE,0x1118,0x1117,0x11e6,0x11e6,0x1120,0x1234,0x1234,0x1122,0x1240,0x1240,0x1018,0x03128d,0x128d,0x1019,0x0312a6,0x12a6,0x101a,0x0312df,0x12df,0x101a,0x0312e3,0x12e3,0x101a,0x0312da,0x12da,0x101e,0x0312a8,0x12a8,0x101f,0x0312c2,0x12c2,0x101f,0x0312dc,0x12dc)
GDI32(NtGdiModifyWorldTransform                         ,  NONE,0x10e4,0x10e3,0x10da,0x10da,0x10ec,0x10db,0x10db,0x10ee,0x10d6,0x10d6,0x104c,0x0c10d6,0x10d6,0x104d,0x0c10d7,0x10d7,0x104e,0x0c10d8,0x10d8,0x104e,0x0c10d8,0x10d8,0x104e,0x0c10d7,0x10d7,0x1055,0x0c10cc,0x10cc,0x1056,0x0c10cc,0x10cc,0x1056,0x0c10cc,0x10cc)
GDI32(NtGdiMonoBitmap                                   ,  NONE,0x10e5,0x10e4,0x11e7,0x11e7,0x10ed,0x1235,0x1235,0x10ef,0x1241,0x1241,0x104b,0x04128e,0x128e,0x104c,0x0412a7,0x12a7,0x104d,0x0412e0,0x12e0,0x104d,0x0412e4,0x12e4,0x104d,0x0412db,0x12db,0x1054,0x0412a9,0x12a9,0x1055,0x0412c3,0x12c3,0x1055,0x0412dd,0x12dd)
GDI32(NtGdiMoveTo                                       ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x1236,  NONE,  NONE,0x1242,  NONE,    NONE,0x128f,  NONE,    NONE,0x12a8,0x104c,0x1412e1,0x12e1,0x104c,0x1412e5,0x12e5,0x104c,0x1412dc,0x12dc,0x1053,0x1412aa,0x12aa,0x1054,0x1412c4,0x12c4,0x1054,0x1412de,0x12de)
GDI32(NtGdiOffsetClipRgn                                ,  NONE,0x10e7,0x10e6,0x11e9,0x11e9,0x10ef,0x1237,0x1237,0x10f1,0x1243,0x1243,0x1049,0x0c1290,0x1290,0x104a,0x0c12a9,0x12a9,0x104b,0x0c12e2,0x12e2,0x104b,0x0c12e6,0x12e6,0x104b,0x0c12dd,0x12dd,0x1052,0x0c12ab,0x12ab,0x1053,0x0c12c5,0x12c5,0x1053,0x0c12df,0x12df)
GDI32(NtGdiOffsetRgn                                    ,  NONE,0x10e8,0x10e7,0x107f,0x107f,0x10f0,0x1080,0x1080,0x10f2,0x107e,0x107e,0x1048,0x0a107e,0x107e,0x1049,0x0a107f,0x107f,0x104a,0x0a1080,0x1080,0x104a,0x0a1080,0x1080,0x104a,0x0a1080,0x1080,0x1051,0x0a107e,0x107e,0x1052,0x0a107e,0x107e,0x1052,0x0a107e,0x107e)
GDI32(NtGdiOpenDCW                                      ,  NONE,0x10e9,0x10e8,0x10de,0x10de,0x10f1,0x10df,0x10df,0x10f3,0x10da,0x10da,0x1047,0x0010da,0x10da,0x1048,0x0010db,0x10db,0x1049,0x0010dc,0x10dc,0x1049,0x0010dc,0x10dc,0x1049,0x0010db,0x10db,0x1050,0x0010d0,0x10d0,0x1051,0x0010d0,0x10d0,0x1051,0x0010d0,0x10d0)
GDI32(NtGdiPATHOBJ_bEnum                                ,0x1276,0x128e,0x128a,0x11ea,0x11ea,0x12af,0x1238,0x1238,0x12cf,0x1244,0x1244,0x12f9,0x001291,0x1291,0x1302,0x0012aa,0x12aa,0x130e,0x0012e3,0x12e3,0x130f,0x0012e7,0x12e7,0x1315,0x0012de,0x12de,0x12d6,0x0012ac,0x12ac,0x12df,0x0012c6,0x12c6,0x12e9,0x0012e0,0x12e0)
GDI32(NtGdiPATHOBJ_bEnumClipLines                       ,0x1279,0x1291,0x128d,0x11eb,0x11eb,0x12b2,0x1239,0x1239,0x12d2,0x1245,0x1245,0x12f6,0x001292,0x1292,0x12ff,0x0012ab,0x12ab,0x130b,0x0012e4,0x12e4,0x130c,0x0012e8,0x12e8,0x1312,0x0012df,0x12df,0x12d3,0x0012ad,0x12ad,0x12dc,0x0012c7,0x12c7,0x12e6,0x0012e1,0x12e1)
GDI32(NtGdiPATHOBJ_vEnumStart                           ,0x1277,0x128f,0x128b,0x11ec,0x11ec,0x12b0,0x123a,0x123a,0x12d0,0x1246,0x1246,0x12f8,0x001293,0x1293,0x1301,0x0012ac,0x12ac,0x130d,0x0012e5,0x12e5,0x130e,0x0012e9,0x12e9,0x1314,0x0012e0,0x12e0,0x12d5,0x0012ae,0x12ae,0x12de,0x0012c8,0x12c8,0x12e8,0x0012e2,0x12e2)
GDI32(NtGdiPATHOBJ_vEnumStartClipLines                  ,0x1278,0x1290,0x128c,0x11ed,0x11ed,0x12b1,0x123b,0x123b,0x12d1,0x1247,0x1247,0x12f7,0x001294,0x1294,0x1300,0x0012ad,0x12ad,0x130c,0x0012e6,0x12e6,0x130d,0x0012ea,0x12ea,0x1313,0x0012e1,0x12e1,0x12d4,0x0012af,0x12af,0x12dd,0x0012c9,0x12c9,0x12e7,0x0012e3,0x12e3)
GDI32(NtGdiPATHOBJ_vGetBounds                           ,0x1275,0x128d,0x1289,0x11ee,0x11ee,0x12ae,0x123c,0x123c,0x12ce,0x1248,0x1248,0x12fa,0x001295,0x1295,0x1303,0x0012ae,0x12ae,0x130f,0x0012e7,0x12e7,0x1310,0x0012eb,0x12eb,0x1316,0x0012e2,0x12e2,0x12d7,0x0012b0,0x12b0,0x12e0,0x0012ca,0x12ca,0x12ea,0x0012e4,0x12e4)
GDI32(NtGdiPatBlt                                       ,  NONE,0x10ea,0x10e9,0x1059,0x1059,0x10f2,0x105a,0x105a,0x10f4,0x105a,0x105a,0x1046,0x00105a,0x105a,0x1047,0x00105b,0x105b,0x1048,0x00105c,0x105c,0x1048,0x00105c,0x105c,0x1048,0x00105c,0x105c,0x104f,0x00105c,0x105c,0x1050,0x00105c,0x105c,0x1050,0x00105c,0x105c)
GDI32(NtGdiPathToRegion                                 ,  NONE,0x10ec,0x10eb,0x11ef,0x11ef,0x10f4,0x123d,0x123d,0x10f6,0x1249,0x1249,0x1044,0x031296,0x1296,0x1045,0x0312af,0x12af,0x1046,0x0312e8,0x12e8,0x1046,0x0312ec,0x12ec,0x1046,0x0312e3,0x12e3,0x104d,0x0312b1,0x12b1,0x104e,0x0312cb,0x12cb,0x104e,0x0312e5,0x12e5)
GDI32(NtGdiPlgBlt                                       ,  NONE,0x10ed,0x10ec,0x11f0,0x11f0,0x10f5,0x123e,0x123e,0x10f7,0x124a,0x124a,0x1043,0x001297,0x1297,0x1044,0x0012b0,0x12b0,0x1045,0x0012e9,0x12e9,0x1045,0x0012ed,0x12ed,0x1045,0x0012e4,0x12e4,0x104c,0x0012b2,0x12b2,0x104d,0x0012cc,0x12cc,0x104d,0x0012e6,0x12e6)
GDI32(NtGdiPolyDraw                                     ,  NONE,0x10ee,0x10ed,0x11f1,0x11f1,0x10f6,0x123f,0x123f,0x10f8,0x124b,0x124b,0x1042,0x141298,0x1298,0x1043,0x1412b1,0x12b1,0x1044,0x1412ea,0x12ea,0x1044,0x1412ee,0x12ee,0x1044,0x1412e5,0x12e5,0x104b,0x1412b3,0x12b3,0x104c,0x1412cd,0x12cd,0x104c,0x1412e7,0x12e7)
GDI32(NtGdiPolyPatBlt                                   ,  NONE,0x10eb,0x10ea,0x106f,0x106f,0x10f3,0x1070,0x1070,0x10f5,0x106f,0x106f,0x1045,0x00106f,0x106f,0x1046,0x001070,0x1070,0x1047,0x001071,0x1071,0x1047,0x001071,0x1071,0x1047,0x001071,0x1071,0x104e,0x001070,0x1070,0x104f,0x001070,0x1070,0x104f,0x001070,0x1070)
GDI32(NtGdiPolyPolyDraw                                 ,  NONE,0x10ef,0x10ee,0x1047,0x1047,0x10f7,0x1048,0x1048,0x10f9,0x1048,0x1048,0x1041,0x001049,0x1049,0x1042,0x00104a,0x104a,0x1043,0x00104b,0x104b,0x1043,0x00104b,0x104b,0x1043,0x00104b,0x104b,0x104a,0x00104b,0x104b,0x104b,0x00104b,0x104b,0x104b,0x00104b,0x104b)
GDI32(NtGdiPolyTextOutW                                 ,  NONE,0x10f0,0x10ef,0x11f2,0x11f2,0x10f8,0x1240,0x1240,0x10fa,0x124c,0x124c,0x1040,0x001299,0x1299,0x1041,0x0012b2,0x12b2,0x1042,0x0012eb,0x12eb,0x1042,0x0012ef,0x12ef,0x1042,0x0012e6,0x12e6,0x1049,0x0012b4,0x12b4,0x104a,0x0012ce,0x12ce,0x104a,0x0012e8,0x12e8)
GDI32(NtGdiPtInRegion                                   ,  NONE,0x10f1,0x10f0,0x11f3,0x11f3,0x10f9,0x1241,0x1241,0x10fb,0x124d,0x124d,0x103f,0x0a129a,0x129a,0x1040,0x0a12b3,0x12b3,0x1041,0x0a12ec,0x12ec,0x1041,0x0a12f0,0x12f0,0x1041,0x0a12e7,0x12e7,0x1048,0x0a12b5,0x12b5,0x1049,0x0a12cf,0x12cf,0x1049,0x0a12e9,0x12e9)
GDI32(NtGdiPtVisible                                    ,  NONE,0x10f2,0x10f1,0x11f4,0x11f4,0x10fa,0x1242,0x1242,0x10fc,0x124e,0x124e,0x103e,0x0c129b,0x129b,0x103f,0x0c12b4,0x12b4,0x1040,0x0c12ed,0x12ed,0x1040,0x0c12f1,0x12f1,0x1040,0x0c12e8,0x12e8,0x1047,0x0c12b6,0x12b6,0x1048,0x0c12d0,0x12d0,0x1048,0x0c12ea,0x12ea)
GDI32(NtGdiQueryFontAssocInfo                           ,  NONE,0x10f4,0x10f3,0x10f9,0x10f9,0x10fc,0x10fa,0x10fa,0x10fe,0x10f5,0x10f5,0x103c,0x0310f5,0x10f5,0x103d,0x0310f6,0x10f6,0x103e,0x0310f7,0x10f7,0x103e,0x0310f7,0x10f7,0x103e,0x0310f6,0x10f6,0x1045,0x0310eb,0x10eb,0x1046,0x0310eb,0x10eb,0x1046,0x0310eb,0x10eb)
GDI32(NtGdiQueryFonts                                   ,0x10eb,0x10f3,0x10f2,0x11f5,0x11f5,0x10fb,0x1243,0x1243,0x10fd,0x124f,0x124f,0x103d,0x0a129c,0x129c,0x103e,0x0a12b5,0x12b5,0x103f,0x0a12ee,0x12ee,0x103f,0x0a12f2,0x12f2,0x103f,0x0a12e9,0x12e9,0x1046,0x0a12b7,0x12b7,0x1047,0x0a12d1,0x12d1,0x1047,0x0a12eb,0x12eb)
GDI32(NtGdiRectInRegion                                 ,  NONE,0x10f6,0x10f5,0x10c1,0x10c1,0x10fe,0x10c2,0x10c2,0x1100,0x10bd,0x10bd,0x103a,0x0510bd,0x10bd,0x103b,0x0510be,0x10be,0x103c,0x0510bf,0x10bf,0x103c,0x0510bf,0x10bf,0x103c,0x0510bf,0x10bf,0x1043,0x0510b8,0x10b8,0x1044,0x0510b8,0x10b8,0x1044,0x0510b8,0x10b8)
GDI32(NtGdiRectVisible                                  ,  NONE,0x10f7,0x10f6,0x1032,0x1032,0x10ff,0x1033,0x1033,0x1101,0x1033,0x1033,0x1039,0x071034,0x1034,0x103a,0x071035,0x1035,0x103b,0x071036,0x1036,0x103b,0x071036,0x1036,0x103b,0x071036,0x1036,0x1042,0x071036,0x1036,0x1043,0x071036,0x1036,0x1043,0x071036,0x1036)
GDI32(NtGdiRectangle                                    ,  NONE,0x10f5,0x10f4,0x1091,0x1091,0x10fd,0x1092,0x1092,0x10ff,0x1090,0x1090,0x103b,0x001090,0x1090,0x103c,0x001091,0x1091,0x103d,0x001092,0x1092,0x103d,0x001092,0x1092,0x103d,0x001092,0x1092,0x1044,0x001090,0x1090,0x1045,0x001090,0x1090,0x1045,0x001090,0x1090)
GDI32(NtGdiRemoveFontMemResourceEx                      ,  NONE,0x10f9,0x10f8,0x1119,0x1119,0x1101,0x111a,0x111a,0x1103,0x1113,0x1113,0x1037,0x031113,0x1113,0x1038,0x031114,0x1114,0x1039,0x031115,0x1115,0x1039,0x031115,0x1115,0x1039,0x031114,0x1114,0x1040,0x031109,0x1109,0x1041,0x031109,0x1109,0x1041,0x031109,0x1109)
GDI32(NtGdiRemoveFontResourceW                          ,  NONE,0x10f8,0x10f7,0x11f6,0x11f6,0x1100,0x1244,0x1244,0x1102,0x1250,0x1250,0x1038,0x00129d,0x129d,0x1039,0x0012b6,0x12b6,0x103a,0x0012ef,0x12ef,0x103a,0x0012f3,0x12f3,0x103a,0x0012ea,0x12ea,0x1041,0x0012b8,0x12b8,0x1042,0x0012d2,0x12d2,0x1042,0x0012ec,0x12ec)
GDI32(NtGdiRemoveMergeFont                              ,  NONE,0x1005,0x1005,0x11f7,0x11f7,0x1005,0x1245,0x1245,0x1005,0x1251,0x1251,0x1137,0x07129e,0x129e,0x1139,0x0712b7,0x12b7,0x113a,0x0712f0,0x12f0,0x113c,0x0712f4,0x12f4,0x113d,0x0712eb,0x12eb,0x10fa,0x0712b9,0x12b9,0x10fb,0x0712d3,0x12d3,0x10fb,0x0712ed,0x12ed)
GDI32(NtGdiResetDC                                      ,  NONE,0x10fa,0x10f9,0x11f8,0x11f8,0x1102,0x1246,0x1246,0x1104,0x1252,0x1252,0x1036,0x00129f,0x129f,0x1037,0x0012b8,0x12b8,0x1038,0x0012f1,0x12f1,0x1038,0x0012f5,0x12f5,0x1038,0x0012ec,0x12ec,0x103f,0x0012ba,0x12ba,0x1040,0x0012d4,0x12d4,0x1040,0x0012ee,0x12ee)
GDI32(NtGdiResizePalette                                ,  NONE,0x10fb,0x10fa,0x11f9,0x11f9,0x1103,0x1247,0x1247,0x1105,0x1253,0x1253,0x1035,0x0512a0,0x12a0,0x1036,0x0512b9,0x12b9,0x1037,0x0512f2,0x12f2,0x1037,0x0512f6,0x12f6,0x1037,0x0512ed,0x12ed,0x103e,0x0512bb,0x12bb,0x103f,0x0512d5,0x12d5,0x103f,0x0512ef,0x12ef)
GDI32(NtGdiRestoreDC                                    ,  NONE,0x10fc,0x10fb,0x1039,0x1039,0x1104,0x103a,0x103a,0x1106,0x103a,0x103a,0x1034,0x08103b,0x103b,0x1035,0x08103c,0x103c,0x1036,0x08103d,0x103d,0x1036,0x08103d,0x103d,0x1036,0x08103d,0x103d,0x103d,0x08103d,0x103d,0x103e,0x08103d,0x103d,0x103e,0x08103d,0x103d)
GDI32(NtGdiRoundRect                                    ,  NONE,0x10fd,0x10fc,0x11fa,0x11fa,0x1105,0x1248,0x1248,0x1107,0x1254,0x1254,0x1033,0x0012a1,0x12a1,0x1034,0x0012ba,0x12ba,0x1035,0x0012f3,0x12f3,0x1035,0x0012f7,0x12f7,0x1035,0x0012ee,0x12ee,0x103c,0x0012bc,0x12bc,0x103d,0x0012d6,0x12d6,0x103d,0x0012f0,0x12f0)
GDI32(NtGdiSTROBJ_bEnum                                 ,0x1270,0x1288,0x1284,0x11fb,0x11fb,0x12a9,0x1249,0x1249,0x12c9,0x1255,0x1255,0x12f5,0x0012a2,0x12a2,0x12fe,0x0012bb,0x12bb,0x130a,0x0012f4,0x12f4,0x130b,0x0012f8,0x12f8,0x1311,0x0012ef,0x12ef,0x12d2,0x0012bd,0x12bd,0x12db,0x0012d7,0x12d7,0x12e5,0x0012f1,0x12f1)
GDI32(NtGdiSTROBJ_bEnumPositionsOnly                    ,0x1271,0x1289,0x1285,0x11fc,0x11fc,0x12aa,0x124a,0x124a,0x12ca,0x1256,0x1256,0x12f4,0x0012a3,0x12a3,0x12fd,0x0012bc,0x12bc,0x1309,0x0012f5,0x12f5,0x130a,0x0012f9,0x12f9,0x1310,0x0012f0,0x12f0,0x12d1,0x0012be,0x12be,0x12da,0x0012d8,0x12d8,0x12e4,0x0012f2,0x12f2)
GDI32(NtGdiSTROBJ_bGetAdvanceWidths                     ,0x1272,0x128a,0x1286,0x11fd,0x11fd,0x12ab,0x124b,0x124b,0x12cb,0x1257,0x1257,0x12f3,0x0012a4,0x12a4,0x12fc,0x0012bd,0x12bd,0x1308,0x0012f6,0x12f6,0x1309,0x0012fa,0x12fa,0x130f,0x0012f1,0x12f1,0x12d0,0x0012bf,0x12bf,0x12d9,0x0012d9,0x12d9,0x12e3,0x0012f3,0x12f3)
GDI32(NtGdiSTROBJ_dwGetCodePage                         ,0x1274,0x128c,0x1288,0x11fe,0x11fe,0x12ad,0x124c,0x124c,0x12cd,0x1258,0x1258,0x12f1,0x0012a5,0x12a5,0x12fa,0x0012be,0x12be,0x1306,0x0012f7,0x12f7,0x1307,0x0012fb,0x12fb,0x130d,0x0012f2,0x12f2,0x12ce,0x0012c0,0x12c0,0x12d7,0x0012da,0x12da,0x12e1,0x0012f4,0x12f4)
GDI32(NtGdiSTROBJ_vEnumStart                            ,0x1273,0x128b,0x1287,0x11ff,0x11ff,0x12ac,0x124d,0x124d,0x12cc,0x1259,0x1259,0x12f2,0x0012a6,0x12a6,0x12fb,0x0012bf,0x12bf,0x1307,0x0012f8,0x12f8,0x1308,0x0012fc,0x12fc,0x130e,0x0012f3,0x12f3,0x12cf,0x0012c1,0x12c1,0x12d8,0x0012db,0x12db,0x12e2,0x0012f5,0x12f5)
GDI32(NtGdiSaveDC                                       ,  NONE,0x10fe,0x10fd,0x103a,0x103a,0x1106,0x103b,0x103b,0x1108,0x103b,0x103b,0x1032,0x03103c,0x103c,0x1033,0x03103d,0x103d,0x1034,0x03103e,0x103e,0x1034,0x03103e,0x103e,0x1034,0x03103e,0x103e,0x103b,0x03103e,0x103e,0x103c,0x03103e,0x103e,0x103c,0x03103e,0x103e)
GDI32(NtGdiScaleRgn                                     ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x100a,0x0012c2,0x12c2,0x100a,0x0012dc,0x12dc,0x100a,0x0012f6,0x12f6)
GDI32(NtGdiScaleValues                                  ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x100b,0x0012c3,0x12c3,0x100b,0x0012dd,0x12dd,0x100b,0x0012f7,0x12f7)
GDI32(NtGdiScaleViewportExtEx                           ,  NONE,0x10ff,0x10fe,0x1200,0x1200,0x1107,0x124e,0x124e,0x1109,0x125a,0x125a,0x1031,0x0012a7,0x12a7,0x1032,0x0012c0,0x12c0,0x1033,0x0012f9,0x12f9,0x1033,0x0012fd,0x12fd,0x1033,0x0012f4,0x12f4,0x103a,0x0012c4,0x12c4,0x103b,0x0012de,0x12de,0x103b,0x0012f8,0x12f8)
GDI32(NtGdiScaleWindowExtEx                             ,  NONE,0x1100,0x10ff,0x1201,0x1201,0x1108,0x124f,0x124f,0x110a,0x125b,0x125b,0x1030,0x0012a8,0x12a8,0x1031,0x0012c1,0x12c1,0x1032,0x0012fa,0x12fa,0x1032,0x0012fe,0x12fe,0x1032,0x0012f5,0x12f5,0x1039,0x0012c5,0x12c5,0x103a,0x0012df,0x12df,0x103a,0x0012f9,0x12f9)
GDI32(NtGdiSelectBitmap                                 ,  NONE,0x1101,0x1100,0x100b,0x100b,0x1109,0x100b,0x100b,0x110b,0x100b,0x100b,0x102f,0x07100c,0x100c,0x1030,0x07100d,0x100d,0x1031,0x07100e,0x100e,0x1031,0x07100e,0x100e,0x1031,0x07100e,0x100e,0x1038,0x07100e,0x100e,0x1039,0x07100e,0x100e,0x1039,0x07100e,0x100e)
GDI32(NtGdiSelectBrush                                  ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x1250,  NONE,  NONE,0x125c,  NONE,    NONE,0x12a9,  NONE,    NONE,0x12c2,0x1030,0x0712fb,0x12fb,0x1030,0x0712ff,0x12ff,0x1030,0x0712f6,0x12f6,0x1037,0x0712c6,0x12c6,0x1038,0x0712e0,0x12e0,0x1038,0x0712fa,0x12fa)
GDI32(NtGdiSelectClipPath                               ,  NONE,0x1103,0x1102,0x1203,0x1203,0x110b,0x1251,0x1251,0x110d,0x125d,0x125d,0x102d,0x0812aa,0x12aa,0x102e,0x0812c3,0x12c3,0x102f,0x0812fc,0x12fc,0x102f,0x081300,0x1300,0x102f,0x0812f7,0x12f7,0x1036,0x0812c7,0x12c7,0x1037,0x0812e1,0x12e1,0x1037,0x0812fb,0x12fb)
GDI32(NtGdiSelectFont                                   ,  NONE,0x1104,0x1103,0x1038,0x1038,0x110c,0x1039,0x1039,0x110e,0x1039,0x1039,0x102c,0x07103a,0x103a,0x102d,0x07103b,0x103b,0x102e,0x07103c,0x103c,0x102e,0x07103c,0x103c,0x102e,0x07103c,0x103c,0x1035,0x07103c,0x103c,0x1036,0x07103c,0x103c,0x1036,0x07103c,0x103c)
GDI32(NtGdiSelectPen                                    ,  NONE,0x1105,0x1104,0x1204,0x1204,0x110d,0x1252,0x1252,0x110f,0x125e,0x125e,0x102b,0x0712ab,0x12ab,0x102c,0x0712c4,0x12c4,0x102d,0x0712fd,0x12fd,0x102d,0x071301,0x1301,0x102d,0x0712f8,0x12f8,0x1034,0x0712c8,0x12c8,0x1035,0x0712e2,0x12e2,0x1035,0x0712fc,0x12fc)
GDI32(NtGdiSetBitmapAttributes                          ,  NONE,0x1106,0x1105,0x1205,0x1205,0x110e,0x1253,0x1253,0x1110,0x125f,0x125f,0x102a,0x0512ac,0x12ac,0x102b,0x0512c5,0x12c5,0x102c,0x0512fe,0x12fe,0x102c,0x051302,0x1302,0x102c,0x0512f9,0x12f9,0x1033,0x0512c9,0x12c9,0x1034,0x0512e3,0x12e3,0x1034,0x0512fd,0x12fd)
GDI32(NtGdiSetBitmapBits                                ,  NONE,0x1107,0x1106,0x10b8,0x10b8,0x110f,0x10b9,0x10b9,0x1111,0x10b4,0x10b4,0x1029,0x0a10b4,0x10b4,0x102a,0x0a10b5,0x10b5,0x102b,0x0a10b6,0x10b6,0x102b,0x0a10b6,0x10b6,0x102b,0x0a10b6,0x10b6,0x1032,0x0a10af,0x10af,0x1033,0x0a10af,0x10af,0x1033,0x0a10af,0x10af)
GDI32(NtGdiSetBitmapDimension                           ,  NONE,0x1108,0x1107,0x111d,0x111d,0x1110,0x111e,0x111e,0x1112,0x1117,0x1117,0x1028,0x111117,0x1117,0x1029,0x111118,0x1118,0x102a,0x111119,0x1119,0x102a,0x111119,0x1119,0x102a,0x111118,0x1118,0x1031,0x11110d,0x110d,0x1032,0x11110d,0x110d,0x1032,0x11110d,0x110d)
GDI32(NtGdiSetBoundsRect                                ,  NONE,0x1109,0x1108,0x10ff,0x10ff,0x1111,0x1100,0x1100,0x1113,0x10fb,0x10fb,0x1027,0x0c10fb,0x10fb,0x1028,0x0c10fc,0x10fc,0x1029,0x0c10fd,0x10fd,0x1029,0x0c10fd,0x10fd,0x1029,0x0c10fc,0x10fc,0x1030,0x0c10f1,0x10f1,0x1031,0x0c10f1,0x10f1,0x1031,0x0c10f1,0x10f1)
GDI32(NtGdiSetBrushAttributes                           ,  NONE,0x110a,0x1109,0x1206,0x1206,0x1112,0x1254,0x1254,0x1114,0x1260,0x1260,0x1026,0x0512ad,0x12ad,0x1027,0x0512c6,0x12c6,0x1028,0x0512ff,0x12ff,0x1028,0x051303,0x1303,0x1028,0x0512fa,0x12fa,0x102f,0x0512ca,0x12ca,0x1030,0x0512e4,0x12e4,0x1030,0x0512fe,0x12fe)
GDI32(NtGdiSetBrushOrg                                  ,  NONE,0x110b,0x110a,0x10b1,0x10b1,0x1113,0x10b2,0x10b2,0x1115,0x10ad,0x10ad,0x1025,0x1410ad,0x10ad,0x1026,0x1410ae,0x10ae,0x1027,0x1410af,0x10af,0x1027,0x1410af,0x10af,0x1027,0x1410af,0x10af,0x102e,0x1410a8,0x10a8,0x102f,0x1410a8,0x10a8,0x102f,0x1410a8,0x10a8)
GDI32(NtGdiSetColorAdjustment                           ,  NONE,0x110c,0x110b,0x1207,0x1207,0x1114,0x1255,0x1255,0x1116,0x1261,0x1261,0x1024,0x0712ae,0x12ae,0x1025,0x0712c7,0x12c7,0x1026,0x071300,0x1300,0x1026,0x071304,0x1304,0x1026,0x0712fb,0x12fb,0x102d,0x0712cb,0x12cb,0x102e,0x0712e5,0x12e5,0x102e,0x0712ff,0x12ff)
GDI32(NtGdiSetColorSpace                                ,  NONE,0x110d,0x110c,0x1208,0x1208,0x1115,0x1256,0x1256,0x1117,0x1262,0x1262,0x1023,0x0712af,0x12af,0x1024,0x0712c8,0x12c8,0x1025,0x071301,0x1301,0x1025,0x071305,0x1305,0x1025,0x0712fc,0x12fc,0x102c,0x0712cc,0x12cc,0x102d,0x0712e6,0x12e6,0x102d,0x071300,0x1300)
GDI32(NtGdiSetDIBitsToDeviceInternal                    ,  NONE,0x110f,0x110e,0x1028,0x1028,0x1117,0x1029,0x1029,0x1119,0x1029,0x1029,0x1021,0x00102a,0x102a,0x1022,0x00102b,0x102b,0x1023,0x00102c,0x102c,0x1023,0x00102c,0x102c,0x1023,0x00102c,0x102c,0x1027,0x00102c,0x102c,0x1028,0x00102c,0x102c,0x1028,0x00102c,0x102c)
GDI32(NtGdiSetDeviceGammaRamp                           ,  NONE,0x110e,0x110d,0x1209,0x1209,0x1116,0x1257,0x1257,0x1118,0x1263,0x1263,0x1022,0x0012b0,0x12b0,0x1023,0x0012c9,0x12c9,0x1024,0x001302,0x1302,0x1024,0x001306,0x1306,0x1024,0x0012fd,0x12fd,0x1028,0x0012cd,0x12cd,0x1029,0x0012e7,0x12e7,0x1029,0x001301,0x1301)
GDI32(NtGdiSetFontEnumeration                           ,  NONE,0x1110,0x110f,0x111e,0x111e,0x1118,0x111f,0x111f,0x111a,0x1118,0x1118,0x1020,0x041118,0x1118,0x1021,0x041119,0x1119,0x1022,0x04111a,0x111a,0x1022,0x04111a,0x111a,0x1022,0x041119,0x1119,0x1026,0x04110e,0x110e,0x1027,0x04110e,0x110e,0x1027,0x04110e,0x110e)
GDI32(NtGdiSetFontXform                                 ,  NONE,0x1111,0x1110,0x120a,0x120a,0x1119,0x1258,0x1258,0x111b,0x1264,0x1264,0x101f,0x0c12b1,0x12b1,0x1020,0x0c12ca,0x12ca,0x1021,0x0c1303,0x1303,0x1021,0x0c1307,0x1307,0x1021,0x0c12fe,0x12fe,0x1025,0x0c12ce,0x12ce,0x1026,0x0c12e8,0x12e8,0x1026,0x0c1302,0x1302)
GDI32(NtGdiSetIcmMode                                   ,  NONE,0x1112,0x1111,0x120b,0x120b,0x111a,0x1259,0x1259,0x111c,0x1265,0x1265,0x101e,0x0c12b2,0x12b2,0x101f,0x0c12cb,0x12cb,0x1020,0x0c1304,0x1304,0x1020,0x0c1308,0x1308,0x1020,0x0c12ff,0x12ff,0x1024,0x0c12cf,0x12cf,0x1025,0x0c12e9,0x12e9,0x1025,0x0c1303,0x1303)
GDI32(NtGdiSetLayout                                    ,  NONE,0x1119,0x1118,0x1097,0x1097,0x1121,0x1098,0x1098,0x1123,0x1096,0x1096,0x1017,0x0c1096,0x1096,0x1018,0x0c1097,0x1097,0x1019,0x0c1098,0x1098,0x1019,0x0c1098,0x1098,0x1019,0x0c1098,0x1098,0x101d,0x0c1096,0x1096,0x101e,0x0c1096,0x1096,0x101e,0x0c1096,0x1096)
GDI32(NtGdiSetLinkedUFIs                                ,  NONE,0x1113,0x1112,0x120c,0x120c,0x111b,0x125a,0x125a,0x111d,0x1266,0x1266,0x101d,0x0c12b3,0x12b3,0x101e,0x0c12cc,0x12cc,0x101f,0x0c1305,0x1305,0x101f,0x0c1309,0x1309,0x101f,0x0c1300,0x1300,0x1023,0x0c12d0,0x12d0,0x1024,0x0c12ea,0x12ea,0x1024,0x0c1304,0x1304)
GDI32(NtGdiSetMagicColors                               ,0x110a,0x1114,0x1113,0x120d,0x120d,0x111c,0x125b,0x125b,0x111e,0x1267,0x1267,0x101c,0x0012b4,0x12b4,0x101d,0x0012cd,0x12cd,0x101e,0x001306,0x1306,0x101e,0x00130a,0x130a,0x101e,0x001301,0x1301,0x1022,0x0012d1,0x12d1,0x1023,0x0012eb,0x12eb,0x1023,0x001305,0x1305)
GDI32(NtGdiSetMetaRgn                                   ,  NONE,0x1115,0x1114,0x10e8,0x10e8,0x111d,0x10e9,0x10e9,0x111f,0x10e4,0x10e4,0x101b,0x0310e4,0x10e4,0x101c,0x0310e5,0x10e5,0x101d,0x0310e6,0x10e6,0x101d,0x0310e6,0x10e6,0x101d,0x0310e5,0x10e5,0x1021,0x0310da,0x10da,0x1022,0x0310da,0x10da,0x1022,0x0310da,0x10da)
GDI32(NtGdiSetMiterLimit                                ,  NONE,0x1116,0x1115,0x10e9,0x10e9,0x111e,0x10ea,0x10ea,0x1120,0x10e5,0x10e5,0x101a,0x0c10e5,0x10e5,0x101b,0x0c10e6,0x10e6,0x101c,0x0c10e7,0x10e7,0x101c,0x0c10e7,0x10e7,0x101c,0x0c10e6,0x10e6,0x1020,0x0c10db,0x10db,0x1021,0x0c10db,0x10db,0x1021,0x0c10db,0x10db)
GDI32(NtGdiSetOPMSigningKeyAndSequenceNumbers           ,  NONE,  NONE,  NONE,  NONE,  NONE,0x1122,0x125c,0x125c,0x1124,0x1268,0x1268,0x1016,0x0012b5,0x12b5,0x1017,0x0012ce,0x12ce,0x1018,0x001307,0x1307,0x1018,0x00130b,0x130b,0x1018,0x001302,0x1302,0x101c,0x0012d2,0x12d2,0x101d,0x0012ec,0x12ec,0x101d,0x001306,0x1306)
GDI32(NtGdiSetPUMPDOBJ                                  ,  NONE,0x1297,0x1293,0x120e,0x120e,0x12b8,0x125d,0x125d,0x12d8,0x1269,0x1269,0x1302,0x0012b6,0x12b6,0x130b,0x0012cf,0x12cf,0x1317,0x001308,0x1308,0x1318,0x00130c,0x130c,0x131e,0x001303,0x1303,0x12df,0x0012d3,0x12d3,0x12e8,0x0012ed,0x12ed,0x12f2,0x001307,0x1307)
GDI32(NtGdiSetPixel                                     ,  NONE,0x111a,0x1119,0x10b3,0x10b3,0x1123,0x10b4,0x10b4,0x1125,0x10af,0x10af,0x1015,0x1410af,0x10af,0x1016,0x1410b0,0x10b0,0x1017,0x1410b1,0x10b1,0x1017,0x1410b1,0x10b1,0x1017,0x1410b1,0x10b1,0x101b,0x1410aa,0x10aa,0x101c,0x1410aa,0x10aa,0x101c,0x1410aa,0x10aa)
GDI32(NtGdiSetPixelFormat                               ,  NONE,0x111b,0x111a,0x120f,0x120f,0x1124,0x125e,0x125e,0x1126,0x126a,0x126a,0x1014,0x0812b7,0x12b7,0x1015,0x0812d0,0x12d0,0x1016,0x081309,0x1309,0x1016,0x08130d,0x130d,0x1016,0x081304,0x1304,0x101a,0x0812d4,0x12d4,0x101b,0x0812ee,0x12ee,0x101b,0x081308,0x1308)
GDI32(NtGdiSetPrivateDeviceGammaRamp                    ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,0x102a,0x0c12d5,0x12d5,0x102b,0x0c12ef,0x12ef,0x102b,0x0c1309,0x1309)
GDI32(NtGdiSetRectRgn                                   ,  NONE,0x111c,0x111b,0x1210,0x1210,0x1125,0x125f,0x125f,0x1127,0x126b,0x126b,0x1013,0x0012b8,0x12b8,0x1014,0x0012d1,0x12d1,0x1015,0x00130a,0x130a,0x1015,0x00130e,0x130e,0x1015,0x001305,0x1305,0x1019,0x0012d6,0x12d6,0x101a,0x0012f0,0x12f0,0x101a,0x00130a,0x130a)
GDI32(NtGdiSetSizeDevice                                ,  NONE,0x1121,0x1120,0x1211,0x1211,0x112a,0x1260,0x1260,0x112b,0x126c,0x126c,0x100f,0x0c12b9,0x12b9,0x1010,0x0c12d2,0x12d2,0x1011,0x0c130b,0x130b,0x1011,0x0c130f,0x130f,0x1011,0x0c1306,0x1306,0x1015,0x0c12d7,0x12d7,0x1016,0x0c12f1,0x12f1,0x1016,0x0c130b,0x130b)
GDI32(NtGdiSetSystemPaletteUse                          ,  NONE,0x111d,0x111c,0x1212,0x1212,0x1126,0x1261,0x1261,0x1128,0x126d,0x126d,0x1012,0x0712ba,0x12ba,0x1013,0x0712d3,0x12d3,0x1014,0x07130c,0x130c,0x1014,0x071310,0x1310,0x1014,0x071307,0x1307,0x1018,0x0712d8,0x12d8,0x1019,0x0712f2,0x12f2,0x1019,0x07130c,0x130c)
GDI32(NtGdiSetTextJustification                         ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x1262,  NONE,  NONE,0x126e,  NONE,    NONE,0x12bb,  NONE,    NONE,0x12d4,0x1013,0x0c130d,0x130d,0x1013,0x0c1311,0x1311,0x1013,0x0c1308,0x1308,0x1017,0x0c12d9,0x12d9,0x1018,0x0c12f3,0x12f3,0x1018,0x0c130d,0x130d)
GDI32(NtGdiSetUMPDSandboxState                          ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x1303,0x0012bc,0x12bc,0x130c,0x0012d5,0x12d5,0x1318,0x00130e,0x130e,0x1319,0x001312,0x1312,0x131f,0x001309,0x1309,0x12e0,0x0012da,0x12da,0x12e9,0x0012f4,0x12f4,0x12f3,0x00130e,0x130e)
GDI32(NtGdiSetVirtualResolution                         ,  NONE,0x1120,0x111f,0x10ea,0x10ea,0x1129,0x10eb,0x10eb,0x112a,0x10e6,0x10e6,0x1010,0x0010e6,0x10e6,0x1011,0x0010e7,0x10e7,0x1012,0x0010e8,0x10e8,0x1012,0x0010e8,0x10e8,0x1012,0x0010e7,0x10e7,0x1016,0x0010dc,0x10dc,0x1017,0x0010dc,0x10dc,0x1017,0x0010dc,0x10dc)
GDI32(NtGdiSetupPublicCFONT                             ,  NONE,0x111f,0x111e,0x110b,0x110b,0x1128,0x110c,0x110c,  NONE,  NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
GDI32(NtGdiSfmGetNotificationTokens                     ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x12dc,0x126f,0x126f,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE,  NONE,    NONE,  NONE)
GDI32(NtGdiStartDoc                                     ,  NONE,0x1122,0x1121,0x1214,0x1214,0x112b,0x1263,0x1263,0x112c,0x1270,0x1270,0x100e,0x0012bd,0x12bd,0x100f,0x0012d6,0x12d6,0x1010,0x00130f,0x130f,0x1010,0x001313,0x1313,0x1010,0x00130a,0x130a,0x1014,0x0012db,0x12db,0x1015,0x0012f5,0x12f5,0x1015,0x00130f,0x130f)
GDI32(NtGdiStartPage                                    ,  NONE,0x1123,0x1122,0x1215,0x1215,0x112c,0x1264,0x1264,0x112d,0x1271,0x1271,0x100d,0x0312be,0x12be,0x100e,0x0312d7,0x12d7,0x100f,0x031310,0x1310,0x100f,0x031314,0x1314,0x100f,0x03130b,0x130b,0x1013,0x0312dc,0x12dc,0x1014,0x0312f6,0x12f6,0x1014,0x031310,0x1310)
GDI32(NtGdiStretchBlt                                   ,  NONE,0x1124,0x1123,0x1030,0x1030,0x112d,0x1031,0x1031,0x112e,0x1031,0x1031,0x100c,0x001032,0x1032,0x100d,0x001033,0x1033,0x100e,0x001034,0x1034,0x100e,0x001034,0x1034,0x100e,0x001034,0x1034,0x1012,0x001034,0x1034,0x1013,0x001034,0x1034,0x1013,0x001034,0x1034)
GDI32(NtGdiStretchDIBitsInternal                        ,  NONE,0x1125,0x1124,0x1082,0x1082,0x112e,0x1083,0x1083,0x112f,0x1081,0x1081,0x100b,0x001081,0x1081,0x100c,0x001082,0x1082,0x100d,0x001083,0x1083,0x100d,0x001083,0x1083,0x100d,0x001083,0x1083,0x1011,0x001081,0x1081,0x1012,0x001081,0x1081,0x1012,0x001081,0x1081)
GDI32(NtGdiStrokeAndFillPath                            ,  NONE,0x1126,0x1125,0x1216,0x1216,0x112f,0x1265,0x1265,0x1130,0x1272,0x1272,0x100a,0x0312bf,0x12bf,0x100b,0x0312d8,0x12d8,0x100c,0x031311,0x1311,0x100c,0x031315,0x1315,0x100c,0x03130c,0x130c,0x1010,0x0312dd,0x12dd,0x1011,0x0312f7,0x12f7,0x1011,0x031311,0x1311)
GDI32(NtGdiStrokePath                                   ,  NONE,0x1127,0x1126,0x1217,0x1217,0x1130,0x1266,0x1266,0x1131,0x1273,0x1273,0x1009,0x0312c0,0x12c0,0x100a,0x0312d9,0x12d9,0x100b,0x031312,0x1312,0x100b,0x031316,0x1316,0x100b,0x03130d,0x130d,0x100f,0x0312de,0x12de,0x1010,0x0312f8,0x12f8,0x1010,0x031312,0x1312)
GDI32(NtGdiSwapBuffers                                  ,  NONE,0x1128,0x1127,0x1218,0x1218,0x1131,0x1267,0x1267,0x1132,0x1274,0x1274,0x1008,0x0312c1,0x12c1,0x1009,0x0312da,0x12da,0x100a,0x031313,0x1313,0x100a,0x031317,0x1317,0x100a,0x03130e,0x130e,0x100e,0x0312df,0x12df,0x100f,0x0312f9,0x12f9,0x100f,0x031313,0x1313)
GDI32(NtGdiTransformPoints                              ,  NONE,0x1129,0x1128,0x1072,0x1072,0x1132,0x1073,0x1073,0x1133,0x1072,0x1072,0x1007,0x001072,0x1072,0x1008,0x001073,0x1073,0x1009,0x001074,0x1074,0x1009,0x001074,0x1074,0x1009,0x001074,0x1074,0x1009,0x001073,0x1073,0x1009,0x001073,0x1073,0x1009,0x001073,0x1073)
GDI32(NtGdiTransparentBlt                               ,  NONE,0x112a,0x1129,0x1219,0x1219,0x1133,0x1268,0x1268,0x1134,0x1275,0x1275,0x1006,0x0012c2,0x12c2,0x1007,0x0012db,0x12db,0x1008,0x001314,0x1314,0x1008,0x001318,0x1318,0x1008,0x00130f,0x130f,0x1008,0x0012e0,0x12e0,0x1008,0x0012fa,0x12fa,0x1008,0x001314,0x1314)
GDI32(NtGdiUMPDEngFreeUserMem                           ,  NONE,0x1299,0x1295,0x121a,0x121a,0x12ba,0x1269,0x1269,0x12da,0x1276,0x1276,0x1300,0x0012c3,0x12c3,0x1309,0x0012dc,0x12dc,0x1315,0x001315,0x1315,0x1316,0x001319,0x1319,0x131c,0x001310,0x1310,0x12dd,0x0012e1,0x12e1,0x12e6,0x0012fb,0x12fb,0x12f0,0x001315,0x1315)
GDI32(NtGdiUnloadPrinterDriver                          ,  NONE,0x112b,0x112a,0x121b,0x121b,0x1134,0x126a,0x126a,0x1135,0x1277,0x1277,0x1005,0x0512c4,0x12c4,0x1006,0x0512dd,0x12dd,0x1007,0x051316,0x1316,0x1007,0x05131a,0x131a,0x1007,0x051311,0x1311,0x1007,0x0512e2,0x12e2,0x1007,0x0512fc,0x12fc,0x1007,0x051316,0x1316)
GDI32(NtGdiUnmapMemFont                                 ,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,  NONE,0x126b,  NONE,  NONE,0x1278,  NONE,    NONE,0x12c5,  NONE,    NONE,0x12de,0x1006,0x001317,0x1317,0x1006,0x00131b,0x131b,0x1006,0x001312,0x1312,0x1006,0x0012e3,0x12e3,0x1006,0x0012fd,0x12fd,0x1006,0x001317,0x1317)
GDI32(NtGdiUnrealizeObject                              ,  NONE,0x112d,0x112c,0x108f,0x108f,0x1136,0x1090,0x1090,0x1137,0x108e,0x108e,0x1003,0x03108e,0x108e,0x1004,0x03108f,0x108f,0x1005,0x031090,0x1090,0x1005,0x031090,0x1090,0x1005,0x031090,0x1090,0x1005,0x03108e,0x108e,0x1005,0x03108e,0x108e,0x1005,0x03108e,0x108e)
GDI32(NtGdiUpdateColors                                 ,  NONE,0x112e,0x112d,0x121d,0x121d,0x1137,0x126c,0x126c,0x1138,0x1279,0x1279,0x1002,0x0312c6,0x12c6,0x1003,0x0312df,0x12df,0x1004,0x031318,0x1318,0x1004,0x03131c,0x131c,0x1004,0x031313,0x1313,0x1004,0x0312e4,0x12e4,0x1004,0x0312fe,0x12fe,0x1004,0x031318,0x1318)
GDI32(NtGdiUpdateTransform                              ,  NONE,0x1296,0x1292,0x121e,0x121e,0x12b7,0x126d,0x126d,0x12d7,0x127a,0x127a,0x12ff,0x0012c7,0x12c7,0x1308,0x0012e0,0x12e0,0x1314,0x001319,0x1319,0x1315,0x00131d,0x131d,0x131b,0x001314,0x1314,0x12dc,0x0012e5,0x12e5,0x12e5,0x0012ff,0x12ff,0x12ef,0x001319,0x1319)
GDI32(NtGdiWidenPath                                    ,  NONE,0x112f,0x112e,0x121f,0x121f,0x1138,0x126e,0x126e,0x1139,0x127b,0x127b,0x1001,0x0312c8,0x12c8,0x1002,0x0312e1,0x12e1,0x1003,0x03131a,0x131a,0x1003,0x03131e,0x131e,0x1003,0x031315,0x1315,0x1003,0x0312e6,0x12e6,0x1003,0x031300,0x1300,0x1003,0x03131a,0x131a)
GDI32(NtGdiXFORMOBJ_bApplyXform                         ,0x1266,0x127e,0x127a,0x1220,0x1220,0x129f,0x126f,0x126f,0x12bf,0x127c,0x127c,0x12e8,0x0012c9,0x12c9,0x12f1,0x0012e2,0x12e2,0x12fd,0x00131b,0x131b,0x12fe,0x00131f,0x131f,0x1304,0x001316,0x1316,0x12c5,0x0012e7,0x12e7,0x12ce,0x001301,0x1301,0x12d8,0x00131b,0x131b)
GDI32(NtGdiXFORMOBJ_iGetXform                           ,0x1267,0x127f,0x127b,0x1221,0x1221,0x12a0,0x1270,0x1270,0x12c0,0x127d,0x127d,0x12e7,0x0012ca,0x12ca,0x12f0,0x0012e3,0x12e3,0x12fc,0x00131c,0x131c,0x12fd,0x001320,0x1320,0x1303,0x001317,0x1317,0x12c4,0x0012e8,0x12e8,0x12cd,0x001302,0x1302,0x12d7,0x00131c,0x131c)
GDI32(NtGdiXLATEOBJ_cGetPalette                         ,0x1259,0x1271,0x126d,0x1222,0x1222,0x1292,0x1271,0x1271,0x12b2,0x127e,0x127e,0x12dc,0x0012cb,0x12cb,0x12e5,0x0012e4,0x12e4,0x12f1,0x00131d,0x131d,0x12f2,0x001321,0x1321,0x12f8,0x001318,0x1318,0x12b9,0x0012e9,0x12e9,0x12c2,0x001303,0x1303,0x12cc,0x00131d,0x131d)
GDI32(NtGdiXLATEOBJ_hGetColorTransform                  ,0x125b,0x1273,0x126f,0x1223,0x1223,0x1294,0x1272,0x1272,0x12b4,0x127f,0x127f,0x12da,0x0012cc,0x12cc,0x12e3,0x0012e5,0x12e5,0x12ef,0x00131e,0x131e,0x12f0,0x001322,0x1322,0x12f6,0x001319,0x1319,0x12b7,0x0012ea,0x12ea,0x12c0,0x001304,0x1304,0x12ca,0x00131e,0x131e)
GDI32(NtGdiXLATEOBJ_iXlate                              ,0x125a,0x1272,0x126e,0x1224,0x1224,0x1293,0x1273,0x1273,0x12b3,0x1280,0x1280,0x12db,0x0012cd,0x12cd,0x12e4,0x0012e6,0x12e6,0x12f0,0x00131f,0x131f,0x12f1,0x001323,0x1323,0x12f7,0x00131a,0x131a,0x12b8,0x0012eb,0x12eb,0x12c1,0x001305,0x1305,0x12cb,0x00131f,0x131f)

KERNEL32(AddConsoleAliasInternal                           ,NONE,NONE,NONE,0x2040,NONE,NONE,0x2046,NONE,NONE,0x2047,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(AllocConsoleInternal                              ,NONE,NONE,NONE,0x203c,NONE,NONE,0x2042,NONE,NONE,0x2042,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(AttachConsoleInternal                             ,NONE,NONE,NONE,0x203e,NONE,NONE,0x2044,NONE,NONE,0x2044,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(CloseConsoleHandle                                ,NONE,NONE,NONE,0x2003,NONE,NONE,0x2003,NONE,NONE,0x2003,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(ConnectConsoleInternal                            ,NONE,NONE,NONE,0x203b,NONE,NONE,0x2041,NONE,NONE,0x2041,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(ConsoleMenuControl                                ,NONE,NONE,NONE,0x2032,NONE,NONE,0x2038,NONE,NONE,0x2038,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(CreateConsoleScreenBuffer                         ,NONE,NONE,NONE,0x2010,NONE,NONE,0x2010,NONE,NONE,0x2010,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(DuplicateConsoleHandle                            ,NONE,NONE,NONE,0x2004,NONE,NONE,0x2004,NONE,NONE,0x2004,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(ExpungeConsoleCommandHistoryInternal              ,NONE,NONE,NONE,0x2046,NONE,NONE,0x204c,NONE,NONE,0x204d,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(FillConsoleOutput                                 ,NONE,NONE,NONE,0x200f,NONE,NONE,0x200f,NONE,NONE,0x200f,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(FlushConsoleInputBuffer                           ,NONE,NONE,NONE,0x2021,NONE,NONE,0x2024,NONE,NONE,0x2024,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(FreeConsoleInternal                               ,NONE,NONE,NONE,0x203d,NONE,NONE,0x2043,NONE,NONE,0x2043,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(GenerateConsoleCtrlEvent                          ,NONE,NONE,NONE,0x201f,NONE,NONE,0x2022,NONE,NONE,0x2022,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(GetConsoleAliasExesInternal                       ,NONE,NONE,NONE,0x2045,NONE,NONE,0x204b,NONE,NONE,0x204c,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(GetConsoleAliasExesLengthInternal                 ,NONE,NONE,NONE,0x2043,NONE,NONE,0x2049,NONE,NONE,0x204a,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(GetConsoleAliasInternal                           ,NONE,NONE,NONE,0x2041,NONE,NONE,0x2047,NONE,NONE,0x2048,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(GetConsoleAliasesInternal                         ,NONE,NONE,NONE,0x2044,NONE,NONE,0x204a,NONE,NONE,0x204b,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(GetConsoleAliasesLengthInternal                   ,NONE,NONE,NONE,0x2042,NONE,NONE,0x2048,NONE,NONE,0x2049,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(GetConsoleCP                                      ,NONE,NONE,NONE,0x202a,NONE,NONE,0x2030,NONE,NONE,0x2030,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(GetConsoleCharType                                ,NONE,NONE,NONE,0x204d,NONE,NONE,0x2052,NONE,NONE,0x2053,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(GetConsoleCommandHistoryInternal                  ,NONE,NONE,NONE,0x2049,NONE,NONE,0x204f,NONE,NONE,0x2050,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(GetConsoleCommandHistoryLengthInternal            ,NONE,NONE,NONE,0x2048,NONE,NONE,0x204e,NONE,NONE,0x204f,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(GetConsoleCursorInfo                              ,NONE,NONE,NONE,0x2018,NONE,NONE,0x201a,NONE,NONE,0x201a,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(GetConsoleCursorMode                              ,NONE,NONE,NONE,0x2050,NONE,NONE,0x2055,NONE,NONE,0x2056,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(GetConsoleDisplayMode                             ,NONE,NONE,NONE,0x2038,NONE,NONE,0x203e,NONE,NONE,0x203e,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(GetConsoleFontInfo                                ,NONE,NONE,NONE,0x201b,NONE,NONE,0x201d,NONE,NONE,0x201d,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(GetConsoleFontSize                                ,NONE,NONE,NONE,0x201c,NONE,NONE,0x201e,NONE,NONE,0x201e,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(GetConsoleHandleInformation                       ,NONE,NONE,NONE,0x2005,NONE,NONE,0x2005,NONE,NONE,0x2005,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(GetConsoleHardwareState                           ,NONE,NONE,NONE,0x2036,NONE,NONE,0x203c,NONE,NONE,0x203c,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(GetConsoleHistoryInfo                             ,NONE,NONE,NONE,  NONE,NONE,NONE,0x2012,NONE,NONE,0x2012,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(GetConsoleInput                                   ,NONE,NONE,NONE,0x2009,NONE,NONE,0x2009,NONE,NONE,0x2009,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(GetConsoleKeyboardLayoutNameWorker                ,NONE,NONE,NONE,0x202e,NONE,NONE,0x2034,NONE,NONE,0x2034,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(GetConsoleLangId                                  ,NONE,NONE,NONE,0x203f,NONE,NONE,0x2045,NONE,NONE,0x2045,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(GetConsoleMode                                    ,NONE,NONE,NONE,0x2012,NONE,NONE,0x2014,NONE,NONE,0x2014,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(GetConsoleNlsMode                                 ,NONE,NONE,NONE,0x2053,NONE,NONE,0x2058,NONE,NONE,0x2059,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(GetConsoleOutputCP                                ,NONE,NONE,NONE,0x202c,NONE,NONE,0x2032,NONE,NONE,0x2032,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(GetConsoleProcessList                             ,NONE,NONE,NONE,0x2013,NONE,NONE,0x2015,NONE,NONE,0x2015,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(GetConsoleScreenBufferInfo                        ,NONE,NONE,NONE,0x2017,NONE,NONE,0x2019,NONE,NONE,0x2019,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(GetConsoleScreenBufferInfoEx                      ,NONE,NONE,NONE,  NONE,NONE,NONE,0x2013,NONE,NONE,0x2013,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(GetConsoleSelectionInfo                           ,NONE,NONE,NONE,0x2019,NONE,NONE,0x201b,NONE,NONE,0x201b,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(GetConsoleTitleInternal                           ,NONE,NONE,NONE,0x204b,NONE,NONE,0x2050,NONE,NONE,0x2051,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(GetConsoleWindow                                  ,NONE,NONE,NONE,0x202f,NONE,NONE,0x2035,NONE,NONE,0x2035,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(GetCurrentConsoleFont                             ,NONE,NONE,NONE,0x201d,NONE,NONE,0x201f,NONE,NONE,0x201f,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(GetCurrentConsoleFontEx                           ,NONE,NONE,NONE,  NONE,NONE,NONE,0x2020,NONE,NONE,0x2020,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(GetLargestConsoleWindowSize                       ,NONE,NONE,NONE,0x2016,NONE,NONE,0x2018,NONE,NONE,0x2018,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(GetNumberOfConsoleFonts                           ,NONE,NONE,NONE,0x2014,NONE,NONE,0x2016,NONE,NONE,0x2016,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(GetNumberOfConsoleInputEvents                     ,NONE,NONE,NONE,0x2015,NONE,NONE,0x2017,NONE,NONE,0x2017,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(GetNumberOfConsoleMouseButtons                    ,NONE,NONE,NONE,0x201a,NONE,NONE,0x201c,NONE,NONE,0x201c,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(InvalidateConsoleDIBits                           ,NONE,NONE,NONE,0x2011,NONE,NONE,0x2011,NONE,NONE,0x2011,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(NtWow64ConsoleLaunchServerProcess                 ,NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,0x300d,NONE,NONE,0x300c,NONE,NONE,0x300c,NONE,NONE,0x300c,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(NtWow64CsrBaseCheckRunApp                         ,NONE,NONE,NONE,0x3011,NONE,NONE,0x300b,NONE,NONE,0x300b,NONE,NONE,0x3009,NONE,NONE,0x3008,NONE,NONE,0x3008,NONE,NONE,0x3008,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(NtWow64CsrBaseClientConnectToServer               ,NONE,NONE,NONE,0x300b,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(NtWow64CsrBaseQueryModuleData                     ,NONE,NONE,NONE,0x3013,NONE,NONE,0x300c,NONE,NONE,0x300c,NONE,NONE,0x300a,NONE,NONE,0x3009,NONE,NONE,0x3009,NONE,NONE,0x3009,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(NtWow64CsrBasepCreateActCtx                       ,NONE,NONE,NONE,0x300f,NONE,NONE,0x300a,NONE,NONE,0x300a,NONE,NONE,0x3008,NONE,NONE,0x3007,NONE,NONE,0x3007,NONE,NONE,0x3007,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(NtWow64CsrBasepCreateProcess                      ,NONE,NONE,NONE,0x3004,NONE,NONE,0x3003,NONE,NONE,0x3003,NONE,NONE,0x3003,NONE,NONE,0x3002,NONE,NONE,0x3002,NONE,NONE,0x3002,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(NtWow64CsrBasepCreateThread                       ,NONE,NONE,NONE,0x300a,NONE,NONE,0x3009,NONE,NONE,0x3009,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(NtWow64CsrBasepDefineDosDevice                    ,NONE,NONE,NONE,0x3002,NONE,NONE,0x3002,NONE,NONE,0x3002,NONE,NONE,0x3002,NONE,NONE,0x3001,NONE,NONE,0x3001,NONE,NONE,0x3001,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(NtWow64CsrBasepExitProcess                        ,NONE,NONE,NONE,0x3005,NONE,NONE,0x3004,NONE,NONE,0x3004,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(NtWow64CsrBasepGetProcessShutdownParam            ,NONE,NONE,NONE,0x3007,NONE,NONE,0x3006,NONE,NONE,0x3006,NONE,NONE,0x3005,NONE,NONE,0x3004,NONE,NONE,0x3004,NONE,NONE,0x3004,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(NtWow64CsrBasepGetTempFile                        ,NONE,NONE,NONE,0x3003,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(NtWow64CsrBasepNlsCreateSection                   ,NONE,NONE,NONE,0x300e,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(NtWow64CsrBasepNlsGetUserInfo                     ,NONE,NONE,NONE,0x3012,NONE,NONE,0x300f,NONE,NONE,0x300e,NONE,NONE,0x300c,NONE,NONE,0x300b,NONE,NONE,0x300b,NONE,NONE,0x300b,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(NtWow64CsrBasepNlsSetMultipleUserInfo             ,NONE,NONE,NONE,0x300d,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(NtWow64CsrBasepNlsSetUserInfo                     ,NONE,NONE,NONE,0x300c,NONE,NONE,0x300d,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(NtWow64CsrBasepNlsUpdateCacheCount                ,NONE,NONE,NONE,0x3010,NONE,NONE,0x300e,NONE,NONE,0x300d,NONE,NONE,0x300b,NONE,NONE,0x300a,NONE,NONE,0x300a,NONE,NONE,0x300a,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(NtWow64CsrBasepRefreshIniFileMapping              ,NONE,NONE,NONE,0x3001,NONE,NONE,0x3001,NONE,NONE,0x3001,NONE,NONE,0x3001,NONE,NONE,0x3000,NONE,NONE,0x3000,NONE,NONE,0x3000,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(NtWow64CsrBasepSetClientTimeZoneInformation       ,NONE,NONE,NONE,0x3009,NONE,NONE,0x3008,NONE,NONE,0x3008,NONE,NONE,0x3007,NONE,NONE,0x3006,NONE,NONE,0x3006,NONE,NONE,0x3006,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(NtWow64CsrBasepSetProcessShutdownParam            ,NONE,NONE,NONE,0x3006,NONE,NONE,0x3005,NONE,NONE,0x3005,NONE,NONE,0x3004,NONE,NONE,0x3003,NONE,NONE,0x3003,NONE,NONE,0x3003,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(NtWow64CsrBasepSetTermsrvAppInstallMode           ,NONE,NONE,NONE,0x3008,NONE,NONE,0x3007,NONE,NONE,0x3007,NONE,NONE,0x3006,NONE,NONE,0x3005,NONE,NONE,0x3005,NONE,NONE,0x3005,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(NtWow64CsrBasepSoundSentryNotification            ,NONE,NONE,NONE,0x3000,NONE,NONE,0x3000,NONE,NONE,0x3000,NONE,NONE,0x3000,NONE,NONE,0x300d,NONE,NONE,0x300d,NONE,NONE,0x300d,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(OpenConsoleWInternal                              ,NONE,NONE,NONE,0x2000,NONE,NONE,0x2000,NONE,NONE,0x2000,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(QueryConsoleIMEInternal                           ,NONE,NONE,NONE,  NONE,NONE,NONE,0x205c,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(ReadConsoleInternal                               ,NONE,NONE,NONE,0x2001,NONE,NONE,0x2001,NONE,NONE,0x2001,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(ReadConsoleOutputInternal                         ,NONE,NONE,NONE,0x200b,NONE,NONE,0x200b,NONE,NONE,0x200b,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(ReadConsoleOutputString                           ,NONE,NONE,NONE,0x200d,NONE,NONE,0x200d,NONE,NONE,0x200d,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(RegisterConsoleIMEInternal                        ,NONE,NONE,NONE,0x2055,NONE,NONE,0x205a,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(RegisterConsoleOS2                                ,NONE,NONE,NONE,0x2051,NONE,NONE,0x2056,NONE,NONE,0x2057,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(RegisterConsoleVDM                                ,NONE,NONE,NONE,0x2034,NONE,NONE,0x203a,NONE,NONE,0x203a,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(ScrollConsoleScreenBufferInternal                 ,NONE,NONE,NONE,0x2026,NONE,NONE,0x202c,NONE,NONE,0x202c,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(SetConsoleActiveScreenBuffer                      ,NONE,NONE,NONE,0x2020,NONE,NONE,0x2023,NONE,NONE,0x2023,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(SetConsoleCP                                      ,NONE,NONE,NONE,0x202b,NONE,NONE,0x2031,NONE,NONE,0x2031,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(SetConsoleCommandHistoryMode                      ,NONE,NONE,NONE,0x204a,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(SetConsoleCursor                                  ,NONE,NONE,NONE,0x2030,NONE,NONE,0x2036,NONE,NONE,0x2036,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(SetConsoleCursorInfo                              ,NONE,NONE,NONE,0x2024,NONE,NONE,0x2027,NONE,NONE,0x2027,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(SetConsoleCursorMode                              ,NONE,NONE,NONE,0x204f,NONE,NONE,0x2054,NONE,NONE,0x2055,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(SetConsoleCursorPosition                          ,NONE,NONE,NONE,0x2023,NONE,NONE,0x2026,NONE,NONE,0x2026,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(SetConsoleDisplayMode                             ,NONE,NONE,NONE,0x2035,NONE,NONE,0x203b,NONE,NONE,0x203b,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(SetConsoleFont                                    ,NONE,NONE,NONE,0x2028,NONE,NONE,0x202e,NONE,NONE,0x202e,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(SetConsoleHandleInformation                       ,NONE,NONE,NONE,0x2006,NONE,NONE,0x2006,NONE,NONE,0x2006,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(SetConsoleHardwareState                           ,NONE,NONE,NONE,0x2037,NONE,NONE,0x203d,NONE,NONE,0x203d,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(SetConsoleHistoryInfo                             ,NONE,NONE,NONE,  NONE,NONE,NONE,0x2028,NONE,NONE,0x2028,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(SetConsoleIcon                                    ,NONE,NONE,NONE,0x2029,NONE,NONE,0x202f,NONE,NONE,0x202f,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(SetConsoleKeyShortcuts                            ,NONE,NONE,NONE,0x2039,NONE,NONE,0x203f,NONE,NONE,0x203f,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(SetConsoleLocalEUDC                               ,NONE,NONE,NONE,0x204e,NONE,NONE,0x2053,NONE,NONE,0x2054,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(SetConsoleMenuClose                               ,NONE,NONE,NONE,0x203a,NONE,NONE,0x2040,NONE,NONE,0x2040,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(SetConsoleMode                                    ,NONE,NONE,NONE,0x201e,NONE,NONE,0x2021,NONE,NONE,0x2021,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(SetConsoleNlsMode                                 ,NONE,NONE,NONE,0x2054,NONE,NONE,0x2059,NONE,NONE,0x205a,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(SetConsoleNumberOfCommandsInternal                ,NONE,NONE,NONE,0x2047,NONE,NONE,0x204d,NONE,NONE,0x204e,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(SetConsoleOS2OemFormat                            ,NONE,NONE,NONE,0x2052,NONE,NONE,0x2057,NONE,NONE,0x2058,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(SetConsoleOutputCPInternal                        ,NONE,NONE,NONE,0x202d,NONE,NONE,0x2033,NONE,NONE,0x2033,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(SetConsolePaletteInternal                         ,NONE,NONE,NONE,0x2033,NONE,NONE,0x2039,NONE,NONE,0x2039,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(SetConsoleScreenBufferInfoEx                      ,NONE,NONE,NONE,  NONE,NONE,NONE,0x2029,NONE,NONE,0x2029,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(SetConsoleScreenBufferSize                        ,NONE,NONE,NONE,0x2022,NONE,NONE,0x2025,NONE,NONE,0x2025,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(SetConsoleTextAttribute                           ,NONE,NONE,NONE,0x2027,NONE,NONE,0x202d,NONE,NONE,0x202d,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(SetConsoleTitleInternal                           ,NONE,NONE,NONE,0x204c,NONE,NONE,0x2051,NONE,NONE,0x2052,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(SetConsoleWindowInfo                              ,NONE,NONE,NONE,0x2025,NONE,NONE,0x202a,NONE,NONE,0x202a,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(SetCurrentConsoleFontEx                           ,NONE,NONE,NONE,  NONE,NONE,NONE,0x202b,NONE,NONE,0x202b,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(SetLastConsoleEventActiveInternal                 ,NONE,NONE,NONE,0x2008,NONE,NONE,0x2008,NONE,NONE,0x2008,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(ShowConsoleCursor                                 ,NONE,NONE,NONE,0x2031,NONE,NONE,0x2037,NONE,NONE,0x2037,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(UnregisterConsoleIMEInternal                      ,NONE,NONE,NONE,0x2056,NONE,NONE,0x205b,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(VerifyConsoleIoHandle                             ,NONE,NONE,NONE,0x2007,NONE,NONE,0x2007,NONE,NONE,0x2007,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(WriteConsoleInputInternal                         ,NONE,NONE,NONE,0x200a,NONE,NONE,0x200a,NONE,NONE,0x200a,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(WriteConsoleInternal                              ,NONE,NONE,NONE,0x2002,NONE,NONE,0x2002,NONE,NONE,0x2002,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(WriteConsoleOutputInternal                        ,NONE,NONE,NONE,0x200c,NONE,NONE,0x200c,NONE,NONE,0x200c,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)
KERNEL32(WriteConsoleOutputString                          ,NONE,NONE,NONE,0x200e,NONE,NONE,0x200e,NONE,NONE,0x200e,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE,  NONE,NONE,NONE, NONE,NONE,NONE,  NONE,NONE, NONE,   NONE, NONE,  NONE,    NONE,  NONE)

/* NTDLL(name,                                          w2k, xpx86,w2003,xpwow, xp64,vis86,viwow,vis64,w7x86,w7wow,w7x64,w8x86,   w8wow,w8x64,w81x86, w81wow,w81x64,w10x86,w10wow,w10x64,w11x86,w11wow,w11x64,w12x86,w12wow,w12x64,w13x86,w13wow,w13x64,w14x86,w14wow,w14x64,w15x86,w15wow,w15x64)*/
NTDLL(NtAcceptConnectPort                               ,0x00,0x000,0x000,0x060,0x060,0x000,0x060,0x060,0x000,0x060,0x060,0x1ac,0x000061,0x061,0x001,0x000001,0x001,0x002,0x000002,0x002,0x002,0x000002,0x002,0x002,0x000002,0x002,0x002,0x000002,0x002,0x002,0x000002,0x002,0x002,0x000002,0x002)
NTDLL(NtAccessCheck                                     ,0x01,0x001,0x001,0x061,0x061,0x001,0x061,0x061,0x001,0x061,0x061,0x1ab,0x000062,0x062,0x1b0,0x000062,0x062,0x000,0x000000,0x000,0x000,0x000000,0x000,0x000,0x000000,0x000,0x000,0x000000,0x000,0x000,0x000000,0x000,0x000,0x000000,0x000)
NTDLL(NtAccessCheckAndAuditAlarm                        ,0x02,0x002,0x002,0x026,0x026,0x002,0x026,0x026,0x002,0x026,0x026,0x1aa,0x000027,0x027,0x1af,0x000028,0x028,0x1b7,0x000029,0x029,0x1ba,0x000029,0x029,0x1bc,0x000029,0x029,0x1c2,0x000029,0x029,0x1c6,0x000029,0x029,0x1c6,0x000029,0x029)
NTDLL(NtAccessCheckByType                               ,0x03,0x003,0x003,0x062,0x062,0x003,0x062,0x062,0x003,0x062,0x062,0x1a9,0x000063,0x063,0x1ae,0x000063,0x063,0x1b6,0x000063,0x063,0x1b9,0x000063,0x063,0x1bb,0x000063,0x063,0x1c1,0x000063,0x063,0x1c5,0x000063,0x063,0x1c5,0x000063,0x063)
NTDLL(NtAccessCheckByTypeAndAuditAlarm                  ,0x04,0x004,0x004,0x056,0x056,0x004,0x056,0x056,0x004,0x056,0x056,0x1a8,0x000057,0x057,0x1ad,0x000058,0x058,0x1b5,0x000059,0x059,0x1b8,0x000059,0x059,0x1ba,0x000059,0x059,0x1c0,0x000059,0x059,0x1c4,0x000059,0x059,0x1c4,0x000059,0x059)
NTDLL(NtAccessCheckByTypeResultList                     ,0x05,0x005,0x005,0x063,0x063,0x005,0x063,0x063,0x005,0x063,0x063,0x1a7,0x000064,0x064,0x1ac,0x000064,0x064,0x1b4,0x000064,0x064,0x1b7,0x000064,0x064,0x1b9,0x000064,0x064,0x1bf,0x000064,0x064,0x1c3,0x000064,0x064,0x1c3,0x000064,0x064)
NTDLL(NtAccessCheckByTypeResultListAndAuditAlarm        ,0x06,0x006,0x006,0x064,0x064,0x006,0x064,0x064,0x006,0x064,0x064,0x1a6,0x000065,0x065,0x1ab,0x000065,0x065,0x1b3,0x000065,0x065,0x1b6,0x000065,0x065,0x1b8,0x000065,0x065,0x1be,0x000065,0x065,0x1c2,0x000065,0x065,0x1c2,0x000065,0x065)
NTDLL(NtAccessCheckByTypeResultListAndAuditAlarmByHandle,0x07,0x007,0x007,0x065,0x065,0x007,0x065,0x065,0x007,0x065,0x065,0x1a5,0x000066,0x066,0x1aa,0x000066,0x066,0x1b2,0x000066,0x066,0x1b5,0x000066,0x066,0x1b7,0x000066,0x066,0x1bd,0x000066,0x066,0x1c1,0x000066,0x066,0x1c1,0x000066,0x066)
NTDLL(NtAcquireCMFViewOwnership                         ,NONE, NONE, NONE, NONE, NONE,0x185,0x066,0x066, NONE, NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE)
NTDLL(NtAcquireProcessActivityReference                 ,NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE,0x1bc,0x000067,0x067,0x1c0,0x000067,0x067,0x1c0,0x000067,0x067)
NTDLL(NtAddAtom                                         ,0x08,0x008,0x008,0x044,0x044,0x008,0x044,0x044,0x008,0x044,0x044,0x1a3,0x0a0045,0x045,0x1a8,0x0a0046,0x046,0x1b0,0x0a0047,0x047,0x1b3,0x0a0047,0x047,0x1b5,0x0a0047,0x047,0x1ba,0x0a0047,0x047,0x1be,0x0a0047,0x047,0x1be,0x0a0047,0x047)
NTDLL(NtAddAtomEx                                       ,NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,0x1a4,0x000067,0x067,0x1a9,0x000067,0x067,0x1b1,0x110067,0x067,0x1b4,0x110067,0x067,0x1b6,0x110067,0x067,0x1bb,0x110068,0x068,0x1bf,0x110068,0x068,0x1bf,0x110068,0x068)
NTDLL(NtAddBootEntry                                    ,NONE,0x009,0x009,0x066,0x066,0x009,0x067,0x067,0x009,0x066,0x066,0x1a2,0x000068,0x068,0x1a7,0x000068,0x068,0x1af,0x000068,0x068,0x1b2,0x000068,0x068,0x1b4,0x000068,0x068,0x1b9,0x000069,0x069,0x1bd,0x000069,0x069,0x1bd,0x000069,0x069)
NTDLL(NtAddDriverEntry                                  ,NONE, NONE,0x00a,0x067,0x067,0x00a,0x068,0x068,0x00a,0x067,0x067,0x1a1,0x000069,0x069,0x1a6,0x000069,0x069,0x1ae,0x000069,0x069,0x1b1,0x000069,0x069,0x1b3,0x000069,0x069,0x1b8,0x00006a,0x06a,0x1bc,0x00006a,0x06a,0x1bc,0x00006a,0x06a)
NTDLL(NtAdjustGroupsToken                               ,0x09,0x00a,0x00b,0x068,0x068,0x00b,0x069,0x069,0x00b,0x068,0x068,0x19f,0x00006a,0x06a,0x1a4,0x00006a,0x06a,0x1ac,0x00006a,0x06a,0x1af,0x00006a,0x06a,0x1b1,0x00006a,0x06a,0x1b6,0x00006b,0x06b,0x1ba,0x00006b,0x06b,0x1ba,0x00006b,0x06b)
NTDLL(NtAdjustPrivilegesToken                           ,0x0a,0x00b,0x00c,0x03e,0x03e,0x00c,0x03e,0x03e,0x00c,0x03e,0x03e,0x19e,0x00003f,0x03f,0x1a3,0x000040,0x040,0x1ab,0x000041,0x041,0x1ae,0x000041,0x041,0x1b0,0x000041,0x041,0x1b5,0x000041,0x041,0x1b9,0x000041,0x041,0x1b9,0x000041,0x041)
NTDLL(NtAdjustTokenClaimsAndDeviceGroups                ,NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,0x1a0,0x00006b,0x06b,0x1a5,0x00006b,0x06b,0x1ad,0x00006b,0x06b,0x1b0,0x00006b,0x06b,0x1b2,0x00006b,0x06b,0x1b7,0x00006c,0x06c,0x1bb,0x00006c,0x06c,0x1bb,0x00006c,0x06c)
NTDLL(NtAlertResumeThread                               ,0x0b,0x00c,0x00d,0x069,0x069,0x00d,0x06a,0x06a,0x00d,0x069,0x069,0x19d,0x07006c,0x06c,0x1a2,0x07006c,0x06c,0x1aa,0x07006c,0x06c,0x1ad,0x07006c,0x06c,0x1af,0x07006c,0x06c,0x1b4,0x07006d,0x06d,0x1b8,0x07006d,0x06d,0x1b8,0x07006d,0x06d)
NTDLL(NtAlertThread                                     ,0x0c,0x00d,0x00e,0x06a,0x06a,0x00e,0x06b,0x06b,0x00e,0x06a,0x06a,0x19c,0x03006d,0x06d,0x1a1,0x03006d,0x06d,0x1a9,0x03006d,0x06d,0x1ac,0x03006d,0x06d,0x1ae,0x03006d,0x06d,0x1b3,0x03006e,0x06e,0x1b7,0x03006e,0x06e,0x1b7,0x03006e,0x06e)
NTDLL(NtAlertThreadByThreadId                           ,NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,0x19b,0x04006e,0x06e,0x1a0,0x04006e,0x06e,0x1a8,0x04006e,0x06e,0x1ab,0x04006e,0x06e,0x1ad,0x04006e,0x06e,0x1b2,0x04006f,0x06f,0x1b6,0x04006f,0x06f,0x1b6,0x04006f,0x06f)
NTDLL(NtAllocateLocallyUniqueId                         ,0x0d,0x00e,0x00f,0x06b,0x06b,0x00f,0x06c,0x06c,0x00f,0x06b,0x06b,0x19a,0x04006f,0x06f,0x19f,0x04006f,0x06f,0x1a7,0x04006f,0x06f,0x1aa,0x04006f,0x06f,0x1ac,0x04006f,0x06f,0x1b1,0x040070,0x070,0x1b5,0x040070,0x070,0x1b5,0x040070,0x070)
NTDLL(NtAllocateReserveObject                           ,NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,0x010,0x06c,0x06c,0x199,0x000070,0x070,0x19e,0x000070,0x070,0x1a6,0x000070,0x070,0x1a9,0x000070,0x070,0x1ab,0x000070,0x070,0x1b0,0x000071,0x071,0x1b4,0x000071,0x071,0x1b4,0x000071,0x071)
NTDLL(NtAllocateUserPhysicalPages                       ,0x0e,0x00f,0x010,0x06c,0x06c,0x010,0x06d,0x06d,0x011,0x06d,0x06d,0x198,0x000071,0x071,0x19d,0x000071,0x071,0x1a5,0x000071,0x071,0x1a8,0x000071,0x071,0x1aa,0x000071,0x071,0x1af,0x000072,0x072,0x1b3,0x000072,0x072,0x1b3,0x000072,0x072)
NTDLL(NtAllocateUuids                                   ,0x0f,0x010,0x011,0x06d,0x06d,0x011,0x06e,0x06e,0x012,0x06e,0x06e,0x197,0x110072,0x072,0x19c,0x110072,0x072,0x1a4,0x110072,0x072,0x1a7,0x110072,0x072,0x1a9,0x110072,0x072,0x1ae,0x110073,0x073,0x1b2,0x110073,0x073,0x1b2,0x110073,0x073)
NTDLL(NtAllocateVirtualMemory                           ,0x10,0x011,0x012,0x015,0x015,0x012,0x015,0x015,0x013,0x015,0x015,0x196,0x000016,0x016,0x19b,0x000017,0x017,0x1a3,0x000018,0x018,0x1a6,0x000018,0x018,0x1a8,0x000018,0x018,0x1ad,0x000018,0x018,0x1b1,0x000018,0x018,0x1b1,0x000018,0x018)
NTDLL(NtAllocateVirtualMemoryEx                         ,NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE,0x1cc,0x000074,0x074)
NTDLL(NtAlpcAcceptConnectPort                           ,NONE, NONE, NONE, NONE, NONE,0x013,0x06f,0x06f,0x014,0x06f,0x06f,0x195,0x000073,0x073,0x19a,0x000073,0x073,0x1a2,0x000073,0x073,0x1a5,0x000073,0x073,0x1a7,0x000073,0x073,0x1ac,0x000074,0x074,0x1b0,0x000074,0x074,0x1b0,0x000075,0x075)
NTDLL(NtAlpcCancelMessage                               ,NONE, NONE, NONE, NONE, NONE,0x014,0x070,0x070,0x015,0x070,0x070,0x194,0x000074,0x074,0x199,0x000074,0x074,0x1a1,0x000074,0x074,0x1a4,0x000074,0x074,0x1a6,0x000074,0x074,0x1ab,0x000075,0x075,0x1af,0x000075,0x075,0x1af,0x000076,0x076)
NTDLL(NtAlpcConnectPort                                 ,NONE, NONE, NONE, NONE, NONE,0x015,0x071,0x071,0x016,0x071,0x071,0x192,0x000075,0x075,0x197,0x000075,0x075,0x19f,0x000075,0x075,0x1a2,0x000075,0x075,0x1a4,0x000075,0x075,0x1a9,0x000076,0x076,0x1ad,0x000076,0x076,0x1ad,0x000077,0x077)
NTDLL(NtAlpcConnectPortEx                               ,NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,0x193,0x000076,0x076,0x198,0x000076,0x076,0x1a0,0x000076,0x076,0x1a3,0x000076,0x076,0x1a5,0x000076,0x076,0x1aa,0x000077,0x077,0x1ae,0x000077,0x077,0x1ae,0x000078,0x078)
NTDLL(NtAlpcCreatePort                                  ,NONE, NONE, NONE, NONE, NONE,0x016,0x072,0x072,0x017,0x072,0x072,0x191,0x000077,0x077,0x196,0x000077,0x077,0x19e,0x000077,0x077,0x1a1,0x000077,0x077,0x1a3,0x000077,0x077,0x1a8,0x000078,0x078,0x1ac,0x000078,0x078,0x1ac,0x000079,0x079)
NTDLL(NtAlpcCreatePortSection                           ,NONE, NONE, NONE, NONE, NONE,0x017,0x073,0x073,0x018,0x073,0x073,0x190,0x000078,0x078,0x195,0x000078,0x078,0x19d,0x000078,0x078,0x1a0,0x000078,0x078,0x1a2,0x000078,0x078,0x1a7,0x000079,0x079,0x1ab,0x000079,0x079,0x1ab,0x00007a,0x07a)
NTDLL(NtAlpcCreateResourceReserve                       ,NONE, NONE, NONE, NONE, NONE,0x018,0x074,0x074,0x019,0x074,0x074,0x18f,0x000079,0x079,0x194,0x000079,0x079,0x19c,0x000079,0x079,0x19f,0x000079,0x079,0x1a1,0x000079,0x079,0x1a6,0x00007a,0x07a,0x1aa,0x00007a,0x07a,0x1aa,0x00007b,0x07b)
NTDLL(NtAlpcCreateSectionView                           ,NONE, NONE, NONE, NONE, NONE,0x019,0x075,0x075,0x01a,0x075,0x075,0x18e,0x00007a,0x07a,0x193,0x00007a,0x07a,0x19b,0x00007a,0x07a,0x19e,0x00007a,0x07a,0x1a0,0x00007a,0x07a,0x1a5,0x00007b,0x07b,0x1a9,0x00007b,0x07b,0x1a9,0x00007c,0x07c)
NTDLL(NtAlpcCreateSecurityContext                       ,NONE, NONE, NONE, NONE, NONE,0x01a,0x076,0x076,0x01b,0x076,0x076,0x18d,0x00007b,0x07b,0x192,0x00007b,0x07b,0x19a,0x00007b,0x07b,0x19d,0x00007b,0x07b,0x19f,0x00007b,0x07b,0x1a4,0x00007c,0x07c,0x1a8,0x00007c,0x07c,0x1a8,0x00007d,0x07d)
NTDLL(NtAlpcDeletePortSection                           ,NONE, NONE, NONE, NONE, NONE,0x01b,0x077,0x077,0x01c,0x077,0x077,0x18c,0x00007c,0x07c,0x191,0x00007c,0x07c,0x199,0x00007c,0x07c,0x19c,0x00007c,0x07c,0x19e,0x00007c,0x07c,0x1a3,0x00007d,0x07d,0x1a7,0x00007d,0x07d,0x1a7,0x00007e,0x07e)
NTDLL(NtAlpcDeleteResourceReserve                       ,NONE, NONE, NONE, NONE, NONE,0x01c,0x078,0x078,0x01d,0x078,0x078,0x18b,0x00007d,0x07d,0x190,0x00007d,0x07d,0x198,0x00007d,0x07d,0x19b,0x00007d,0x07d,0x19d,0x00007d,0x07d,0x1a2,0x00007e,0x07e,0x1a6,0x00007e,0x07e,0x1a6,0x00007f,0x07f)
NTDLL(NtAlpcDeleteSectionView                           ,NONE, NONE, NONE, NONE, NONE,0x01d,0x079,0x079,0x01e,0x079,0x079,0x18a,0x00007e,0x07e,0x18f,0x00007e,0x07e,0x197,0x00007e,0x07e,0x19a,0x00007e,0x07e,0x19c,0x00007e,0x07e,0x1a1,0x00007f,0x07f,0x1a5,0x00007f,0x07f,0x1a5,0x000080,0x080)
NTDLL(NtAlpcDeleteSecurityContext                       ,NONE, NONE, NONE, NONE, NONE,0x01e,0x07a,0x07a,0x01f,0x07a,0x07a,0x189,0x00007f,0x07f,0x18e,0x00007f,0x07f,0x196,0x00007f,0x07f,0x199,0x00007f,0x07f,0x19b,0x00007f,0x07f,0x1a0,0x000080,0x080,0x1a4,0x000080,0x080,0x1a4,0x000081,0x081)
NTDLL(NtAlpcDisconnectPort                              ,NONE, NONE, NONE, NONE, NONE,0x01f,0x07b,0x07b,0x020,0x07b,0x07b,0x188,0x000080,0x080,0x18d,0x000080,0x080,0x195,0x000080,0x080,0x198,0x000080,0x080,0x19a,0x000080,0x080,0x19f,0x000081,0x081,0x1a3,0x000081,0x081,0x1a3,0x000082,0x082)
NTDLL(NtAlpcImpersonateClientContainerOfPort            ,NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE,0x194,0x000081,0x081,0x197,0x000081,0x081,0x199,0x000081,0x081,0x19e,0x000082,0x082,0x1a2,0x000082,0x082,0x1a2,0x000083,0x083)
NTDLL(NtAlpcImpersonateClientOfPort                     ,NONE, NONE, NONE, NONE, NONE,0x020,0x07c,0x07c,0x021,0x07c,0x07c,0x187,0x000081,0x081,0x18c,0x000081,0x081,0x193,0x000082,0x082,0x196,0x000082,0x082,0x198,0x000082,0x082,0x19d,0x000083,0x083,0x1a1,0x000083,0x083,0x1a1,0x000084,0x084)
NTDLL(NtAlpcOpenSenderProcess                           ,NONE, NONE, NONE, NONE, NONE,0x021,0x07d,0x07d,0x022,0x07d,0x07d,0x186,0x000082,0x082,0x18b,0x000082,0x082,0x192,0x000083,0x083,0x195,0x000083,0x083,0x197,0x000083,0x083,0x19c,0x000084,0x084,0x1a0,0x000084,0x084,0x1a0,0x000085,0x085)
NTDLL(NtAlpcOpenSenderThread                            ,NONE, NONE, NONE, NONE, NONE,0x022,0x07e,0x07e,0x023,0x07e,0x07e,0x185,0x000083,0x083,0x18a,0x000083,0x083,0x191,0x000084,0x084,0x194,0x000084,0x084,0x196,0x000084,0x084,0x19b,0x000085,0x085,0x19f,0x000085,0x085,0x19f,0x000086,0x086)
NTDLL(NtAlpcQueryInformation                            ,NONE, NONE, NONE, NONE, NONE,0x023,0x07f,0x07f,0x024,0x07f,0x07f,0x184,0x000084,0x084,0x189,0x000084,0x084,0x190,0x000085,0x085,0x193,0x000085,0x085,0x195,0x000085,0x085,0x19a,0x000086,0x086,0x19e,0x000086,0x086,0x19e,0x000087,0x087)
NTDLL(NtAlpcQueryInformationMessage                     ,NONE, NONE, NONE, NONE, NONE,0x024,0x080,0x080,0x025,0x080,0x080,0x183,0x000085,0x085,0x188,0x000085,0x085,0x18f,0x000086,0x086,0x192,0x000086,0x086,0x194,0x000086,0x086,0x199,0x000087,0x087,0x19d,0x000087,0x087,0x19d,0x000088,0x088)
NTDLL(NtAlpcRevokeSecurityContext                       ,NONE, NONE, NONE, NONE, NONE,0x025,0x081,0x081,0x026,0x081,0x081,0x182,0x000086,0x086,0x187,0x000086,0x086,0x18e,0x000087,0x087,0x191,0x000087,0x087,0x193,0x000087,0x087,0x198,0x000088,0x088,0x19c,0x000088,0x088,0x19c,0x000089,0x089)
NTDLL(NtAlpcSendWaitReceivePort                         ,NONE, NONE, NONE, NONE, NONE,0x026,0x082,0x082,0x027,0x082,0x082,0x181,0x000087,0x087,0x186,0x000087,0x087,0x18d,0x000088,0x088,0x190,0x000088,0x088,0x192,0x000088,0x088,0x197,0x000089,0x089,0x19b,0x000089,0x089,0x19b,0x00008a,0x08a)
NTDLL(NtAlpcSetInformation                              ,NONE, NONE, NONE, NONE, NONE,0x027,0x083,0x083,0x028,0x083,0x083,0x180,0x000088,0x088,0x185,0x000088,0x088,0x18c,0x000089,0x089,0x18f,0x000089,0x089,0x191,0x000089,0x089,0x196,0x00008a,0x08a,0x19a,0x00008a,0x08a,0x19a,0x00008b,0x08b)
NTDLL(NtApphelpCacheControl                             ,NONE, NONE,0x013,0x049,0x049,0x028,0x049,0x049,0x029,0x049,0x049,0x17f,0x00004a,0x04a,0x184,0x00004b,0x04b,0x18b,0x00004c,0x04c,0x18e,0x00004c,0x04c,0x190,0x00004c,0x04c,0x195,0x00004c,0x04c,0x199,0x00004c,0x04c,0x199,0x00004c,0x04c)
NTDLL(NtAreMappedFilesTheSame                           ,0x11,0x012,0x014,0x06e,0x06e,0x029,0x084,0x084,0x02a,0x084,0x084,0x17e,0x050089,0x089,0x183,0x050089,0x089,0x18a,0x05008a,0x08a,0x18d,0x05008a,0x08a,0x18f,0x05008a,0x08a,0x194,0x05008b,0x08b,0x198,0x05008b,0x08b,0x198,0x05008c,0x08c)
NTDLL(NtAssignProcessToJobObject                        ,0x12,0x013,0x015,0x06f,0x06f,0x02a,0x085,0x085,0x02b,0x085,0x085,0x17d,0x08008a,0x08a,0x182,0x08008a,0x08a,0x189,0x08008b,0x08b,0x18c,0x08008b,0x08b,0x18e,0x08008b,0x08b,0x193,0x08008c,0x08c,0x197,0x08008c,0x08c,0x197,0x08008d,0x08d)
NTDLL(NtAssociateWaitCompletionPacket                   ,NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,0x17c,0x00008b,0x08b,0x181,0x00008b,0x08b,0x188,0x00008c,0x08c,0x18b,0x00008c,0x08c,0x18d,0x00008c,0x08c,0x192,0x00008d,0x08d,0x196,0x00008d,0x08d,0x196,0x00008e,0x08e)
NTDLL(NtCallEnclave                                     ,NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE,0x194,0x00008e,0x08e,0x194,0x00008f,0x08f)
NTDLL(NtCallbackReturn                                  ,0x13,0x014,0x016,0x002,0x002,0x02b,0x002,0x002,0x02c,0x002,0x002,0x17b,0x000003,0x003,0x180,0x000004,0x004,0x187,0x000005,0x005,0x18a,0x000005,0x005,0x18c,0x000005,0x005,0x191,0x000005,0x005,0x195,0x000005,0x005,0x195,0x000005,0x005)
NTDLL(NtCancelDeviceWakeupRequest                       ,0x16,0x015,0x017,0x070,0x070,0x02c,0x086,0x086, NONE, NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE)
NTDLL(NtCancelIoFile                                    ,0x14,0x016,0x018,0x05a,0x05a,0x02d,0x05a,0x05a,0x02d,0x05a,0x05a,0x17a,0x00005b,0x05b,0x17f,0x00005c,0x05c,0x186,0x00005d,0x05d,0x189,0x00005d,0x05d,0x18b,0x00005d,0x05d,0x190,0x00005d,0x05d,0x193,0x00005d,0x05d,0x193,0x00005d,0x05d)
NTDLL(NtCancelIoFileEx                                  ,NONE, NONE, NONE, NONE, NONE,0x171,0x087,0x087,0x02e,0x086,0x086,0x179,0x00008c,0x08c,0x17e,0x00008c,0x08c,0x185,0x00008d,0x08d,0x188,0x00008d,0x08d,0x18a,0x00008d,0x08d,0x18f,0x00008e,0x08e,0x192,0x00008f,0x08f,0x192,0x000090,0x090)
NTDLL(NtCancelSynchronousIoFile                         ,NONE, NONE, NONE, NONE, NONE,0x172,0x088,0x088,0x02f,0x087,0x087,0x178,0x00008d,0x08d,0x17d,0x00008d,0x08d,0x184,0x00008e,0x08e,0x187,0x00008e,0x08e,0x189,0x00008e,0x08e,0x18e,0x00008f,0x08f,0x191,0x000090,0x090,0x191,0x000091,0x091)
NTDLL(NtCancelTimer                                     ,0x15,0x017,0x019,0x05e,0x05e,0x02e,0x05e,0x05e,0x030,0x05e,0x05e,0x177,0x07005f,0x05f,0x17c,0x070060,0x060,0x183,0x070061,0x061,0x186,0x070061,0x061,0x188,0x070061,0x061,0x18d,0x070061,0x061,0x190,0x070061,0x061,0x190,0x070061,0x061)
NTDLL(NtCancelTimer2                                    ,NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,    NONE, NONE,0x043,0x00008e,0x08e,0x044,0x00008f,0x08f,0x044,0x00008f,0x08f,0x044,0x00008f,0x08f,0x044,0x000090,0x090,0x045,0x000091,0x091,0x045,0x000092,0x092)
NTDLL(NtCancelWaitCompletionPacket                      ,NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,0x176,0x00008e,0x08e,0x17b,0x00008f,0x08f,0x182,0x000090,0x090,0x185,0x000090,0x090,0x187,0x000090,0x090,0x18c,0x000091,0x091,0x18f,0x000092,0x092,0x18f,0x000093,0x093)
NTDLL(NtClearEvent                                      ,0x17,0x018,0x01a,0x03b,0x03b,0x02f,0x03b,0x03b,0x031,0x03b,0x03b,0x175,0x03003c,0x03c,0x17a,0x03003d,0x03d,0x181,0x03003e,0x03e,0x184,0x03003e,0x03e,0x186,0x03003e,0x03e,0x18b,0x03003e,0x03e,0x18e,0x03003e,0x03e,0x18e,0x03003e,0x03e)
NTDLL(NtClose                                           ,0x18,0x019,0x01b,0x00c,0x00c,0x030,0x00c,0x00c,0x032,0x00c,0x00c,0x174,0x00000d,0x00d,0x179,0x03000e,0x00e,0x180,0x03000f,0x00f,0x183,0x03000f,0x00f,0x185,0x03000f,0x00f,0x18a,0x03000f,0x00f,0x18d,0x03000f,0x00f,0x18d,0x03000f,0x00f)
NTDLL(NtCloseObjectAuditAlarm                           ,0x19,0x01a,0x01c,0x038,0x038,0x031,0x038,0x038,0x033,0x038,0x038,0x173,0x000039,0x039,0x178,0x00003a,0x03a,0x17f,0x00003b,0x03b,0x182,0x00003b,0x03b,0x184,0x00003b,0x03b,0x189,0x00003b,0x03b,0x18c,0x00003b,0x03b,0x18c,0x00003b,0x03b)
NTDLL(NtCommitComplete                                  ,NONE, NONE, NONE, NONE, NONE,0x05f,0x089,0x089,0x034,0x088,0x088,0x172,0x00008f,0x08f,0x177,0x000090,0x090,0x17e,0x000091,0x091,0x181,0x000091,0x091,0x183,0x000091,0x091,0x188,0x000092,0x092,0x18b,0x000093,0x093,0x18b,0x000094,0x094)
NTDLL(NtCommitEnlistment                                ,NONE, NONE, NONE, NONE, NONE,0x057,0x08a,0x08a,0x035,0x089,0x089,0x171,0x000090,0x090,0x176,0x000091,0x091,0x17d,0x000092,0x092,0x180,0x000092,0x092,0x182,0x000092,0x092,0x187,0x000093,0x093,0x18a,0x000094,0x094,0x18a,0x000095,0x095)
NTDLL(NtCommitRegistryTransaction                       ,NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE,0x1c0,0x000093,0x093,0x1c6,0x000094,0x094,0x1ca,0x000095,0x095,0x1ca,0x000096,0x096)
NTDLL(NtCommitTransaction                               ,NONE, NONE, NONE, NONE, NONE,0x05b,0x08b,0x08b,0x036,0x08a,0x08a,0x170,0x000091,0x091,0x175,0x000092,0x092,0x17c,0x000093,0x093,0x17f,0x000093,0x093,0x181,0x000094,0x094,0x186,0x000095,0x095,0x189,0x000096,0x096,0x189,0x000097,0x097)
NTDLL(NtCompactKeys                                     ,NONE,0x01b,0x01d,0x071,0x071,0x032,0x08c,0x08c,0x037,0x08b,0x08b,0x16f,0x000092,0x092,0x174,0x000093,0x093,0x17b,0x000094,0x094,0x17e,0x000094,0x094,0x180,0x000095,0x095,0x185,0x000096,0x096,0x188,0x000097,0x097,0x188,0x000098,0x098)
NTDLL(NtCompareObjects                                  ,NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE,0x17a,0x000095,0x095,0x17d,0x000095,0x095,0x17f,0x000096,0x096,0x184,0x000097,0x097,0x187,0x000098,0x098,0x187,0x000099,0x099)
NTDLL(NtCompareSigningLevels                            ,NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE,0x183,0x000098,0x098,0x186,0x000099,0x099,0x186,0x00009a,0x09a)
NTDLL(NtCompareTokens                                   ,NONE,0x01c,0x01e,0x072,0x072,0x033,0x08d,0x08d,0x038,0x08c,0x08c,0x16e,0x000093,0x093,0x173,0x000094,0x094,0x179,0x000096,0x096,0x17c,0x000096,0x096,0x17e,0x000097,0x097,0x182,0x000099,0x099,0x185,0x00009a,0x09a,0x185,0x00009b,0x09b)
NTDLL(NtCompleteConnectPort                             ,0x1a,0x01d,0x01f,0x073,0x073,0x034,0x08e,0x08e,0x039,0x08d,0x08d,0x16d,0x000094,0x094,0x172,0x000095,0x095,0x178,0x000097,0x097,0x17b,0x000097,0x097,0x17d,0x000098,0x098,0x181,0x00009a,0x09a,0x184,0x00009b,0x09b,0x184,0x00009c,0x09c)
NTDLL(NtCompressKey                                     ,NONE,0x01e,0x020,0x074,0x074,0x035,0x08f,0x08f,0x03a,0x08e,0x08e,0x16c,0x030095,0x095,0x171,0x030096,0x096,0x177,0x030098,0x098,0x17a,0x030098,0x098,0x17c,0x030099,0x099,0x180,0x03009b,0x09b,0x183,0x03009c,0x09c,0x183,0x03009d,0x09d)
NTDLL(NtConnectPort                                     ,0x1b,0x01f,0x021,0x075,0x075,0x036,0x090,0x090,0x03b,0x08f,0x08f,0x16b,0x000096,0x096,0x170,0x000097,0x097,0x176,0x000099,0x099,0x179,0x000099,0x099,0x17b,0x00009a,0x09a,0x17f,0x00009c,0x09c,0x182,0x00009d,0x09d,0x182,0x00009e,0x09e)
NTDLL(NtContinue                                        ,0x1c,0x020,0x022,0x040,0x040,0x037,0x040,0x040,0x03c,0x040,0x040,0x16a,0x000041,0x041,0x16f,0x000042,0x042,0x175,0x000043,0x043,0x178,0x000043,0x043,0x17a,0x000043,0x043,0x17e,0x000043,0x043,0x181,0x000043,0x043,0x181,0x000043,0x043)
NTDLL(NtConvertBetweenAuxiliaryCounterAndPerformanceCounter,NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE,0x17d,0x00009d,0x09d,0x180,0x00009e,0x09e,0x180,0x00009f,0x09f)
NTDLL(NtCreateChannel                                   ,0xf1, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE)
NTDLL(NtCreateDebugObject                               ,NONE,0x021,0x023,0x076,0x076,0x038,0x091,0x091,0x03d,0x090,0x090,0x169,0x000097,0x097,0x16e,0x000098,0x098,0x174,0x00009a,0x09a,0x177,0x00009a,0x09a,0x179,0x00009b,0x09b,0x17c,0x00009e,0x09e,0x17f,0x00009f,0x09f,0x17f,0x0000a0,0x0a0)
NTDLL(NtCreateDirectoryObject                           ,0x1d,0x022,0x024,0x077,0x077,0x039,0x092,0x092,0x03e,0x091,0x091,0x168,0x000098,0x098,0x16d,0x000099,0x099,0x173,0x00009b,0x09b,0x176,0x00009b,0x09b,0x178,0x00009c,0x09c,0x17b,0x00009f,0x09f,0x17e,0x0000a0,0x0a0,0x17e,0x0000a1,0x0a1)
NTDLL(NtCreateDirectoryObjectEx                         ,NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,0x167,0x000099,0x099,0x16c,0x00009a,0x09a,0x172,0x00009c,0x09c,0x175,0x00009c,0x09c,0x177,0x00009d,0x09d,0x17a,0x0000a0,0x0a0,0x17d,0x0000a1,0x0a1,0x17d,0x0000a2,0x0a2)
NTDLL(NtCreateEnclave                                   ,NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE,0x174,0x00009d,0x09d,0x176,0x00009e,0x09e,0x179,0x0000a1,0x0a1,0x17c,0x0000a2,0x0a2,0x17c,0x0000a3,0x0a3)
NTDLL(NtCreateEnlistment                                ,NONE, NONE, NONE, NONE, NONE,0x06f,0x093,0x093,0x03f,0x092,0x092,0x166,0x00009a,0x09a,0x16b,0x00009b,0x09b,0x171,0x00009d,0x09d,0x173,0x00009e,0x09e,0x175,0x00009f,0x09f,0x178,0x0000a2,0x0a2,0x17b,0x0000a3,0x0a3,0x17b,0x0000a4,0x0a4)
NTDLL(NtCreateEvent                                     ,0x1e,0x023,0x025,0x045,0x045,0x03a,0x045,0x045,0x040,0x045,0x045,0x165,0x000046,0x046,0x16a,0x000047,0x047,0x170,0x000048,0x048,0x172,0x000048,0x048,0x174,0x000048,0x048,0x177,0x000048,0x048,0x17a,0x000048,0x048,0x17a,0x000048,0x048)
NTDLL(NtCreateEventPair                                 ,0x1f,0x024,0x026,0x078,0x078,0x03b,0x094,0x094,0x041,0x093,0x093,0x164,0x00009b,0x09b,0x169,0x00009c,0x09c,0x16f,0x00009e,0x09e,0x171,0x00009f,0x09f,0x173,0x0000a0,0x0a0,0x176,0x0000a3,0x0a3,0x179,0x0000a4,0x0a4,0x179,0x0000a5,0x0a5)
NTDLL(NtCreateFile                                      ,0x20,0x025,0x027,0x052,0x052,0x03c,0x052,0x052,0x042,0x052,0x052,0x163,0x000053,0x053,0x168,0x000054,0x054,0x16e,0x000055,0x055,0x170,0x000055,0x055,0x172,0x000055,0x055,0x175,0x000055,0x055,0x178,0x000055,0x055,0x178,0x000055,0x055)
NTDLL(NtCreateIRTimer                                   ,NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,0x161,0x00009c,0x09c,0x165,0x00009d,0x09d,0x16b,0x00009f,0x09f,0x16d,0x0000a0,0x0a0,0x16f,0x0000a1,0x0a1,0x172,0x0000a4,0x0a4,0x175,0x0000a5,0x0a5,0x175,0x0000a6,0x0a6)
NTDLL(NtCreateIoCompletion                              ,0x21,0x026,0x028,0x079,0x079,0x03d,0x095,0x095,0x043,0x094,0x094,0x162,0x00009d,0x09d,0x167,0x00009e,0x09e,0x16d,0x0000a0,0x0a0,0x16f,0x0000a1,0x0a1,0x171,0x0000a2,0x0a2,0x174,0x0000a5,0x0a5,0x177,0x0000a6,0x0a6,0x177,0x0000a7,0x0a7)
NTDLL(NtCreateJobObject                                 ,0x22,0x027,0x029,0x07a,0x07a,0x03e,0x096,0x096,0x044,0x095,0x095,0x160,0x00009e,0x09e,0x164,0x00009f,0x09f,0x16a,0x0000a1,0x0a1,0x16c,0x0000a2,0x0a2,0x16e,0x0000a3,0x0a3,0x171,0x0000a6,0x0a6,0x174,0x0000a7,0x0a7,0x174,0x0000a8,0x0a8)
NTDLL(NtCreateJobSet                                    ,NONE,0x028,0x02a,0x07b,0x07b,0x03f,0x097,0x097,0x045,0x096,0x096,0x15f,0x00009f,0x09f,0x163,0x0000a0,0x0a0,0x169,0x0000a2,0x0a2,0x16b,0x0000a3,0x0a3,0x16d,0x0000a4,0x0a4,0x170,0x0000a7,0x0a7,0x173,0x0000a8,0x0a8,0x173,0x0000a9,0x0a9)
NTDLL(NtCreateKey                                       ,0x23,0x029,0x02b,0x01a,0x01a,0x040,0x01a,0x01a,0x046,0x01a,0x01a,0x15e,0x00001b,0x01b,0x162,0x00001c,0x01c,0x168,0x00001d,0x01d,0x16a,0x00001d,0x01d,0x16c,0x00001d,0x01d,0x16f,0x00001d,0x01d,0x172,0x00001d,0x01d,0x172,0x00001d,0x01d)
NTDLL(NtCreateKeyTransacted                             ,NONE, NONE, NONE, NONE, NONE,0x041,0x098,0x098,0x048,0x097,0x097,0x15c,0x0000a0,0x0a0,0x160,0x0000a1,0x0a1,0x166,0x0000a3,0x0a3,0x168,0x0000a4,0x0a4,0x16a,0x0000a5,0x0a5,0x16d,0x0000a8,0x0a8,0x170,0x0000a9,0x0a9,0x170,0x0000aa,0x0aa)
NTDLL(NtCreateKeyedEvent                                ,NONE,0x117,0x121,0x07c,0x07c,0x168,0x099,0x099,0x047,0x098,0x098,0x15d,0x0000a1,0x0a1,0x161,0x0000a2,0x0a2,0x167,0x0000a4,0x0a4,0x169,0x0000a5,0x0a5,0x16b,0x0000a6,0x0a6,0x16e,0x0000a9,0x0a9,0x171,0x0000aa,0x0aa,0x171,0x0000ab,0x0ab)
NTDLL(NtCreateLowBoxToken                               ,NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,0x149,0x0000a2,0x0a2,0x14d,0x0000a3,0x0a3,0x153,0x0000a5,0x0a5,0x155,0x0000a6,0x0a6,0x157,0x0000a7,0x0a7,0x15a,0x0000aa,0x0aa,0x15d,0x0000ab,0x0ab,0x15d,0x0000ac,0x0ac)
NTDLL(NtCreateMailslotFile                              ,0x24,0x02a,0x02c,0x07d,0x07d,0x042,0x09a,0x09a,0x049,0x099,0x099,0x15b,0x0000a3,0x0a3,0x15f,0x0000a4,0x0a4,0x165,0x0000a6,0x0a6,0x167,0x0000a7,0x0a7,0x169,0x0000a8,0x0a8,0x16c,0x0000ab,0x0ab,0x16f,0x0000ac,0x0ac,0x16f,0x0000ad,0x0ad)
NTDLL(NtCreateMutant                                    ,0x25,0x02b,0x02d,0x07e,0x07e,0x043,0x09b,0x09b,0x04a,0x09a,0x09a,0x15a,0x0000a4,0x0a4,0x15e,0x0000a5,0x0a5,0x164,0x0000a7,0x0a7,0x166,0x0000a8,0x0a8,0x168,0x0000a9,0x0a9,0x16b,0x0000ac,0x0ac,0x16e,0x0000ad,0x0ad,0x16e,0x0000ae,0x0ae)
NTDLL(NtCreateNamedPipeFile                             ,0x26,0x02c,0x02e,0x07f,0x07f,0x044,0x09c,0x09c,0x04b,0x09b,0x09b,0x159,0x0000a5,0x0a5,0x15d,0x0000a6,0x0a6,0x163,0x0000a8,0x0a8,0x165,0x0000a9,0x0a9,0x167,0x0000aa,0x0aa,0x16a,0x0000ad,0x0ad,0x16d,0x0000ae,0x0ae,0x16d,0x0000af,0x0af)
NTDLL(NtCreatePagingFile                                ,0x27,0x02d,0x02f,0x080,0x080,0x046,0x09d,0x09d,0x04c,0x09c,0x09c,0x158,0x0000a6,0x0a6,0x15c,0x0000a7,0x0a7,0x162,0x0000a9,0x0a9,0x164,0x0000aa,0x0aa,0x166,0x0000ab,0x0ab,0x169,0x0000ae,0x0ae,0x16c,0x0000af,0x0af,0x16c,0x0000b0,0x0b0)
NTDLL(NtCreatePartition                                 ,NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE,0x14a,0x0000aa,0x0aa,0x14c,0x0000ab,0x0ab,0x14e,0x0000ac,0x0ac,0x151,0x0000af,0x0af,0x154,0x0000b0,0x0b0,0x154,0x0000b1,0x0b1)
NTDLL(NtCreatePort                                      ,0x28,0x02e,0x030,0x081,0x081,0x047,0x09e,0x09e,0x04d,0x09d,0x09d,0x157,0x0000a7,0x0a7,0x15b,0x0000a8,0x0a8,0x161,0x0000ab,0x0ab,0x163,0x0000ac,0x0ac,0x165,0x0000ad,0x0ad,0x168,0x0000b0,0x0b0,0x16b,0x0000b1,0x0b1,0x16b,0x0000b2,0x0b2)
NTDLL(NtCreatePrivateNamespace                          ,NONE, NONE, NONE, NONE, NONE,0x045,0x09f,0x09f,0x04e,0x09e,0x09e,0x156,0x0000a8,0x0a8,0x15a,0x0000a9,0x0a9,0x160,0x0000ac,0x0ac,0x162,0x0000ad,0x0ad,0x164,0x0000ae,0x0ae,0x167,0x0000b1,0x0b1,0x16a,0x0000b2,0x0b2,0x16a,0x0000b3,0x0b3)
NTDLL(NtCreateProcess                                   ,0x29,0x02f,0x031,0x082,0x082,0x048,0x0a0,0x0a0,0x04f,0x09f,0x09f,0x155,0x0000a9,0x0a9,0x159,0x0000aa,0x0aa,0x15f,0x0000ad,0x0ad,0x161,0x0000ae,0x0ae,0x163,0x0000af,0x0af,0x166,0x0000b2,0x0b2,0x169,0x0000b3,0x0b3,0x169,0x0000b4,0x0b4)
NTDLL(NtCreateProcessEx                                 ,NONE,0x030,0x032,0x04a,0x04a,0x049,0x04a,0x04a,0x050,0x04a,0x04a,0x154,0x00004b,0x04b,0x158,0x00004c,0x04c,0x15e,0x00004d,0x04d,0x160,0x00004d,0x04d,0x162,0x00004d,0x04d,0x165,0x00004d,0x04d,0x168,0x00004d,0x04d,0x168,0x00004d,0x04d)
NTDLL(NtCreateProfile                                   ,0x2a,0x031,0x033,0x083,0x083,0x04a,0x0a1,0x0a1,0x051,0x0a0,0x0a0,0x153,0x0000aa,0x0aa,0x157,0x0000ab,0x0ab,0x15d,0x0000ae,0x0ae,0x15f,0x0000af,0x0af,0x161,0x0000b0,0x0b0,0x164,0x0000b3,0x0b3,0x167,0x0000b4,0x0b4,0x167,0x0000b5,0x0b5)
NTDLL(NtCreateProfileEx                                 ,NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,0x052,0x0a1,0x0a1,0x152,0x0000ab,0x0ab,0x156,0x0000ac,0x0ac,0x15c,0x0000af,0x0af,0x15e,0x0000b0,0x0b0,0x160,0x0000b1,0x0b1,0x163,0x0000b4,0x0b4,0x166,0x0000b5,0x0b5,0x166,0x0000b6,0x0b6)
NTDLL(NtCreateRegistryTransaction                       ,NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE,0x1be,0x0000b2,0x0b2,0x1c4,0x0000b5,0x0b5,0x1c8,0x0000b6,0x0b6,0x1c8,0x0000b7,0x0b7)
NTDLL(NtCreateResourceManager                           ,NONE, NONE, NONE, NONE, NONE,0x06b,0x0a2,0x0a2,0x053,0x0a2,0x0a2,0x151,0x0000ac,0x0ac,0x155,0x0000ad,0x0ad,0x15b,0x0000b0,0x0b0,0x15d,0x0000b1,0x0b1,0x15f,0x0000b3,0x0b3,0x162,0x0000b6,0x0b6,0x165,0x0000b7,0x0b7,0x165,0x0000b8,0x0b8)
NTDLL(NtCreateSection                                   ,0x2b,0x032,0x034,0x047,0x047,0x04b,0x047,0x047,0x054,0x047,0x047,0x150,0x000048,0x048,0x154,0x000049,0x049,0x15a,0x00004a,0x04a,0x15c,0x00004a,0x04a,0x15e,0x00004a,0x04a,0x161,0x00004a,0x04a,0x164,0x00004a,0x04a,0x164,0x00004a,0x04a)
NTDLL(NtCreateSemaphore                                 ,0x2c,0x033,0x035,0x084,0x084,0x04c,0x0a3,0x0a3,0x055,0x0a3,0x0a3,0x14f,0x0000ad,0x0ad,0x153,0x0000ae,0x0ae,0x159,0x0000b1,0x0b1,0x15b,0x0000b2,0x0b2,0x15d,0x0000b4,0x0b4,0x160,0x0000b7,0x0b7,0x163,0x0000b8,0x0b8,0x163,0x0000b9,0x0b9)
NTDLL(NtCreateSymbolicLinkObject                        ,0x2d,0x034,0x036,0x085,0x085,0x04d,0x0a4,0x0a4,0x056,0x0a4,0x0a4,0x14e,0x0000ae,0x0ae,0x152,0x0000af,0x0af,0x158,0x0000b2,0x0b2,0x15a,0x0000b3,0x0b3,0x15c,0x0000b5,0x0b5,0x15f,0x0000b8,0x0b8,0x162,0x0000b9,0x0b9,0x162,0x0000ba,0x0ba)
NTDLL(NtCreateThread                                    ,0x2e,0x035,0x037,0x04b,0x04b,0x04e,0x04b,0x04b,0x057,0x04b,0x04b,0x14d,0x00004c,0x04c,0x151,0x00004d,0x04d,0x157,0x00004e,0x04e,0x159,0x00004e,0x04e,0x15b,0x00004e,0x04e,0x15e,0x00004e,0x04e,0x161,0x00004e,0x04e,0x161,0x00004e,0x04e)
NTDLL(NtCreateThreadEx                                  ,NONE, NONE, NONE, NONE, NONE,0x17e,0x0a5,0x0a5,0x058,0x0a5,0x0a5,0x14c,0x0000af,0x0af,0x150,0x0000b0,0x0b0,0x156,0x0000b3,0x0b3,0x158,0x0000b4,0x0b4,0x15a,0x0000b6,0x0b6,0x15d,0x0000b9,0x0b9,0x160,0x0000ba,0x0ba,0x160,0x0000bb,0x0bb)
NTDLL(NtCreateTimer                                     ,0x2f,0x036,0x038,0x086,0x086,0x04f,0x0a6,0x0a6,0x059,0x0a6,0x0a6,0x14b,0x0000b0,0x0b0,0x14f,0x0000b1,0x0b1,0x155,0x0000b4,0x0b4,0x157,0x0000b5,0x0b5,0x159,0x0000b7,0x0b7,0x15c,0x0000ba,0x0ba,0x15f,0x0000bb,0x0bb,0x15f,0x0000bc,0x0bc)
NTDLL(NtCreateTimer2                                    ,NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,    NONE, NONE,0x166,0x0000b2,0x0b2,0x16c,0x0000b5,0x0b5,0x16e,0x0000b6,0x0b6,0x170,0x0000b8,0x0b8,0x173,0x0000bb,0x0bb,0x176,0x0000bc,0x0bc,0x176,0x0000bd,0x0bd)
NTDLL(NtCreateToken                                     ,0x30,0x037,0x039,0x087,0x087,0x050,0x0a7,0x0a7,0x05a,0x0a7,0x0a7,0x148,0x0000b1,0x0b1,0x14c,0x0000b3,0x0b3,0x152,0x0000b6,0x0b6,0x154,0x0000b7,0x0b7,0x156,0x0000b9,0x0b9,0x159,0x0000bc,0x0bc,0x15c,0x0000bd,0x0bd,0x15c,0x0000be,0x0be)
NTDLL(NtCreateTokenEx                                   ,NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,0x14a,0x0000b2,0x0b2,0x14e,0x0000b4,0x0b4,0x154,0x0000b7,0x0b7,0x156,0x0000b8,0x0b8,0x158,0x0000ba,0x0ba,0x15b,0x0000bd,0x0bd,0x15e,0x0000be,0x0be,0x15e,0x0000bf,0x0bf)
NTDLL(NtCreateTransaction                               ,NONE, NONE, NONE, NONE, NONE,0x051,0x0a8,0x0a8,0x05b,0x0a8,0x0a8,0x147,0x0000b3,0x0b3,0x14b,0x0000b5,0x0b5,0x151,0x0000b8,0x0b8,0x153,0x0000b9,0x0b9,0x155,0x0000bb,0x0bb,0x158,0x0000be,0x0be,0x15b,0x0000bf,0x0bf,0x15b,0x0000c0,0x0c0)
NTDLL(NtCreateTransactionManager                        ,NONE, NONE, NONE, NONE, NONE,0x064,0x0a9,0x0a9,0x05c,0x0a9,0x0a9,0x146,0x0000b4,0x0b4,0x14a,0x0000b6,0x0b6,0x150,0x0000b9,0x0b9,0x152,0x0000ba,0x0ba,0x154,0x0000bc,0x0bc,0x157,0x0000bf,0x0bf,0x15a,0x0000c0,0x0c0,0x15a,0x0000c1,0x0c1)
NTDLL(NtCreateUserProcess                               ,NONE, NONE, NONE, NONE, NONE,0x17f,0x0aa,0x0aa,0x05d,0x0aa,0x0aa,0x145,0x0000b5,0x0b5,0x149,0x0000b7,0x0b7,0x14f,0x0000ba,0x0ba,0x151,0x0000bb,0x0bb,0x153,0x0000bd,0x0bd,0x156,0x0000c0,0x0c0,0x159,0x0000c1,0x0c1,0x159,0x0000c2,0x0c2)
NTDLL(NtCreateWaitCompletionPacket                      ,NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,0x143,0x0000b6,0x0b6,0x147,0x0000b8,0x0b8,0x14d,0x0000bb,0x0bb,0x14f,0x0000bc,0x0bc,0x151,0x0000be,0x0be,0x154,0x0000c1,0x0c1,0x157,0x0000c2,0x0c2,0x157,0x0000c3,0x0c3)
NTDLL(NtCreateWaitablePort                              ,0x31,0x038,0x03a,0x088,0x088,0x073,0x0ab,0x0ab,0x05e,0x0ab,0x0ab,0x144,0x0000b7,0x0b7,0x148,0x0000b9,0x0b9,0x14e,0x0000bc,0x0bc,0x150,0x0000bd,0x0bd,0x152,0x0000bf,0x0bf,0x155,0x0000c2,0x0c2,0x158,0x0000c3,0x0c3,0x158,0x0000c4,0x0c4)
NTDLL(NtCreateWnfStateName                              ,NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,0x142,0x0000b8,0x0b8,0x146,0x0000ba,0x0ba,0x14c,0x0000bd,0x0bd,0x14e,0x0000be,0x0be,0x150,0x0000c0,0x0c0,0x153,0x0000c3,0x0c3,0x156,0x0000c4,0x0c4,0x156,0x0000c5,0x0c5)
NTDLL(NtCreateWorkerFactory                             ,NONE, NONE, NONE, NONE, NONE,0x177,0x0ac,0x0ac,0x05f,0x0ac,0x0ac,0x141,0x0000b9,0x0b9,0x145,0x0000bb,0x0bb,0x14b,0x0000be,0x0be,0x14d,0x0000bf,0x0bf,0x14f,0x0000c1,0x0c1,0x152,0x0000c4,0x0c4,0x155,0x0000c5,0x0c5,0x155,0x0000c6,0x0c6)
NTDLL(NtDebugActiveProcess                              ,NONE,0x039,0x03b,0x089,0x089,0x074,0x0ad,0x0ad,0x060,0x0ad,0x0ad,0x140,0x0800ba,0x0ba,0x144,0x0800bc,0x0bc,0x149,0x0800bf,0x0bf,0x14b,0x0800c0,0x0c0,0x14d,0x0800c2,0x0c2,0x150,0x0800c5,0x0c5,0x153,0x0800c6,0x0c6,0x153,0x0800c7,0x0c7)
NTDLL(NtDebugContinue                                   ,NONE,0x03a,0x03c,0x08a,0x08a,0x075,0x0ae,0x0ae,0x061,0x0ae,0x0ae,0x13f,0x0000bb,0x0bb,0x143,0x0000bd,0x0bd,0x148,0x0000c0,0x0c0,0x14a,0x0000c1,0x0c1,0x14c,0x0000c3,0x0c3,0x14f,0x0000c6,0x0c6,0x152,0x0000c7,0x0c7,0x152,0x0000c8,0x0c8)
NTDLL(NtDelayExecution                                  ,0x32,0x03b,0x03d,0x031,0x031,0x076,0x031,0x031,0x062,0x031,0x031,0x13e,0x060032,0x032,0x142,0x060033,0x033,0x147,0x060034,0x034,0x149,0x060034,0x034,0x14b,0x060034,0x034,0x14e,0x060034,0x034,0x151,0x060034,0x034,0x151,0x060034,0x034)
NTDLL(NtDeleteAtom                                      ,0x33,0x03c,0x03e,0x08b,0x08b,0x077,0x0af,0x0af,0x063,0x0af,0x0af,0x13d,0x0400bc,0x0bc,0x141,0x0400be,0x0be,0x146,0x0400c1,0x0c1,0x148,0x0400c2,0x0c2,0x14a,0x0400c4,0x0c4,0x14d,0x0400c7,0x0c7,0x150,0x0400c8,0x0c8,0x150,0x0400c9,0x0c9)
NTDLL(NtDeleteBootEntry                                 ,NONE,0x03d,0x03f,0x08c,0x08c,0x078,0x0b0,0x0b0,0x064,0x0b0,0x0b0,0x13c,0x0000bd,0x0bd,0x140,0x0000bf,0x0bf,0x145,0x0000c2,0x0c2,0x147,0x0000c3,0x0c3,0x149,0x0000c5,0x0c5,0x14c,0x0000c8,0x0c8,0x14f,0x0000c9,0x0c9,0x14f,0x0000ca,0x0ca)
NTDLL(NtDeleteDriverEntry                               ,NONE, NONE,0x040,0x08d,0x08d,0x079,0x0b1,0x0b1,0x065,0x0b1,0x0b1,0x13b,0x0000be,0x0be,0x13f,0x0000c0,0x0c0,0x144,0x0000c3,0x0c3,0x146,0x0000c4,0x0c4,0x148,0x0000c6,0x0c6,0x14b,0x0000c9,0x0c9,0x14e,0x0000ca,0x0ca,0x14e,0x0000cb,0x0cb)
NTDLL(NtDeleteFile                                      ,0x34,0x03e,0x041,0x08e,0x08e,0x07a,0x0b2,0x0b2,0x066,0x0b2,0x0b2,0x13a,0x0000bf,0x0bf,0x13e,0x0000c1,0x0c1,0x143,0x0000c4,0x0c4,0x145,0x0000c5,0x0c5,0x147,0x0000c7,0x0c7,0x14a,0x0000ca,0x0ca,0x14d,0x0000cb,0x0cb,0x14d,0x0000cc,0x0cc)
NTDLL(NtDeleteKey                                       ,0x35,0x03f,0x042,0x08f,0x08f,0x07b,0x0b3,0x0b3,0x067,0x0b3,0x0b3,0x139,0x0000c0,0x0c0,0x13d,0x0000c2,0x0c2,0x142,0x0000c5,0x0c5,0x144,0x0000c6,0x0c6,0x146,0x0000c8,0x0c8,0x149,0x0000cb,0x0cb,0x14c,0x0000cc,0x0cc,0x14c,0x0000cd,0x0cd)
NTDLL(NtDeleteObjectAuditAlarm                          ,0x36,0x040,0x043,0x090,0x090,0x07d,0x0b4,0x0b4,0x068,0x0b4,0x0b4,0x138,0x0000c1,0x0c1,0x13c,0x0000c3,0x0c3,0x141,0x0000c6,0x0c6,0x143,0x0000c7,0x0c7,0x145,0x0000c9,0x0c9,0x148,0x0000cc,0x0cc,0x14b,0x0000cd,0x0cd,0x14b,0x0000ce,0x0ce)
NTDLL(NtDeletePrivateNamespace                          ,NONE, NONE, NONE, NONE, NONE,0x07c,0x0b5,0x0b5,0x069,0x0b5,0x0b5,0x137,0x0000c2,0x0c2,0x13b,0x0000c4,0x0c4,0x140,0x0000c7,0x0c7,0x142,0x0000c8,0x0c8,0x144,0x0000ca,0x0ca,0x147,0x0000cd,0x0cd,0x14a,0x0000ce,0x0ce,0x14a,0x0000cf,0x0cf)
NTDLL(NtDeleteValueKey                                  ,0x37,0x041,0x044,0x091,0x091,0x07e,0x0b6,0x0b6,0x06a,0x0b6,0x0b6,0x136,0x0000c3,0x0c3,0x13a,0x0000c5,0x0c5,0x13f,0x0000c8,0x0c8,0x141,0x0000c9,0x0c9,0x143,0x0000cb,0x0cb,0x146,0x0000ce,0x0ce,0x149,0x0000cf,0x0cf,0x149,0x0000d0,0x0d0)
NTDLL(NtDeleteWnfStateData                              ,NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,0x135,0x0000c4,0x0c4,0x139,0x0000c6,0x0c6,0x13e,0x0000c9,0x0c9,0x140,0x0000ca,0x0ca,0x142,0x0000cc,0x0cc,0x145,0x0000cf,0x0cf,0x148,0x0000d0,0x0d0,0x148,0x0000d1,0x0d1)
NTDLL(NtDeleteWnfStateName                              ,NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,0x134,0x0000c5,0x0c5,0x138,0x0000c7,0x0c7,0x13d,0x0000ca,0x0ca,0x13f,0x0000cb,0x0cb,0x141,0x0000cd,0x0cd,0x144,0x0000d0,0x0d0,0x147,0x0000d1,0x0d1,0x147,0x0000d2,0x0d2)
NTDLL(NtDeviceIoControlFile                             ,0x38,0x042,0x045,0x004,0x004,0x07f,0x004,0x004,0x06b,0x004,0x004,0x133,0x1b0005,0x005,0x137,0x1b0006,0x006,0x13c,0x1b0007,0x007,0x13e,0x1b0007,0x007,0x140,0x1b0007,0x007,0x143,0x1b0007,0x007,0x146,0x1b0007,0x007,0x146,0x1b0007,0x007)
NTDLL(NtDisableLastKnownGood                            ,NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,0x06c,0x0b7,0x0b7,0x132,0x0000c6,0x0c6,0x136,0x0000c8,0x0c8,0x13b,0x0000cb,0x0cb,0x13d,0x0000cc,0x0cc,0x13f,0x0000ce,0x0ce,0x142,0x0000d1,0x0d1,0x145,0x0000d2,0x0d2,0x145,0x0000d3,0x0d3)
NTDLL(NtDisplayString                                   ,0x39,0x043,0x046,0x092,0x092,0x080,0x0b7,0x0b7,0x06d,0x0b8,0x0b8,0x131,0x0000c7,0x0c7,0x135,0x0000c9,0x0c9,0x13a,0x0000cc,0x0cc,0x13c,0x0000cd,0x0cd,0x13e,0x0000cf,0x0cf,0x141,0x0000d2,0x0d2,0x144,0x0000d3,0x0d3,0x144,0x0000d4,0x0d4)
NTDLL(NtDrawText                                        ,NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,0x06e,0x0b9,0x0b9,0x130,0x0000c8,0x0c8,0x134,0x0000ca,0x0ca,0x139,0x0000cd,0x0cd,0x13b,0x0000ce,0x0ce,0x13d,0x0000d0,0x0d0,0x140,0x0000d3,0x0d3,0x143,0x0000d4,0x0d4,0x143,0x0000d5,0x0d5)
NTDLL(NtDuplicateObject                                 ,0x3a,0x044,0x047,0x039,0x039,0x081,0x039,0x039,0x06f,0x039,0x039,0x12f,0x00003a,0x03a,0x133,0x00003b,0x03b,0x138,0x00003c,0x03c,0x13a,0x00003c,0x03c,0x13c,0x00003c,0x03c,0x13f,0x00003c,0x03c,0x142,0x00003c,0x03c,0x142,0x00003c,0x03c)
NTDLL(NtDuplicateToken                                  ,0x3b,0x045,0x048,0x03f,0x03f,0x082,0x03f,0x03f,0x070,0x03f,0x03f,0x12e,0x000040,0x040,0x132,0x000041,0x041,0x137,0x000042,0x042,0x139,0x000042,0x042,0x13b,0x000042,0x042,0x13e,0x000042,0x042,0x141,0x000042,0x042,0x141,0x000042,0x042)
NTDLL(NtEnableLastKnownGood                             ,NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,0x071,0x0ba,0x0ba,0x12d,0x0000c9,0x0c9,0x131,0x0000cb,0x0cb,0x136,0x0000ce,0x0ce,0x138,0x0000cf,0x0cf,0x13a,0x0000d1,0x0d1,0x13d,0x0000d4,0x0d4,0x140,0x0000d5,0x0d5,0x140,0x0000d6,0x0d6)
NTDLL(NtEnumerateBootEntries                            ,NONE,0x046,0x049,0x093,0x093,0x083,0x0b8,0x0b8,0x072,0x0bb,0x0bb,0x12c,0x0000ca,0x0ca,0x130,0x0000cc,0x0cc,0x135,0x0000cf,0x0cf,0x137,0x0000d0,0x0d0,0x139,0x0000d2,0x0d2,0x13c,0x0000d5,0x0d5,0x13f,0x0000d6,0x0d6,0x13f,0x0000d7,0x0d7)
NTDLL(NtEnumerateDriverEntries                          ,NONE, NONE,0x04a,0x094,0x094,0x084,0x0b9,0x0b9,0x073,0x0bc,0x0bc,0x12b,0x0000cb,0x0cb,0x12f,0x0000cd,0x0cd,0x134,0x0000d0,0x0d0,0x136,0x0000d1,0x0d1,0x138,0x0000d3,0x0d3,0x13b,0x0000d6,0x0d6,0x13e,0x0000d7,0x0d7,0x13e,0x0000d8,0x0d8)
NTDLL(NtEnumerateKey                                    ,0x3c,0x047,0x04b,0x02f,0x02f,0x085,0x02f,0x02f,0x074,0x02f,0x02f,0x12a,0x000030,0x030,0x12e,0x000031,0x031,0x133,0x000032,0x032,0x135,0x000032,0x032,0x137,0x000032,0x032,0x13a,0x000032,0x032,0x13d,0x000032,0x032,0x13d,0x000032,0x032)
NTDLL(NtEnumerateSystemEnvironmentValuesEx              ,NONE,0x048,0x04c,0x095,0x095,0x086,0x0ba,0x0ba,0x075,0x0bd,0x0bd,0x129,0x0000cc,0x0cc,0x12d,0x0000ce,0x0ce,0x132,0x0000d1,0x0d1,0x134,0x0000d2,0x0d2,0x136,0x0000d4,0x0d4,0x139,0x0000d7,0x0d7,0x13c,0x0000d8,0x0d8,0x13c,0x0000d9,0x0d9)
NTDLL(NtEnumerateTransactionObject                      ,NONE, NONE, NONE, NONE, NONE,0x087,0x0bb,0x0bb,0x076,0x0be,0x0be,0x128,0x0000cd,0x0cd,0x12c,0x0000cf,0x0cf,0x131,0x0000d2,0x0d2,0x133,0x0000d3,0x0d3,0x135,0x0000d5,0x0d5,0x138,0x0000d8,0x0d8,0x13b,0x0000d9,0x0d9,0x13b,0x0000da,0x0da)
NTDLL(NtEnumerateValueKey                               ,0x3d,0x049,0x04d,0x010,0x010,0x088,0x010,0x010,0x077,0x010,0x010,0x127,0x000011,0x011,0x12b,0x000012,0x012,0x130,0x000013,0x013,0x132,0x000013,0x013,0x134,0x000013,0x013,0x137,0x000013,0x013,0x13a,0x000013,0x013,0x13a,0x000013,0x013)
NTDLL(NtExtendSection                                   ,0x3e,0x04a,0x04e,0x096,0x096,0x089,0x0bc,0x0bc,0x078,0x0bf,0x0bf,0x126,0x0000ce,0x0ce,0x12a,0x0000d0,0x0d0,0x12f,0x0000d3,0x0d3,0x131,0x0000d4,0x0d4,0x133,0x0000d6,0x0d6,0x136,0x0000d9,0x0d9,0x139,0x0000da,0x0da,0x139,0x0000db,0x0db)
NTDLL(NtFilterBootOption                                ,NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,0x125,0x0000cf,0x0cf,0x129,0x0000d1,0x0d1,0x12e,0x0000d4,0x0d4,0x130,0x0000d5,0x0d5,0x132,0x0000d7,0x0d7,0x135,0x0000da,0x0da,0x138,0x0000db,0x0db,0x138,0x0000dc,0x0dc)
NTDLL(NtFilterToken                                     ,0x3f,0x04b,0x04f,0x097,0x097,0x08a,0x0bd,0x0bd,0x079,0x0c0,0x0c0,0x123,0x0000d0,0x0d0,0x127,0x0000d2,0x0d2,0x12c,0x0000d5,0x0d5,0x12e,0x0000d6,0x0d6,0x130,0x0000d8,0x0d8,0x133,0x0000db,0x0db,0x136,0x0000dc,0x0dc,0x136,0x0000dd,0x0dd)
NTDLL(NtFilterTokenEx                                   ,NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,0x124,0x0000d1,0x0d1,0x128,0x0000d3,0x0d3,0x12d,0x0000d6,0x0d6,0x12f,0x0000d7,0x0d7,0x131,0x0000d9,0x0d9,0x134,0x0000dc,0x0dc,0x137,0x0000dd,0x0dd,0x137,0x0000de,0x0de)
NTDLL(NtFindAtom                                        ,0x40,0x04c,0x050,0x011,0x011,0x08b,0x011,0x011,0x07a,0x011,0x011,0x122,0x0a0012,0x012,0x126,0x0a0013,0x013,0x12b,0x0a0014,0x014,0x12d,0x0a0014,0x014,0x12f,0x0a0014,0x014,0x132,0x0a0014,0x014,0x135,0x0a0014,0x014,0x135,0x0a0014,0x014)
NTDLL(NtFlushBuffersFile                                ,0x41,0x04d,0x051,0x048,0x048,0x08c,0x048,0x048,0x07b,0x048,0x048,0x120,0x000049,0x049,0x124,0x00004a,0x04a,0x129,0x00004b,0x04b,0x12b,0x00004b,0x04b,0x12d,0x00004b,0x04b,0x130,0x00004b,0x04b,0x133,0x00004b,0x04b,0x133,0x00004b,0x04b)
NTDLL(NtFlushBuffersFileEx                              ,NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,0x121,0x0000d2,0x0d2,0x125,0x0000d4,0x0d4,0x12a,0x0000d7,0x0d7,0x12c,0x0000d8,0x0d8,0x12e,0x0000da,0x0da,0x131,0x0000dd,0x0dd,0x134,0x0000de,0x0de,0x134,0x0000df,0x0df)
NTDLL(NtFlushInstallUILanguage                          ,NONE, NONE, NONE, NONE, NONE,0x183,0x0be,0x0be,0x07c,0x0c1,0x0c1,0x11f,0x0000d3,0x0d3,0x123,0x0000d5,0x0d5,0x128,0x0000d8,0x0d8,0x12a,0x0000d9,0x0d9,0x12c,0x0000db,0x0db,0x12f,0x0000de,0x0de,0x132,0x0000df,0x0df,0x132,0x0000e0,0x0e0)
NTDLL(NtFlushInstructionCache                           ,0x42,0x04e,0x052,0x098,0x098,0x08d,0x0bf,0x0bf,0x07d,0x0c2,0x0c2,0x11e,0x0c00d4,0x0d4,0x122,0x0c00d6,0x0d6,0x127,0x0c00d9,0x0d9,0x129,0x0c00da,0x0da,0x12b,0x0c00dc,0x0dc,0x12e,0x0c00df,0x0df,0x131,0x0c00e0,0x0e0,0x131,0x0c00e1,0x0e1)
NTDLL(NtFlushKey                                        ,0x43,0x04f,0x053,0x099,0x099,0x08e,0x0c0,0x0c0,0x07e,0x0c3,0x0c3,0x11d,0x0300d5,0x0d5,0x121,0x0300d7,0x0d7,0x126,0x0300da,0x0da,0x128,0x0300db,0x0db,0x12a,0x0300dd,0x0dd,0x12d,0x0300e0,0x0e0,0x130,0x0300e1,0x0e1,0x130,0x0300e2,0x0e2)
NTDLL(NtFlushProcessWriteBuffers                        ,NONE, NONE, NONE, NONE, NONE,0x08f,0x0c1,0x0c1,0x07f,0x0c4,0x0c4,0x11c,0x0000d6,0x0d6,0x120,0x0000d8,0x0d8,0x125,0x0000db,0x0db,0x127,0x0000dc,0x0dc,0x129,0x0000de,0x0de,0x12c,0x0000e1,0x0e1,0x12f,0x0000e2,0x0e2,0x12f,0x0000e3,0x0e3)
NTDLL(NtFlushVirtualMemory                              ,0x44,0x050,0x054,0x09a,0x09a,0x090,0x0c2,0x0c2,0x080,0x0c5,0x0c5,0x11b,0x0000d7,0x0d7,0x11f,0x0000d9,0x0d9,0x124,0x0000dc,0x0dc,0x126,0x0000dd,0x0dd,0x128,0x0000df,0x0df,0x12b,0x0000e2,0x0e2,0x12e,0x0000e3,0x0e3,0x12e,0x0000e4,0x0e4)
NTDLL(NtFlushWriteBuffer                                ,0x45,0x051,0x055,0x09b,0x09b,0x091,0x0c3,0x0c3,0x081,0x0c6,0x0c6,0x11a,0x0100d8,0x0d8,0x11e,0x0100da,0x0da,0x123,0x0100dd,0x0dd,0x125,0x0100de,0x0de,0x127,0x0100e0,0x0e0,0x12a,0x0100e3,0x0e3,0x12d,0x0100e4,0x0e4,0x12d,0x0100e5,0x0e5)
NTDLL(NtFreeUserPhysicalPages                           ,0x46,0x052,0x056,0x09c,0x09c,0x092,0x0c4,0x0c4,0x082,0x0c7,0x0c7,0x119,0x0000d9,0x0d9,0x11d,0x0000db,0x0db,0x122,0x0000de,0x0de,0x124,0x0000df,0x0df,0x126,0x0000e1,0x0e1,0x129,0x0000e4,0x0e4,0x12c,0x0000e5,0x0e5,0x12c,0x0000e6,0x0e6)
NTDLL(NtFreeVirtualMemory                               ,0x47,0x053,0x057,0x01b,0x01b,0x093,0x01b,0x01b,0x083,0x01b,0x01b,0x118,0x00001c,0x01c,0x11c,0x00001d,0x01d,0x121,0x00001e,0x01e,0x123,0x00001e,0x01e,0x125,0x00001e,0x01e,0x128,0x00001e,0x01e,0x12b,0x00001e,0x01e,0x12b,0x00001e,0x01e)
NTDLL(NtFreezeRegistry                                  ,NONE, NONE, NONE, NONE, NONE,0x094,0x0c5,0x0c5,0x084,0x0c8,0x0c8,0x117,0x0000da,0x0da,0x11b,0x0000dc,0x0dc,0x120,0x0000df,0x0df,0x122,0x0000e0,0x0e0,0x124,0x0000e2,0x0e2,0x127,0x0000e5,0x0e5,0x12a,0x0000e6,0x0e6,0x12a,0x0000e7,0x0e7)
NTDLL(NtFreezeTransactions                              ,NONE, NONE, NONE, NONE, NONE,0x095,0x0c6,0x0c6,0x085,0x0c9,0x0c9,0x116,0x0000db,0x0db,0x11a,0x0000dd,0x0dd,0x11f,0x0000e0,0x0e0,0x121,0x0000e1,0x0e1,0x123,0x0000e3,0x0e3,0x126,0x0000e6,0x0e6,0x129,0x0000e7,0x0e7,0x129,0x0000e8,0x0e8)
NTDLL(NtFsControlFile                                   ,0x48,0x054,0x058,0x036,0x036,0x096,0x036,0x036,0x086,0x036,0x036,0x115,0x1b0037,0x037,0x119,0x1b0038,0x038,0x11e,0x1b0039,0x039,0x120,0x1b0039,0x039,0x122,0x1b0039,0x039,0x125,0x1b0039,0x039,0x128,0x1b0039,0x039,0x128,0x1b0039,0x039)
NTDLL(NtGetCachedSigningLevel                           ,NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,0x114,0x0000dc,0x0dc,0x118,0x0000de,0x0de,0x11d,0x0000e1,0x0e1,0x11f,0x0000e2,0x0e2,0x121,0x0000e4,0x0e4,0x124,0x0000e7,0x0e7,0x127,0x0000e8,0x0e8,0x127,0x0000e9,0x0e9)
NTDLL(NtGetCompleteWnfStateSubscription                 ,NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,    NONE, NONE,0x117,0x0000df,0x0df,0x11c,0x0000e2,0x0e2,0x11e,0x0000e3,0x0e3,0x120,0x0000e5,0x0e5,0x123,0x0000e8,0x0e8,0x126,0x0000e9,0x0e9,0x126,0x0000ea,0x0ea)
NTDLL(NtGetContextThread                                ,0x49,0x055,0x059,0x09d,0x09d,0x097,0x0c7,0x0c7,0x087,0x0ca,0x0ca,0x113,0x0000dd,0x0dd,0x116,0x0000e0,0x0e0,0x11b,0x0000e3,0x0e3,0x11d,0x0000e4,0x0e4,0x11f,0x0000e6,0x0e6,0x122,0x0000e9,0x0e9,0x125,0x0000ea,0x0ea,0x125,0x0000eb,0x0eb)
NTDLL(NtGetCurrentProcessorNumber                       ,NONE, NONE,0x126,0x09e,0x09e,0x16d,0x0c8,0x0c8,0x088,0x0cb,0x0cb,0x112,0x1900de,0x0de,0x115,0x1900e1,0x0e1,0x11a,0x1900e4,0x0e4,0x11c,0x1900e5,0x0e5,0x11e,0x1900e7,0x0e7,0x121,0x1900ea,0x0ea,0x124,0x1900eb,0x0eb,0x124,0x1900ec,0x0ec)
NTDLL(NtGetCurrentProcessorNumberEx                     ,NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE,0x119,0x0000e5,0x0e5,0x11b,0x0000e6,0x0e6,0x11d,0x0000e8,0x0e8,0x120,0x0000eb,0x0eb,0x123,0x0000ec,0x0ec,0x123,0x0000ed,0x0ed)
NTDLL(NtGetDevicePowerState                             ,0x4a,0x056,0x05a,0x09f,0x09f,0x098,0x0c9,0x0c9,0x089,0x0cc,0x0cc,0x111,0x0700df,0x0df,0x114,0x0700e2,0x0e2,0x118,0x0700e6,0x0e6,0x11a,0x0700e7,0x0e7,0x11c,0x0700e9,0x0e9,0x11f,0x0700ec,0x0ec,0x122,0x0700ed,0x0ed,0x122,0x0700ee,0x0ee)
NTDLL(NtGetMUIRegistryInfo                              ,NONE, NONE, NONE, NONE, NONE,0x184,0x0ca,0x0ca,0x08a,0x0cd,0x0cd,0x110,0x0000e0,0x0e0,0x113,0x0000e3,0x0e3,0x117,0x0000e7,0x0e7,0x119,0x0000e8,0x0e8,0x11b,0x0000ea,0x0ea,0x11e,0x0000ed,0x0ed,0x121,0x0000ee,0x0ee,0x121,0x0000ef,0x0ef)
NTDLL(NtGetNextProcess                                  ,NONE, NONE, NONE, NONE, NONE,0x16f,0x0cb,0x0cb,0x08b,0x0ce,0x0ce,0x10f,0x0000e1,0x0e1,0x112,0x0000e4,0x0e4,0x116,0x0000e8,0x0e8,0x118,0x0000e9,0x0e9,0x11a,0x0000eb,0x0eb,0x11d,0x0000ee,0x0ee,0x120,0x0000ef,0x0ef,0x120,0x0000f0,0x0f0)
NTDLL(NtGetNextThread                                   ,NONE, NONE, NONE, NONE, NONE,0x170,0x0cc,0x0cc,0x08c,0x0cf,0x0cf,0x10e,0x0000e2,0x0e2,0x111,0x0000e5,0x0e5,0x115,0x0000e9,0x0e9,0x117,0x0000ea,0x0ea,0x119,0x0000ec,0x0ec,0x11c,0x0000ef,0x0ef,0x11f,0x0000f0,0x0f0,0x11f,0x0000f1,0x0f1)
NTDLL(NtGetNlsSectionPtr                                ,NONE, NONE, NONE, NONE, NONE,0x099,0x0cd,0x0cd,0x08d,0x0d0,0x0d0,0x10d,0x0000e3,0x0e3,0x110,0x0000e6,0x0e6,0x114,0x0000ea,0x0ea,0x116,0x0000eb,0x0eb,0x118,0x0000ed,0x0ed,0x11b,0x0000f0,0x0f0,0x11e,0x0000f1,0x0f1,0x11e,0x0000f2,0x0f2)
NTDLL(NtGetNotificationResourceManager                  ,NONE, NONE, NONE, NONE, NONE,0x06d,0x0ce,0x0ce,0x08e,0x0d1,0x0d1,0x10c,0x0000e4,0x0e4,0x10f,0x0000e7,0x0e7,0x113,0x0000eb,0x0eb,0x115,0x0000ec,0x0ec,0x117,0x0000ee,0x0ee,0x11a,0x0000f1,0x0f1,0x11d,0x0000f2,0x0f2,0x11d,0x0000f3,0x0f3)
NTDLL(NtGetPlugPlayEvent                                ,0x4b,0x057,0x05b,0x0a0,0x0a0,0x09a,0x0cf,0x0cf,0x08f,0x0d2,0x0d2, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE)
NTDLL(NtGetTickCount                                    ,0x4c, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE)
NTDLL(NtGetWriteWatch                                   ,0x4d,0x058,0x05c,0x0a1,0x0a1,0x09b,0x0d0,0x0d0,0x090,0x0d3,0x0d3,0x10b,0x0000e5,0x0e5,0x10e,0x0000e8,0x0e8,0x112,0x0000ec,0x0ec,0x114,0x0000ed,0x0ed,0x116,0x0000ef,0x0ef,0x119,0x0000f2,0x0f2,0x11c,0x0000f3,0x0f3,0x11c,0x0000f4,0x0f4)
NTDLL(NtImpersonateAnonymousToken                       ,0x4e,0x059,0x05d,0x0a2,0x0a2,0x09c,0x0d1,0x0d1,0x091,0x0d4,0x0d4,0x10a,0x0300e6,0x0e6,0x10d,0x0300e9,0x0e9,0x111,0x0300ed,0x0ed,0x113,0x0300ee,0x0ee,0x115,0x0300f0,0x0f0,0x118,0x0300f3,0x0f3,0x11b,0x0300f4,0x0f4,0x11b,0x0300f5,0x0f5)
NTDLL(NtImpersonateClientOfPort                         ,0x4f,0x05a,0x05e,0x01c,0x01c,0x09d,0x01c,0x01c,0x092,0x01c,0x01c,0x109,0x07001d,0x01d,0x10c,0x07001e,0x01e,0x110,0x07001f,0x01f,0x112,0x07001f,0x01f,0x114,0x07001f,0x01f,0x117,0x07001f,0x01f,0x11a,0x07001f,0x01f,0x11a,0x07001f,0x01f)
NTDLL(NtImpersonateThread                               ,0x50,0x05b,0x05f,0x0a3,0x0a3,0x09e,0x0d2,0x0d2,0x093,0x0d5,0x0d5,0x108,0x0000e7,0x0e7,0x10b,0x0000ea,0x0ea,0x10f,0x0000ee,0x0ee,0x111,0x0000ef,0x0ef,0x113,0x0000f1,0x0f1,0x116,0x0000f4,0x0f4,0x119,0x0000f5,0x0f5,0x119,0x0000f6,0x0f6)
NTDLL(NtInitializeEnclave                               ,NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE,0x110,0x0000f0,0x0f0,0x112,0x0000f2,0x0f2,0x115,0x0000f5,0x0f5,0x118,0x0000f6,0x0f6,0x118,0x0000f7,0x0f7)
NTDLL(NtInitializeNlsFiles                              ,NONE, NONE, NONE, NONE, NONE,0x09f,0x0d3,0x0d3,0x094,0x0d6,0x0d6,0x107,0x0000e8,0x0e8,0x10a,0x0000eb,0x0eb,0x10e,0x0000ef,0x0ef,0x10f,0x0000f1,0x0f1,0x111,0x0000f3,0x0f3,0x114,0x0000f6,0x0f6,0x117,0x0000f7,0x0f7,0x117,0x0000f8,0x0f8)
NTDLL(NtInitializeRegistry                              ,0x51,0x05c,0x060,0x0a4,0x0a4,0x0a0,0x0d4,0x0d4,0x095,0x0d7,0x0d7,0x106,0x0000e9,0x0e9,0x109,0x0000ec,0x0ec,0x10d,0x0000f0,0x0f0,0x10e,0x0000f2,0x0f2,0x110,0x0000f4,0x0f4,0x113,0x0000f7,0x0f7,0x116,0x0000f8,0x0f8,0x116,0x0000f9,0x0f9)
NTDLL(NtInitiatePowerAction                             ,0x52,0x05d,0x061,0x0a5,0x0a5,0x0a1,0x0d5,0x0d5,0x096,0x0d8,0x0d8,0x105,0x1100ea,0x0ea,0x108,0x1100ed,0x0ed,0x10c,0x1100f1,0x0f1,0x10d,0x1100f3,0x0f3,0x10f,0x1100f5,0x0f5,0x112,0x1100f8,0x0f8,0x115,0x1100f9,0x0f9,0x115,0x1100fa,0x0fa)
NTDLL(NtIsProcessInJob                                  ,NONE,0x05e,0x062,0x04c,0x04c,0x0a2,0x04c,0x04c,0x097,0x04c,0x04c,0x104,0x08004d,0x04d,0x107,0x08004e,0x04e,0x10b,0x08004f,0x04f,0x10c,0x08004f,0x04f,0x10e,0x08004f,0x04f,0x111,0x08004f,0x04f,0x114,0x08004f,0x04f,0x114,0x08004f,0x04f)
NTDLL(NtIsSystemResumeAutomatic                         ,0x53,0x05f,0x063,0x0a6,0x0a6,0x0a3,0x0d6,0x0d6,0x098,0x0d9,0x0d9,0x103,0x0100eb,0x0eb,0x106,0x0100ee,0x0ee,0x10a,0x0100f2,0x0f2,0x10b,0x0100f4,0x0f4,0x10d,0x0100f6,0x0f6,0x110,0x0100f9,0x0f9,0x113,0x0100fa,0x0fa,0x113,0x0100fb,0x0fb)
NTDLL(NtIsUILanguageComitted                            ,NONE, NONE, NONE, NONE, NONE,0x182,0x0d7,0x0d7,0x099,0x0da,0x0da,0x102,0x0000ec,0x0ec,0x105,0x0000ef,0x0ef,0x109,0x0000f3,0x0f3,0x10a,0x0000f5,0x0f5,0x10c,0x0000f7,0x0f7,0x10f,0x0000fa,0x0fa,0x112,0x0000fb,0x0fb,0x112,0x0000fc,0x0fc)
NTDLL(NtListenChannel                                   ,0xf2, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE)
NTDLL(NtListenPort                                      ,0x54,0x060,0x064,0x0a7,0x0a7,0x0a4,0x0d8,0x0d8,0x09a,0x0db,0x0db,0x101,0x0000ed,0x0ed,0x104,0x0000f0,0x0f0,0x108,0x0000f4,0x0f4,0x109,0x0000f6,0x0f6,0x10b,0x0000f8,0x0f8,0x10e,0x0000fb,0x0fb,0x111,0x0000fc,0x0fc,0x111,0x0000fd,0x0fd)
NTDLL(NtLoadDriver                                      ,0x55,0x061,0x065,0x0a8,0x0a8,0x0a5,0x0d9,0x0d9,0x09b,0x0dc,0x0dc,0x100,0x0000ee,0x0ee,0x103,0x0000f1,0x0f1,0x107,0x0000f5,0x0f5,0x108,0x0000f7,0x0f7,0x10a,0x0000f9,0x0f9,0x10d,0x0000fc,0x0fc,0x110,0x0000fd,0x0fd,0x110,0x0000fe,0x0fe)
NTDLL(NtLoadEnclaveData                                 ,NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE,0x107,0x0000f8,0x0f8,0x109,0x0000fa,0x0fa,0x10c,0x0000fd,0x0fd,0x10f,0x0000fe,0x0fe,0x10f,0x0000ff,0x0ff)
NTDLL(NtLoadHotPatch                                    ,NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE,0x10b,0x0000fe,0x0fe,0x10e,0x0000ff,0x0ff,0x10e,0x000100,0x100)
NTDLL(NtLoadKey                                         ,0x56,0x062,0x066,0x0a9,0x0a9,0x0a6,0x0da,0x0da,0x09c,0x0dd,0x0dd,0x0ff,0x0000ef,0x0ef,0x102,0x0000f2,0x0f2,0x106,0x0000f6,0x0f6,0x106,0x0000f9,0x0f9,0x108,0x0000fb,0x0fb,0x10a,0x0000ff,0x0ff,0x10d,0x000100,0x100,0x10d,0x000101,0x101)
NTDLL(NtLoadKey2                                        ,0x57,0x063,0x067,0x0aa,0x0aa,0x0a7,0x0db,0x0db,0x09d,0x0de,0x0de,0x0fe,0x0000f0,0x0f0,0x101,0x0000f3,0x0f3,0x105,0x0000f7,0x0f7,0x105,0x0000fa,0x0fa,0x107,0x0000fc,0x0fc,0x109,0x000100,0x100,0x10c,0x000101,0x101,0x10c,0x000102,0x102)
NTDLL(NtLoadKeyEx                                       ,NONE, NONE,0x068,0x0ab,0x0ab,0x0a8,0x0dc,0x0dc,0x09e,0x0df,0x0df,0x0fd,0x0000f1,0x0f1,0x100,0x0000f4,0x0f4,0x104,0x0000f8,0x0f8,0x104,0x0000fb,0x0fb,0x106,0x0000fd,0x0fd,0x108,0x000101,0x101,0x10b,0x000102,0x102,0x10b,0x000103,0x103)
NTDLL(NtLockFile                                        ,0x58,0x064,0x069,0x0ac,0x0ac,0x0a9,0x0dd,0x0dd,0x09f,0x0e0,0x0e0,0x0fc,0x0000f2,0x0f2,0x0ff,0x0000f5,0x0f5,0x103,0x0000f9,0x0f9,0x103,0x0000fc,0x0fc,0x105,0x0000fe,0x0fe,0x107,0x000102,0x102,0x10a,0x000103,0x103,0x10a,0x000104,0x104)
NTDLL(NtLockProductActivationKeys                       ,NONE,0x065,0x06a,0x0ad,0x0ad,0x0aa,0x0de,0x0de,0x0a0,0x0e1,0x0e1,0x0fb,0x0500f3,0x0f3,0x0fe,0x0500f6,0x0f6,0x102,0x0500fa,0x0fa,0x102,0x0500fd,0x0fd,0x104,0x0500ff,0x0ff,0x106,0x050103,0x103,0x109,0x050104,0x104,0x109,0x050105,0x105)
NTDLL(NtLockRegistryKey                                 ,NONE,0x066,0x06b,0x0ae,0x0ae,0x0ab,0x0df,0x0df,0x0a1,0x0e2,0x0e2,0x0fa,0x0300f4,0x0f4,0x0fd,0x0300f7,0x0f7,0x101,0x0300fb,0x0fb,0x101,0x0300fe,0x0fe,0x103,0x030100,0x100,0x105,0x030104,0x104,0x108,0x030105,0x105,0x108,0x030106,0x106)
NTDLL(NtLockVirtualMemory                               ,0x59,0x067,0x06c,0x0af,0x0af,0x0ac,0x0e0,0x0e0,0x0a2,0x0e3,0x0e3,0x0f9,0x0000f5,0x0f5,0x0fc,0x0000f8,0x0f8,0x100,0x0000fc,0x0fc,0x100,0x0000ff,0x0ff,0x102,0x000101,0x101,0x104,0x000105,0x105,0x107,0x000106,0x106,0x107,0x000107,0x107)
NTDLL(NtMakePermanentObject                             ,NONE,0x068,0x06d,0x0b0,0x0b0,0x0ad,0x0e1,0x0e1,0x0a3,0x0e4,0x0e4,0x0f8,0x0300f6,0x0f6,0x0fb,0x0300f9,0x0f9,0x0ff,0x0300fd,0x0fd,0x0ff,0x030100,0x100,0x101,0x030102,0x102,0x103,0x030106,0x106,0x106,0x030107,0x107,0x106,0x030108,0x108)
NTDLL(NtMakeTemporaryObject                             ,0x5a,0x069,0x06e,0x0b1,0x0b1,0x0ae,0x0e2,0x0e2,0x0a4,0x0e5,0x0e5,0x0f7,0x0300f7,0x0f7,0x0fa,0x0300fa,0x0fa,0x0fe,0x0300fe,0x0fe,0x0fe,0x030101,0x101,0x100,0x030103,0x103,0x102,0x030107,0x107,0x105,0x030108,0x108,0x105,0x030109,0x109)
NTDLL(NtManagePartition                                 ,NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE,0x0f7,0x0000ff,0x0ff,0x0f7,0x000102,0x102,0x0f9,0x000104,0x104,0x0fb,0x000108,0x108,0x0fe,0x000109,0x109,0x0fe,0x00010a,0x10a)
NTDLL(NtMapCMFModule                                    ,NONE, NONE, NONE, NONE, NONE,0x181,0x0e3,0x0e3,0x0a5,0x0e6,0x0e6,0x0f6,0x0000f8,0x0f8,0x0f9,0x0000fb,0x0fb,0x0fd,0x000100,0x100,0x0fd,0x000103,0x103,0x0ff,0x000105,0x105,0x101,0x000109,0x109,0x104,0x00010a,0x10a,0x104,0x00010b,0x10b)
NTDLL(NtMapUserPhysicalPages                            ,0x5b,0x06a,0x06f,0x0b2,0x0b2,0x0af,0x0e4,0x0e4,0x0a6,0x0e7,0x0e7,0x0f5,0x0a00f9,0x0f9,0x0f8,0x0a00fc,0x0fc,0x0fc,0x0a0101,0x101,0x0fc,0x0a0104,0x104,0x0fe,0x0a0106,0x106,0x100,0x0a010a,0x10a,0x103,0x0a010b,0x10b,0x103,0x0a010c,0x10c)
NTDLL(NtMapUserPhysicalPagesScatter                     ,0x5c,0x06b,0x070,0x000,0x000,0x0b0,0x000,0x000,0x0a7,0x000,0x000,0x0f4,0x0a0001,0x001,0x0f7,0x0a0002,0x002,0x0fb,0x0a0003,0x003,0x0fb,0x0a0003,0x003,0x0fd,0x0a0003,0x003,0x0ff,0x0a0003,0x003,0x102,0x0a0003,0x003,0x102,0x0a0003,0x003)
NTDLL(NtMapViewOfSection                                ,0x5d,0x06c,0x071,0x025,0x025,0x0b1,0x025,0x025,0x0a8,0x025,0x025,0x0f3,0x000026,0x026,0x0f6,0x000027,0x027,0x0fa,0x000028,0x028,0x0fa,0x000028,0x028,0x0fc,0x000028,0x028,0x0fe,0x000028,0x028,0x101,0x000028,0x028,0x101,0x000028,0x028)
NTDLL(NtMapViewOfSectionEx                              ,NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE,0x1cd,0x00010d,0x10d)
NTDLL(NtModifyBootEntry                                 ,NONE,0x06d,0x072,0x0b3,0x0b3,0x0b2,0x0e5,0x0e5,0x0a9,0x0e8,0x0e8,0x0f2,0x0000fa,0x0fa,0x0f5,0x0000fd,0x0fd,0x0f9,0x000102,0x102,0x0f9,0x000105,0x105,0x0fb,0x000107,0x107,0x0fd,0x00010b,0x10b,0x100,0x00010c,0x10c,0x100,0x00010e,0x10e)
NTDLL(NtModifyDriverEntry                               ,NONE, NONE,0x073,0x0b4,0x0b4,0x0b3,0x0e6,0x0e6,0x0aa,0x0e9,0x0e9,0x0f1,0x0000fb,0x0fb,0x0f4,0x0000fe,0x0fe,0x0f8,0x000103,0x103,0x0f8,0x000106,0x106,0x0fa,0x000108,0x108,0x0fc,0x00010c,0x10c,0x0ff,0x00010d,0x10d,0x0ff,0x00010f,0x10f)
NTDLL(NtNotifyChangeDirectoryFile                       ,0x5e,0x06e,0x074,0x0b5,0x0b5,0x0b4,0x0e7,0x0e7,0x0ab,0x0ea,0x0ea,0x0f0,0x0000fc,0x0fc,0x0f3,0x0000ff,0x0ff,0x0f6,0x000104,0x104,0x0f6,0x000107,0x107,0x0f8,0x000109,0x109,0x0fa,0x00010d,0x10d,0x0fc,0x00010e,0x10e,0x0fc,0x000110,0x110)
NTDLL(NtNotifyChangeDirectoryFileEx                     ,NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE,0x0fd,0x00010f,0x10f,0x0fd,0x000111,0x111)
NTDLL(NtNotifyChangeKey                                 ,0x5f,0x06f,0x075,0x0b6,0x0b6,0x0b5,0x0e8,0x0e8,0x0ac,0x0eb,0x0eb,0x0ef,0x0000fd,0x0fd,0x0f2,0x000100,0x100,0x0f5,0x000105,0x105,0x0f5,0x000108,0x108,0x0f7,0x00010a,0x10a,0x0f9,0x00010e,0x10e,0x0fb,0x000110,0x110,0x0fb,0x000112,0x112)
NTDLL(NtNotifyChangeMultipleKeys                        ,0x60,0x070,0x076,0x0b7,0x0b7,0x0b6,0x0e9,0x0e9,0x0ad,0x0ec,0x0ec,0x0ee,0x0000fe,0x0fe,0x0f1,0x000101,0x101,0x0f4,0x000106,0x106,0x0f4,0x000109,0x109,0x0f6,0x00010b,0x10b,0x0f8,0x00010f,0x10f,0x0fa,0x000111,0x111,0x0fa,0x000113,0x113)
NTDLL(NtNotifyChangeSession                             ,NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,0x0ae,0x0ed,0x0ed,0x0ed,0x0000ff,0x0ff,0x0f0,0x000102,0x102,0x0f3,0x000107,0x107,0x0f3,0x00010a,0x10a,0x0f5,0x00010c,0x10c,0x0f7,0x000110,0x110,0x0f9,0x000112,0x112,0x0f9,0x000114,0x114)
NTDLL(NtOpenChannel                                     ,0xf3, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE)
NTDLL(NtOpenDirectoryObject                             ,0x61,0x071,0x077,0x055,0x055,0x0b7,0x055,0x055,0x0af,0x055,0x055,0x0ec,0x000056,0x056,0x0ef,0x000057,0x057,0x0f2,0x000058,0x058,0x0f2,0x000058,0x058,0x0f4,0x000058,0x058,0x0f6,0x000058,0x058,0x0f8,0x000058,0x058,0x0f8,0x000058,0x058)
NTDLL(NtOpenEnlistment                                  ,NONE, NONE, NONE, NONE, NONE,0x070,0x0ea,0x0ea,0x0b0,0x0ee,0x0ee,0x0eb,0x000100,0x100,0x0ee,0x000103,0x103,0x0f1,0x000108,0x108,0x0f1,0x00010b,0x10b,0x0f3,0x00010d,0x10d,0x0f5,0x000111,0x111,0x0f7,0x000113,0x113,0x0f7,0x000115,0x115)
NTDLL(NtOpenEvent                                       ,0x62,0x072,0x078,0x03d,0x03d,0x0b8,0x03d,0x03d,0x0b1,0x03d,0x03d,0x0ea,0x00003e,0x03e,0x0ed,0x00003f,0x03f,0x0f0,0x000040,0x040,0x0f0,0x000040,0x040,0x0f2,0x000040,0x040,0x0f4,0x000040,0x040,0x0f6,0x000040,0x040,0x0f6,0x000040,0x040)
NTDLL(NtOpenEventPair                                   ,0x63,0x073,0x079,0x0b8,0x0b8,0x0b9,0x0eb,0x0eb,0x0b2,0x0ef,0x0ef,0x0e9,0x000101,0x101,0x0ec,0x000104,0x104,0x0ef,0x000109,0x109,0x0ef,0x00010c,0x10c,0x0f1,0x00010e,0x10e,0x0f3,0x000112,0x112,0x0f5,0x000114,0x114,0x0f5,0x000116,0x116)
NTDLL(NtOpenFile                                        ,0x64,0x074,0x07a,0x030,0x030,0x0ba,0x030,0x030,0x0b3,0x030,0x030,0x0e8,0x000031,0x031,0x0eb,0x000032,0x032,0x0ee,0x000033,0x033,0x0ee,0x000033,0x033,0x0f0,0x000033,0x033,0x0f2,0x000033,0x033,0x0f4,0x000033,0x033,0x0f4,0x000033,0x033)
NTDLL(NtOpenIoCompletion                                ,0x65,0x075,0x07b,0x0b9,0x0b9,0x0bb,0x0ec,0x0ec,0x0b4,0x0f0,0x0f0,0x0e7,0x000102,0x102,0x0ea,0x000105,0x105,0x0ed,0x00010a,0x10a,0x0ed,0x00010d,0x10d,0x0ef,0x00010f,0x10f,0x0f1,0x000113,0x113,0x0f3,0x000115,0x115,0x0f3,0x000117,0x117)
NTDLL(NtOpenJobObject                                   ,0x66,0x076,0x07c,0x0ba,0x0ba,0x0bc,0x0ed,0x0ed,0x0b5,0x0f1,0x0f1,0x0e6,0x000103,0x103,0x0e9,0x000106,0x106,0x0ec,0x00010b,0x10b,0x0ec,0x00010e,0x10e,0x0ee,0x000110,0x110,0x0f0,0x000114,0x114,0x0f2,0x000116,0x116,0x0f2,0x000118,0x118)
NTDLL(NtOpenKey                                         ,0x67,0x077,0x07d,0x00f,0x00f,0x0bd,0x00f,0x00f,0x0b6,0x00f,0x00f,0x0e5,0x000010,0x010,0x0e8,0x000011,0x011,0x0eb,0x000012,0x012,0x0eb,0x000012,0x012,0x0ed,0x000012,0x012,0x0ef,0x000012,0x012,0x0f1,0x000012,0x012,0x0f1,0x000012,0x012)
NTDLL(NtOpenKeyEx                                       ,NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,0x0b7,0x0f2,0x0f2,0x0e3,0x000104,0x104,0x0e6,0x000107,0x107,0x0e9,0x00010c,0x10c,0x0e9,0x00010f,0x10f,0x0eb,0x000111,0x111,0x0ed,0x000115,0x115,0x0ef,0x000117,0x117,0x0ef,0x000119,0x119)
NTDLL(NtOpenKeyTransacted                               ,NONE, NONE, NONE, NONE, NONE,0x0be,0x0ee,0x0ee,0x0b9,0x0f3,0x0f3,0x0e2,0x000105,0x105,0x0e5,0x000108,0x108,0x0e8,0x00010d,0x10d,0x0e8,0x000110,0x110,0x0ea,0x000112,0x112,0x0ec,0x000116,0x116,0x0ee,0x000118,0x118,0x0ee,0x00011a,0x11a)
NTDLL(NtOpenKeyTransactedEx                             ,NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,0x0ba,0x0f4,0x0f4,0x0e1,0x000106,0x106,0x0e4,0x000109,0x109,0x0e7,0x00010e,0x10e,0x0e7,0x000111,0x111,0x0e9,0x000113,0x113,0x0eb,0x000117,0x117,0x0ed,0x000119,0x119,0x0ed,0x00011b,0x11b)
NTDLL(NtOpenKeyedEvent                                  ,NONE,0x118,0x122,0x0bb,0x0bb,0x169,0x0ef,0x0ef,0x0b8,0x0f5,0x0f5,0x0e4,0x000107,0x107,0x0e7,0x00010a,0x10a,0x0ea,0x00010f,0x10f,0x0ea,0x000112,0x112,0x0ec,0x000114,0x114,0x0ee,0x000118,0x118,0x0f0,0x00011a,0x11a,0x0f0,0x00011c,0x11c)
NTDLL(NtOpenMutant                                      ,0x68,0x078,0x07e,0x0bc,0x0bc,0x0bf,0x0f0,0x0f0,0x0bb,0x0f6,0x0f6,0x0e0,0x000108,0x108,0x0e3,0x00010b,0x10b,0x0e6,0x000110,0x110,0x0e6,0x000113,0x113,0x0e8,0x000115,0x115,0x0ea,0x000119,0x119,0x0ec,0x00011b,0x11b,0x0ec,0x00011d,0x11d)
NTDLL(NtOpenObjectAuditAlarm                            ,0x69,0x079,0x07f,0x0bd,0x0bd,0x0c1,0x0f1,0x0f1,0x0bc,0x0f7,0x0f7,0x0df,0x000109,0x109,0x0e2,0x00010c,0x10c,0x0e5,0x000111,0x111,0x0e5,0x000114,0x114,0x0e7,0x000116,0x116,0x0e9,0x00011a,0x11a,0x0eb,0x00011c,0x11c,0x0eb,0x00011e,0x11e)
NTDLL(NtOpenPartition                                   ,NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE,0x0e0,0x000112,0x112,0x0e0,0x000115,0x115,0x0e2,0x000117,0x117,0x0e4,0x00011b,0x11b,0x0e6,0x00011d,0x11d,0x0e6,0x00011f,0x11f)
NTDLL(NtOpenPrivateNamespace                            ,NONE, NONE, NONE, NONE, NONE,0x0c0,0x0f2,0x0f2,0x0bd,0x0f8,0x0f8,0x0de,0x00010a,0x10a,0x0e1,0x00010d,0x10d,0x0e4,0x000113,0x113,0x0e4,0x000116,0x116,0x0e6,0x000118,0x118,0x0e8,0x00011c,0x11c,0x0ea,0x00011e,0x11e,0x0ea,0x000120,0x120)
NTDLL(NtOpenProcess                                     ,0x6a,0x07a,0x080,0x023,0x023,0x0c2,0x023,0x023,0x0be,0x023,0x023,0x0dd,0x000024,0x024,0x0e0,0x000025,0x025,0x0e3,0x000026,0x026,0x0e3,0x000026,0x026,0x0e5,0x000026,0x026,0x0e7,0x000026,0x026,0x0e9,0x000026,0x026,0x0e9,0x000026,0x026)
NTDLL(NtOpenProcessToken                                ,0x6b,0x07b,0x081,0x0be,0x0be,0x0c3,0x0f3,0x0f3,0x0bf,0x0f9,0x0f9,0x0dc,0x00010b,0x10b,0x0df,0x00010e,0x10e,0x0e2,0x000114,0x114,0x0e2,0x000117,0x117,0x0e4,0x000119,0x119,0x0e6,0x00011d,0x11d,0x0e8,0x00011f,0x11f,0x0e8,0x000121,0x121)
NTDLL(NtOpenProcessTokenEx                              ,NONE,0x07c,0x082,0x02d,0x02d,0x0c4,0x02d,0x02d,0x0c0,0x02d,0x02d,0x0db,0x00002e,0x02e,0x0de,0x00002f,0x02f,0x0e1,0x000030,0x030,0x0e1,0x000030,0x030,0x0e3,0x000030,0x030,0x0e5,0x000030,0x030,0x0e7,0x000030,0x030,0x0e7,0x000030,0x030)
NTDLL(NtOpenRegistryTransaction                         ,NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE,0x1bf,0x00011a,0x11a,0x1c5,0x00011e,0x11e,0x1c9,0x000120,0x120,0x1c9,0x000122,0x122)
NTDLL(NtOpenResourceManager                             ,NONE, NONE, NONE, NONE, NONE,0x06c,0x0f4,0x0f4,0x0c1,0x0fa,0x0fa,0x0da,0x00010c,0x10c,0x0dd,0x00010f,0x10f,0x0df,0x000115,0x115,0x0df,0x000118,0x118,0x0e1,0x00011b,0x11b,0x0e3,0x00011f,0x11f,0x0e5,0x000121,0x121,0x0e5,0x000123,0x123)
NTDLL(NtOpenSection                                     ,0x6c,0x07d,0x083,0x034,0x034,0x0c5,0x034,0x034,0x0c2,0x034,0x034,0x0d9,0x000035,0x035,0x0dc,0x000036,0x036,0x0de,0x000037,0x037,0x0de,0x000037,0x037,0x0e0,0x000037,0x037,0x0e2,0x000037,0x037,0x0e4,0x000037,0x037,0x0e4,0x000037,0x037)
NTDLL(NtOpenSemaphore                                   ,0x6d,0x07e,0x084,0x0bf,0x0bf,0x0c6,0x0f5,0x0f5,0x0c3,0x0fb,0x0fb,0x0d8,0x00010d,0x10d,0x0db,0x000110,0x110,0x0dd,0x000116,0x116,0x0dd,0x000119,0x119,0x0df,0x00011c,0x11c,0x0e1,0x000120,0x120,0x0e3,0x000122,0x122,0x0e3,0x000124,0x124)
NTDLL(NtOpenSession                                     ,NONE, NONE, NONE, NONE, NONE,0x0c7,0x0f6,0x0f6,0x0c4,0x0fc,0x0fc,0x0d7,0x00010e,0x10e,0x0da,0x000111,0x111,0x0dc,0x000117,0x117,0x0dc,0x00011a,0x11a,0x0de,0x00011d,0x11d,0x0e0,0x000121,0x121,0x0e2,0x000123,0x123,0x0e2,0x000125,0x125)
NTDLL(NtOpenSymbolicLinkObject                          ,0x6e,0x07f,0x085,0x0c0,0x0c0,0x0c8,0x0f7,0x0f7,0x0c5,0x0fd,0x0fd,0x0d6,0x00010f,0x10f,0x0d9,0x000112,0x112,0x0db,0x000118,0x118,0x0db,0x00011b,0x11b,0x0dd,0x00011e,0x11e,0x0df,0x000122,0x122,0x0e1,0x000124,0x124,0x0e1,0x000126,0x126)
NTDLL(NtOpenThread                                      ,0x6f,0x080,0x086,0x0c1,0x0c1,0x0c9,0x0f8,0x0f8,0x0c6,0x0fe,0x0fe,0x0d5,0x000110,0x110,0x0d8,0x000113,0x113,0x0da,0x000119,0x119,0x0da,0x00011c,0x11c,0x0dc,0x00011f,0x11f,0x0de,0x000123,0x123,0x0e0,0x000125,0x125,0x0e0,0x000127,0x127)
NTDLL(NtOpenThreadToken                                 ,0x70,0x081,0x087,0x021,0x021,0x0ca,0x021,0x021,0x0c7,0x021,0x021,0x0d4,0x000022,0x022,0x0d7,0x000023,0x023,0x0d9,0x000024,0x024,0x0d9,0x000024,0x024,0x0db,0x000024,0x024,0x0dd,0x000024,0x024,0x0df,0x000024,0x024,0x0df,0x000024,0x024)
NTDLL(NtOpenThreadTokenEx                               ,NONE,0x082,0x088,0x02c,0x02c,0x0cb,0x02c,0x02c,0x0c8,0x02c,0x02c,0x0d3,0x00002d,0x02d,0x0d6,0x00002e,0x02e,0x0d8,0x00002f,0x02f,0x0d8,0x00002f,0x02f,0x0da,0x00002f,0x02f,0x0dc,0x00002f,0x02f,0x0de,0x00002f,0x02f,0x0de,0x00002f,0x02f)
NTDLL(NtOpenTimer                                       ,0x71,0x083,0x089,0x0c2,0x0c2,0x0cc,0x0f9,0x0f9,0x0c9,0x0ff,0x0ff,0x0d2,0x000111,0x111,0x0d5,0x000114,0x114,0x0d7,0x00011a,0x11a,0x0d7,0x00011d,0x11d,0x0d9,0x000120,0x120,0x0db,0x000124,0x124,0x0dd,0x000126,0x126,0x0dd,0x000128,0x128)
NTDLL(NtOpenTransaction                                 ,NONE, NONE, NONE, NONE, NONE,0x052,0x0fa,0x0fa,0x0ca,0x100,0x100,0x0d1,0x000112,0x112,0x0d4,0x000115,0x115,0x0d6,0x00011b,0x11b,0x0d6,0x00011e,0x11e,0x0d8,0x000121,0x121,0x0da,0x000125,0x125,0x0dc,0x000127,0x127,0x0dc,0x000129,0x129)
NTDLL(NtOpenTransactionManager                          ,NONE, NONE, NONE, NONE, NONE,0x065,0x0fb,0x0fb,0x0cb,0x101,0x101,0x0d0,0x000113,0x113,0x0d3,0x000116,0x116,0x0d5,0x00011c,0x11c,0x0d5,0x00011f,0x11f,0x0d7,0x000122,0x122,0x0d9,0x000126,0x126,0x0db,0x000128,0x128,0x0db,0x00012a,0x12a)
NTDLL(NtPlugPlayControl                                 ,0x72,0x084,0x08a,0x0c3,0x0c3,0x0cd,0x0fc,0x0fc,0x0cc,0x102,0x102,0x0cf,0x000114,0x114,0x0d2,0x000117,0x117,0x0d4,0x00011d,0x11d,0x0d4,0x000120,0x120,0x0d6,0x000123,0x123,0x0d8,0x000127,0x127,0x0da,0x000129,0x129,0x0da,0x00012b,0x12b)
NTDLL(NtPowerInformation                                ,0x73,0x085,0x08b,0x05c,0x05c,0x0ce,0x05c,0x05c,0x0cd,0x05c,0x05c,0x0ce,0x00005d,0x05d,0x0d1,0x00005e,0x05e,0x0d3,0x00005f,0x05f,0x0d3,0x00005f,0x05f,0x0d5,0x00005f,0x05f,0x0d7,0x00005f,0x05f,0x0d9,0x00005f,0x05f,0x0d9,0x00005f,0x05f)
NTDLL(NtPrePrepareComplete                              ,NONE, NONE, NONE, NONE, NONE,0x05d,0x0fd,0x0fd,0x0d0,0x103,0x103,0x0cb,0x000115,0x115,0x0ce,0x000118,0x118,0x0d0,0x00011e,0x11e,0x0d0,0x000121,0x121,0x0d2,0x000124,0x124,0x0d4,0x000128,0x128,0x0d6,0x00012a,0x12a,0x0d6,0x00012c,0x12c)
NTDLL(NtPrePrepareEnlistment                            ,NONE, NONE, NONE, NONE, NONE,0x055,0x0fe,0x0fe,0x0d1,0x104,0x104,0x0ca,0x000116,0x116,0x0cd,0x000119,0x119,0x0cf,0x00011f,0x11f,0x0cf,0x000122,0x122,0x0d1,0x000125,0x125,0x0d3,0x000129,0x129,0x0d5,0x00012b,0x12b,0x0d5,0x00012d,0x12d)
NTDLL(NtPrepareComplete                                 ,NONE, NONE, NONE, NONE, NONE,0x05e,0x0ff,0x0ff,0x0ce,0x105,0x105,0x0cd,0x000117,0x117,0x0d0,0x00011a,0x11a,0x0d2,0x000120,0x120,0x0d2,0x000123,0x123,0x0d4,0x000126,0x126,0x0d6,0x00012a,0x12a,0x0d8,0x00012c,0x12c,0x0d8,0x00012e,0x12e)
NTDLL(NtPrepareEnlistment                               ,NONE, NONE, NONE, NONE, NONE,0x056,0x100,0x100,0x0cf,0x106,0x106,0x0cc,0x000118,0x118,0x0cf,0x00011b,0x11b,0x0d1,0x000121,0x121,0x0d1,0x000124,0x124,0x0d3,0x000127,0x127,0x0d5,0x00012b,0x12b,0x0d7,0x00012d,0x12d,0x0d7,0x00012f,0x12f)
NTDLL(NtPrivilegeCheck                                  ,0x74,0x086,0x08c,0x0c4,0x0c4,0x0cf,0x101,0x101,0x0d2,0x107,0x107,0x0c8,0x0c0119,0x119,0x0cb,0x0c011c,0x11c,0x0cd,0x0c0122,0x122,0x0cd,0x0c0125,0x125,0x0cf,0x0c0128,0x128,0x0d1,0x0c012c,0x12c,0x0d3,0x0c012e,0x12e,0x0d3,0x0c0130,0x130)
NTDLL(NtPrivilegeObjectAuditAlarm                       ,0x76,0x087,0x08d,0x0c5,0x0c5,0x0d0,0x102,0x102,0x0d4,0x108,0x108,0x0c6,0x00011a,0x11a,0x0c9,0x00011d,0x11d,0x0cb,0x000123,0x123,0x0cb,0x000126,0x126,0x0cd,0x000129,0x129,0x0cf,0x00012d,0x12d,0x0d1,0x00012f,0x12f,0x0d1,0x000131,0x131)
NTDLL(NtPrivilegedServiceAuditAlarm                     ,0x75,0x088,0x08e,0x0c6,0x0c6,0x0d1,0x103,0x103,0x0d3,0x109,0x109,0x0c7,0x00011b,0x11b,0x0ca,0x00011e,0x11e,0x0cc,0x000124,0x124,0x0cc,0x000127,0x127,0x0ce,0x00012a,0x12a,0x0d0,0x00012e,0x12e,0x0d2,0x000130,0x130,0x0d2,0x000132,0x132)
NTDLL(NtPropagationComplete                             ,NONE, NONE, NONE, NONE, NONE,0x175,0x104,0x104,0x0d5,0x10a,0x10a,0x0c5,0x00011c,0x11c,0x0c8,0x00011f,0x11f,0x0ca,0x000125,0x125,0x0ca,0x000128,0x128,0x0cc,0x00012b,0x12b,0x0ce,0x00012f,0x12f,0x0d0,0x000131,0x131,0x0d0,0x000133,0x133)
NTDLL(NtPropagationFailed                               ,NONE, NONE, NONE, NONE, NONE,0x176,0x105,0x105,0x0d6,0x10b,0x10b,0x0c4,0x00011d,0x11d,0x0c7,0x000120,0x120,0x0c9,0x000126,0x126,0x0c9,0x000129,0x129,0x0cb,0x00012c,0x12c,0x0cd,0x000130,0x130,0x0cf,0x000132,0x132,0x0cf,0x000134,0x134)
NTDLL(NtProtectVirtualMemory                            ,0x77,0x089,0x08f,0x04d,0x04d,0x0d2,0x04d,0x04d,0x0d7,0x04d,0x04d,0x0c3,0x00004e,0x04e,0x0c6,0x00004f,0x04f,0x0c8,0x000050,0x050,0x0c8,0x000050,0x050,0x0ca,0x000050,0x050,0x0cc,0x000050,0x050,0x0ce,0x000050,0x050,0x0ce,0x000050,0x050)
NTDLL(NtPulseEvent                                      ,0x78,0x08a,0x090,0x0c7,0x0c7,0x0d3,0x106,0x106,0x0d8,0x10c,0x10c,0x0c2,0x07011e,0x11e,0x0c5,0x070121,0x121,0x0c7,0x070127,0x127,0x0c7,0x07012a,0x12a,0x0c9,0x07012d,0x12d,0x0cb,0x070131,0x131,0x0cd,0x070133,0x133,0x0cd,0x070135,0x135)
NTDLL(NtQueryAttributesFile                             ,0x7a,0x08b,0x091,0x03a,0x03a,0x0d4,0x03a,0x03a,0x0d9,0x03a,0x03a,0x0c1,0x00003b,0x03b,0x0c4,0x00003c,0x03c,0x0c6,0x00003d,0x03d,0x0c6,0x00003d,0x03d,0x0c8,0x00003d,0x03d,0x0ca,0x00003d,0x03d,0x0cc,0x00003d,0x03d,0x0cc,0x00003d,0x03d)
NTDLL(NtQueryAuxiliaryCounterFrequency                  ,NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE,0x092,0x000132,0x132,0x093,0x000134,0x134,0x093,0x000136,0x136)
NTDLL(NtQueryBootEntryOrder                             ,NONE,0x08c,0x092,0x0c8,0x0c8,0x0d5,0x107,0x107,0x0da,0x10d,0x10d,0x0c0,0x00011f,0x11f,0x0c3,0x000122,0x122,0x0c5,0x000128,0x128,0x0c5,0x00012b,0x12b,0x0c7,0x00012e,0x12e,0x0c9,0x000133,0x133,0x0cb,0x000135,0x135,0x0cb,0x000137,0x137)
NTDLL(NtQueryBootOptions                                ,NONE,0x08d,0x093,0x0c9,0x0c9,0x0d6,0x108,0x108,0x0db,0x10e,0x10e,0x0bf,0x000120,0x120,0x0c2,0x000123,0x123,0x0c4,0x000129,0x129,0x0c4,0x00012c,0x12c,0x0c6,0x00012f,0x12f,0x0c8,0x000134,0x134,0x0ca,0x000136,0x136,0x0ca,0x000138,0x138)
NTDLL(NtQueryDebugFilterState                           ,NONE,0x08e,0x094,0x0ca,0x0ca,0x0d7,0x109,0x109,0x0dc,0x10f,0x10f,0x0be,0x050121,0x121,0x0c1,0x050124,0x124,0x0c3,0x05012a,0x12a,0x0c3,0x05012d,0x12d,0x0c5,0x050130,0x130,0x0c7,0x050135,0x135,0x0c9,0x050137,0x137,0x0c9,0x050139,0x139)
NTDLL(NtQueryDefaultLocale                              ,0x7b,0x08f,0x095,0x012,0x012,0x0d8,0x012,0x012,0x0dd,0x012,0x012,0x0bd,0x050013,0x013,0x0c0,0x050014,0x014,0x0c2,0x050015,0x015,0x0c2,0x050015,0x015,0x0c4,0x050015,0x015,0x0c6,0x050015,0x015,0x0c8,0x050015,0x015,0x0c8,0x050015,0x015)
NTDLL(NtQueryDefaultUILanguage                          ,0x7c,0x090,0x096,0x041,0x041,0x0d9,0x041,0x041,0x0de,0x041,0x041,0x0bc,0x040042,0x042,0x0bf,0x040043,0x043,0x0c1,0x040044,0x044,0x0c1,0x040044,0x044,0x0c3,0x040044,0x044,0x0c5,0x040044,0x044,0x0c7,0x040044,0x044,0x0c7,0x040044,0x044)
NTDLL(NtQueryDirectoryFile                              ,0x7d,0x091,0x097,0x032,0x032,0x0da,0x032,0x032,0x0df,0x032,0x032,0x0bb,0x000033,0x033,0x0be,0x000034,0x034,0x0c0,0x000035,0x035,0x0c0,0x000035,0x035,0x0c2,0x000035,0x035,0x0c4,0x000035,0x035,0x0c5,0x000035,0x035,0x0c5,0x000035,0x035)
NTDLL(NtQueryDirectoryFileEx                            ,NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE,0x0c6,0x000138,0x138,0x0c6,0x00013a,0x13a)
NTDLL(NtQueryDirectoryObject                            ,0x7e,0x092,0x098,0x0cb,0x0cb,0x0db,0x10a,0x10a,0x0e0,0x110,0x110,0x0ba,0x000122,0x122,0x0bd,0x000125,0x125,0x0bf,0x00012b,0x12b,0x0bf,0x00012e,0x12e,0x0c1,0x000131,0x131,0x0c3,0x000136,0x136,0x0c4,0x000139,0x139,0x0c4,0x00013b,0x13b)
NTDLL(NtQueryDriverEntryOrder                           ,NONE, NONE,0x099,0x0cc,0x0cc,0x0dc,0x10b,0x10b,0x0e1,0x111,0x111,0x0b9,0x000123,0x123,0x0bc,0x000126,0x126,0x0be,0x00012c,0x12c,0x0be,0x00012f,0x12f,0x0c0,0x000132,0x132,0x0c2,0x000137,0x137,0x0c3,0x00013a,0x13a,0x0c3,0x00013c,0x13c)
NTDLL(NtQueryEaFile                                     ,0x7f,0x093,0x09a,0x0cd,0x0cd,0x0dd,0x10c,0x10c,0x0e2,0x112,0x112,0x0b8,0x000124,0x124,0x0bb,0x000127,0x127,0x0bd,0x00012d,0x12d,0x0bd,0x000130,0x130,0x0bf,0x000133,0x133,0x0c1,0x000138,0x138,0x0c2,0x00013b,0x13b,0x0c2,0x00013d,0x13d)
NTDLL(NtQueryEvent                                      ,0x80,0x094,0x09b,0x053,0x053,0x0de,0x053,0x053,0x0e3,0x053,0x053,0x0b7,0x000054,0x054,0x0ba,0x000055,0x055,0x0bc,0x000056,0x056,0x0bc,0x000056,0x056,0x0be,0x000056,0x056,0x0c0,0x000056,0x056,0x0c1,0x000056,0x056,0x0c1,0x000056,0x056)
NTDLL(NtQueryFullAttributesFile                         ,0x81,0x095,0x09c,0x0ce,0x0ce,0x0df,0x10d,0x10d,0x0e4,0x113,0x113,0x0b6,0x000125,0x125,0x0b9,0x000128,0x128,0x0bb,0x00012e,0x12e,0x0bb,0x000131,0x131,0x0bd,0x000134,0x134,0x0bf,0x000139,0x139,0x0c0,0x00013c,0x13c,0x0c0,0x00013e,0x13e)
NTDLL(NtQueryInformationAtom                            ,0x79,0x096,0x09d,0x0cf,0x0cf,0x0e0,0x10e,0x10e,0x0e5,0x114,0x114,0x0b5,0x000126,0x126,0x0b8,0x000129,0x129,0x0ba,0x00012f,0x12f,0x0ba,0x000132,0x132,0x0bc,0x000135,0x135,0x0be,0x00013a,0x13a,0x0bf,0x00013d,0x13d,0x0bf,0x00013f,0x13f)
NTDLL(NtQueryInformationByName                          ,NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE,0x0bd,0x00013b,0x13b,0x0be,0x00013e,0x13e,0x0be,0x000140,0x140)
NTDLL(NtQueryInformationEnlistment                      ,NONE, NONE, NONE, NONE, NONE,0x072,0x10f,0x10f,0x0e6,0x115,0x115,0x0b4,0x000127,0x127,0x0b7,0x00012a,0x12a,0x0b9,0x000130,0x130,0x0b9,0x000133,0x133,0x0bb,0x000136,0x136,0x0bc,0x00013c,0x13c,0x0bd,0x00013f,0x13f,0x0bd,0x000141,0x141)
NTDLL(NtQueryInformationFile                            ,0x82,0x097,0x09e,0x00e,0x00e,0x0e1,0x00e,0x00e,0x0e7,0x00e,0x00e,0x0b3,0x00000f,0x00f,0x0b6,0x000010,0x010,0x0b8,0x000011,0x011,0x0b8,0x000011,0x011,0x0ba,0x000011,0x011,0x0bb,0x000011,0x011,0x0bc,0x000011,0x011,0x0bc,0x000011,0x011)
NTDLL(NtQueryInformationJobObject                       ,0x83,0x098,0x09f,0x0d0,0x0d0,0x0e2,0x110,0x110,0x0e8,0x116,0x116,0x0b2,0x000128,0x128,0x0b5,0x00012b,0x12b,0x0b7,0x000131,0x131,0x0b7,0x000134,0x134,0x0b9,0x000137,0x137,0x0ba,0x00013d,0x13d,0x0bb,0x000140,0x140,0x0bb,0x000142,0x142)
NTDLL(NtQueryInformationPort                            ,0x85,0x099,0x0a0,0x0d1,0x0d1,0x0e3,0x111,0x111,0x0e9,0x117,0x117,0x0b1,0x000129,0x129,0x0b4,0x00012c,0x12c,0x0b6,0x000132,0x132,0x0b6,0x000135,0x135,0x0b8,0x000138,0x138,0x0b9,0x00013e,0x13e,0x0ba,0x000141,0x141,0x0ba,0x000143,0x143)
NTDLL(NtQueryInformationProcess                         ,0x86,0x09a,0x0a1,0x016,0x016,0x0e4,0x016,0x016,0x0ea,0x016,0x016,0x0b0,0x000017,0x017,0x0b3,0x000018,0x018,0x0b5,0x000019,0x019,0x0b5,0x000019,0x019,0x0b7,0x000019,0x019,0x0b8,0x000019,0x019,0x0b9,0x000019,0x019,0x0b9,0x000019,0x019)
NTDLL(NtQueryInformationResourceManager                 ,NONE, NONE, NONE, NONE, NONE,0x06e,0x112,0x112,0x0eb,0x118,0x118,0x0af,0x00012a,0x12a,0x0b2,0x00012d,0x12d,0x0b4,0x000133,0x133,0x0b4,0x000136,0x136,0x0b6,0x000139,0x139,0x0b7,0x00013f,0x13f,0x0b8,0x000142,0x142,0x0b8,0x000144,0x144)
NTDLL(NtQueryInformationThread                          ,0x87,0x09b,0x0a2,0x022,0x022,0x0e5,0x022,0x022,0x0ec,0x022,0x022,0x0ae,0x000023,0x023,0x0b1,0x000024,0x024,0x0b3,0x000025,0x025,0x0b3,0x000025,0x025,0x0b5,0x000025,0x025,0x0b6,0x000025,0x025,0x0b7,0x000025,0x025,0x0b7,0x000025,0x025)
NTDLL(NtQueryInformationToken                           ,0x88,0x09c,0x0a3,0x01e,0x01e,0x0e6,0x01e,0x01e,0x0ed,0x01e,0x01e,0x0ad,0x00001f,0x01f,0x0b0,0x000020,0x020,0x0b2,0x000021,0x021,0x0b2,0x000021,0x021,0x0b4,0x000021,0x021,0x0b5,0x000021,0x021,0x0b6,0x000021,0x021,0x0b6,0x000021,0x021)
NTDLL(NtQueryInformationTransaction                     ,NONE, NONE, NONE, NONE, NONE,0x053,0x113,0x113,0x0ee,0x119,0x119,0x0ac,0x00012b,0x12b,0x0af,0x00012e,0x12e,0x0b1,0x000134,0x134,0x0b1,0x000137,0x137,0x0b3,0x00013a,0x13a,0x0b4,0x000140,0x140,0x0b5,0x000143,0x143,0x0b5,0x000145,0x145)
NTDLL(NtQueryInformationTransactionManager              ,NONE, NONE, NONE, NONE, NONE,0x054,0x114,0x114,0x0ef,0x11a,0x11a,0x0ab,0x00012c,0x12c,0x0ae,0x00012f,0x12f,0x0b0,0x000135,0x135,0x0b0,0x000138,0x138,0x0b2,0x00013b,0x13b,0x0b3,0x000141,0x141,0x0b4,0x000144,0x144,0x0b4,0x000146,0x146)
NTDLL(NtQueryInformationWorkerFactory                   ,NONE, NONE, NONE, NONE, NONE,0x17b,0x115,0x115,0x0f0,0x11b,0x11b,0x0aa,0x00012d,0x12d,0x0ad,0x000130,0x130,0x0af,0x000136,0x136,0x0af,0x000139,0x139,0x0b1,0x00013c,0x13c,0x0b2,0x000142,0x142,0x0b3,0x000145,0x145,0x0b3,0x000147,0x147)
NTDLL(NtQueryInstallUILanguage                          ,0x89,0x09d,0x0a4,0x0d2,0x0d2,0x0e7,0x116,0x116,0x0f1,0x11c,0x11c,0x0a9,0x04012e,0x12e,0x0ac,0x040131,0x131,0x0ae,0x040137,0x137,0x0ae,0x04013a,0x13a,0x0b0,0x04013d,0x13d,0x0b1,0x040143,0x143,0x0b2,0x040146,0x146,0x0b2,0x040148,0x148)
NTDLL(NtQueryIntervalProfile                            ,0x8a,0x09e,0x0a5,0x0d3,0x0d3,0x0e8,0x117,0x117,0x0f2,0x11d,0x11d,0x0a8,0x05012f,0x12f,0x0ab,0x050132,0x132,0x0ad,0x050138,0x138,0x0ad,0x05013b,0x13b,0x0af,0x05013e,0x13e,0x0b0,0x050144,0x144,0x0b1,0x050147,0x147,0x0b1,0x050149,0x149)
NTDLL(NtQueryIoCompletion                               ,0x84,0x09f,0x0a6,0x0d4,0x0d4,0x0e9,0x118,0x118,0x0f3,0x11e,0x11e,0x0a7,0x000130,0x130,0x0aa,0x000133,0x133,0x0ac,0x000139,0x139,0x0ac,0x00013c,0x13c,0x0ae,0x00013f,0x13f,0x0af,0x000145,0x145,0x0b0,0x000148,0x148,0x0b0,0x00014a,0x14a)
NTDLL(NtQueryKey                                        ,0x8b,0x0a0,0x0a7,0x013,0x013,0x0ea,0x013,0x013,0x0f4,0x013,0x013,0x0a6,0x000014,0x014,0x0a9,0x000015,0x015,0x0ab,0x000016,0x016,0x0ab,0x000016,0x016,0x0ad,0x000016,0x016,0x0ae,0x000016,0x016,0x0af,0x000016,0x016,0x0af,0x000016,0x016)
NTDLL(NtQueryLicenseValue                               ,NONE, NONE, NONE, NONE, NONE,0x180,0x119,0x119,0x0f5,0x11f,0x11f,0x0a5,0x000131,0x131,0x0a8,0x000134,0x134,0x0aa,0x00013a,0x13a,0x0aa,0x00013d,0x13d,0x0ac,0x000140,0x140,0x0ad,0x000146,0x146,0x0ae,0x000149,0x149,0x0ae,0x00014b,0x14b)
NTDLL(NtQueryMultipleValueKey                           ,0x8c,0x0a1,0x0a8,0x0d5,0x0d5,0x0eb,0x11a,0x11a,0x0f6,0x120,0x120,0x0a4,0x000132,0x132,0x0a7,0x000135,0x135,0x0a9,0x00013b,0x13b,0x0a9,0x00013e,0x13e,0x0ab,0x000141,0x141,0x0ac,0x000147,0x147,0x0ad,0x00014a,0x14a,0x0ad,0x00014c,0x14c)
NTDLL(NtQueryMutant                                     ,0x8d,0x0a2,0x0a9,0x0d6,0x0d6,0x0ec,0x11b,0x11b,0x0f7,0x121,0x121,0x0a3,0x000133,0x133,0x0a6,0x000136,0x136,0x0a8,0x00013c,0x13c,0x0a8,0x00013f,0x13f,0x0aa,0x000142,0x142,0x0ab,0x000148,0x148,0x0ac,0x00014b,0x14b,0x0ac,0x00014d,0x14d)
NTDLL(NtQueryObject                                     ,0x8e,0x0a3,0x0aa,0x00d,0x00d,0x0ed,0x00d,0x00d,0x0f8,0x00d,0x00d,0x0a2,0x00000e,0x00e,0x0a5,0x00000f,0x00f,0x0a7,0x000010,0x010,0x0a7,0x000010,0x010,0x0a9,0x000010,0x010,0x0aa,0x000010,0x010,0x0ab,0x000010,0x010,0x0ab,0x000010,0x010)
NTDLL(NtQueryOpenSubKeys                                ,0x8f,0x0a4,0x0ab,0x0d7,0x0d7,0x0ee,0x11c,0x11c,0x0f9,0x122,0x122,0x0a1,0x000134,0x134,0x0a4,0x000137,0x137,0x0a6,0x00013d,0x13d,0x0a6,0x000140,0x140,0x0a8,0x000143,0x143,0x0a9,0x000149,0x149,0x0aa,0x00014c,0x14c,0x0aa,0x00014e,0x14e)
NTDLL(NtQueryOpenSubKeysEx                              ,NONE, NONE,0x0ac,0x0d8,0x0d8,0x0ef,0x11d,0x11d,0x0fa,0x123,0x123,0x0a0,0x000135,0x135,0x0a3,0x000138,0x138,0x0a5,0x00013e,0x13e,0x0a5,0x000141,0x141,0x0a7,0x000144,0x144,0x0a8,0x00014a,0x14a,0x0a9,0x00014d,0x14d,0x0a9,0x00014f,0x14f)
NTDLL(NtQueryPerformanceCounter                         ,0x90,0x0a5,0x0ad,0x02e,0x02e,0x0f0,0x02e,0x02e,0x0fb,0x02e,0x02e,0x09f,0x05002f,0x02f,0x0a2,0x050030,0x030,0x0a4,0x050031,0x031,0x0a4,0x050031,0x031,0x0a6,0x050031,0x031,0x0a7,0x050031,0x031,0x0a8,0x050031,0x031,0x0a8,0x050031,0x031)
NTDLL(NtQueryPortInformationProcess                     ,NONE,0x11b,0x125,0x0d9,0x0d9,0x16c,0x11e,0x11e,0x0fc,0x124,0x124,0x09e,0x010136,0x136,0x0a1,0x010139,0x139,0x0a3,0x01013f,0x13f,0x0a3,0x010142,0x142,0x0a5,0x010145,0x145,0x0a6,0x01014b,0x14b,0x0a7,0x01014e,0x14e,0x0a7,0x010150,0x150)
NTDLL(NtQueryQuotaInformationFile                       ,0x91,0x0a6,0x0ae,0x0da,0x0da,0x0f1,0x11f,0x11f,0x0fd,0x125,0x125,0x09d,0x000137,0x137,0x0a0,0x00013a,0x13a,0x0a2,0x000140,0x140,0x0a2,0x000143,0x143,0x0a4,0x000146,0x146,0x0a5,0x00014c,0x14c,0x0a6,0x00014f,0x14f,0x0a6,0x000151,0x151)
NTDLL(NtQuerySection                                    ,0x92,0x0a7,0x0af,0x04e,0x04e,0x0f2,0x04e,0x04e,0x0fe,0x04e,0x04e,0x09c,0x00004f,0x04f,0x09f,0x000050,0x050,0x0a1,0x000051,0x051,0x0a1,0x000051,0x051,0x0a3,0x000051,0x051,0x0a4,0x000051,0x051,0x0a5,0x000051,0x051,0x0a5,0x000051,0x051)
NTDLL(NtQuerySecurityAttributesToken                    ,NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,0x0ff,0x126,0x126,0x09b,0x000138,0x138,0x09e,0x00013b,0x13b,0x0a0,0x000141,0x141,0x0a0,0x000144,0x144,0x0a2,0x000147,0x147,0x0a3,0x00014d,0x14d,0x0a4,0x000150,0x150,0x0a4,0x000152,0x152)
NTDLL(NtQuerySecurityObject                             ,0x93,0x0a8,0x0b0,0x0db,0x0db,0x0f3,0x120,0x120,0x100,0x127,0x127,0x09a,0x000139,0x139,0x09d,0x00013c,0x13c,0x09f,0x000142,0x142,0x09f,0x000145,0x145,0x0a1,0x000148,0x148,0x0a2,0x00014e,0x14e,0x0a3,0x000151,0x151,0x0a3,0x000153,0x153)
NTDLL(NtQuerySecurityPolicy                             ,NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE,0x0a0,0x000149,0x149,0x0a1,0x00014f,0x14f,0x0a2,0x000152,0x152,0x0a2,0x000154,0x154)
NTDLL(NtQuerySemaphore                                  ,0x94,0x0a9,0x0b1,0x0dc,0x0dc,0x0f4,0x121,0x121,0x101,0x128,0x128,0x099,0x00013a,0x13a,0x09c,0x00013d,0x13d,0x09e,0x000143,0x143,0x09e,0x000146,0x146,0x09f,0x00014a,0x14a,0x0a0,0x000150,0x150,0x0a1,0x000153,0x153,0x0a1,0x000155,0x155)
NTDLL(NtQuerySymbolicLinkObject                         ,0x95,0x0aa,0x0b2,0x0dd,0x0dd,0x0f5,0x122,0x122,0x102,0x129,0x129,0x098,0x00013b,0x13b,0x09b,0x00013e,0x13e,0x09d,0x000144,0x144,0x09d,0x000147,0x147,0x09e,0x00014b,0x14b,0x09f,0x000151,0x151,0x0a0,0x000154,0x154,0x0a0,0x000156,0x156)
NTDLL(NtQuerySystemEnvironmentValue                     ,0x96,0x0ab,0x0b3,0x0de,0x0de,0x0f6,0x123,0x123,0x103,0x12a,0x12a,0x097,0x00013c,0x13c,0x09a,0x00013f,0x13f,0x09c,0x000145,0x145,0x09c,0x000148,0x148,0x09d,0x00014c,0x14c,0x09e,0x000152,0x152,0x09f,0x000155,0x155,0x09f,0x000157,0x157)
NTDLL(NtQuerySystemEnvironmentValueEx                   ,NONE,0x0ac,0x0b4,0x0df,0x0df,0x0f7,0x124,0x124,0x104,0x12b,0x12b,0x096,0x00013d,0x13d,0x099,0x000140,0x140,0x09b,0x000146,0x146,0x09b,0x000149,0x149,0x09c,0x00014d,0x14d,0x09d,0x000153,0x153,0x09e,0x000156,0x156,0x09e,0x000158,0x158)
NTDLL(NtQuerySystemInformation                          ,0x97,0x0ad,0x0b5,0x033,0x033,0x0f8,0x033,0x033,0x105,0x033,0x033,0x095,0x000034,0x034,0x098,0x000035,0x035,0x09a,0x000036,0x036,0x09a,0x000036,0x036,0x09b,0x000036,0x036,0x09c,0x000036,0x036,0x09d,0x000036,0x036,0x09d,0x000036,0x036)
NTDLL(NtQuerySystemInformationEx                        ,NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,0x106,0x12c,0x12c,0x094,0x00013e,0x13e,0x097,0x000141,0x141,0x099,0x000147,0x147,0x099,0x00014a,0x14a,0x09a,0x00014e,0x14e,0x09b,0x000154,0x154,0x09c,0x000157,0x157,0x09c,0x000159,0x159)
NTDLL(NtQuerySystemTime                                 ,0x98,0x0ae,0x0b6,0x057, NONE,0x0f9,0x057, NONE,0x107,0x057, NONE,0x093,0x180058, NONE,0x096,0x180059, NONE,0x098,0x18005a, NONE,0x098,0x18005a, NONE,0x099,0x18005a, NONE,0x09a,0x18005a, NONE,0x09b,0x18005a, NONE,0x09b,0x18005a, NONE)
NTDLL(NtQueryTimer                                      ,0x99,0x0af,0x0b7,0x035,0x035,0x0fa,0x035,0x035,0x108,0x035,0x035,0x092,0x000036,0x036,0x095,0x000037,0x037,0x097,0x000038,0x038,0x097,0x000038,0x038,0x098,0x000038,0x038,0x099,0x000038,0x038,0x09a,0x000038,0x038,0x09a,0x000038,0x038)
NTDLL(NtQueryTimerResolution                            ,0x9a,0x0b0,0x0b8,0x0e0,0x0e0,0x0fb,0x125,0x125,0x109,0x12d,0x12d,0x091,0x0a013f,0x13f,0x094,0x0a0142,0x142,0x096,0x0a0148,0x148,0x096,0x0a014b,0x14b,0x097,0x0a014f,0x14f,0x098,0x0a0155,0x155,0x099,0x0a0158,0x158,0x099,0x0a015a,0x15a)
NTDLL(NtQueryValueKey                                   ,0x9b,0x0b1,0x0b9,0x014,0x014,0x0fc,0x014,0x014,0x10a,0x014,0x014,0x090,0x000015,0x015,0x093,0x000016,0x016,0x095,0x000017,0x017,0x095,0x000017,0x017,0x096,0x000017,0x017,0x097,0x000017,0x017,0x098,0x000017,0x017,0x098,0x000017,0x017)
NTDLL(NtQueryVirtualMemory                              ,0x9c,0x0b2,0x0ba,0x020,0x020,0x0fd,0x020,0x020,0x10b,0x020,0x020,0x08f,0x000021,0x021,0x092,0x000022,0x022,0x094,0x000023,0x023,0x094,0x000023,0x023,0x095,0x000023,0x023,0x096,0x000023,0x023,0x097,0x000023,0x023,0x097,0x000023,0x023)
NTDLL(NtQueryVolumeInformationFile                      ,0x9d,0x0b3,0x0bb,0x046,0x046,0x0fe,0x046,0x046,0x10c,0x046,0x046,0x08e,0x000047,0x047,0x091,0x000048,0x048,0x093,0x000049,0x049,0x093,0x000049,0x049,0x094,0x000049,0x049,0x095,0x000049,0x049,0x096,0x000049,0x049,0x096,0x000049,0x049)
NTDLL(NtQueryWnfStateData                               ,NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,0x08c,0x000140,0x140,0x08f,0x000143,0x143,0x091,0x000149,0x149,0x091,0x00014c,0x14c,0x092,0x000150,0x150,0x093,0x000156,0x156,0x094,0x000159,0x159,0x094,0x00015b,0x15b)
NTDLL(NtQueryWnfStateNameInformation                    ,NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,0x08d,0x000141,0x141,0x090,0x000144,0x144,0x092,0x00014a,0x14a,0x092,0x00014d,0x14d,0x093,0x000151,0x151,0x094,0x000157,0x157,0x095,0x00015a,0x15a,0x095,0x00015c,0x15c)
NTDLL(NtQueueApcThread                                  ,0x9e,0x0b4,0x0bc,0x042,0x042,0x0ff,0x042,0x042,0x10d,0x042,0x042,0x08b,0x000043,0x043,0x08e,0x000044,0x044,0x090,0x000045,0x045,0x090,0x000045,0x045,0x091,0x000045,0x045,0x091,0x000045,0x045,0x092,0x000045,0x045,0x092,0x000045,0x045)
NTDLL(NtQueueApcThreadEx                                ,NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,0x10e,0x12e,0x12e,0x08a,0x000142,0x142,0x08d,0x000145,0x145,0x08f,0x00014b,0x14b,0x08f,0x00014e,0x14e,0x090,0x000152,0x152,0x090,0x000158,0x158,0x091,0x00015b,0x15b,0x091,0x00015d,0x15d)
NTDLL(NtRaiseException                                  ,0x9f,0x0b5,0x0bd,0x0e1,0x0e1,0x100,0x126,0x126,0x10f,0x12f,0x12f,0x089,0x000143,0x143,0x08c,0x000146,0x146,0x08e,0x00014c,0x14c,0x08e,0x00014f,0x14f,0x08f,0x000153,0x153,0x08f,0x000159,0x159,0x090,0x00015c,0x15c,0x090,0x00015e,0x15e)
NTDLL(NtRaiseHardError                                  ,0xa0,0x0b6,0x0be,0x0e2,0x0e2,0x101,0x127,0x127,0x110,0x130,0x130,0x088,0x000144,0x144,0x08b,0x000147,0x147,0x08d,0x00014d,0x14d,0x08d,0x000150,0x150,0x08e,0x000154,0x154,0x08e,0x00015a,0x15a,0x08f,0x00015d,0x15d,0x08f,0x00015f,0x15f)
NTDLL(NtReadFile                                        ,0xa1,0x0b7,0x0bf,0x003,0x003,0x102,0x003,0x003,0x111,0x003,0x003,0x087,0x1a0004,0x004,0x08a,0x1a0005,0x005,0x08c,0x1a0006,0x006,0x08c,0x1a0006,0x006,0x08d,0x1a0006,0x006,0x08d,0x1a0006,0x006,0x08e,0x1a0006,0x006,0x08e,0x1a0006,0x006)
NTDLL(NtReadFileScatter                                 ,0xa2,0x0b8,0x0c0,0x02b,0x02b,0x103,0x02b,0x02b,0x112,0x02b,0x02b,0x086,0x1a002c,0x02c,0x089,0x1a002d,0x02d,0x08b,0x1a002e,0x02e,0x08b,0x1a002e,0x02e,0x08c,0x1a002e,0x02e,0x08c,0x1a002e,0x02e,0x08d,0x1a002e,0x02e,0x08d,0x1a002e,0x02e)
NTDLL(NtReadOnlyEnlistment                              ,NONE, NONE, NONE, NONE, NONE,0x058,0x128,0x128,0x113,0x131,0x131,0x085,0x000145,0x145,0x088,0x000148,0x148,0x08a,0x00014e,0x14e,0x08a,0x000151,0x151,0x08b,0x000155,0x155,0x08b,0x00015b,0x15b,0x08c,0x00015e,0x15e,0x08c,0x000160,0x160)
NTDLL(NtReadRequestData                                 ,0xa3,0x0b9,0x0c1,0x051,0x051,0x104,0x051,0x051,0x114,0x051,0x051,0x084,0x000052,0x052,0x087,0x000053,0x053,0x089,0x000054,0x054,0x089,0x000054,0x054,0x08a,0x000054,0x054,0x08a,0x000054,0x054,0x08b,0x000054,0x054,0x08b,0x000054,0x054)
NTDLL(NtReadVirtualMemory                               ,0xa4,0x0ba,0x0c2,0x03c,0x03c,0x105,0x03c,0x03c,0x115,0x03c,0x03c,0x083,0x00003d,0x03d,0x086,0x00003e,0x03e,0x088,0x00003f,0x03f,0x088,0x00003f,0x03f,0x089,0x00003f,0x03f,0x089,0x00003f,0x03f,0x08a,0x00003f,0x03f,0x08a,0x00003f,0x03f)
NTDLL(NtRecoverEnlistment                               ,NONE, NONE, NONE, NONE, NONE,0x068,0x129,0x129,0x116,0x132,0x132,0x082,0x000146,0x146,0x085,0x000149,0x149,0x087,0x00014f,0x14f,0x087,0x000152,0x152,0x088,0x000156,0x156,0x088,0x00015c,0x15c,0x089,0x00015f,0x15f,0x089,0x000161,0x161)
NTDLL(NtRecoverResourceManager                          ,NONE, NONE, NONE, NONE, NONE,0x069,0x12a,0x12a,0x117,0x133,0x133,0x081,0x000147,0x147,0x084,0x00014a,0x14a,0x086,0x000150,0x150,0x086,0x000153,0x153,0x087,0x000157,0x157,0x087,0x00015d,0x15d,0x088,0x000160,0x160,0x088,0x000162,0x162)
NTDLL(NtRecoverTransactionManager                       ,NONE, NONE, NONE, NONE, NONE,0x06a,0x12b,0x12b,0x118,0x134,0x134,0x080,0x000148,0x148,0x083,0x00014b,0x14b,0x085,0x000151,0x151,0x085,0x000154,0x154,0x086,0x000158,0x158,0x086,0x00015e,0x15e,0x087,0x000161,0x161,0x087,0x000163,0x163)
NTDLL(NtRegisterProtocolAddressInformation              ,NONE, NONE, NONE, NONE, NONE,0x174,0x12c,0x12c,0x119,0x135,0x135,0x07f,0x000149,0x149,0x082,0x00014c,0x14c,0x084,0x000152,0x152,0x084,0x000155,0x155,0x085,0x000159,0x159,0x085,0x00015f,0x15f,0x086,0x000162,0x162,0x086,0x000164,0x164)
NTDLL(NtRegisterThreadTerminatePort                     ,0xa5,0x0bb,0x0c3,0x0e3,0x0e3,0x106,0x12d,0x12d,0x11a,0x136,0x136,0x07e,0x03014a,0x14a,0x081,0x03014d,0x14d,0x083,0x030153,0x153,0x083,0x030156,0x156,0x084,0x03015a,0x15a,0x084,0x030160,0x160,0x085,0x030163,0x163,0x085,0x030165,0x165)
NTDLL(NtReleaseCMFViewOwnership                         ,NONE, NONE, NONE, NONE, NONE,0x186,0x12e,0x12e, NONE, NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE)
NTDLL(NtReleaseKeyedEvent                               ,NONE,0x119,0x123,0x0e4,0x0e4,0x16a,0x12f,0x12f,0x11b,0x137,0x137,0x07d,0x14014b,0x14b,0x080,0x14014e,0x14e,0x082,0x140154,0x154,0x082,0x140157,0x157,0x083,0x14015b,0x15b,0x083,0x140161,0x161,0x084,0x140164,0x164,0x084,0x140166,0x166)
NTDLL(NtReleaseMutant                                   ,0xa6,0x0bc,0x0c4,0x01d,0x01d,0x107,0x01d,0x01d,0x11c,0x01d,0x01d,0x07c,0x07001e,0x01e,0x07f,0x07001f,0x01f,0x081,0x070020,0x020,0x081,0x070020,0x020,0x082,0x070020,0x020,0x082,0x070020,0x020,0x083,0x070020,0x020,0x083,0x070020,0x020)
NTDLL(NtReleaseSemaphore                                ,0xa7,0x0bd,0x0c5,0x007,0x007,0x108,0x007,0x007,0x11d,0x007,0x007,0x07b,0x0c0008,0x008,0x07e,0x0c0009,0x009,0x080,0x0c000a,0x00a,0x080,0x0c000a,0x00a,0x081,0x0c000a,0x00a,0x081,0x0c000a,0x00a,0x082,0x0c000a,0x00a,0x082,0x0c000a,0x00a)
NTDLL(NtReleaseWorkerFactoryWorker                      ,NONE, NONE, NONE, NONE, NONE,0x178,0x130,0x130,0x11e,0x138,0x138,0x07a,0x00014c,0x14c,0x07d,0x03014f,0x14f,0x07f,0x030155,0x155,0x07f,0x030158,0x158,0x080,0x03015c,0x15c,0x080,0x030162,0x162,0x081,0x030165,0x165,0x081,0x030167,0x167)
NTDLL(NtRemoveIoCompletion                              ,0xa8,0x0be,0x0c6,0x006,0x006,0x109,0x006,0x006,0x11f,0x006,0x006,0x079,0x1c0007,0x007,0x07c,0x1c0008,0x008,0x07e,0x1c0009,0x009,0x07e,0x1c0009,0x009,0x07f,0x1c0009,0x009,0x07f,0x1c0009,0x009,0x080,0x1c0009,0x009,0x080,0x1c0009,0x009)
NTDLL(NtRemoveIoCompletionEx                            ,NONE, NONE, NONE, NONE, NONE,0x173,0x131,0x131,0x120,0x139,0x139,0x078,0x00014d,0x14d,0x07b,0x000150,0x150,0x07d,0x000156,0x156,0x07d,0x000159,0x159,0x07e,0x00015d,0x15d,0x07e,0x000163,0x163,0x07f,0x000166,0x166,0x07f,0x000168,0x168)
NTDLL(NtRemoveProcessDebug                              ,NONE,0x0bf,0x0c7,0x0e5,0x0e5,0x10a,0x132,0x132,0x121,0x13a,0x13a,0x077,0x08014e,0x14e,0x07a,0x080151,0x151,0x07c,0x080157,0x157,0x07c,0x08015a,0x15a,0x07d,0x08015e,0x15e,0x07d,0x080164,0x164,0x07e,0x080167,0x167,0x07e,0x080169,0x169)
NTDLL(NtRenameKey                                       ,NONE,0x0c0,0x0c8,0x0e6,0x0e6,0x10b,0x133,0x133,0x122,0x13b,0x13b,0x076,0x00014f,0x14f,0x079,0x000152,0x152,0x07b,0x000158,0x158,0x07b,0x00015b,0x15b,0x07c,0x00015f,0x15f,0x07c,0x000165,0x165,0x07d,0x000168,0x168,0x07d,0x00016a,0x16a)
NTDLL(NtRenameTransactionManager                        ,NONE, NONE, NONE, NONE, NONE,0x066,0x134,0x134,0x123,0x13c,0x13c,0x075,0x000150,0x150,0x078,0x000153,0x153,0x07a,0x000159,0x159,0x07a,0x00015c,0x15c,0x07b,0x000160,0x160,0x07b,0x000166,0x166,0x07c,0x000169,0x169,0x07c,0x00016b,0x16b)
NTDLL(NtReplaceKey                                      ,0xa9,0x0c1,0x0c9,0x0e7,0x0e7,0x10c,0x135,0x135,0x124,0x13d,0x13d,0x074,0x000151,0x151,0x077,0x000154,0x154,0x079,0x00015a,0x15a,0x079,0x00015d,0x15d,0x07a,0x000161,0x161,0x07a,0x000167,0x167,0x07b,0x00016a,0x16a,0x07b,0x00016c,0x16c)
NTDLL(NtReplacePartitionUnit                            ,NONE, NONE, NONE, NONE, NONE,0x10d,0x136,0x136,0x125,0x13e,0x13e,0x073,0x000152,0x152,0x076,0x000155,0x155,0x078,0x00015b,0x15b,0x078,0x00015e,0x15e,0x079,0x000162,0x162,0x079,0x000168,0x168,0x07a,0x00016b,0x16b,0x07a,0x00016d,0x16d)
NTDLL(NtReplyPort                                       ,0xaa,0x0c2,0x0ca,0x009,0x009,0x10e,0x009,0x009,0x126,0x009,0x009,0x072,0x00000a,0x00a,0x075,0x00000b,0x00b,0x077,0x00000c,0x00c,0x077,0x00000c,0x00c,0x078,0x00000c,0x00c,0x078,0x00000c,0x00c,0x079,0x00000c,0x00c,0x079,0x00000c,0x00c)
NTDLL(NtReplyWaitReceivePort                            ,0xab,0x0c3,0x0cb,0x008,0x008,0x10f,0x008,0x008,0x127,0x008,0x008,0x071,0x000009,0x009,0x074,0x00000a,0x00a,0x076,0x00000b,0x00b,0x076,0x00000b,0x00b,0x077,0x00000b,0x00b,0x077,0x00000b,0x00b,0x078,0x00000b,0x00b,0x078,0x00000b,0x00b)
NTDLL(NtReplyWaitReceivePortEx                          ,0xac,0x0c4,0x0cc,0x028,0x028,0x110,0x028,0x028,0x128,0x028,0x028,0x070,0x000029,0x029,0x073,0x00002a,0x02a,0x075,0x00002b,0x02b,0x075,0x00002b,0x02b,0x076,0x00002b,0x02b,0x076,0x00002b,0x02b,0x077,0x00002b,0x02b,0x077,0x00002b,0x02b)
NTDLL(NtReplyWaitReplyPort                              ,0xad,0x0c5,0x0cd,0x0e8,0x0e8,0x111,0x137,0x137,0x129,0x13f,0x13f,0x06f,0x000153,0x153,0x072,0x000156,0x156,0x074,0x00015c,0x15c,0x074,0x00015f,0x15f,0x075,0x000163,0x163,0x075,0x000169,0x169,0x076,0x00016c,0x16c,0x076,0x00016e,0x16e)
NTDLL(NtReplyWaitSendChannel                            ,0xf4, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE)
NTDLL(NtRequestDeviceWakeup                             ,0xae,0x0c6,0x0ce,0x0e9,0x0e9,0x112,0x138,0x138, NONE, NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE)
NTDLL(NtRequestPort                                     ,0xaf,0x0c7,0x0cf,0x0ea,0x0ea,0x113,0x139,0x139,0x12a,0x140,0x140,0x06e,0x000154,0x154,0x071,0x000157,0x157,0x073,0x00015d,0x15d,0x073,0x000160,0x160,0x074,0x000164,0x164,0x074,0x00016a,0x16a,0x075,0x00016d,0x16d,0x075,0x00016f,0x16f)
NTDLL(NtRequestWaitReplyPort                            ,0xb0,0x0c8,0x0d0,0x01f,0x01f,0x114,0x01f,0x01f,0x12b,0x01f,0x01f,0x06d,0x000020,0x020,0x070,0x000021,0x021,0x072,0x000022,0x022,0x072,0x000022,0x022,0x073,0x000022,0x022,0x073,0x000022,0x022,0x074,0x000022,0x022,0x074,0x000022,0x022)
NTDLL(NtRequestWakeupLatency                            ,0xb1,0x0c9,0x0d1,0x0eb,0x0eb,0x115,0x13a,0x13a, NONE, NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE)
NTDLL(NtResetEvent                                      ,0xb2,0x0ca,0x0d2,0x0ec,0x0ec,0x116,0x13b,0x13b,0x12c,0x141,0x141,0x06c,0x070155,0x155,0x06f,0x070158,0x158,0x071,0x07015e,0x15e,0x071,0x070161,0x161,0x072,0x070165,0x165,0x072,0x07016b,0x16b,0x073,0x07016e,0x16e,0x073,0x070170,0x170)
NTDLL(NtResetWriteWatch                                 ,0xb3,0x0cb,0x0d3,0x0ed,0x0ed,0x117,0x13c,0x13c,0x12d,0x142,0x142,0x06b,0x0c0156,0x156,0x06e,0x0c0159,0x159,0x070,0x0c015f,0x15f,0x070,0x0c0162,0x162,0x071,0x0c0166,0x166,0x071,0x0c016c,0x16c,0x072,0x0c016f,0x16f,0x072,0x0c0171,0x171)
NTDLL(NtRestoreKey                                      ,0xb4,0x0cc,0x0d4,0x0ee,0x0ee,0x118,0x13d,0x13d,0x12e,0x143,0x143,0x06a,0x000157,0x157,0x06d,0x00015a,0x15a,0x06f,0x000160,0x160,0x06f,0x000163,0x163,0x070,0x000167,0x167,0x070,0x00016d,0x16d,0x071,0x000170,0x170,0x071,0x000172,0x172)
NTDLL(NtResumeProcess                                   ,NONE,0x0cd,0x0d5,0x0ef,0x0ef,0x119,0x13e,0x13e,0x12f,0x144,0x144,0x069,0x030158,0x158,0x06c,0x03015b,0x15b,0x06e,0x030161,0x161,0x06e,0x030164,0x164,0x06f,0x030168,0x168,0x06f,0x03016e,0x16e,0x070,0x030171,0x171,0x070,0x030173,0x173)
NTDLL(NtResumeThread                                    ,0xb5,0x0ce,0x0d6,0x04f,0x04f,0x11a,0x04f,0x04f,0x130,0x04f,0x04f,0x068,0x070050,0x050,0x06b,0x070051,0x051,0x06d,0x070052,0x052,0x06d,0x070052,0x052,0x06e,0x070052,0x052,0x06e,0x070052,0x052,0x06f,0x070052,0x052,0x06f,0x070052,0x052)
NTDLL(NtRevertContainerImpersonation                    ,NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE,0x06c,0x000162,0x162,0x06c,0x000165,0x165,0x06d,0x000169,0x169,0x06d,0x00016f,0x16f,0x06e,0x000172,0x172,0x06e,0x000174,0x174)
NTDLL(NtRollbackComplete                                ,NONE, NONE, NONE, NONE, NONE,0x059,0x13f,0x13f,0x131,0x145,0x145,0x067,0x000159,0x159,0x06a,0x00015c,0x15c,0x06b,0x000163,0x163,0x06b,0x000166,0x166,0x06c,0x00016a,0x16a,0x06c,0x000170,0x170,0x06d,0x000173,0x173,0x06d,0x000175,0x175)
NTDLL(NtRollbackEnlistment                              ,NONE, NONE, NONE, NONE, NONE,0x05a,0x140,0x140,0x132,0x146,0x146,0x066,0x00015a,0x15a,0x069,0x00015d,0x15d,0x06a,0x000164,0x164,0x06a,0x000167,0x167,0x06b,0x00016b,0x16b,0x06b,0x000171,0x171,0x06c,0x000174,0x174,0x06c,0x000176,0x176)
NTDLL(NtRollbackRegistryTransaction                     ,NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE,0x1c1,0x00016c,0x16c,0x1c7,0x000172,0x172,0x1cb,0x000175,0x175,0x1cb,0x000177,0x177)
NTDLL(NtRollbackTransaction                             ,NONE, NONE, NONE, NONE, NONE,0x05c,0x141,0x141,0x133,0x147,0x147,0x065,0x00015b,0x15b,0x068,0x00015e,0x15e,0x069,0x000165,0x165,0x069,0x000168,0x168,0x06a,0x00016d,0x16d,0x06a,0x000173,0x173,0x06b,0x000176,0x176,0x06b,0x000178,0x178)
NTDLL(NtRollforwardTransactionManager                   ,NONE, NONE, NONE, NONE, NONE,0x067,0x142,0x142,0x134,0x148,0x148,0x064,0x00015c,0x15c,0x067,0x00015f,0x15f,0x068,0x000166,0x166,0x068,0x000169,0x169,0x069,0x00016e,0x16e,0x069,0x000174,0x174,0x06a,0x000177,0x177,0x06a,0x000179,0x179)
NTDLL(NtSaveKey                                         ,0xb6,0x0cf,0x0d7,0x0f0,0x0f0,0x11b,0x143,0x143,0x135,0x149,0x149,0x063,0x00015d,0x15d,0x066,0x080160,0x160,0x067,0x080167,0x167,0x067,0x08016a,0x16a,0x068,0x08016f,0x16f,0x068,0x080175,0x175,0x069,0x080178,0x178,0x069,0x08017a,0x17a)
NTDLL(NtSaveKeyEx                                       ,NONE,0x0d0,0x0d8,0x0f1,0x0f1,0x11c,0x144,0x144,0x136,0x14a,0x14a,0x062,0x00015e,0x15e,0x065,0x0e0161,0x161,0x066,0x0e0168,0x168,0x066,0x0e016b,0x16b,0x067,0x0e0170,0x170,0x067,0x0e0176,0x176,0x068,0x0e0179,0x179,0x068,0x0e017b,0x17b)
NTDLL(NtSaveMergedKeys                                  ,0xb7,0x0d1,0x0d9,0x0f2,0x0f2,0x11d,0x145,0x145,0x137,0x14b,0x14b,0x061,0x0b015f,0x15f,0x064,0x0b0162,0x162,0x065,0x0b0169,0x169,0x065,0x0b016c,0x16c,0x066,0x0b0171,0x171,0x066,0x0b0177,0x177,0x067,0x0b017a,0x17a,0x067,0x0b017c,0x17c)
NTDLL(NtSecureConnectPort                               ,0xb8,0x0d2,0x0da,0x0f3,0x0f3,0x11e,0x146,0x146,0x138,0x14c,0x14c,0x060,0x000160,0x160,0x063,0x000163,0x163,0x064,0x00016a,0x16a,0x064,0x00016d,0x16d,0x065,0x000172,0x172,0x065,0x000178,0x178,0x066,0x00017b,0x17b,0x066,0x00017d,0x17d)
NTDLL(NtSendWaitReplyChannel                            ,0xf5, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE)
NTDLL(NtSerializeBoot                                   ,NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,0x139,0x14d,0x14d,0x05f,0x000161,0x161,0x062,0x000164,0x164,0x063,0x00016b,0x16b,0x063,0x00016e,0x16e,0x064,0x000173,0x173,0x064,0x000179,0x179,0x065,0x00017c,0x17c,0x065,0x00017e,0x17e)
NTDLL(NtSetBootEntryOrder                               ,NONE,0x0d3,0x0db,0x0f4,0x0f4,0x11f,0x147,0x147,0x13a,0x14e,0x14e,0x05e,0x000162,0x162,0x061,0x000165,0x165,0x062,0x00016c,0x16c,0x062,0x00016f,0x16f,0x063,0x000174,0x174,0x063,0x00017a,0x17a,0x064,0x00017d,0x17d,0x064,0x00017f,0x17f)
NTDLL(NtSetBootOptions                                  ,NONE,0x0d4,0x0dc,0x0f5,0x0f5,0x120,0x148,0x148,0x13b,0x14f,0x14f,0x05d,0x000163,0x163,0x060,0x000166,0x166,0x061,0x00016d,0x16d,0x061,0x000170,0x170,0x062,0x000175,0x175,0x062,0x00017b,0x17b,0x063,0x00017e,0x17e,0x063,0x000180,0x180)
NTDLL(NtSetCachedSigningLevel                           ,NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,0x05c,0x000164,0x164,0x05f,0x000167,0x167,0x060,0x00016e,0x16e,0x060,0x000171,0x171,0x061,0x000176,0x176,0x061,0x00017c,0x17c,0x062,0x00017f,0x17f,0x062,0x000181,0x181)
NTDLL(NtSetCachedSigningLevel2                          ,NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE,0x060,0x000177,0x177,0x060,0x00017d,0x17d,0x061,0x000180,0x180,0x061,0x000182,0x182)
NTDLL(NtSetContextChannel                               ,0xf6, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE)
NTDLL(NtSetContextThread                                ,0xba,0x0d5,0x0dd,0x0f6,0x0f6,0x121,0x149,0x149,0x13c,0x150,0x150,0x05b,0x000165,0x165,0x05e,0x000168,0x168,0x05f,0x00016f,0x16f,0x05f,0x000172,0x172,0x05f,0x000178,0x178,0x05f,0x00017e,0x17e,0x060,0x000181,0x181,0x060,0x000183,0x183)
NTDLL(NtSetDebugFilterState                             ,NONE,0x0d6,0x0de,0x0f7,0x0f7,0x122,0x14a,0x14a,0x13d,0x151,0x151,0x05a,0x0a0166,0x166,0x05d,0x0a0169,0x169,0x05e,0x0a0170,0x170,0x05e,0x0a0173,0x173,0x05e,0x0a0179,0x179,0x05e,0x0a017f,0x17f,0x05f,0x0a0182,0x182,0x05f,0x0a0184,0x184)
NTDLL(NtSetDefaultHardErrorPort                         ,0xbb,0x0d7,0x0df,0x0f8,0x0f8,0x123,0x14b,0x14b,0x13e,0x152,0x152,0x059,0x030167,0x167,0x05c,0x03016a,0x16a,0x05d,0x030171,0x171,0x05d,0x030174,0x174,0x05d,0x03017a,0x17a,0x05d,0x030180,0x180,0x05e,0x030183,0x183,0x05e,0x030185,0x185)
NTDLL(NtSetDefaultLocale                                ,0xbc,0x0d8,0x0e0,0x0f9,0x0f9,0x124,0x14c,0x14c,0x13f,0x153,0x153,0x058,0x050168,0x168,0x05b,0x05016b,0x16b,0x05c,0x050172,0x172,0x05c,0x050175,0x175,0x05c,0x05017b,0x17b,0x05c,0x050181,0x181,0x05d,0x050184,0x184,0x05d,0x050186,0x186)
NTDLL(NtSetDefaultUILanguage                            ,0xbd,0x0d9,0x0e1,0x0fa,0x0fa,0x125,0x14d,0x14d,0x140,0x154,0x154,0x057,0x040169,0x169,0x05a,0x04016c,0x16c,0x05b,0x040173,0x173,0x05b,0x040176,0x176,0x05b,0x04017c,0x17c,0x05b,0x040182,0x182,0x05c,0x040185,0x185,0x05c,0x040187,0x187)
NTDLL(NtSetDriverEntryOrder                             ,NONE, NONE,0x0e2,0x0fb,0x0fb,0x126,0x14e,0x14e,0x141,0x155,0x155,0x056,0x00016a,0x16a,0x059,0x00016d,0x16d,0x05a,0x000174,0x174,0x05a,0x000177,0x177,0x05a,0x00017d,0x17d,0x05a,0x000183,0x183,0x05b,0x000186,0x186,0x05b,0x000188,0x188)
NTDLL(NtSetEaFile                                       ,0xbe,0x0da,0x0e3,0x0fc,0x0fc,0x127,0x14f,0x14f,0x142,0x156,0x156,0x055,0x00016b,0x16b,0x058,0x00016e,0x16e,0x059,0x000175,0x175,0x059,0x000178,0x178,0x059,0x00017e,0x17e,0x059,0x000184,0x184,0x05a,0x000187,0x187,0x05a,0x000189,0x189)
NTDLL(NtSetEvent                                        ,0xbf,0x0db,0x0e4,0x00b,0x00b,0x128,0x00b,0x00b,0x143,0x00b,0x00b,0x054,0x07000c,0x00c,0x057,0x07000d,0x00d,0x058,0x07000e,0x00e,0x058,0x07000e,0x00e,0x058,0x07000e,0x00e,0x058,0x07000e,0x00e,0x059,0x07000e,0x00e,0x059,0x07000e,0x00e)
NTDLL(NtSetEventBoostPriority                           ,NONE,0x0dc,0x0e5,0x02a,0x02a,0x129,0x02a,0x02a,0x144,0x02a,0x02a,0x053,0x03002b,0x02b,0x056,0x03002c,0x02c,0x057,0x03002d,0x02d,0x057,0x03002d,0x02d,0x057,0x03002d,0x02d,0x057,0x03002d,0x02d,0x058,0x03002d,0x02d,0x058,0x03002d,0x02d)
NTDLL(NtSetHighEventPair                                ,0xc0,0x0dd,0x0e6,0x0fd,0x0fd,0x12a,0x150,0x150,0x145,0x157,0x157,0x052,0x03016c,0x16c,0x055,0x03016f,0x16f,0x056,0x030176,0x176,0x056,0x030179,0x179,0x056,0x03017f,0x17f,0x056,0x030185,0x185,0x057,0x030188,0x188,0x057,0x03018a,0x18a)
NTDLL(NtSetHighWaitLowEventPair                         ,0xc1,0x0de,0x0e7,0x0fe,0x0fe,0x12b,0x151,0x151,0x146,0x158,0x158,0x051,0x03016d,0x16d,0x054,0x030170,0x170,0x055,0x030177,0x177,0x055,0x03017a,0x17a,0x055,0x030180,0x180,0x055,0x030186,0x186,0x056,0x030189,0x189,0x056,0x03018b,0x18b)
NTDLL(NtSetIRTimer                                      ,NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,0x040,0x00016e,0x16e,0x041,0x070171,0x171,0x042,0x070178,0x178,0x042,0x07017b,0x17b,0x042,0x070181,0x181,0x042,0x070187,0x187,0x043,0x07018a,0x18a,0x043,0x07018c,0x18c)
NTDLL(NtSetInformationDebugObject                       ,NONE,0x0df,0x0e8,0x0ff,0x0ff,0x12c,0x152,0x152,0x147,0x159,0x159,0x050,0x00016f,0x16f,0x053,0x000172,0x172,0x054,0x000179,0x179,0x054,0x00017c,0x17c,0x054,0x000182,0x182,0x054,0x000188,0x188,0x055,0x00018b,0x18b,0x055,0x00018d,0x18d)
NTDLL(NtSetInformationEnlistment                        ,NONE, NONE, NONE, NONE, NONE,0x071,0x153,0x153,0x148,0x15a,0x15a,0x04f,0x000170,0x170,0x052,0x000173,0x173,0x053,0x00017a,0x17a,0x053,0x00017d,0x17d,0x053,0x000183,0x183,0x053,0x000189,0x189,0x054,0x00018c,0x18c,0x054,0x00018e,0x18e)
NTDLL(NtSetInformationFile                              ,0xc2,0x0e0,0x0e9,0x024,0x024,0x12d,0x024,0x024,0x149,0x024,0x024,0x04e,0x000025,0x025,0x051,0x000026,0x026,0x052,0x000027,0x027,0x052,0x000027,0x027,0x052,0x000027,0x027,0x052,0x000027,0x027,0x053,0x000027,0x027,0x053,0x000027,0x027)
NTDLL(NtSetInformationJobObject                         ,0xc3,0x0e1,0x0ea,0x100,0x100,0x12e,0x154,0x154,0x14a,0x15b,0x15b,0x04d,0x000171,0x171,0x050,0x000174,0x174,0x051,0x00017b,0x17b,0x051,0x00017e,0x17e,0x051,0x000184,0x184,0x051,0x00018a,0x18a,0x052,0x00018d,0x18d,0x052,0x00018f,0x18f)
NTDLL(NtSetInformationKey                               ,0xc4,0x0e2,0x0eb,0x101,0x101,0x12f,0x155,0x155,0x14b,0x15c,0x15c,0x04c,0x000172,0x172,0x04f,0x000175,0x175,0x050,0x00017c,0x17c,0x050,0x00017f,0x17f,0x050,0x000185,0x185,0x050,0x00018b,0x18b,0x051,0x00018e,0x18e,0x051,0x000190,0x190)
NTDLL(NtSetInformationObject                            ,0xc5,0x0e3,0x0ec,0x059,0x059,0x130,0x059,0x059,0x14c,0x059,0x059,0x04b,0x00005a,0x05a,0x04e,0x00005b,0x05b,0x04f,0x00005c,0x05c,0x04f,0x00005c,0x05c,0x04f,0x00005c,0x05c,0x04f,0x00005c,0x05c,0x050,0x00005c,0x05c,0x050,0x00005c,0x05c)
NTDLL(NtSetInformationProcess                           ,0xc6,0x0e4,0x0ed,0x019,0x019,0x131,0x019,0x019,0x14d,0x019,0x019,0x04a,0x00001a,0x01a,0x04d,0x00001b,0x01b,0x04e,0x00001c,0x01c,0x04e,0x00001c,0x01c,0x04e,0x00001c,0x01c,0x04e,0x00001c,0x01c,0x04f,0x00001c,0x01c,0x04f,0x00001c,0x01c)
NTDLL(NtSetInformationResourceManager                   ,NONE, NONE, NONE, NONE, NONE,0x063,0x156,0x156,0x14e,0x15d,0x15d,0x049,0x000173,0x173,0x04c,0x000176,0x176,0x04d,0x00017d,0x17d,0x04d,0x000180,0x180,0x04d,0x000186,0x186,0x04d,0x00018c,0x18c,0x04e,0x00018f,0x18f,0x04e,0x000191,0x191)
NTDLL(NtSetInformationSymbolicLink                      ,NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE,0x1b8,0x00017e,0x17e,0x1bb,0x000181,0x181,0x1bd,0x000187,0x187,0x1c3,0x00018d,0x18d,0x1c7,0x000190,0x190,0x1c7,0x000192,0x192)
NTDLL(NtSetInformationThread                            ,0xc7,0x0e5,0x0ee,0x00a,0x00a,0x132,0x00a,0x00a,0x14f,0x00a,0x00a,0x048,0x00000b,0x00b,0x04b,0x00000c,0x00c,0x04c,0x00000d,0x00d,0x04c,0x00000d,0x00d,0x04c,0x00000d,0x00d,0x04c,0x00000d,0x00d,0x04d,0x00000d,0x00d,0x04d,0x00000d,0x00d)
NTDLL(NtSetInformationToken                             ,0xc8,0x0e6,0x0ef,0x102,0x102,0x133,0x157,0x157,0x150,0x15e,0x15e,0x047,0x000174,0x174,0x04a,0x000177,0x177,0x04b,0x00017f,0x17f,0x04b,0x000182,0x182,0x04b,0x000188,0x188,0x04b,0x00018e,0x18e,0x04c,0x000191,0x191,0x04c,0x000193,0x193)
NTDLL(NtSetInformationTransaction                       ,NONE, NONE, NONE, NONE, NONE,0x061,0x158,0x158,0x151,0x15f,0x15f,0x046,0x000175,0x175,0x049,0x000178,0x178,0x04a,0x000180,0x180,0x04a,0x000183,0x183,0x04a,0x000189,0x189,0x04a,0x00018f,0x18f,0x04b,0x000192,0x192,0x04b,0x000194,0x194)
NTDLL(NtSetInformationTransactionManager                ,NONE, NONE, NONE, NONE, NONE,0x062,0x159,0x159,0x152,0x160,0x160,0x045,0x000176,0x176,0x048,0x000179,0x179,0x049,0x000181,0x181,0x049,0x000184,0x184,0x049,0x00018a,0x18a,0x049,0x000190,0x190,0x04a,0x000193,0x193,0x04a,0x000195,0x195)
NTDLL(NtSetInformationVirtualMemory                     ,NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,0x0c9,0x000177,0x177,0x0cc,0x00017a,0x17a,0x0ce,0x000182,0x182,0x0ce,0x000185,0x185,0x0d0,0x00018b,0x18b,0x0d2,0x000191,0x191,0x0d4,0x000194,0x194,0x0d4,0x000196,0x196)
NTDLL(NtSetInformationWorkerFactory                     ,NONE, NONE, NONE, NONE, NONE,0x17a,0x15a,0x15a,0x153,0x161,0x161,0x044,0x000178,0x178,0x047,0x00017b,0x17b,0x048,0x000183,0x183,0x048,0x000186,0x186,0x048,0x00018c,0x18c,0x048,0x000192,0x192,0x049,0x000195,0x195,0x049,0x000197,0x197)
NTDLL(NtSetIntervalProfile                              ,0xc9,0x0e7,0x0f0,0x103,0x103,0x134,0x15b,0x15b,0x154,0x162,0x162,0x043,0x050179,0x179,0x046,0x05017c,0x17c,0x047,0x050184,0x184,0x047,0x050187,0x187,0x047,0x05018d,0x18d,0x047,0x050193,0x193,0x048,0x050196,0x196,0x048,0x050198,0x198)
NTDLL(NtSetIoCompletion                                 ,0xb9,0x0e8,0x0f1,0x104,0x104,0x135,0x15c,0x15c,0x155,0x163,0x163,0x042,0x00017a,0x17a,0x045,0x00017d,0x17d,0x046,0x000185,0x185,0x046,0x000188,0x188,0x046,0x00018e,0x18e,0x046,0x000194,0x194,0x047,0x000197,0x197,0x047,0x000199,0x199)
NTDLL(NtSetIoCompletionEx                               ,NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,0x156,0x164,0x164,0x041,0x00017b,0x17b,0x044,0x00017e,0x17e,0x045,0x000186,0x186,0x045,0x000189,0x189,0x045,0x00018f,0x18f,0x045,0x000195,0x195,0x046,0x000198,0x198,0x046,0x00019a,0x19a)
NTDLL(NtSetLdtEntries                                   ,0xca,0x0e9,0x0f2,0x105,0x105,0x136,0x15d,0x15d,0x157,0x165,0x165,0x03f,0x00017c,0x17c,0x040,0x00017f,0x17f,0x041,0x000187,0x187,0x041,0x00018a,0x18a,0x041,0x000190,0x190,0x041,0x000196,0x196,0x042,0x000199,0x199,0x042,0x00019b,0x19b)
NTDLL(NtSetLowEventPair                                 ,0xcb,0x0ea,0x0f3,0x106,0x106,0x137,0x15e,0x15e,0x158,0x166,0x166,0x03e,0x03017d,0x17d,0x03f,0x030180,0x180,0x040,0x030188,0x188,0x040,0x03018b,0x18b,0x040,0x030191,0x191,0x040,0x030197,0x197,0x041,0x03019a,0x19a,0x041,0x03019c,0x19c)
NTDLL(NtSetLowWaitHighEventPair                         ,0xcc,0x0eb,0x0f4,0x107,0x107,0x138,0x15f,0x15f,0x159,0x167,0x167,0x03d,0x03017e,0x17e,0x03e,0x030181,0x181,0x03f,0x030189,0x189,0x03f,0x03018c,0x18c,0x03f,0x030192,0x192,0x03f,0x030198,0x198,0x040,0x03019b,0x19b,0x040,0x03019d,0x19d)
NTDLL(NtSetQuotaInformationFile                         ,0xcd,0x0ec,0x0f5,0x108,0x108,0x139,0x160,0x160,0x15a,0x168,0x168,0x03c,0x00017f,0x17f,0x03d,0x000182,0x182,0x03e,0x00018a,0x18a,0x03e,0x00018d,0x18d,0x03e,0x000193,0x193,0x03e,0x000199,0x199,0x03f,0x00019c,0x19c,0x03f,0x00019e,0x19e)
NTDLL(NtSetSecurityObject                               ,0xce,0x0ed,0x0f6,0x109,0x109,0x13a,0x161,0x161,0x15b,0x169,0x169,0x03b,0x000180,0x180,0x03c,0x000183,0x183,0x03d,0x00018b,0x18b,0x03d,0x00018e,0x18e,0x03d,0x000194,0x194,0x03d,0x00019a,0x19a,0x03e,0x00019d,0x19d,0x03e,0x00019f,0x19f)
NTDLL(NtSetSystemEnvironmentValue                       ,0xcf,0x0ee,0x0f7,0x10a,0x10a,0x13b,0x162,0x162,0x15c,0x16a,0x16a,0x03a,0x000181,0x181,0x03b,0x000184,0x184,0x03c,0x00018c,0x18c,0x03c,0x00018f,0x18f,0x03c,0x000195,0x195,0x03c,0x00019b,0x19b,0x03d,0x00019e,0x19e,0x03d,0x0001a0,0x1a0)
NTDLL(NtSetSystemEnvironmentValueEx                     ,NONE,0x0ef,0x0f8,0x10b,0x10b,0x13c,0x163,0x163,0x15d,0x16b,0x16b,0x039,0x000182,0x182,0x03a,0x000185,0x185,0x03b,0x00018d,0x18d,0x03b,0x000190,0x190,0x03b,0x000196,0x196,0x03b,0x00019c,0x19c,0x03c,0x00019f,0x19f,0x03c,0x0001a1,0x1a1)
NTDLL(NtSetSystemInformation                            ,0xd0,0x0f0,0x0f9,0x10c,0x10c,0x13d,0x164,0x164,0x15e,0x16c,0x16c,0x038,0x000183,0x183,0x039,0x000186,0x186,0x03a,0x00018e,0x18e,0x03a,0x000191,0x191,0x03a,0x000197,0x197,0x03a,0x00019d,0x19d,0x03b,0x0001a0,0x1a0,0x03b,0x0001a2,0x1a2)
NTDLL(NtSetSystemPowerState                             ,0xd1,0x0f1,0x0fa,0x10d,0x10d,0x13e,0x165,0x165,0x15f,0x16d,0x16d,0x037,0x000184,0x184,0x038,0x000187,0x187,0x039,0x00018f,0x18f,0x039,0x000192,0x192,0x039,0x000198,0x198,0x039,0x00019e,0x19e,0x03a,0x0001a1,0x1a1,0x03a,0x0001a3,0x1a3)
NTDLL(NtSetSystemTime                                   ,0xd2,0x0f2,0x0fb,0x10e,0x10e,0x13f,0x166,0x166,0x160,0x16e,0x16e,0x036,0x050185,0x185,0x037,0x050188,0x188,0x038,0x050190,0x190,0x038,0x050193,0x193,0x038,0x050199,0x199,0x038,0x05019f,0x19f,0x039,0x0501a2,0x1a2,0x039,0x0501a4,0x1a4)
NTDLL(NtSetThreadExecutionState                         ,0xd3,0x0f3,0x0fc,0x10f,0x10f,0x140,0x167,0x167,0x161,0x16f,0x16f,0x035,0x050186,0x186,0x036,0x050189,0x189,0x037,0x050191,0x191,0x037,0x050194,0x194,0x037,0x05019a,0x19a,0x037,0x0501a0,0x1a0,0x038,0x0501a3,0x1a3,0x038,0x0501a5,0x1a5)
NTDLL(NtSetTimer                                        ,0xd4,0x0f4,0x0fd,0x05f,0x05f,0x141,0x05f,0x05f,0x162,0x05f,0x05f,0x034,0x000060,0x060,0x035,0x000061,0x061,0x036,0x000062,0x062,0x036,0x000062,0x062,0x036,0x000062,0x062,0x036,0x000062,0x062,0x037,0x000062,0x062,0x037,0x000062,0x062)
NTDLL(NtSetTimer2                                       ,NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,    NONE, NONE,0x042,0x00018a,0x18a,0x043,0x000192,0x192,0x043,0x000195,0x195,0x043,0x00019b,0x19b,0x043,0x0001a1,0x1a1,0x044,0x0001a4,0x1a4,0x044,0x0001a6,0x1a6)
NTDLL(NtSetTimerEx                                      ,NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,0x163,0x170,0x170,0x033,0x000187,0x187,0x034,0x00018b,0x18b,0x035,0x000193,0x193,0x035,0x000196,0x196,0x035,0x00019c,0x19c,0x035,0x0001a2,0x1a2,0x036,0x0001a5,0x1a5,0x036,0x0001a7,0x1a7)
NTDLL(NtSetTimerResolution                              ,0xd5,0x0f5,0x0fe,0x110,0x110,0x142,0x168,0x168,0x164,0x171,0x171,0x032,0x0a0188,0x188,0x033,0x0a018c,0x18c,0x034,0x0a0194,0x194,0x034,0x0a0197,0x197,0x034,0x0a019d,0x19d,0x034,0x0a01a3,0x1a3,0x035,0x0a01a6,0x1a6,0x035,0x0a01a8,0x1a8)
NTDLL(NtSetUuidSeed                                     ,0xd6,0x0f6,0x0ff,0x111,0x111,0x143,0x169,0x169,0x165,0x172,0x172,0x031,0x040189,0x189,0x032,0x04018d,0x18d,0x033,0x040195,0x195,0x033,0x040198,0x198,0x033,0x04019e,0x19e,0x033,0x0401a4,0x1a4,0x034,0x0401a7,0x1a7,0x034,0x0401a9,0x1a9)
NTDLL(NtSetValueKey                                     ,0xd7,0x0f7,0x100,0x05d,0x05d,0x144,0x05d,0x05d,0x166,0x05d,0x05d,0x030,0x00005e,0x05e,0x031,0x00005f,0x05f,0x032,0x000060,0x060,0x032,0x000060,0x060,0x032,0x000060,0x060,0x032,0x000060,0x060,0x033,0x000060,0x060,0x033,0x000060,0x060)
NTDLL(NtSetVolumeInformationFile                        ,0xd8,0x0f8,0x101,0x112,0x112,0x145,0x16a,0x16a,0x167,0x173,0x173,0x02f,0x00018a,0x18a,0x030,0x00018e,0x18e,0x031,0x000196,0x196,0x031,0x000199,0x199,0x031,0x00019f,0x19f,0x031,0x0001a5,0x1a5,0x032,0x0001a8,0x1a8,0x032,0x0001aa,0x1aa)
NTDLL(NtSetWnfProcessNotificationEvent                  ,NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,    NONE, NONE,0x02f,0x00018f,0x18f,0x030,0x000197,0x197,0x030,0x00019a,0x19a,0x030,0x0001a0,0x1a0,0x030,0x0001a6,0x1a6,0x031,0x0001a9,0x1a9,0x031,0x0001ab,0x1ab)
NTDLL(NtShutdownSystem                                  ,0xd9,0x0f9,0x102,0x113,0x113,0x146,0x16b,0x16b,0x168,0x174,0x174,0x02e,0x04018b,0x18b,0x02e,0x040190,0x190,0x02f,0x040198,0x198,0x02f,0x04019b,0x19b,0x02f,0x0401a1,0x1a1,0x02f,0x0401a7,0x1a7,0x030,0x0401aa,0x1aa,0x030,0x0401ac,0x1ac)
NTDLL(NtShutdownWorkerFactory                           ,NONE, NONE, NONE, NONE, NONE,0x17d,0x16c,0x16c,0x169,0x175,0x175,0x02d,0x00018c,0x18c,0x02d,0x000191,0x191,0x02e,0x000199,0x199,0x02e,0x00019c,0x19c,0x02e,0x0001a2,0x1a2,0x02e,0x0001a8,0x1a8,0x02f,0x0001ab,0x1ab,0x02f,0x0001ad,0x1ad)
NTDLL(NtSignalAndWaitForSingleObject                    ,0xda,0x0fa,0x103,0x114,0x114,0x147,0x16d,0x16d,0x16a,0x176,0x176,0x02c,0x13018d,0x18d,0x02c,0x130192,0x192,0x02d,0x13019a,0x19a,0x02d,0x13019d,0x19d,0x02d,0x1301a3,0x1a3,0x02d,0x1301a9,0x1a9,0x02e,0x1301ac,0x1ac,0x02e,0x1301ae,0x1ae)
NTDLL(NtSinglePhaseReject                               ,NONE, NONE, NONE, NONE, NONE,0x060,0x16e,0x16e,0x16b,0x177,0x177,0x02b,0x00018e,0x18e,0x02b,0x000193,0x193,0x02c,0x00019b,0x19b,0x02c,0x00019e,0x19e,0x02c,0x0001a4,0x1a4,0x02c,0x0001aa,0x1aa,0x02d,0x0001ad,0x1ad,0x02d,0x0001af,0x1af)
NTDLL(NtStartProfile                                    ,0xdb,0x0fb,0x104,0x115,0x115,0x148,0x16f,0x16f,0x16c,0x178,0x178,0x02a,0x03018f,0x18f,0x02a,0x030194,0x194,0x02b,0x03019c,0x19c,0x02b,0x03019f,0x19f,0x02b,0x0301a5,0x1a5,0x02b,0x0301ab,0x1ab,0x02c,0x0301ae,0x1ae,0x02c,0x0301b0,0x1b0)
NTDLL(NtStopProfile                                     ,0xdc,0x0fc,0x105,0x116,0x116,0x149,0x170,0x170,0x16d,0x179,0x179,0x029,0x030190,0x190,0x029,0x030195,0x195,0x02a,0x03019d,0x19d,0x02a,0x0301a0,0x1a0,0x02a,0x0301a6,0x1a6,0x02a,0x0301ac,0x1ac,0x02b,0x0301af,0x1af,0x02b,0x0301b1,0x1b1)
NTDLL(NtSubscribeWnfStateChange                         ,NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,0x028,0x000191,0x191,0x028,0x000196,0x196,0x029,0x00019e,0x19e,0x029,0x0001a1,0x1a1,0x029,0x0001a7,0x1a7,0x029,0x0001ad,0x1ad,0x02a,0x0001b0,0x1b0,0x02a,0x0001b2,0x1b2)
NTDLL(NtSuspendProcess                                  ,NONE,0x0fd,0x106,0x117,0x117,0x14a,0x171,0x171,0x16e,0x17a,0x17a,0x027,0x030192,0x192,0x027,0x030197,0x197,0x028,0x03019f,0x19f,0x028,0x0301a2,0x1a2,0x028,0x0301a8,0x1a8,0x028,0x0301ae,0x1ae,0x029,0x0301b1,0x1b1,0x029,0x0301b3,0x1b3)
NTDLL(NtSuspendThread                                   ,0xdd,0x0fe,0x107,0x118,0x118,0x14b,0x172,0x172,0x16f,0x17b,0x17b,0x026,0x070193,0x193,0x026,0x070198,0x198,0x027,0x0701a0,0x1a0,0x027,0x0701a3,0x1a3,0x027,0x0701a9,0x1a9,0x027,0x0701af,0x1af,0x028,0x0701b2,0x1b2,0x028,0x0701b4,0x1b4)
NTDLL(NtSystemDebugControl                              ,0xde,0x0ff,0x108,0x119,0x119,0x14c,0x173,0x173,0x170,0x17c,0x17c,0x025,0x000194,0x194,0x025,0x000199,0x199,0x026,0x0001a1,0x1a1,0x026,0x0001a4,0x1a4,0x026,0x0001aa,0x1aa,0x026,0x0001b0,0x1b0,0x027,0x0001b3,0x1b3,0x027,0x0001b5,0x1b5)
NTDLL(NtTerminateEnclave                                ,NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE,0x026,0x0001b4,0x1b4,0x026,0x0001b6,0x1b6)
NTDLL(NtTerminateJobObject                              ,0xdf,0x100,0x109,0x11a,0x11a,0x14d,0x174,0x174,0x171,0x17d,0x17d,0x024,0x070195,0x195,0x024,0x07019a,0x19a,0x025,0x0701a2,0x1a2,0x025,0x0701a5,0x1a5,0x025,0x0701ab,0x1ab,0x025,0x0701b1,0x1b1,0x025,0x0701b5,0x1b5,0x025,0x0701b7,0x1b7)
NTDLL(NtTerminateProcess                                ,0xe0,0x101,0x10a,0x029,0x029,0x14e,0x029,0x029,0x172,0x029,0x029,0x023,0x00002a,0x02a,0x023,0x07002b,0x02b,0x024,0x07002c,0x02c,0x024,0x07002c,0x02c,0x024,0x07002c,0x02c,0x024,0x07002c,0x02c,0x024,0x07002c,0x02c,0x024,0x07002c,0x02c)
NTDLL(NtTerminateThread                                 ,0xe1,0x102,0x10b,0x050,0x050,0x14f,0x050,0x050,0x173,0x050,0x050,0x022,0x000051,0x051,0x022,0x070052,0x052,0x023,0x070053,0x053,0x023,0x070053,0x053,0x023,0x070053,0x053,0x023,0x070053,0x053,0x023,0x070053,0x053,0x023,0x070053,0x053)
NTDLL(NtTestAlert                                       ,0xe2,0x103,0x10c,0x11b,0x11b,0x150,0x175,0x175,0x174,0x17e,0x17e,0x021,0x020196,0x196,0x021,0x02019b,0x19b,0x022,0x0201a3,0x1a3,0x022,0x0201a6,0x1a6,0x022,0x0201ac,0x1ac,0x022,0x0201b2,0x1b2,0x022,0x0201b6,0x1b6,0x022,0x0201b8,0x1b8)
NTDLL(NtThawRegistry                                    ,NONE, NONE, NONE, NONE, NONE,0x151,0x176,0x176,0x175,0x17f,0x17f,0x020,0x000197,0x197,0x020,0x00019c,0x19c,0x021,0x0001a4,0x1a4,0x021,0x0001a7,0x1a7,0x021,0x0001ad,0x1ad,0x021,0x0001b3,0x1b3,0x021,0x0001b7,0x1b7,0x021,0x0001b9,0x1b9)
NTDLL(NtThawTransactions                                ,NONE, NONE, NONE, NONE, NONE,0x152,0x177,0x177,0x176,0x180,0x180,0x01f,0x000198,0x198,0x01f,0x00019d,0x19d,0x020,0x0001a5,0x1a5,0x020,0x0001a8,0x1a8,0x020,0x0001ae,0x1ae,0x020,0x0001b4,0x1b4,0x020,0x0001b8,0x1b8,0x020,0x0001ba,0x1ba)
NTDLL(NtTraceControl                                    ,NONE, NONE, NONE, NONE, NONE,0x154,0x178,0x178,0x177,0x181,0x181,0x01e,0x000199,0x199,0x01e,0x00019e,0x19e,0x01f,0x0001a6,0x1a6,0x01f,0x0001a9,0x1a9,0x01f,0x0001af,0x1af,0x01f,0x0001b5,0x1b5,0x01f,0x0001b9,0x1b9,0x01f,0x0001bb,0x1bb)
NTDLL(NtTraceEvent                                      ,NONE,0x104,0x10d,0x05b,0x05b,0x153,0x05b,0x05b,0x178,0x05b,0x05b,0x01d,0x00005c,0x05c,0x01d,0x00005d,0x05d,0x01e,0x00005e,0x05e,0x01e,0x00005e,0x05e,0x01e,0x00005e,0x05e,0x01e,0x00005e,0x05e,0x01e,0x00005e,0x05e,0x01e,0x00005e,0x05e)
NTDLL(NtTranslateFilePath                               ,NONE,0x105,0x10e,0x11c,0x11c,0x155,0x179,0x179,0x179,0x182,0x182,0x01c,0x11019a,0x19a,0x01c,0x11019f,0x19f,0x01d,0x1101a7,0x1a7,0x01d,0x1101aa,0x1aa,0x01d,0x1101b0,0x1b0,0x01d,0x1101b6,0x1b6,0x01d,0x1101ba,0x1ba,0x01d,0x1101bc,0x1bc)
NTDLL(NtUmsThreadYield                                  ,NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,0x17a,0x183,0x183,0x01b,0x00019b,0x19b,0x01b,0x0001a0,0x1a0,0x01c,0x0001a8,0x1a8,0x01c,0x0001ab,0x1ab,0x01c,0x0001b1,0x1b1,0x01c,0x0001b7,0x1b7,0x01c,0x0001bb,0x1bb,0x01c,0x0001bd,0x1bd)
NTDLL(NtUnloadDriver                                    ,0xe3,0x106,0x10f,0x11d,0x11d,0x156,0x17a,0x17a,0x17b,0x184,0x184,0x01a,0x00019c,0x19c,0x01a,0x0001a1,0x1a1,0x01b,0x0001a9,0x1a9,0x01b,0x0001ac,0x1ac,0x01b,0x0001b2,0x1b2,0x01b,0x0001b8,0x1b8,0x01b,0x0001bc,0x1bc,0x01b,0x0001be,0x1be)
NTDLL(NtUnloadKey                                       ,0xe4,0x107,0x110,0x11e,0x11e,0x157,0x17b,0x17b,0x17c,0x185,0x185,0x019,0x00019d,0x19d,0x019,0x0001a2,0x1a2,0x01a,0x0001aa,0x1aa,0x01a,0x0001ad,0x1ad,0x01a,0x0001b3,0x1b3,0x01a,0x0001b9,0x1b9,0x01a,0x0001bd,0x1bd,0x01a,0x0001bf,0x1bf)
NTDLL(NtUnloadKey2                                      ,NONE, NONE,0x111,0x11f,0x11f,0x158,0x17c,0x17c,0x17d,0x186,0x186,0x018,0x00019e,0x19e,0x018,0x0001a3,0x1a3,0x019,0x0001ab,0x1ab,0x019,0x0001ae,0x1ae,0x019,0x0001b4,0x1b4,0x019,0x0001ba,0x1ba,0x019,0x0001be,0x1be,0x019,0x0001c0,0x1c0)
NTDLL(NtUnloadKeyEx                                     ,NONE,0x108,0x112,0x120,0x120,0x159,0x17d,0x17d,0x17e,0x187,0x187,0x017,0x00019f,0x19f,0x017,0x0001a4,0x1a4,0x018,0x0001ac,0x1ac,0x018,0x0001af,0x1af,0x018,0x0001b5,0x1b5,0x018,0x0001bb,0x1bb,0x018,0x0001bf,0x1bf,0x018,0x0001c1,0x1c1)
NTDLL(NtUnlockFile                                      ,0xe5,0x109,0x113,0x121,0x121,0x15a,0x17e,0x17e,0x17f,0x188,0x188,0x016,0x0001a0,0x1a0,0x016,0x0001a5,0x1a5,0x017,0x0001ad,0x1ad,0x017,0x0001b0,0x1b0,0x017,0x0001b6,0x1b6,0x017,0x0001bc,0x1bc,0x017,0x0001c0,0x1c0,0x017,0x0001c2,0x1c2)
NTDLL(NtUnlockVirtualMemory                             ,0xe6,0x10a,0x114,0x122,0x122,0x15b,0x17f,0x17f,0x180,0x189,0x189,0x015,0x0001a1,0x1a1,0x015,0x0001a6,0x1a6,0x016,0x0001ae,0x1ae,0x016,0x0001b1,0x1b1,0x016,0x0001b7,0x1b7,0x016,0x0001bd,0x1bd,0x016,0x0001c1,0x1c1,0x016,0x0001c3,0x1c3)
NTDLL(NtUnmapViewOfSection                              ,0xe7,0x10b,0x115,0x027,0x027,0x15c,0x027,0x027,0x181,0x027,0x027,0x013,0x000028,0x028,0x013,0x000029,0x029,0x014,0x00002a,0x02a,0x014,0x00002a,0x02a,0x014,0x00002a,0x02a,0x014,0x00002a,0x02a,0x014,0x00002a,0x02a,0x014,0x00002a,0x02a)
NTDLL(NtUnmapViewOfSectionEx                            ,NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,0x014,0x0001a2,0x1a2,0x014,0x0001a7,0x1a7,0x015,0x0001af,0x1af,0x015,0x0001b2,0x1b2,0x015,0x0001b8,0x1b8,0x015,0x0001be,0x1be,0x015,0x0001c2,0x1c2,0x015,0x0001c4,0x1c4)
NTDLL(NtUnsubscribeWnfStateChange                       ,NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,0x011,0x0001a3,0x1a3,0x011,0x0001a8,0x1a8,0x012,0x0001b0,0x1b0,0x012,0x0001b3,0x1b3,0x012,0x0001b9,0x1b9,0x012,0x0001bf,0x1bf,0x012,0x0001c3,0x1c3,0x012,0x0001c5,0x1c5)
NTDLL(NtUpdateWnfStateData                              ,NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,0x012,0x0001a4,0x1a4,0x012,0x0001a9,0x1a9,0x013,0x0001b1,0x1b1,0x013,0x0001b4,0x1b4,0x013,0x0001ba,0x1ba,0x013,0x0001c0,0x1c0,0x013,0x0001c4,0x1c4,0x013,0x0001c6,0x1c6)
NTDLL(NtVdmControl                                      ,0xe8,0x10c,0x116,0x123,0x123,0x15d,0x180,0x180,0x182,0x18a,0x18a,0x010,0x0001a5,0x1a5,0x010,0x0001aa,0x1aa,0x011,0x0001b2,0x1b2,0x011,0x0001b5,0x1b5,0x011,0x0001bb,0x1bb,0x011,0x0001c1,0x1c1,0x011,0x0001c5,0x1c5,0x011,0x0001c7,0x1c7)
NTDLL(NtWaitForAlertByThreadId                          ,NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,0x00f,0x0601a6,0x1a6,0x00f,0x0601ab,0x1ab,0x010,0x0601b3,0x1b3,0x010,0x0601b6,0x1b6,0x010,0x0601bc,0x1bc,0x010,0x0601c2,0x1c2,0x010,0x0601c6,0x1c6,0x010,0x0601c8,0x1c8)
NTDLL(NtWaitForDebugEvent                               ,NONE,0x10d,0x117,0x124,0x124,0x15e,0x181,0x181,0x183,0x18b,0x18b,0x00e,0x0001a7,0x1a7,0x00e,0x0001ac,0x1ac,0x00f,0x0001b4,0x1b4,0x00f,0x0001b7,0x1b7,0x00f,0x0001bd,0x1bd,0x00f,0x0001c3,0x1c3,0x00f,0x0001c7,0x1c7,0x00f,0x0001c9,0x1c9)
NTDLL(NtWaitForKeyedEvent                               ,NONE,0x11a,0x124,0x125,0x125,0x16b,0x182,0x182,0x184,0x18c,0x18c,0x00d,0x1501a8,0x1a8,0x00d,0x1501ad,0x1ad,0x00e,0x1501b5,0x1b5,0x00e,0x1501b8,0x1b8,0x00e,0x1501be,0x1be,0x00e,0x1501c4,0x1c4,0x00e,0x1501c8,0x1c8,0x00e,0x1501ca,0x1ca)
NTDLL(NtWaitForMultipleObjects                          ,0xe9,0x10e,0x118,0x058,0x058,0x15f,0x058,0x058,0x185,0x058,0x058,0x00c,0x1d0059,0x059,0x00c,0x1d005a,0x05a,0x00d,0x1d005b,0x05b,0x00d,0x1d005b,0x05b,0x00d,0x1d005b,0x05b,0x00d,0x1d005b,0x05b,0x00d,0x1d005b,0x05b,0x00d,0x1d005b,0x05b)
NTDLL(NtWaitForMultipleObjects32                        ,NONE, NONE,0x127,0x017,0x017,0x16e,0x017,0x017,0x186,0x017,0x017,0x00b,0x1e0018,0x018,0x00b,0x1e0019,0x019,0x00c,0x1e001a,0x01a,0x00c,0x1e001a,0x01a,0x00c,0x1e001a,0x01a,0x00c,0x1e001a,0x01a,0x00c,0x1e001a,0x01a,0x00c,0x1e001a,0x01a)
NTDLL(NtWaitForSingleObject                             ,0xea,0x10f,0x119,0x001,0x001,0x160,0x001,0x001,0x187,0x001,0x001,0x00a,0x0d0002,0x002,0x00a,0x0d0003,0x003,0x00b,0x0d0004,0x004,0x00b,0x0d0004,0x004,0x00b,0x0d0004,0x004,0x00b,0x0d0004,0x004,0x00b,0x0d0004,0x004,0x00b,0x0d0004,0x004)
NTDLL(NtWaitForWnfNotifications                         ,NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,0x009,0x0001a9,0x1a9, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE)
NTDLL(NtWaitForWorkViaWorkerFactory                     ,NONE, NONE, NONE, NONE, NONE,0x179,0x183,0x183,0x188,0x18d,0x18d,0x008,0x0001aa,0x1aa,0x009,0x0001ae,0x1ae,0x00a,0x0001b6,0x1b6,0x00a,0x0001b9,0x1b9,0x00a,0x0001bf,0x1bf,0x00a,0x0001c5,0x1c5,0x00a,0x0001c9,0x1c9,0x00a,0x0001cb,0x1cb)
NTDLL(NtWaitHighEventPair                               ,0xeb,0x110,0x11a,0x126,0x126,0x161,0x184,0x184,0x189,0x18e,0x18e,0x007,0x0301ab,0x1ab,0x008,0x0301af,0x1af,0x009,0x0301b7,0x1b7,0x009,0x0301ba,0x1ba,0x009,0x0301c0,0x1c0,0x009,0x0301c6,0x1c6,0x009,0x0301ca,0x1ca,0x009,0x0301cc,0x1cc)
NTDLL(NtWaitLowEventPair                                ,0xec,0x111,0x11b,0x127,0x127,0x162,0x185,0x185,0x18a,0x18f,0x18f,0x006,0x0301ac,0x1ac,0x007,0x0301b0,0x1b0,0x008,0x0301b8,0x1b8,0x008,0x0301bb,0x1bb,0x008,0x0301c1,0x1c1,0x008,0x0301c7,0x1c7,0x008,0x0301cb,0x1cb,0x008,0x0301cd,0x1cd)
NTDLL(NtWorkerFactoryWorkerReady                        ,NONE, NONE, NONE, NONE, NONE,0x17c,0x186,0x186,0x18b,0x190,0x190,0x000,0x000000,0x000,0x000,0x030000,0x000,0x001,0x030001,0x001,0x001,0x030001,0x001,0x001,0x030001,0x001,0x001,0x030001,0x001,0x001,0x030001,0x001,0x001,0x030001,0x001)
NTDLL(NtWow64AllocateVirtualMemory64                    ,NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,0x0001bb, NONE, NONE,0x0001bf, NONE, NONE,0x0001c7, NONE, NONE,0x0001ca, NONE, NONE,0x0001d0, NONE, NONE,0x0001d6, NONE, NONE,0x0001da, NONE, NONE,0x0001dc, NONE)
NTDLL(NtWow64CallFunction64                             ,NONE, NONE, NONE, NONE, NONE, NONE,0x197, NONE, NONE,0x1a3, NONE, NONE,0x0001bf, NONE, NONE,0x0001c3, NONE, NONE,0x0001ca, NONE, NONE,0x0001cd, NONE, NONE,0x0001d3, NONE, NONE,0x0001d9, NONE, NONE,0x0001dd, NONE, NONE,0x0001df, NONE)
NTDLL(NtWow64CsrAllocateCaptureBuffer                   ,NONE, NONE, NONE,0x12c, NONE, NONE,0x18a, NONE, NONE,0x194, NONE, NONE,0x0001b0, NONE, NONE,0x0001b4, NONE, NONE,0x0001bc, NONE, NONE,0x0001bf, NONE, NONE,0x0001c5, NONE, NONE,0x0001cb, NONE, NONE,0x0001cf, NONE, NONE,0x0001d1, NONE)
NTDLL(NtWow64CsrAllocateMessagePointer                  ,NONE, NONE, NONE,0x12e, NONE, NONE,0x18c, NONE, NONE,0x196, NONE, NONE,0x0001b2, NONE, NONE,0x0001b6, NONE, NONE,0x0001be, NONE, NONE,0x0001c1, NONE, NONE,0x0001c7, NONE, NONE,0x0001cd, NONE, NONE,0x0001d1, NONE, NONE,0x0001d3, NONE)
NTDLL(NtWow64CsrCaptureMessageBuffer                    ,NONE, NONE, NONE,0x12f, NONE, NONE,0x18d, NONE, NONE,0x197, NONE, NONE,0x0001b3, NONE, NONE,0x0001b7, NONE, NONE,0x0001bf, NONE, NONE,0x0001c2, NONE, NONE,0x0001c8, NONE, NONE,0x0001ce, NONE, NONE,0x0001d2, NONE, NONE,0x0001d4, NONE)
NTDLL(NtWow64CsrCaptureMessageString                    ,NONE, NONE, NONE,0x130, NONE, NONE,0x18e, NONE, NONE,0x198, NONE, NONE,0x0001b4, NONE, NONE,0x0001b8, NONE, NONE,0x0001c0, NONE, NONE,0x0001c3, NONE, NONE,0x0001c9, NONE, NONE,0x0001cf, NONE, NONE,0x0001d3, NONE, NONE,0x0001d5, NONE)
NTDLL(NtWow64CsrClientCallServer                        ,NONE, NONE, NONE,0x12b, NONE, NONE,0x189, NONE, NONE,0x193, NONE, NONE,0x0001af, NONE, NONE,0x0001b3, NONE, NONE,0x0001bb, NONE, NONE,0x0001be, NONE, NONE,0x0001c4, NONE, NONE,0x0001ca, NONE, NONE,0x0001ce, NONE, NONE,0x0001d0, NONE)
NTDLL(NtWow64CsrClientConnectToServer                   ,NONE, NONE, NONE,0x128, NONE, NONE,0x187, NONE, NONE,0x191, NONE, NONE,0x0001ad, NONE, NONE,0x0001b1, NONE, NONE,0x0001b9, NONE, NONE,0x0001bc, NONE, NONE,0x0001c2, NONE, NONE,0x0001c8, NONE, NONE,0x0001cc, NONE, NONE,0x0001ce, NONE)
NTDLL(NtWow64CsrFreeCaptureBuffer                       ,NONE, NONE, NONE,0x12d, NONE, NONE,0x18b, NONE, NONE,0x195, NONE, NONE,0x0001b1, NONE, NONE,0x0001b5, NONE, NONE,0x0001bd, NONE, NONE,0x0001c0, NONE, NONE,0x0001c6, NONE, NONE,0x0001cc, NONE, NONE,0x0001d0, NONE, NONE,0x0001d2, NONE)
NTDLL(NtWow64CsrGetProcessId                            ,NONE, NONE, NONE,0x132, NONE, NONE,0x18f, NONE, NONE,0x199, NONE, NONE,0x0001b5, NONE, NONE,0x0001b9, NONE, NONE,0x0001c1, NONE, NONE,0x0001c4, NONE, NONE,0x0001ca, NONE, NONE,0x0001d0, NONE, NONE,0x0001d4, NONE, NONE,0x0001d6, NONE)
NTDLL(NtWow64CsrIdentifyAlertableThread                 ,NONE, NONE, NONE,0x12a, NONE, NONE,0x188, NONE, NONE,0x192, NONE, NONE,0x0001ae, NONE, NONE,0x0001b2, NONE, NONE,0x0001ba, NONE, NONE,0x0001bd, NONE, NONE,0x0001c3, NONE, NONE,0x0001c9, NONE, NONE,0x0001cd, NONE, NONE,0x0001cf, NONE)
NTDLL(NtWow64CsrNewThread                               ,NONE, NONE, NONE,0x129, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE)
NTDLL(NtWow64CsrSetPriorityClass                        ,NONE, NONE, NONE,0x131, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE)
NTDLL(NtWow64CsrVerifyRegion                            ,NONE, NONE, NONE, NONE, NONE, NONE,0x190, NONE, NONE,0x19a, NONE, NONE,0x0001b6, NONE, NONE,0x0001ba, NONE, NONE,0x0001c2, NONE, NONE,0x0001c5, NONE, NONE,0x0001cb, NONE, NONE,0x0001d1, NONE, NONE,0x0001d5, NONE, NONE,0x0001d7, NONE)
NTDLL(NtWow64DebuggerCall                               ,NONE, NONE, NONE,0x133, NONE, NONE,0x191, NONE, NONE,0x19b, NONE, NONE,0x0001b7, NONE, NONE,0x0001bb, NONE, NONE,0x0001c3, NONE, NONE,0x0001c6, NONE, NONE,0x0001cc, NONE, NONE,0x0001d2, NONE, NONE,0x0001d6, NONE, NONE,0x0001d8, NONE)
NTDLL(NtWow64GetCurrentProcessorNumberEx                ,NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,0x19c, NONE, NONE,0x0001b8, NONE, NONE,0x0001bc, NONE, NONE,0x0001c4, NONE, NONE,0x0001c7, NONE, NONE,0x0001cd, NONE, NONE,0x0001d3, NONE, NONE,0x0001d7, NONE, NONE,0x0001d9, NONE)
NTDLL(NtWow64GetNativeSystemInformation                 ,NONE, NONE, NONE,0x134, NONE, NONE,0x192, NONE, NONE,0x19d, NONE, NONE,0x0001b9, NONE, NONE,0x0001bd, NONE, NONE,0x0001c5, NONE, NONE,0x0001c8, NONE, NONE,0x0001ce, NONE, NONE,0x0001d4, NONE, NONE,0x0001d8, NONE, NONE,0x0001da, NONE)
NTDLL(NtWow64InterlockedPopEntrySList                   ,NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,0x19e, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE)
NTDLL(NtWow64IsProcessorFeaturePresent                  ,NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,0x0001cb, NONE, NONE,0x0001ce, NONE, NONE,0x0001d4, NONE, NONE,0x0001da, NONE, NONE,0x0001de, NONE, NONE,0x0001e0, NONE)
NTDLL(NtWow64QueryInformationProcess64                  ,NONE, NONE, NONE,0x135, NONE, NONE,0x193, NONE, NONE,0x19f, NONE, NONE,0x0001ba, NONE, NONE,0x0001be, NONE, NONE,0x0001c6, NONE, NONE,0x0001c9, NONE, NONE,0x0001cf, NONE, NONE,0x0001d5, NONE, NONE,0x0001d9, NONE, NONE,0x0001db, NONE)
NTDLL(NtWow64QueryVirtualMemory64                       ,NONE, NONE, NONE,0x137, NONE, NONE,0x196, NONE, NONE,0x1a2, NONE, NONE,0x0001be, NONE, NONE,0x0001c2, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE, NONE,    NONE, NONE)
NTDLL(NtWow64ReadVirtualMemory64                        ,NONE, NONE, NONE,0x136, NONE, NONE,0x194, NONE, NONE,0x1a0, NONE, NONE,0x0001bc, NONE, NONE,0x0001c0, NONE, NONE,0x0001c8, NONE, NONE,0x0001cb, NONE, NONE,0x0001d1, NONE, NONE,0x0001d7, NONE, NONE,0x0001db, NONE, NONE,0x0001dd, NONE)
NTDLL(NtWow64WriteVirtualMemory64                       ,NONE, NONE, NONE, NONE, NONE, NONE,0x195, NONE, NONE,0x1a1, NONE, NONE,0x0001bd, NONE, NONE,0x0001c1, NONE, NONE,0x0001c9, NONE, NONE,0x0001cc, NONE, NONE,0x0001d2, NONE, NONE,0x0001d8, NONE, NONE,0x0001dc, NONE, NONE,0x0001de, NONE)
NTDLL(NtWriteFile                                       ,0xed,0x112,0x11c,0x005,0x005,0x163,0x005,0x005,0x18c,0x005,0x005,0x005,0x1a0006,0x006,0x006,0x1a0007,0x007,0x007,0x1a0008,0x008,0x007,0x1a0008,0x008,0x007,0x1a0008,0x008,0x007,0x1a0008,0x008,0x007,0x1a0008,0x008,0x007,0x1a0008,0x008)
NTDLL(NtWriteFileGather                                 ,0xee,0x113,0x11d,0x018,0x018,0x164,0x018,0x018,0x18d,0x018,0x018,0x004,0x1a0019,0x019,0x005,0x1a001a,0x01a,0x006,0x1a001b,0x01b,0x006,0x1a001b,0x01b,0x006,0x1a001b,0x01b,0x006,0x1a001b,0x01b,0x006,0x1a001b,0x01b,0x006,0x1a001b,0x01b)
NTDLL(NtWriteRequestData                                ,0xef,0x114,0x11e,0x054,0x054,0x165,0x054,0x054,0x18e,0x054,0x054,0x003,0x000055,0x055,0x004,0x000056,0x056,0x005,0x000057,0x057,0x005,0x000057,0x057,0x005,0x000057,0x057,0x005,0x000057,0x057,0x005,0x000057,0x057,0x005,0x000057,0x057)
NTDLL(NtWriteVirtualMemory                              ,0xf0,0x115,0x11f,0x037,0x037,0x166,0x037,0x037,0x18f,0x037,0x037,0x002,0x000038,0x038,0x003,0x000039,0x039,0x004,0x00003a,0x03a,0x004,0x00003a,0x03a,0x004,0x00003a,0x03a,0x004,0x00003a,0x03a,0x004,0x00003a,0x03a,0x004,0x00003a,0x03a)
NTDLL(NtYieldExecution                                  ,0xf7,0x116,0x120,0x043,0x043,0x167,0x043,0x043,0x190,0x043,0x043,0x001,0x010044,0x044,0x002,0x010045,0x045,0x003,0x010046,0x046,0x003,0x010046,0x046,0x003,0x010046,0x046,0x003,0x010046,0x046,0x003,0x010046,0x046,0x003,0x010046,0x046)
